/*
Aether the Gazelle
Battle Support
	by Krabs
*/

local SKIN = "aether"
local WEIGHT = 70
local SHIELDS = 1
local Aether = Silverhorn.Aether

local scriptloaded = false

local STATE_NONE = 0
local STATE_CASTING = 1

local BLESSING_NAME = "Blessing"	-- kinda redundant but lmao
local BLESSING_RINGCOST = 15
local BLESSING_COOLDOWN = TICRATE*15/2

local BLESSING_STARTUP = TICRATE/4
local BLESSING_DURATION = BLESSING_STARTUP + TICRATE*8/9
local BLESSING_FRAMETICS = {3, 3, 2, 2, 2}
local BLESSING_LOOPTICS = 2	-- loop the final frames
local BLESSING_TOTALTICS = 12	-- make sure this is a sum of BLESSING_FRAMETICS

local function AetherBlessBurst(mo, scale)
	local bless = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
	bless.target = mo
	bless.state = S_AETHER_BLESSBURST
	bless.scale = FixedMul(scale, $)
	bless.blendmode = AST_ADD
end

rawset(_G, "AetherBlessSparkles", function(mo, scale, count, speed, variance)	-- gonna reuse this elsewhere
	if not (speed) then return end
	scale = $ or FRACUNIT
	count = $ or 1
	variance = $ or 0
	
	-- spawn sparkles
	for i = 1, count
		local hangle = P_RandomRange(1, 360) * ANG1
		local vangle = P_RandomRange(1, 360) * ANG1
				
		local momentum = speed + (P_RandomRange(-variance, variance) * FRACUNIT)
		local xthrust = FixedMul(cos(hangle), FixedMul(sin(vangle), speed))
		local ythrust = FixedMul(sin(hangle), FixedMul(cos(vangle), speed))
		local zthrust = FixedMul(sin(vangle), speed)
		
		local spark = AetherParticle(mo, xthrust, ythrust, zthrust, scale)
		spark.color = SKINCOLOR_GOLDENROD
		spark.destscale = 1
		spark.scalespeed = FixedMul(spark.scale, FRACUNIT/18)
		spark.momx = xthrust
		spark.momy = ythrust
		spark.momz = zthrust
	end
end)

-- co
rawset(_G, "SPELLNOGUARD", {})
SPELLNOGUARD[SPL_NONE] = false
SPELLNOGUARD[SPL_ARCANE] = false
SPELLNOGUARD[SPL_SWORDRAIN] = false
SPELLNOGUARD[SPL_BARRIER] = true
SPELLNOGUARD[SPL_AURA] = false
SPELLNOGUARD[SPL_GENESIS] = true
	
local function HandleSpecial(mo, doaction)
	local B = CBW_Battle
	
	local p = mo.player
	local aether = p.aether
	if not (aether) then return end
	
	-- might as well make sure we cant guard while using a spell
	for i = 0, SPL_MAX
		if not (SPELLNOGUARD[i]) then continue end	-- we should be able to guard with this
		
		local spell = aether.spells[i]
		if not (spell.active) then continue end
		
		p.canguard = false
		break
	end
	
	-- set name and rings
	p.actiontext = BLESSING_NAME
	p.actionrings = BLESSING_RINGCOST
	
	-- action states
	if not (p.actionstate)
		-- activate the ability
		if (doaction == 1)	-- just pressed it
			B.PayRings(p)
			p.actionstate = STATE_CASTING
			p.actiontime = 0	-- timer for abilities
			p.canguard = false
					
			-- set state
			P_ResetPlayer(p)
			aether.flags = $ & ~Aether.FLAGS.FLEW
			aether.flight = 0
			p.pflags = ($|PF_FULLSTASIS) & ~(PF_JUMPED|PF_SHIELDABILITY)
			mo.state = S_AETHER_BLESSING
			
			-- start sound
			S_StartSound(mo, sfx_aeblss)
		end
	else
		-- cancel state early
		if (P_PlayerInPain(p))
			p.actionstate = STATE_NONE
			B.ApplyCooldown(p, BLESSING_COOLDOWN)
			return
		end
		
		-- set state and frame
		if (mo.state != S_AETHER_BLESSING) then mo.state = S_AETHER_BLESSING end
		local frame = 0
		
		local actiontime = p.actiontime - BLESSING_STARTUP
		if (actiontime >= BLESSING_TOTALTICS)	-- time to do the move!
			if (actiontime == BLESSING_TOTALTICS)
				-- find spells to reset
				local resetspells = {}
				for i = 0, SPL_MAX
					local spell = aether.spells[i]
					if not (spell.cooldown) then continue end
					
					-- add it to the list of spells to reset
					spell.cooldown = 0
					table.insert(resetspells, i)
				end
				
				-- reset all the spells that we found
				if (#resetspells)
					-- reset icons
					for i = 1, #resetspells
						local spell_id = resetspells[i]
						local spell_i = (#resetspells % 2) and i or i+1
						
						local angle = mo.angle + ANGLE_90
						local offset = (FRACUNIT*32 * (spell_i/2)) * ((i%2) and -1 or 1)
						local xpos = FixedMul(cos(angle), offset)
						local ypos = FixedMul(sin(angle), offset)
						
						local thok = P_SpawnMobjFromMobj(mo, xpos, ypos, p.height*9/8, MT_THOK)
						thok.aether = mo
						thok.sprite = SPR_APOW
						thok.frame = spell_id|FF_FULLBRIGHT
						thok.color = AETHERHUD[spell_id].color
						thok.fuse = thok.tics
						thok.aetheroffset = offset
						
						thok.scalespeed = thok.scale/12
						P_SetScale(thok, thok.scale/2)
						
						if mo.eflags & MFE_VERTICALFLIP
							thok.z = $ - thok.height + FixedMul(thok.info.height, mo.scale)
						end
					end
					
					-- reset sounds
					S_StartSound(mo, sfx_spcded, p)
				end
				
				-- visual burst
				AetherBlessBurst(mo, FRACUNIT*4/3)
				AetherBlessSparkles(mo, FRACUNIT*4/5, 24, FRACUNIT*24, 2)
				S_StartSound(mo, sfx_aeblsc)
				
				-- apply a buff to nearby allies!
				local range = FixedMul(mo.scale, FRACUNIT*256)
				searchBlockmap("objects", function(ref, found)
					if (ref == found) then return end	-- can't buff ourself
					if not ((found and found.valid) or found.health) then return end	-- gotta be alive
					
					-- player checks
					if not (found.player) then return end	-- only for players
					if not (SpellSympathy(ref.player, found.player)) then return end
					if (found.skin == SKIN) then return end	-- uhhh... nuh uh!
					
					-- LoS + dist check
					if not (P_CheckSight(ref, found)) then return end	-- no LoS
					local dist = FixedHypot(FixedHypot(found.x - ref.x, found.y - ref.y), found.z - ref.z)
					if not (dist <= range) then return end	-- too far away
					
					-- buff the player we found
					found.aether_blessed = true
					found.aether_blessed_check = nil	-- reset this, just incase
					AetherBlessBurst(found, FRACUNIT*10/9)
					AetherBlessSparkles(found, FRACUNIT*2/3, 16, FRACUNIT*18, 2)
				end, mo, mo.x - range, mo.x + range, mo.y - range, mo.y + range)
			end
			
			-- animate the move
			actiontime = $ - BLESSING_TOTALTICS
			local frame_subtract = (actiontime % (BLESSING_LOOPTICS * 2)) / BLESSING_LOOPTICS
			frame = #BLESSING_FRAMETICS - 1 - frame_subtract
		elseif (actiontime > 0)	-- animating
			local tics = 0
			for i = 0, #BLESSING_FRAMETICS - 1
				tics = $ + BLESSING_FRAMETICS[i + 1]
				if (tics >= actiontime)
					frame = i
					break
				end
			end
		end
		mo.frame = frame
		
		-- functionality
		local modifier = FRACUNIT*96/100
		
		p.pflags = $|(PF_FULLSTASIS)
		mo.flags = $|(MF_NOGRAVITY)
		p.canguard = false
		mo.aether_resetflags = true
		
		mo.momx = FixedMul(modifier, $)
		mo.momy = FixedMul(modifier, $)
		mo.momz = FixedMul(modifier, $)
		if (mo.momz * P_MobjFlip(mo) < 0)
			mo.momz = 0
		end
		
		if (P_IsObjectOnGround(mo)) then mo.z = $ + P_MobjFlip(mo) end	-- dont be grounded
		
		-- end state naturally
		p.actiontime = $ + 1
		if (p.actiontime >= BLESSING_DURATION)
			p.actionstate = STATE_NONE
			B.ApplyCooldown(p, BLESSING_COOLDOWN)
			
			-- reset state
			local thokked_reset = nil
			if not (P_IsObjectOnGround(mo)) then thokked_reset = false end
						
			p.actionstate = STATE_NONE
			B.ResetPlayerProperties(p, not P_IsObjectOnGround(mo), thokked_reset)
			if (P_IsObjectOnGround(mo)) 
				local state = S_PLAY_STND
				if (p.speed) then state = (p.speed > p.runspeed) and S_PLAY_RUN or S_PLAY_WALK end
				mo.state = state
			end
			
			return
		end
	end
	
	if (p.actionstate == STATE_NONE)
	and (mo.aether_resetflags)
		mo.flags = $ & ~(MF_NOGRAVITY)
		mo.aether_resetflags = nil
	end
end

-- thinkframe to handle buffed players!
addHook("ThinkFrame", do
	for p in players.iterate
		local mo = p.mo
		if not (mo and mo.valid) then continue end
		
		-- make sure they're blessed
		if not (mo.aether_blessed) then continue end
		
		-- instantly reset cooldown
		if not (p.actioncooldown)
			mo.aether_blessed_check = true
		else
			if (mo.aether_blessed_check)
				p.actioncooldown = 0
				mo.aether_blessed = nil
				mo.aether_blessed_check = nil
				AetherBlessSparkles(mo, FRACUNIT/2, 12, FRACUNIT*13, 2)
				S_StartSound(mo, sfx_spcded)
				continue
			end
		end
		
		-- sparkles while active
		local r = mo.radius >> FRACBITS
		local h = (mo.height >> FRACBITS) + 4
		local x = P_RandomRange(-r, r)*FRACUNIT
		local y = P_RandomRange(-r, r)*FRACUNIT
		local z = P_RandomRange(0, h)*FRACUNIT
		
		local spark = AetherParticle(mo, x, y, z, FRACUNIT*2/3)
		spark.color = SKINCOLOR_GOLDENROD
	end
end)

-- exhaust function
local halted_tics = TICRATE*5/3
local slowfall_tics = TICRATE*9/4
local flight_mintics, flight_maxtics = TICRATE*10/3, TICRATE*2	-- will scale to this depending on Aether.FLIGHT_THRUST

local function HandleExhaust(p)
	local aether = p.aether
	if not aether return end
	local mo = p.mo
	
	if (P_IsObjectOnGround(mo))
		-- reset exhaust on the ground
		p.exhaustmeter = FRACUNIT
	else
		-- get maxtime
		local maxtime
		if (aether.halted)
			maxtime = halted_tics
		elseif (aether.slowfall)
			maxtime = slowfall_tics
		elseif (aether.flight)
			maxtime = flight_mintics	-- start at the minimum exhaust
			
			local momz = mo.momz * P_MobjFlip(mo)
			if (momz > 0)
				-- scale it depending on how fast we're ascending
				local exhaust_multiplier = min(FRACUNIT, FixedDiv(Aether.FLIGHT_THRUST, FixedDiv(momz, mo.scale)))
				local exhaust_range = flight_maxtics - flight_mintics
				maxtime = flight_mintics + FixedMul(exhaust_multiplier, exhaust_range)
			end
		end
		
		-- we got a maxtime to work with
		if (maxtime != nil)
			local exhaust_remove = FRACUNIT / maxtime
			p.exhaustmeter = max(0, $ - exhaust_remove)
		end
		
		if p.exhaustmeter <= 0
			if aether.flight
				aether.flight = 0
				aether.wings = nil
				S_StopSoundByID(mo, sfx_flight)
			end
			aether.jumped = false
		end
	end
	
	return true
end

-- priority/collision function
local function HandlePriority(p)
	local aether = p.aether
	if not (aether) then return end
	
	-- laziness
	local B = CBW_Battle
	
	-- aura burst stuff
	if aether.spells
	and aether.spells[SPL_BARRIER].active
	and aether.barrierbursttics
		-- just for the hurt message
		B.SetPriority(p, 0, 0, nil, 0, 0, "Aura Burst")
		return
	end
	
	-- adjust pre-existing priority to remove attack
	if not (p.powers[pw_super] or p.powers[pw_invulnerability])	-- we want to keep it
		-- store her old attack priority
		aether.battle_atk = p.battle_atk
		aether.battle_satk = p.battle_satk
		
		-- now force them to 0
		p.battle_atk = 0
		p.battle_satk = 0
		p.battle_hurttxt = "frail attack"	-- might as well set this, too
	end
end

local collision_tics = TICRATE*2/3
local collision_height = FRACUNIT*6

local function HandleCollision(s, t, p, mo, atk, def, weight, hurt, pain, ground, angle, thrust, thrust2, collisiontype)
	if not (p[s] and mo[s]) then return end
	if not (p[t] and mo[t]) then return end
	
	-- invert hurt for easier checks
	if (s == 2 and abs(hurt) == 1)
		hurt = $ * -1
	end
	
	-- shell deflections should tumble mighty!
	if (p[s].battle_hurttxt == "frail attack")
	and (hurt >= 0)	-- just incase, maybe someone could get hit by 0 attack??
	and not (P_PlayerInPain(p[s]) or p[s].tumble)	-- okay lets not bully her now
		-- check how much attack aether *wouldve* had
		local aether_atk = 0
		if (p[s].aether)
			aether_atk = p[s].aether.battle_atk	-- use this, instead
			
			-- try get sweetspot attack
			if (p.battle_sfunc != nil)
				if (CBW_Battle.PriorityFunction[p.battle_sfunc](mo[s], mo[t]))
					aether_atk = aether.battle_satk
				end
			end
		end
		
		-- first check for conditions where we wouldnt of taken damage from the collision
		if (aether_atk)	-- aether tried to attack
		or (atk[t] and not p[s].powers[pw_flashing])	-- they attacked and we blocked it
			-- tumble the player
			local tumble_tics = collision_tics
			if (aether_atk > def[t])	-- reduce tumble if we wouldve won the collision!
				local divider = 1 + (aether_atk - def[t])
				tumble_tics = $ / divider
			end
			
			CBW_Battle.DoPlayerTumble(p[s], tumble_tics)
			P_SetObjectMomZ(mo[s], CBW_Battle.WaterFactor(mo[s], collision_height), true)
			return
		end
	end
end

local function LoadAether()
	//Projectile names
	mobjinfo[MT_AETHER_SWORD].name = "Sword Rain"
	mobjinfo[MT_GENESISSPAWNPOINT].name = "Genesis Beam"
	
	//Nerf Sword Rain
	local sword_spawntics = TICRATE*5/36
	MAXSPELLTIME[SPL_SWORDRAIN] = sword_spawntics * 5
	mobjinfo[MT_AETHER_SWORD].speed = $*15/16
	
	-- adjust cooldowns
	SPELLCOOLDOWNS[SPL_NONE] = TICRATE*5
	SPELLCOOLDOWNS[SPL_BARRIER] = TICRATE*8
	
	-- set costs
	SPELLCOSTS[SPL_NONE] = 4
	SPELLCOSTS[SPL_ARCANE] = 2
	SPELLCOSTS[SPL_SWORDRAIN] = 0
	SPELLCOSTS[SPL_BARRIER] = 12
	SPELLCOSTS[SPL_AURA] = 6
	SPELLCOSTS[SPL_GENESIS] = 10
		
	CBW_Battle.SkinVars[SKIN] = {
		special = HandleSpecial,
		func_priority_ext = HandlePriority,
		func_collide = HandleCollision,
		func_exhaust = HandleExhaust,
		weight = WEIGHT,
		shields = SHIELDS
	}
	
	print("\140Aether\128 is ready to spell chaos!")
end


addHook("ThinkFrame", do
	if not scriptloaded
	and CBW_Battle
		scriptloaded = true
		LoadAether()
	end
end)