/*
Aether the Gazelle
HUD management
	by Tatsuru
	& SMS Alfredo
	& Snu
*/

rawset(_G, "AETHERHUD", {})

for i = 0, SPL_MAX + 1
	AETHERHUD[i] = {
		cooldown = 0
	}
end

AETHERHUD[SPL_NONE].name = "Light Orb"
AETHERHUD[SPL_BARRIER].name = "Aura Burst"
AETHERHUD[SPL_ARCANE].name = "Arcane Rush"
AETHERHUD[SPL_SWORDRAIN].name = "Sword Rain"
AETHERHUD[SPL_AURA].name = "Speedwave"
AETHERHUD[SPL_GENESIS].name = "Genesis"

AETHERHUD[SPL_NONE].color = SKINCOLOR_GREY
AETHERHUD[SPL_BARRIER].color = SKINCOLOR_COBALT
AETHERHUD[SPL_ARCANE].color = SKINCOLOR_MAGENTA
AETHERHUD[SPL_SWORDRAIN].color = SKINCOLOR_GOLD
AETHERHUD[SPL_AURA].color = Silverhorn.EmeraldColor
AETHERHUD[SPL_GENESIS].color = SKINCOLOR_SUPERGOLD3

//Intialize the aetherhud_disabled variable
local aetherhud_disabled = Silverhorn.LocalData["setting"].aetherhud_disabled

//aetherhud command
COM_AddCommand("aetherhud", function(player, arg)
	if arg
		//Args are reversed from the user's perspective because internally it's
		//for disabling and NOT enabling
		//This is in case the variable gets intialized to nil somehow
		if arg == "0" or arg == "false" or arg == "off"
			aetherhud_disabled = true
		elseif arg == "1" or arg == "true" or arg == "on" or arg == "default"
			aetherhud_disabled = false
		else
			CONS_Printf(player, "aetherhud [on/off] - Toggles Aether's spell HUD on the right side of the screen.")
		end
		
		Silverhorn.SaveValue("setting", "aetherhud_disabled", aetherhud_disabled)
	else
		CONS_Printf(player, "aetherhud [on/off] - Toggles Aether's spell HUD on the right side of the screen.")
	end
end, COM_LOCAL)

-- Converts numbers into the cute continue countdown patches
local function DigitToPatch(v, n)
	local d = n / 10
	local i = n - d * 10

	local patches = d and {
		v.cachePatch("CONTNUM"..d),
		v.cachePatch("CONTNUM"..i)
	} or {
		v.cachePatch("CONTNUM"..i)
	}

	return patches
end

-- Draw the menu!
hud.add(function(v, p)
	if not p.aether or p.powers[pw_carry] == CR_NIGHTSMODE
	or p.powers[pw_carry] == CR_NIGHTSFALL
	or mapheaderinfo[gamemap].silverhorn_endingmap
	or srb2p
	return end

	local aether = p.aether
	local spell = aether.currentspell
	local menuspell = aether.selectspell
	local time = aether.menutime
	local iconscale = aether.iconscale

	if time == MENUTIME return end

	local maxdist = 80
	local dist = maxdist - (time * (maxdist/MENUTIME))
	local vertscale = FRACUNIT*2/3	-- reduces the vertical draw distance
	local basescale = 2*FRACUNIT/3
	
	local basewidth = 320
	local baseheight = 200

	-- Draw the menu background
	local strength = 16 - (time * (16/MENUTIME))
	-- v.fadeScreen(0xFF00, strength)	-- do NOT use normal transparency, what the fuck, this would lag the game...

	-- Draw the spell icons!
	for i = 0, SPL_MAX-1
		-- get the spell we're drawing
		local drawspell = i + 1
		if (i == SPL_MAX-1)
			drawspell = SPL_NONE
			
			-- genesis drawing
			local can_genesis = (p.powers[pw_super] or CBW_Battle)
			if (can_genesis and not P_IsObjectOnGround(p.realmo))
				drawspell = SPL_GENESIS
			end
		end
		
		local vertical = not (i%2) and true or false
		local cooldown = aether.spells[drawspell].cooldown

		local x = basewidth/2 * FRACUNIT 
		local y = baseheight/2 * FRACUNIT 
		if (i != SPL_MAX-1)
			x = $ + dist * cos(FixedAngle((i-1)*90*FRACUNIT))
			y = $ + FixedMul(vertical and vertscale or FRACUNIT, dist * sin(FixedAngle((i-1)*90*FRACUNIT)))
		end
		
		local flags = V_PERPLAYER
		local trans = V_20TRANS
		local scale = basescale

		-- Big if true
		if drawspell == menuspell
			scale = $ + (iconscale + (time*FRACUNIT/2)) / (splitscreen and 2 or 1)
			
			-- Keep the image in place and make it grow if the menu closes while it's selected!
			x = basewidth/2 * FRACUNIT 
			y = baseheight/2 * FRACUNIT 
			if (i != SPL_MAX-1)
				x = $ + maxdist * cos(FixedAngle((i-1)*90*FRACUNIT))
				y = $ + FixedMul(vertical and vertscale or FRACUNIT, maxdist * sin(FixedAngle((i-1)*90*FRACUNIT)))
			end
			
			if not cooldown
				trans = 0
			end
		end

		local colormap = v.getColormap(TC_DEFAULT, AETHERHUD[drawspell].color)
		if not CanUseSpellWithFlag(p, drawspell)
			colormap = v.getColormap(TC_DEFAULT, SKINCOLOR_BLACK)
		end
		
		//We max the scale so as to prevent a negative scale error
		v.drawScaled(x, y, max(scale - time*FRACUNIT/10, 1), v.getSpritePatch(SPR_APOW, drawspell), flags|trans, colormap)

		-- spell name:
		trans = (max(0, time - 1))<<FRACBITS
		v.drawString(x/FRACUNIT, (y/FRACUNIT)+20, AETHERHUD[drawspell].name, V_ALLOWLOWERCASE|flags|trans, "center")

		-- Draw numbers!
		if drawspell ~= menuspell	-- don't do it right after casting
			if cooldown	-- cooldown comes first
				local patches = DigitToPatch(v, cooldown/TICRATE + 1)
				local numscale = scale
				local offs = #patches > 1 and FixedMul(-7*FRACUNIT, numscale) or 0
				local ddoffs = #patches > 1 and FixedMul(3*FRACUNIT, numscale) or 0
				trans = (max(0, time - 1))<<FRACBITS

				for k, j in ipairs(patches)
					v.drawScaled(x + k*offs - ddoffs, y - (j.height/2)*numscale, numscale, j, flags|trans)
					offs = $ * -1
				end
			elseif SPELLCOSTS[drawspell]	-- otherwise... draw the cost!
			and not aether.spells[drawspell].active	-- but not while it's active
				local patches = DigitToPatch(v, SPELLCOSTS[drawspell])
				local numscale = scale
				local offs = #patches > 1 and FixedMul(-7*FRACUNIT, numscale) or 0
				local ddoffs = #patches > 1 and FixedMul(3*FRACUNIT, numscale) or 0
				trans = (max(0, time - 1))<<FRACBITS

				for k, j in ipairs(patches)
					v.drawScaled(x + k*offs - ddoffs, y - (j.height/2)*numscale, numscale, j, flags|trans, v.getColormap(TC_RAINBOW, SKINCOLOR_GOLDENROD))
					offs = $ * -1
				end
			end
		end
	end

	-- draw the current spell...
	/*if spell
		local colormap = v.getColormap(-1, AETHERHUD[spell].color)
		local trans = (max(0, time - 1)/2)<<FRACBITS

		v.drawScaled(160<<FRACBITS, 100<<FRACBITS, FRACUNIT - time*FRACUNIT/10, v.getSpritePatch(SPR_APOW, spell), trans, colormap)
	end*/

end, "game")

-- On-screen spell display
hud.add(function(v, p)
	if not p.aether or p.powers[pw_carry] == CR_NIGHTSMODE
	or p.powers[pw_carry] == CR_NIGHTSFALL or aetherhud_disabled or srb2p return end

	local aether = p.aether
	local iconscale = aether.iconscale
	local can_genesis = (p.powers[pw_super] or CBW_Battle)
	and (aether.currentspell == SPL_NONE or aether.currentspell == SPL_GENESIS)

	-- Step 1: compute the initial Y coordinate depending on the amount of spells active or on cooldown
	local spelly = 100
	local HEIGHT = 24

	for i = 0, SPL_MAX
		local spell = aether.spells[i]
		if spell.active
		or spell.cooldown
		or (can_genesis and (i == SPL_NONE or i == SPL_GENESIS))
			spelly = $ - HEIGHT/2
		end
	end
	
	for i = 0, SPL_MAX -- for all the spells!
		local spell = i
		local time = aether.menutime
		local cooldown = aether.spells[spell].cooldown

		if not time return end

		local color = v.getColormap(TC_DEFAULT, AETHERHUD[spell].color)
		if not aether.spells[spell].active
		and (not CanUseSpells(p, true) or not CanUseSpellWithFlag(p, i)
		or (can_genesis and i == SPL_NONE and not P_IsObjectOnGround(p.mo))
		or (can_genesis and i == SPL_GENESIS and P_IsObjectOnGround(p.mo)))
			color = v.getColormap(TC_DEFAULT, SKINCOLOR_BLACK)
		end
		
		local flags = V_SNAPTORIGHT|V_PERPLAYER
		local trans = (min(10, 10 - time))<<FRACBITS

		-- Active spell sparkle handling
		local sparkles = aether.HUDSPARKLES[i]

		if sparkles
			for _, i in ipairs(sparkles)
				local t = trans + (i.time - 10) * FRACUNIT
				local p = v.getSpritePatch("AETP", i.frame, 0, i.rot)
				local x = i.x * FRACUNIT
				local y = (spelly + i.y) * FRACUNIT

				v.drawScaled(x, y, 2*FRACUNIT/3, p, flags|trans, color)
			end
		end

		-- Draw current spell icon
		if sparkles and #sparkles or cooldown or i == aether.currentspell
		or (can_genesis and (i == SPL_NONE or i == SPL_GENESIS))
			local patch = v.getSpritePatch(SPR_APOW, i)
			local x = (320 - patch.width/3) * FRACUNIT
			local y = spelly * FRACUNIT
			local scale = FRACUNIT/2
			
			if i == aether.currentspell
			or (can_genesis and (i == SPL_NONE or i == SPL_GENESIS))
				scale = $+($-iconscale)*2
			end

			v.drawScaled(x, y, scale, patch, flags|trans, color)

			-- Draw cooldown counter
			if cooldown
				local patches = DigitToPatch(v, cooldown/TICRATE + 1)
				local numscale = FRACUNIT/2
				local offs = #patches > 1 and FixedMul(-7*FRACUNIT, numscale) or 0
				local ddoffs = #patches > 1 and FixedMul(3*FRACUNIT, numscale) or 0

				for k, j in ipairs(patches)
					v.drawScaled(x + k*offs - ddoffs, y - (j.height/2)*numscale, numscale, j, flags|trans)
					offs = $ * -1
				end
			elseif i == aether.currentspell	-- selected spell
			and SPELLCOSTS[aether.currentspell]	-- got a cost for it
			and not aether.spells[aether.currentspell].active	-- not active
				local patches = DigitToPatch(v, SPELLCOSTS[aether.currentspell])
				local numscale = FRACUNIT/2
				local offs = #patches > 1 and FixedMul(-7*FRACUNIT, numscale) or 0
				local ddoffs = #patches > 1 and FixedMul(3*FRACUNIT, numscale) or 0

				for k, j in ipairs(patches)
					v.drawScaled(x + k*offs - ddoffs, y - (j.height/2)*numscale, numscale, j, flags|trans, v.getColormap(TC_RAINBOW, SKINCOLOR_GOLDENROD))
					offs = $ * -1
				end
			end

			spelly = $+ HEIGHT
		end
	end
end, "game")