/*
Aether the Gazelle
Main thinker
	by Tatsuru
	& Lach
	& Lat' (?!?!)
	& SMS Alfredo
	& Snu
beta
*/

local SKIN = "aether"
local Silverhorn = Silverhorn // performance hax
local Aether = Silverhorn.Aether

rawset(_G, "MENUTIME", 6)	-- menu time
-- this variable alone changes everything related to how long the button must be held, all is accounted including hud

Aether.DASHMODE_TIME = 3*TICRATE	-- time before Aether's dashmode activates
Aether.FLIGHT_TIME = 3*TICRATE		-- How much Aether's flight lasts overall
Aether.FLIGHT_THRUST = 6*FRACUNIT	-- How much Aether ascends per tic when holding jump

local BUTTONS = {BT_JUMP, BT_SPIN, BT_WEAPONNEXT, BT_WEAPONPREV, BT_WEAPONMASK}

Aether.FLAGS = {}
Aether.FLAGS.FLEW = 1 << 0
Aether.FLAGS.SOL = 1 << 1

local SPELL_SOUNDS = {
	[SPL_NONE] = sfx_spsel,
	[SPL_ARCANE] = sfx_sparcr,
	[SPL_SWORDRAIN] = sfx_spswrn,
	[SPL_BARRIER] = sfx_spaubt,
	[SPL_AURA] = sfx_spdwv,
	[SPL_GENESIS] = sfx_spsel,
}

// undoes menu-stasis stuff
local function AetherDehalt(p)
	local mo = p.mo
	
	if p.playerstate == PST_LIVE
		if p.aether.haltedinfo
		and CanUseSpells(p, true)
			mo.momx = p.aether.haltedinfo.x + p.cmomx
			mo.momy = p.aether.haltedinfo.y + p.cmomy
			mo.momz = p.aether.haltedinfo.z
			
			if mo.state == S_PLAY_SPINDASH
				mo.state = p.aether.haltedinfo.state
				mo.sprite = p.aether.haltedinfo.sprite
				mo.sprite2 = p.aether.haltedinfo.sprite2
				mo.frame = p.aether.haltedinfo.frame
				mo.tics = p.aether.haltedinfo.tics
			end
			
			if (mo.momx or mo.momy)
			and p.panim == PA_IDLE
				mo.state = S_PLAY_WALK
			end
		end
	end
	mo.blendmode = AST_TRANSLUCENT
	
	p.aether.halted = false
	p.aether.haltedinfo = nil
end

-- Initializes Aether's variables.
local function AetherInit(p)
	local aether = {}

	aether.buttons = {}
	for i = 1, #BUTTONS
		aether.buttons[BUTTONS[i]] = 0
	end

	aether.timer = -1
	aether.flags = 0

	aether.lastposition = {p.mo.x, p.mo.y, p.mo.z}
	aether.dashmode = 0

	aether.currentspell = SPL_NONE
	aether.spells = {}

	for i = 0, SPL_MAX
		aether.spells[i] = {
			cooldown = 0,
			timer = 0,
			active = false,
			debt = 0,
		}
	end

	aether.menutime = 10
	aether.iconscale = FRACUNIT/2
	aether.cooldownanim = 0
	aether.swordrain_aim = p.mo.angle
	
	if p.prevaetherspell
		aether.currentspell = p.prevaetherspell
		aether.selectspell = p.prevaetherspell
	end

	aether.HUDSPARKLES = {}
	for i = 0, SPL_MAX
		aether.HUDSPARKLES[i] = {}
	end	-- for the 5 spells

	p.aether = aether
	if p.mrce
		p.mrce.hyperimages = true
		p.mrce.canhyper = true
	end

	if Silverhorn.IsSol(p)
		Aether.SetSol(p)
	end

	return aether
end

local function AetherReset(p, delete)
	local aether = p.aether
	
	if aether == nil
		return
	end
	
	// Abort all spells
	local spells = aether.spells
	
	for i = 0, SPL_MAX
		DeactivateSpell(p, i, ABORT_RESET)
		spells[i].cooldown = 0
		spells[i].debt = 0
	end
	
	// Unhalt to reset translucency & gravity
	if aether.halted
		AetherDehalt(p)
	end
	
	if delete
		p.aether = nil
		return
	end
	
	for i = 0, SPL_MAX
		aether.HUDSPARKLES[i] = {}
	end
end

rawset(_G, "AetherParticle", function(mo, xoffs, yoffs, zoffs, scale)
	xoffs = $ or 0
	yoffs = $ or 0
	zoffs = $ or 0

	local spark = P_SpawnMobjFromMobj(mo, xoffs, yoffs, zoffs, MT_THOK)
	
	local oldheight = spark.height
	spark.scale = scale or mo.scale
	
	if spark.eflags & MFE_VERTICALFLIP
		spark.z = $ - (spark.height - oldheight)
	end
	
	if mo.color == SKINCOLOR_AETHER
		spark.color = SKINCOLOR_PASTEL
	else
		spark.color = mo.color
	end
	spark.state = S_AETHER_PARTICLE

	return spark
end)

local function SpawnWingParticles(mo)
	local target = mo.target or mo
	local particle
	for i = 0, FixedDiv(mo.height, mo.scale), 6*FRACUNIT
		for j = 1, 4
			if P_RandomChance(FRACUNIT/3) continue end

			local x = 4*j*cos(mo.angle)
			local y = 4*j*sin(mo.angle)
			local colors = {SKINCOLOR_DUSK, SKINCOLOR_AZURE, SKINCOLOR_SKY, SKINCOLOR_BLUEBELL}
			if mo.frame & FF_FRAMEMASK > 0
				colors = {SKINCOLOR_ROSY, SKINCOLOR_VIOLET, SKINCOLOR_PLUM, SKINCOLOR_TAN}
			end
			particle = P_SpawnMobjFromMobj(mo, x, y, i, MT_THOK)

			particle.state = S_AETHER_PARTICLE2
			particle.color = colors[P_RandomRange(1, #colors)]
			particle.frame = P_RandomChance(FRACUNIT/2) and A or B
			particle.momx = target.momx + P_RandomRange(-2, 2) * FRACUNIT
			particle.momy = target.momy + P_RandomRange(-2, 2) * FRACUNIT
			particle.momz = target.momz + P_RandomRange(-2, 2) * FRACUNIT
		end
	end
	return particle
end

Aether.SpawnWings = function(player)
	local aether = player.aether
	if not aether then return end
	
	if aether.wings
		for i, wing in ipairs(aether.wings)
			if wing and wing.valid then return end
		end
	end
	
	local mo = player.mo
	aether.wings = {
		P_SpawnMobjFromMobj(mo, 0, 0, FixedDiv(mo.height/2, mo.scale), MT_AETHER_WING),
		P_SpawnMobjFromMobj(mo, 0, 0, FixedDiv(mo.height/2, mo.scale), MT_AETHER_WING)
	}

	for i, wing in ipairs(aether.wings)
		wing.target = mo
		wing.flipped = (i == 2) and true
		wing.flap = false
		wing.flapangle = 0

		wing.scale = FRACUNIT/24
		wing.destscale = mo.scale
		wing.scalespeed = FRACUNIT/12
	end
end

Aether.StartFlight = function(player)
	local aether = player.aether
	local mo = player.mo
	
	if not aether.flight
		if player.pflags & PF_JUMPED
			player.pflags = $ | PF_NOJUMPDAMAGE
		end
		mo.state = S_PLAY_FLOAT
		aether.flags = $ | Aether.FLAGS.FLEW
	end
	aether.flight = Aether.FLIGHT_TIME
end

rawset(_G, "AetherCooldownAnim", function(p, forcedspell)
	local aether = p.aether
	local mo = p.mo
	local currentspell = forcedspell or aether.currentspell
	
	S_StartSound(mo, sfx_spelcd, p)

	if aether.cooldownanim return end

	aether.cooldownanim = 3*TICRATE
	mo.state = S_PLAY_WAIT
	if not P_IsObjectOnGround(mo)
		p.panim = PA_FALL
	end
	
	local camangle = R_PointToAngle(mo.x, mo.y) + ANG1 * 140
	local drip = P_SpawnMobjFromMobj(mo, 14*cos(camangle), 14*sin(camangle), P_MobjFlip(mo) * (p.height), MT_THOK)
	drip.sprite = SPR_DRIP
	drip.frame = D
	drip.tics = 20
	drip.fuse = drip.tics
	drip.aether = mo
	S_StartSound(mo, sfx_adderr)
	
	local spark = AetherParticle(mo, 32*cos(mo.angle), 32*sin(mo.angle), p.height/2, mo.scale*2)	
	spark.momx = mo.momx
	spark.momy = mo.momy
	spark.momz = mo.momz
	
	if currentspell == SPL_GENESIS
		spark.color = AETHERHUD[SPL_GENESIS].color or $
		P_ColorTeamMissile(spark, p)
	else
		spark.color = AETHERHUD[currentspell].color or $
	end
end)

rawset(_G, "AetherToggleSpell", function(p, forcedspell)
    local aether = p.aether
    local spells = p.aether.spells
	local currentspell = forcedspell or aether.currentspell
	
	//Turn the light orb into Genesis or nothing midair
	/*if currentspell == SPL_NONE and p.mo and p.mo.valid
	and not P_IsObjectOnGround(p.mo)
		if p.powers[pw_super] and not CBW_Battle
			currentspell = SPL_GENESIS
		else
			return false
		end
	end*/
	
	//Spell on cooldown
	if spells[currentspell].cooldown
	or not CanUseSpellWithFlag(p, currentspell)
		AetherCooldownAnim(p, currentspell)
		return false
		
	//Spell already active, do its spin functionality
	elseif spells[currentspell].active
		if SpellSpin[currentspell]
			SpellSpin[currentspell](p)
		end
		
	//Activate spell
	else
		spells[currentspell].active = true
		
		-- handle ring cost + debt
		if SPELLCOSTS[currentspell]
			local remaining_rings = p.rings - SPELLCOSTS[currentspell]
			local sound = sfx_cdfm66
			
			if (remaining_rings < 0)
				spells[currentspell].debt = abs(remaining_rings)
				if (_G["sfx_noring"]) then sound = _G["sfx_noring"] end
			end
			
			p.rings = max(0, remaining_rings)
			S_StartSound(nil, sound, p)
		end
    end
	
	return true
end)

local function AetherHalt(p)
	local mo = p.mo
	
	if p.exhaustmeter == 0
		return
	end
	
	mo.blendmode = AST_MODULATE
	if mo.state != S_PLAY_SPINDASH
		mo.state = S_PLAY_SPINDASH
	end
	
	-- Make sure we actually get crushed if we get crushed
	if mo.ceilingz - mo.floorz < min(mo.height, P_GetPlayerSpinHeight(p))
		P_DamageMobj(mo, nil, nil, 1, DMG_CRUSHED)
	end
	
	-- Null vertical momentum if we touched the floor while halted
	if P_IsObjectOnGround(mo)
	or mo.eflags & MFE_JUSTHITFLOOR
		p.aether.haltedinfo.z = 0
	else
		-- Slow down vertically
		if abs(mo.momz) > 2*mo.scale
			mo.momz = ($ - $ / 16) - P_GetMobjGravity(mo)
		else
			mo.momz = 0
		end
		
		-- If we somehow get into the air...count that as a jump!
		if not (p.pflags & PF_JUMPED)
			p.pflags = $|P_GetJumpFlags(p)
		end
		
		if p.panim == PA_IDLE
			mo.state = S_PLAY_WALK
		end
	end
	
	local speed = FixedHypot(p.rmomx, p.rmomy)
	
	-- Slow down horizontally
	if speed > 2*mo.scale
		local dir = R_PointToAngle2(0, 0, p.rmomx, p.rmomy)
		
		mo.momx = $ - p.cmomx
		mo.momy = $ - p.cmomy
		
		P_InstaThrust(mo, dir, speed - speed/16)
		
		mo.momx = $ + p.cmomx
		mo.momy = $ + p.cmomy
	else
		mo.momx = p.cmomx
		mo.momy = p.cmomy
	end
	
	p.aether.halted = true
end

rawset(_G, "CanUseSpells", function(p, ignorerelease, ignorejump, ignoretrans)
	local aether = p.aether
	local mo = p.mo
	
	if not (mo and mo.valid)
	or p.exiting
	or p.actionstate
	or p.battlespawning
	or p.guard or p.airdodge
	or p.bot == 1
	//or (p.exhaustmeter != nil and p.exhaustmeter <= 0)
	or (aether.releasespin and not ignorerelease)
	or p.powers[pw_nocontrol]
		and not (aether.menutime < (ignorerelease and MENUTIME+1 or MENUTIME)) // I don't know what this variable MEANS
	or p.pflags & (PF_STASIS | PF_SLIDING)
	or P_PlayerInPain(p)
	or (mo.state >= S_PLAY_SUPER_TRANS1 and mo.state <= S_PLAY_SUPER_TRANS6 and not ignoretrans)
	or p.powers[pw_carry] or aether.prevcarry
	or not (P_IsObjectOnGround(p.mo) or aether.jumped or aether.halted or ignorejump)
	or (srb2p and not ignorerelease)
	or aether.spells[SPL_GENESIS].active
		return false
	end
	return true
end)

local function AetherWeaponChange(p)
	local didchange = false
	local aether = p.aether
	local weaponmask = p.cmd.buttons & BT_WEAPONMASK
	
	local maxspells = SPL_TRUEMAX
	local can_genesis = (p.powers[pw_super] or CBW_Battle)
	if (can_genesis and not P_IsObjectOnGround(p.realmo))
		maxspells = SPL_MAX
	end
	
	-- Don't bother with these controls in ringslinger or when you can't use spells
	if G_RingSlingerGametype()
	or not CanUseSpells(p)
		return didchange
	-- adjust SPL_NONE and SPL_GENESIS to switch between eachother
	elseif (aether.currentspell == SPL_NONE and can_genesis and not P_IsObjectOnGround(p.realmo))
	or (aether.currentspell == SPL_GENESIS and (not can_genesis or P_IsObjectOnGround(p.realmo)))
		local correctspell = (aether.currentspell == SPL_NONE) and SPL_GENESIS or SPL_NONE
		aether.currentspell = correctspell
		didchange = 1
	-- Select specific spells
	elseif aether.buttons[BT_WEAPONMASK] == 1
	and weaponmask <= maxspells+1
	and p.pwepbtn == nil //Support for KrabsServ emotes?
		local spell = weaponmask <= maxspells and weaponmask or 0
		aether.currentspell = spell
		didchange = true
	-- Advance one spell when pressing weapon next
	elseif aether.buttons[BT_WEAPONNEXT] == 1
		if aether.currentspell >= maxspells
			aether.currentspell = $ - SPL_TRUEMAX
		else
			aether.currentspell = aether.currentspell+1
		end
		didchange = true
	-- Go back one spell when pressing weapon prev
	elseif aether.buttons[BT_WEAPONPREV] == 1
		if aether.currentspell <= 0
			aether.currentspell = maxspells
		else
			aether.currentspell = aether.currentspell-1
		end
		didchange = true
	end
	
	if didchange
		aether.selectspell = aether.currentspell
		if didchange == true
			S_StartSound(nil, SPELL_SOUNDS[aether.selectspell], p)
			aether.iconscale = 0
		end
		
		aether.buttons[BT_SPIN] = 0
		aether.releasespin = true
	end
	return didchange
end

function Aether.SetSol(p)
	p.aether.flags = $ | Aether.FLAGS.SOL
	Aether.SpawnWings(p)
end

addHook("PlayerSpawn", function(p)
	if p.mo and p.mo.valid and p.mo.health
		if p.mo.skin == SKIN
			--reset schtuff when you spawn
			AetherReset(p, true)
			if p.mo.state == S_PLAY_STND
				p.mo.state = $ // call S_PLAY_STND again to get the animation working
			end
		end
		if (p.wasinaura or p.isinaura)
			p.wasinaura = false
			p.isinaura = false
			-- Reset stats
			p.jumpfactor = skins[p.skin].jumpfactor
			p.normalspeed = skins[p.skin].normalspeed
			p.fakenormalspeed = p.normalspeed
		end
	end
end)
addHook("PlayerThink", function(p)
	local mo = p.mo

	if not mo
	or mo.skin != SKIN
	or p.playerstate ~= PST_LIVE
	or p.iseggrobo
	or p.isjettysyn
		AetherReset(p, true)
		return
	end

	-- Set up Aether variables
	local aether = p.aether
	if not aether
		aether = AetherInit(p)
	end
	
	-- Keep track of your current spell to keep it while respawning
	p.prevaetherspell = aether.currentspell

	-- Aether's personal timer
	aether.timer = max($ + 1, 0)

	-- Check if Aether has jumped once already.
	if P_IsObjectOnGround(mo) or (mo.eflags & MFE_SPRUNG) or p.powers[pw_carry]
		aether.jumped = 0
	elseif (p.pflags & PF_JUMPED) or aether.halted
		if aether.jumped != 2 and (p.pflags & PF_JUMPDOWN)
			aether.jumped = 1
		else
			aether.jumped = 2
		end
	end
	
	-- Jump to fall transition state
	if (mo.state == S_PLAY_JUMP or mo.state == S_PLAY_SPRING)
	and not P_IsObjectOnGround(mo)
	and mo.momz*P_MobjFlip(mo) <= -2*P_GetMobjGravity(mo)
		mo.state = S_PLAY_MELEE_FINISH
		p.panim = PA_FALL
		mo.tics = 4
	end
	
	// spawn wings on Sol transformation
	if p.solchar and p.solchar.istransformed == 2
		Aether.SetSol(p)
	elseif aether.flags & Aether.FLAGS.SOL
	and not Silverhorn.IsSol(p)
		aether.flags = $ & ~Aether.FLAGS.SOL
		aether.wings = nil
	end

	-- If Aether is being carried, just hit the floor or is exiting,
	-- reset her flight.
	if (mo.eflags & MFE_JUSTHITFLOOR)
	or p.powers[pw_carry]
	or P_InQuicksand(mo)
	or p.exiting
		if aether.flags & Aether.FLAGS.FLEW
			aether.flags = $ & ~Aether.FLAGS.FLEW
			S_StopSoundByID(mo, sfx_flight)
			if Silverhorn.IsSol(p)
				Aether.SpawnWings(p)
			else
				aether.wings = nil
			end
		end
		aether.flight = 0
	end

	-- If Aether is flying, decrease her flight time.
	if aether.flight
		if not ((Aether.FLIGHT_TIME-aether.flight+39)%40)
			S_StartSound(mo, sfx_flight)
			if p.powers[pw_super]
				aether.flight = $ + 40
			end
		end

		if not aether.halted
			aether.flight = $ - 1
			if aether.flight == 0
				aether.wings = nil
				S_StopSoundByID(mo, sfx_flight)
			end
		end
	end
	
	-- Slow fall sound effect
	if aether.slowfall
		if aether.lastslowfall == aether.slowfall
			if mo.state == S_PLAY_SPINDASH
				mo.state = S_PLAY_MELEE_FINISH
				p.panim = PA_FALL
				mo.tics = 4
			end
			S_StopSoundByID(mo, sfx_sparkl)
			aether.slowfall = 0
		elseif not ((aether.slowfall+33)%34)
			S_StopSoundByID(mo, sfx_sparkl)
			S_StartSound(mo, sfx_sparkl)
		end
	end
	aether.lastslowfall = aether.slowfall

	-- If Aether is flying, in speedwave, or her speed is past her run speed,
	-- start increasing her dashmode counter.
	if (aether.flight or p.wasinaura
	or p.speed > FixedMul(p.runspeed, mo.scale))
	and not (aether.flags & Aether.FLAGS.FLEW)
	and not Silverhorn.IsSol(p)
	and not p.powers[pw_carry]
	and not srb2p
		if not (p.pflags & PF_THOKKED)
			if aether.dashmode == Aether.DASHMODE_TIME - 1
				S_StartSound(mo, sfx_aethdm)
			elseif (p.powers[pw_super] or p.powers[pw_sneakers])
			and aether.dashmode == Aether.DASHMODE_TIME/2
				aether.dashmode = Aether.DASHMODE_TIME
				S_StartSound(mo, sfx_aethdm)
			end
			if p.wasinaura and aether.dashmode != Aether.DASHMODE_TIME
				aether.dashmode = max($ + 1, Aether.DASHMODE_TIME - 1)
			else
				aether.dashmode = min($ + 1, Aether.DASHMODE_TIME)
			end
		end
	else
	-- If she isn't, decrease it.
		aether.dashmode = max($ - 2, 0)
	end

	-- If Aether's dashmode timer is past the threshold,
	-- do the sparkles and ghosts.
	if Silverhorn.IsSol(p) and aether.flight
	or aether.dashmode >= Aether.DASHMODE_TIME
		local r = mo.radius >> FRACBITS
		local h = (mo.height >> FRACBITS) + 4
		local x = P_RandomRange(-r, r)*FRACUNIT
		local y = P_RandomRange(-r, r)*FRACUNIT
		local z = P_RandomRange(0, h)*FRACUNIT

		AetherParticle(mo, x, y, z)

		if aether.timer % 2 and not aether.flight
			local ghost = P_SpawnGhostMobj(mo)
			ghost.fuse = 2
			P_MoveOrigin(ghost, unpack(aether.lastposition))
			if ghost.tracer
				ghost.tracer.fuse = ghost.fuse
				P_MoveOrigin(ghost.tracer, ghost.x, ghost.y, ghost.z)
			end

			-- P_ElementalFire but gold
			-- See Inazuma's code for this
			if P_IsObjectOnGround(mo) and p.powers[pw_super]
				Silverhorn.Inazuma.doColorSpinFire = mo
				P_ElementalFire(p)
				Silverhorn.Inazuma.doColorSpinFire = nil
			end
		end
	end

	local rate = (aether.swordrain_aim > p.drawangle) and ANG10 or -ANG10
	local func = (aether.swordrain_aim > p.drawangle) and min or max
	aether.swordrain_aim = func(p.drawangle, $ + rate)

	-- Spell tickers
	for i = 0, SPL_MAX
		local spell = aether.spells[i]
		spell.timer = spell.active and min($ + 1, MAXSPELLTIME[i]) or 0
		
		if not aether.halted // No cheesing the cooldown by waiting in the menu
			spell.cooldown = $ and $ - 1 or 0
		elseif spell.cooldown == nil
			spell.cooldown = 0
		end

		if spell.active
			p.canguard = false
			SpellThinker[i](p)
		end

		if spell.timer >= MAXSPELLTIME[i]
			DeactivateSpell(p, i, ABORT_MAXSPELLTIME)
		end
	end

	-- Active spell sparkle handler
	for i = 0, SPL_MAX	-- we only have 5 spells
		if aether.spells[i].active
			if (leveltime % 2)
				local sparkle = {
					x = 320 + P_RandomRange(1, 10),
					y = P_RandomRange(-10, 10),
					frame = 0,
					rot = 0,
					time = 0
				}

				table.insert(aether.HUDSPARKLES[i], sparkle)
			end
		end

		for k, sparkle in ipairs(aether.HUDSPARKLES[i])
			sparkle.x = $ - 4
			sparkle.frame = ($ + 1)%9
			sparkle.rot = $ + ANG1*30
			sparkle.time = $ + 1

			if sparkle.time >= 20
				table.remove(aether.HUDSPARKLES[i], k)
			end
		end
	end

	-- retain control whilst in spell menu
	if aether.menutime < MENUTIME
		p.powers[pw_nocontrol] = 3
	end
	
	-- Cooldown animation
	aether.cooldownanim = $ and $ - 1 or 0
	
	local minang, minid, mindiff
	local cangle = GetRelativeControlDirection(p)
	
	if cangle != nil
		//print(FixedInt(AngleFixed(cangle)))
		
		--Check for the smallest difference
		for i = 0, 3
			local ang = FixedAngle(i*90*FRACUNIT)
			local diff = min(360*FRACUNIT - abs(AngleFixed(ang - cangle)), abs(AngleFixed(ang - cangle)))
			
			if (minang == nil and diff < 40*FRACUNIT)
			or (minang != nil and diff < mindiff)
				minang = ang
				minid = i + 1
				mindiff = diff
			end
		end
	end
	
	-- Menu controls
	if aether.buttons[BT_SPIN] > 1 and aether.buttons[BT_JUMP] == 1
	and aether.menutime < MENUTIME-1
		aether.releasespin = true -- can't open menu til we let go off of spin again
	elseif minid
	and not aether.waitneutral	-- waiting for neutral input
	and aether.presstimes[minid] == 1
	and aether.buttons[BT_SPIN] > 1 -- held spin
	and not aether.releasespin -- fresh spin input
		if aether.currentspell != minid
			aether.currentspell = minid
			aether.selectspell = minid
			aether.iconscale = 0
			S_StartSound(nil, SPELL_SOUNDS[aether.currentspell], p)
		end
		AetherToggleSpell(p)
		aether.releasespin = true -- can't open menu til we let go off of spin again
	elseif not minid
	and not (p.cmd.buttons & BT_SPIN) and p.lastbuttons & BT_SPIN
	and not aether.releasespin -- fresh spin input
	and aether.menutime < MENUTIME-1
		-- neutral release, get the neutral spell!
		local neutralspell = SPL_NONE
		local can_genesis = (p.powers[pw_super] or CBW_Battle)
		if (can_genesis and not P_IsObjectOnGround(mo))
			neutralspell = SPL_GENESIS
		end
		
		if aether.currentspell != neutralspell
			aether.currentspell = neutralspell
			aether.selectspell = neutralspell
			aether.iconscale = 0
		end
		
		if (can_genesis or P_IsObjectOnGround(mo))
			AetherToggleSpell(p)
		end
		aether.releasespin = true -- can't open menu til we let go off of spin again
	end

	aether.iconscale = min(FRACUNIT/2, $ + FRACUNIT/10)

	-- tldr: bring up menu if we hold any direction
	-- otherwise, cast the last used spell

	-- Menu thinker
	
	if p.exhaustmeter == 0 and aether.halted
		AetherDehalt(p)
	end
	
	local weaponchange = AetherWeaponChange(p)
	
	--Cancel the menu if we aren't holding spin anymore
	if not aether.buttons[BT_SPIN]
	or weaponchange
	or not CanUseSpells(p)
	or (aether.halted and mo.eflags & MFE_SPRUNG)
		if srb2p then return end
		if aether.halted then AetherDehalt(p) end
		
		aether.menutime = min(MENUTIME, $ + 1)
		
		-- Ya can't use spells ya doofus!
		if not CanUseSpells(p) and not weaponchange
			if (mo.state < S_PLAY_SUPER_TRANS1
			or mo.state > S_PLAY_SUPER_TRANS6)
			and (aether.buttons[BT_SPIN] == 1
			or aether.buttons[BT_WEAPONMASK] == 1
			or aether.buttons[BT_WEAPONNEXT] == 1
			or aether.buttons[BT_WEAPONPREV] == 1)
			and p.powers[pw_carry] != CR_NIGHTSMODE
			and p.powers[pw_carry] != CR_NIGHTSFALL
				if not mapheaderinfo[gamemap].silverhorn_endingmap
					S_StartSound(mo, sfx_spelcd, p)
				end
			end
			
			aether.releasespin = true
		end
		
	-- Wait for neutral D-input for pre-menu inputs
	elseif p.cmd.buttons & BT_SPIN and aether.buttons[BT_SPIN] == 1
	and not aether.releasespin
		aether.waitneutral = 4
		
	-- make sure we aren't still holding spin, this has to be a fresh press to start opening the thing
	elseif aether.buttons[BT_SPIN] >= MENUTIME
	and not aether.releasespin
		aether.menutime = max(0, $ - 1)
		aether.selectspell = nil	-- remove this, the menu's being opened
		aether.waitneutral = 4 -- Wait for neutral D-input for menu inputs
		
		-- Save previous info
		aether.haltedinfo = $ or {
			x = p.rmomx,
			y = p.rmomy,
			z = p.mo.momz,
			state = p.mo.state,
			sprite = p.mo.sprite,
			sprite2 = p.mo.sprite2,
			frame = p.mo.frame,
			tics = p.mo.tics,
		}
		
		-- Play the menu opening sound 
		if aether.buttons[BT_SPIN] == MENUTIME
			S_StartSound(nil, sfx_spsel, p)
		end

		AetherHalt(p)
	-- Activate a spell!
	elseif not (p.cmd.buttons & BT_SPIN)
	and (aether.currentspell or P_IsObjectOnGround(mo))
		AetherDehalt(p)
		AetherToggleSpell(p)
	end
	
	-- Count how long buttons have been held for
	for i = 1, #BUTTONS
		aether.buttons[BUTTONS[i]] = (p.cmd.buttons & BUTTONS[i]) and $ + 1 or 0
	end
	
	-- check for neutral inputs, don't allow the selecting of spells until we do!
	if aether.waitneutral
		for i = 1, 4
			if aether.presstimes and aether.presstimes[i] == 0
				aether.waitneutral = $-1
			end
		end
	end

	-- release spin to open the menu again.
	if not (p.cmd.buttons & BT_SPIN)
	and aether.releasespin
		aether.releasespin = nil
		aether.buttons[BT_SPIN] = 0
	end
	
	-- Allow shield abilities and turning super with no spell selected
	if BT_SHIELD
		-- Do nothing
	elseif not aether.currentspell
	or not (P_IsObjectOnGround(mo) or p.pflags & PF_JUMPED)
		p.charflags = $ & ~SF_NOSHIELDABILITY | SF_SUPER
	elseif p.powers[pw_super]
		p.charflags = $ | SF_NOSHIELDABILITY | SF_SUPER
	else
		p.charflags = $ | SF_NOSHIELDABILITY & ~SF_SUPER
	end

	-- Update her last position.
	aether.lastposition = {mo.x, mo.y, mo.z}
	
	-- Keep track of your carry to prevent instant spell casting
	p.aether.prevcarry = p.powers[pw_carry]
end)

addHook("JumpSpecial", function(p)
	local mo = p.mo
	local aether = p.aether

	if not (mo and aether)
	or p.gotflag or p.gotcrystal
	or P_InQuicksand(mo)
	or not CanUseSpells(p, true)
	or p.charability != skins[SKIN].ability
	or not aether.jumped or aether.jumped == 1
	or p.exhaustmeter == 0
		return
	end

	p.pflags = $ | PF_THOKKED | PF_CANCARRY

	if aether.flight
		if FixedDiv(mo.momz, mo.scale*P_MobjFlip(mo)) < Aether.FLIGHT_THRUST
			P_SetObjectMomZ(mo, Aether.FLIGHT_THRUST, false)
		end
		Aether.SpawnWings(p)
		return true
	end

	if (aether.dashmode >= Aether.DASHMODE_TIME or Silverhorn.IsSol(p))
	and not (aether.flags & Aether.FLAGS.FLEW)
		-- Simulate flight
		if FixedDiv(mo.momz, mo.scale*P_MobjFlip(mo)) < Aether.FLIGHT_THRUST
			P_SetObjectMomZ(mo, Aether.FLIGHT_THRUST, false)
		end
		Aether.StartFlight(p)
	else
		if p.pflags & PF_JUMPED
			p.pflags = $ | PF_NOJUMPDAMAGE
		end
		
		-- Simulated slow fall
		aether.slowfall = $ and $+1 or 1
		if mo.state != S_AETHER_ARCANERUSH and mo.state != S_PLAY_SPINDASH
			mo.state = S_PLAY_SPINDASH
		end
		if P_MobjFlip(mo)*mo.momz < 4*P_GetMobjGravity(mo)*P_MobjFlip(mo)
			mo.momz = 4*P_GetMobjGravity(mo)
		end

		-- Mini sparkles!
		if (aether.timer % 2)
			local r = FixedInt(mo.radius * 2)
			local x = P_RandomRange(-r, r)*FRACUNIT
			local y = P_RandomRange(-r, r)*FRACUNIT

			local spark = AetherParticle(mo, x, y, -8*FRACUNIT, 2*mo.scale/3)
			spark.momx = mo.momx
			spark.momy = mo.momy
			spark.tics = P_RandomRange(9, 15)
			P_SetObjectMomZ(spark, 8*FRACUNIT, false)
		end
	end

	return true
end)

-- Wing magic!
addHook("MobjThinker", function(mo)
	local p = mo.target and mo.target.valid and mo.target.player
	local aether = p and p.aether
	
	if not aether
	or not (aether.flight or Silverhorn.IsSol(p) and not (aether.flags & Aether.FLAGS.FLEW))
	or mo.target.skin != SKIN
		local particle = SpawnWingParticles(mo)
		S_StartSound(mo.target or particle, sfx_aethwd)

		P_RemoveMobj(mo)
		return
	end
	
	if not aether.halted
		mo.blendmode = p.powers[pw_super] and AST_ADD or AST_TRANSLUCENT
		
		if not mo.flap
			mo.flapangle = min(120, max($*2, 1))

			if mo.flapangle == 120
				mo.flap = true
			end
		else
			mo.flapangle = max(0, min($/2, 1))

			if mo.flapangle == 0
			and aether.flight // don't flap unless flying
				mo.flap = false
			end
		end
	else
		mo.blendmode = AST_MODULATE
	end
	
	if p.powers[pw_super]
		mo.color = SKINCOLOR_SUPERTAN1 + abs( ( (p.powers[pw_super] >> 1) % 9) - 4)
		mo.frame = FF_PAPERSPRITE|FF_FULLBRIGHT|D + ((p.powers[pw_super] >> 1) % 4)
		mo.spritexscale = FRACUNIT/2
		mo.spriteyscale = FRACUNIT
	else
		mo.color = mo.target.color
		mo.frame = FF_PAPERSPRITE
		mo.spritexscale = FRACUNIT
		mo.spriteyscale = FRACUNIT
	end

	local angle

	if mo.flipped
		angle = p.drawangle + FixedAngle(200*FRACUNIT + mo.flapangle*FRACUNIT)
	else
		angle = p.drawangle + FixedAngle(160*FRACUNIT - mo.flapangle*FRACUNIT)
	end

	local x = mo.target.x + FixedMul(mo.radius, cos(angle))
	local y = mo.target.y + FixedMul(mo.radius, sin(angle))
	local z = mo.target.z + mo.target.height/2
	
	mo.eflags = $ & ~MFE_VERTICALFLIP | (mo.target.eflags & MFE_VERTICALFLIP)
	mo.destscale = mo.target.scale
	
	if mo.eflags & MFE_VERTICALFLIP
		z = $ - mo.height
	end

	P_MoveOrigin(mo, x, y, z)
	mo.angle = R_PointToAngle2(mo.target.x, mo.target.y, mo.x, mo.y)
end, MT_AETHER_WING)

-- Drip and Spell icon
addHook("MobjThinker", function(mo)
	if not (mo.aether and mo.aether.valid and mo.aether.player) return end
	
	if mo.sprite == SPR_APOW
		local xoffset = 0
		local yoffset = 0
		if (mo.aetheroffset != nil)
			local angle = mo.aether.angle + ANGLE_90
			xoffset = FixedMul(cos(angle), mo.aetheroffset)
			yoffset = FixedMul(sin(angle), mo.aetheroffset)
		end
		
		local camangle = R_PointToAngle(mo.aether.x, mo.aether.y) + ANG1 * 140
		P_MoveOrigin(mo, mo.aether.x + P_ReturnThrustX(nil, camangle, mo.aether.scale) + xoffset,
		mo.aether.y + P_ReturnThrustY(nil, camangle, mo.aether.scale) + yoffset,
		mo.aether.z + P_MobjFlip(mo.aether) * mo.aether.height*9/8)
		
		if mo.aether.eflags & MFE_VERTICALFLIP
			mo.z = $ - mo.height + FixedMul(mo.info.height, mo.aether.scale)
		end
	else
		local camangle = R_PointToAngle(mo.aether.x, mo.aether.y) + ANG1 * 140
		P_MoveOrigin(mo, mo.aether.x + P_ReturnThrustX(nil, camangle, 14*mo.aether.scale),
		mo.aether.y + P_ReturnThrustY(nil, camangle, 14*mo.aether.scale),
		mo.aether.z + P_MobjFlip(mo.aether) * (mo.aether.height + mo.scale/4*mo.fuse))
	end
end, MT_THOK)

//Thinkframe for the glow horn overlay
addHook("PostThinkFrame", do for player in players.iterate
	local horn = player.followmobj
	local mo = player.mo
	local aether = player.aether
	local hasglowhorns = aether
	and player.powers[pw_carry] != CR_NIGHTSMODE
	and player.powers[pw_carry] != CR_NIGHTSFALL
	
	if not (mo and mo.valid and mo.skin == SKIN and aether)
		continue
	end
	
	//Add PF_SLIDING to prevcarry because SRB2 is stupid
	if not aether.prevcarry and player.pflags & PF_SLIDING
		player.aether.prevcarry = -1
	end
	
	if srb2p or not (horn and horn.valid and horn.type == skins[SKIN].followitem)
		continue
	end
	
	//Super form halo
	local halo = horn.halo
	if hasglowhorns and player.powers[pw_super]
		if halo and halo.valid
			A_CapeChase(halo, 68*FRACUNIT + 1, 0)
			halo.color = SKINCOLOR_SUPERTAN1 + abs( ( (player.powers[pw_super] >> 1) % 9) - 4)
			halo.flags2 = ($ & ~MF2_DONTDRAW) | (mo.flags2 & MF2_DONTDRAW)
			
			if mo.state >= S_PLAY_SUPER_TRANS1 and mo.state < S_PLAY_SUPER_TRANS6
				local trans = (S_PLAY_SUPER_TRANS6 - mo.state)*3/2
				halo.frame = ($ & ~FF_TRANSMASK)|(trans<<FF_TRANSSHIFT)
			else
				halo.frame = $ & ~FF_TRANSMASK
			end
		else
			halo = P_SpawnMobjFromMobj(mo, 0,0, 68*FRACUNIT, MT_INAZUMA_SILVERHORN)
			horn.halo = halo
			halo.tracer = mo
			halo.flags2 = $|MF2_SPLAT|MF2_DONTDRAW
			halo.renderflags = RF_FLOORSPRITE|RF_NOSPLATBILLBOARD
			halo.state = S_AETHER_WING
			halo.frame = ($|FF_FULLBRIGHT) + 2
		end
	elseif halo and halo.valid
		P_RemoveMobj(halo)
	end
	
	if not (hasglowhorns)
		horn.state = S_INVISIBLE
		continue
	end
	
	local spells = aether.spells
	local currentspell = aether.currentspell
	if spells[SPL_GENESIS].active
		currentspell = SPL_GENESIS
	end
	
	if not horn.color
		horn.color = AETHERHUD[currentspell].color
		horn.reactiontime = currentspell
		horn.threshold = 225/2
		horn.spellwasactive = false
	end
	
	/*if horn.lastlook
		if horn.threshold < 225/2
			currentspell = horn.lastlook
		else
			horn.lastlook = 0
			horn.color = AETHERHUD[currentspell].color
			horn.reactiontime = currentspell
			horn.spellwasactive = spells[currentspell].active
		end
	end*/
	
	for i = 0, SPL_MAX
		if spells[i].cooldown == 1 and not aether.halted
			//horn.threshold = 135/2
			//horn.lastlook = i
			//currentspell = i
			
			local thok = P_SpawnMobjFromMobj(mo, 0,0, player.height*9/8, MT_THOK)
			thok.aether = mo
			thok.sprite = SPR_APOW
			thok.frame = i|FF_FULLBRIGHT
			thok.color = AETHERHUD[i].color
			thok.fuse = thok.tics
			
			thok.scalespeed = thok.scale/12
			P_SetScale(thok, thok.scale/2)
			
			if mo.eflags & MFE_VERTICALFLIP
				thok.z = $ - thok.height + FixedMul(thok.info.height, mo.scale)
			end
			
			S_StartSound(mo, sfx_spcded, player)
		end
	end
	
	//Switch spell and horn color!
	if horn.reactiontime != currentspell
	or (spells[currentspell].active and spells[currentspell].timer == 1)
	or spells[currentspell].cooldown == 1 or aether.cooldownanim == 3*TICRATE
	or (currentspell and horn.reactiontime == currentspell and horn.spellwasactive != spells[currentspell].active)
		horn.color = AETHERHUD[currentspell].color
		if currentspell == SPL_GENESIS then P_ColorTeamMissile(horn, player) end
		horn.threshold = 135/2
		
	//Pulsate fast during the spell
	elseif currentspell and spells[currentspell].active
		horn.threshold = $+8
		
	//Pulsate slightly faster for the intial pulse
	elseif horn.threshold < 225/2
		horn.threshold = $+2
		
	//Don't glow with the default spell!
	elseif horn.threshold >= 225/2 and not currentspell
		horn.threshold = 225/2
		
	//Normal pulsation
	else
		horn.threshold = $+1
	end
	horn.reactiontime = currentspell
	horn.spellwasactive = spells[currentspell].active
	
	local trans = min((sin(FixedAngle(horn.threshold*4*FRACUNIT))+FRACUNIT)/(FRACUNIT*2/11), 11)
	if spells[currentspell].cooldown
	or (not CanUseSpells(player) and not player.actionstate)
	or not CanUseSpellWithFlag(player, currentspell)
		trans = min($/2+6, 11)
	end
	
	if trans >= NUMTRANSMAPS
		horn.state = S_INVISIBLE
		continue
	end
	
	trans = $<<FF_TRANSSHIFT
	
	horn.tracer = mo
	horn.skin = mo.skin
	
	local FLAGS2 = (MF2_SPLAT or 0)|MF2_DONTDRAW|MF2_OBJECTFLIP
	local EFLAGS = MFE_VERTICALFLIP
	
	horn.flags2 = ($ & ~FLAGS2) | (mo.flags2 & FLAGS2)
	horn.eflags = ($ & ~EFLAGS) | (mo.eflags & EFLAGS)
	horn.renderflags = mo.renderflags
	horn.blendmode = mo.blendmode
	horn.rollangle = mo.rollangle
	horn.mirrored = mo.mirrored
	horn.spritexscale = mo.spritexscale
	horn.spriteyscale = mo.spriteyscale
	horn.spritexoffset = mo.spritexoffset
	horn.spriteyoffset = mo.spriteyoffset
	
	horn.state = S_INVISIBLE
	horn.sprite = mo.sprite
	horn.frame = mo.frame|trans|FF_FULLBRIGHT
	horn.sprite2 = mo.sprite2|FF_SPR2SUPER
	
	horn.angle = player.drawangle
	horn.scale = player.mo.scale
	horn.radius = player.mo.radius
	horn.height = player.mo.height
end end)

addHook("MobjRemoved", function(horn)
	if horn and horn.valid and horn.halo and horn.halo.valid
		P_RemoveMobj(horn.halo)
	end
end, MT_INAZUMA_SILVERHORN)

-- Button check shenanigans! Counterclockwise
addHook("PreThinkFrame", do
	for p in players.iterate
		if not p.aether continue end

		if not p.aether.presstimes
			p.aether.presstimes = {
				0,
				0,
				0,
				0,
			}
		end

		-- 30 is arbitrary but basically this is meant for controllers so that selecting spells isn't downright shitty
		p.aether.pressing = {
			p.cmd.forwardmove > 30 and p.cmd.forwardmove or 0, -- up
			p.cmd.sidemove > 30 and p.cmd.sidemove or 0, -- right
			p.cmd.forwardmove < -30 and p.cmd.forwardmove or 0, -- down
			p.cmd.sidemove < -30 and p.cmd.sidemove or 0, -- left
		}
		
		p.aether.cmd = {
			forwardmove = p.cmd.forwardmove,
			sidemove = p.cmd.sidemove
		}
		
		-- Used mostly for menu...
		for i = 1, 4
			if p.aether.pressing[i]
				p.aether.presstimes[i] = $+1
			else
				p.aether.presstimes[i] = 0
			end
		end
	end
end)