-- Persona support but for Aether

if not srb2p return end

-- basic linear interpolation. Percent must be between 0 and FRACUNIT
local function lerp(start, finish, percent)
	return start + FixedMul(finish-start, FRACUNIT-percent)
end

freeslot("SPR_KGYA", "SPR_ATSH")

local atr_spellslist = {
	"PSPELL_ARCANERUSH",
	"PSPELL_SWORDRAIN",
	"PSPELL_SPEEDWAVE",
	"PSPELL_AURABURST",
}

for i = 1, #atr_spellslist
	rawset(_G, atr_spellslist[i], i)
end

local spell_2_skill = {
	[PSPELL_ARCANERUSH] = "atr_arcanebuster",
	[PSPELL_SWORDRAIN] = "atr_shiningswords",
	[PSPELL_SPEEDWAVE] = "atr_speedwave",
	[PSPELL_AURABURST] = "atr_auraburst",
}

local function ATR_removeSpells(mo)
	--print("ATR: Remove...")
	local i = #mo.skills
	while i
		local atk = attackDefs[mo.skills[i]]
		if atk and atk.aetherspell
			--print("ATR: remove "..i)
			table.remove(mo.skills, i)
		else
			i = $-1
		end	
	end

	i = #mo.saveskills
	while i
		local atk = attackDefs[mo.saveskills[i]]
		if atk and atk.aetherspell
			--print("ATR: remove (s)"..i)
			table.remove(mo.saveskills, i)
		else
			i = $-1
		end	
	end	
end

local function ATR_giveSpell(mo, spell)	
	--print("ATR: get "..spell)
	mo.skills[#mo.skills+1] = spell
	mo.saveskills[#mo.saveskills+1] = spell
end

-- Let aether switch spells...
addHook("MobjThinker", function(mo)
	if not mo.battlen return end	-- how
	local btl = server.P_BattleStatus[mo.battlen]
	
	if btl and btl.running
		if btl.battlestate == BS_DOTURN
		and btl.turnorder[1] == mo	-- my turn!
		
			mo.t_atrhudtimer = $ or 0
			
			if mo.t_act ~= ACT_NONE
			and not server.P_BattleStatus[mo.battlen].hudtimer.out
				mo.t_atrhudtimer = min(7, $+1)
			else	
				if mo.t_atrhudtimer
					mo.t_atrhudtimer = max(0, $-1)
				end	
			end	
			
			if mo.t_act == ACT_NONE	-- selecting action				
				-- check if we have "atr_spells" in our skills list
				local spells
				for i = 1, #mo.passiveskills
					if mo.passiveskills[i] == "atr_spells"
						spells = true
						break
					end	
				end
				
				if spells
					mo.t_spellangle = $ or 0
					mo.t_spell = $ or 1
					
					local p = mo.control
					local pmo = p.mo
					
					if mo.t_spellangle > 0
						mo.t_spellangle = $-1
					
					elseif mo.t_spellangle < 0
						mo.t_spellangle = $+1
					
					end
					
					local nbutt = p.invertcommand and BT_CUSTOM1 or BT_CUSTOM2
					local pbutt = p.invertcommand and BT_CUSTOM2 or BT_CUSTOM1
					
					if pmo.P_inputs[nbutt] == 1
						mo.t_spell = $+1
						if mo.t_spell > #spell_2_skill
							mo.t_spell = 1
						end	
						
						mo.t_spellangle = -3
						
						--print("New spell: "..mo.t_spell)
						ATR_removeSpells(mo)
						ATR_giveSpell(mo, spell_2_skill[mo.t_spell])
						playSound(mo.battlen, sfx_turn, mo)
					
					elseif pmo.P_inputs[pbutt] == 1
						mo.t_spell = $-1
						if not mo.t_spell
							mo.t_spell = #spell_2_skill
						end	
						
						mo.t_spellangle = 3
						
						--print("New spell: "..mo.t_spell)
						ATR_removeSpells(mo)
						ATR_giveSpell(mo, spell_2_skill[mo.t_spell])
						playSound(mo.battlen, sfx_turn, mo)
						
					end
				end
			end	
		else	-- not your turn / different state
			
			local p = mo.control
			if p and p.valid
				if not p.cursormemory
					mo.t_spell = 1
				end
			end
			
		end
	end	
	
end, MT_PFIGHTER)

attackDefs["atr_arcanebuster"] = {
	name = "Arcane Buster",
	type = ATK_ALMIGHTY,
	target = TGT_ENEMY,
	costtype = CST_SPPERCENT,
	cost = 3,
	accuracy = 9999,
	desc = "Almighty dmg to one enemy.\nDestroys barriers.\nDmg scales with equipped weapon.",
	aetherspell = true,
	aetherspellmult = FRACUNIT*14/10,
	power = 20,
	anim = 	function(mo, target, hittargets, timer)
				
				local target = hittargets[1]
				
				if timer == 10
					playSound(mo.battlen, sfx_hama3)
					local m = P_SpawnMobj(target.x, target.y, target.z, MT_DUMMY)
					m.scale = 1
					m.destscale = FRACUNIT*2
					m.state = S_AETHER_ARCANEBOOSTER
					m.flags2 = $|MF2_SPLAT
					m.renderflags = $|RF_NOSPLATBILLBOARD|RF_SLOPESPLAT
					
					mo.splatshit = m
				
				elseif timer >= 20
				and timer < TICRATE +10
					local zpos = target.z + (timer-20)*FRACUNIT*3/2
					local m = P_SpawnMobj(target.x + P_RandomRange(-40, 40)*FRACUNIT, target.y + P_RandomRange(-40, 40)*FRACUNIT, zpos, MT_DUMMY)
					m.state = S_AETHER_PARTICLE				
					m.color = SKINCOLOR_VIOLET
				
				elseif timer >= TICRATE + 10
				and timer < TICRATE*3	
					
					if timer == TICRATE + 10
						playSound(mo.battlen, sfx_abxp)
						damageObject(target)
						
						playSound(mo.battlen, sfx_bufu6)
						localquake(mo.battlen, FRACUNIT*30, 5)						
					end
					
					for i = 1, 2
						local m = P_SpawnMobj(target.x + P_RandomRange(-40, 40)*FRACUNIT, target.y + P_RandomRange(-40, 40)*FRACUNIT, target.z, MT_DUMMY)
						m.state = S_AETHER_ARCANEPARTICLE
						m.momz = P_RandomRange(14, 40)*FRACUNIT
					end	
					
				elseif timer == TICRATE*3
					local g = P_SpawnGhostMobj(mo.splatshit)
					g.destscale = FRACUNIT*12
					g.flags2 = $|MF2_SPLAT
					g.renderflags = $|RF_NOSPLATBILLBOARD|RF_SLOPESPLAT					
					P_RemoveMobj(mo.splatshit)
					mo.splatshit = nil
					return true
				end	
			end,
}

attackDefs["atr_shiningswords"] = {
	name = "Shining Swords",
	type = ATK_BLESS,
	target = TGT_ALLENEMIES,
	costtype = CST_SPPERCENT,
	cost = 6,
	accuracy = 95,
	desc = "Bless dmg 8x to random enemies.\nDmg scales with equipped weapon\nand is reduced if hitting 1 target.",
	aetherspell = true,
	aetherspellmult = FRACUNIT/5,
	aetherswords = true,	-- used to know we must neuter this attack if #targets > 1
	power = 20*8,
	hits = 8,
	anim = 	function(mo, target, hittargets, timer)

				local avgx = 0
				local avgy = 0
				
				for i = 1, #hittargets
					local t = hittargets[i]

					avgx = $+ t.x/FRACUNIT
					avgy = $+ t.y/FRACUNIT
				end
				
				avgx = $ / #hittargets
				avgy = $ / #hittargets
				avgx = $*FRACUNIT
				avgy = $*FRACUNIT
				
							
				if timer == 10
					mo.swords = {}
					mo.critted = false
					
				elseif timer > 10
				and timer < TICRATE*3
					
					if timer
						local sword = P_SpawnMobj(avgx + P_RandomRange(-300, 300)*FRACUNIT, avgy + P_RandomRange(-300, 300)*FRACUNIT, mo.z + FRACUNIT*1024, MT_DUMMY)
						
						sword.sprite = SPR_SWRD
						sword.tics = -1
						sword.frame = F|FF_FULLBRIGHT|FF_ADD
						sword.renderflags = $ | RF_NOCOLORMAPS
						mo.swords[#mo.swords+1] = sword
						sword.momz = -FRACUNIT*120
						
						if timer%4 == 0
							playSound(mo.battlen, sfx_hamas1)
						end	
					end
				end
				
				if timer >= 25
				and timer <= 25 + 10*8
					if (timer-25) % 10 == 0
						
						-- pick a random target...
						local tgt = P_RandomRange(1, #target)
						local t = target[tgt]
						local downstate = t.down
						damageObject(t)
						
						playSound(mo.battlen, sfx_hit)
						localquake(mo.battlen, FRACUNIT*30, 5)

						local d = P_SpawnMobj(t.x + P_RandomRange(-96, 96)*FRACUNIT, t.y + P_RandomRange(-96, 96)*FRACUNIT, t.z+P_RandomRange(64, 160)*FRACUNIT, MT_THOK)
						d.sprite = SPR_THMP
						d.frame = A|FF_FULLBRIGHT
						d.scale = FRACUNIT/4 + P_RandomRange(-FRACUNIT/8, FRACUNIT/8)
						d.destscale = FRACUNIT*4
						d.scalespeed = FRACUNIT/8
						d.fuse = TICRATE-10
						d.tics = d.fuse
						d.momz = FRACUNIT*2
						d.momx = P_RandomRange(0, FRACUNIT-1)*P_RandomRange(-8, 8)
						d.momy = P_RandomRange(0, FRACUNIT-1)*P_RandomRange(-8, 8)
						d.color = SKINCOLOR_SUPERTAN3
						
						if t.damagestate == DMG_WEAK
						and t.down			-- make sure the target was knocked down
						and not downstate	-- the target wasn't down before.
							mo.critted = true
						end
					end
				end
				
				if mo.swords and #mo.swords
					
					-- spawn lazy lerp trail
					local i = #mo.swords
					while i
						local a = mo.swords[i]
						if not a or not a.valid
							i = $-1
							continue
						end	
						
						a.interp_prevx = $ or a.x
						a.interp_prevy = $ or a.y
						a.interp_prevz = $ or a.z
						
						for i = 0, 9
							local percent = FRACUNIT * (10-i) /10
							local x = lerp(a.interp_prevx, a.x, percent)
							local y = lerp(a.interp_prevy, a.y, percent)
							local z = lerp(a.interp_prevz, a.z, percent)
							local trail = P_SpawnMobjFromMobj(a, 0, 0, 0, MT_DUMMY)
							P_MoveOrigin(trail, x, y, z)
							--local offs = (mo.scale) / 8
							trail.sprite = SPR_NTHK
							trail.frame = A|FF_FULLBRIGHT
							trail.color = SKINCOLOR_SUPERTAN3
							trail.scale = a.scale/2 - ((10-i)*a.scale / 240)
							trail.destscale = 1
							trail.scalespeed = a.scale/24
							trail.target = a
						end	
						
						a.interp_prevx = a.x
						a.interp_prevy = a.y
						a.interp_prevz = a.z
						
						if a.z <= mo.z
							for i = 1, 16
								local sp = P_SpawnMobj(a.x, a.y, a.z, MT_SUPERSPARK)
								sp.momx = P_RandomRange(-8, 8)*FRACUNIT
								sp.momy = P_RandomRange(-8, 8)*FRACUNIT
								sp.momz = P_RandomRange(8, 18)*FRACUNIT
							end	
							P_RemoveMobj(a)
							table.remove(mo.swords, i)
						end
						
						i = $-1
						
					end
				end	
				
				if timer == TICRATE*5
					mo.swords = nil
					mo.atr_neuterdmg = nil
					
					-- do not trust damage calculations because we aren't garanteed
					-- to hit every enemy when using this move
					mo.setonemore = mo.critted
					
					return true
				end	
				
			end,
}

attackDefs["atr_speedwave"] = {
	name = "Speed Wave",
	type = ATK_SUPPORT,
	target = TGT_ALLALLIES,
	costtype = CST_SPPERCENT,
	cost = 4,
	aetherspell = true,
	aetherspellmult = 0,
	desc = "Raise party agility by 1 stage\nand heals HP based on equipped weapon.",
	power = 0,
	anim = 	function(mo, target, hittargets, timer)
				
				local NUMPARTS = 16
				if timer == 10
					-- spawn a speedwave
					-- since the radius will be dynamic we won't use the same code as aether's spells...
					
					mo.speedwave = {}
					mo.waveangle = 0
					mo.waveradius = 0
					
					playSound(mo.battlen, sfx_swon)
					
					for i = 1, NUMPARTS
						
						local part = P_SpawnMobj(mo.x, mo.y, mo.z, MT_DUMMY)
						part.state = S_AETHER_RUNES
						part.frame = P_RandomRange(0, 7) | FF_PAPERSPRITE
						part.blendmode = AST_ADD
						part.color = Silverhorn.EmeraldColor
						part.colorized = true
						mo.speedwave[#mo.speedwave+1] = part
					end
				end
				
				if timer >= 10 and mo.speedwave
					
					if timer <= TICRATE
						mo.waveradius = min(128, $ + 20)
					elseif timer > TICRATE
						mo.waveradius = $ + 10
						mo.waveangle = $ + ANG1 * (timer - TICRATE)/10
					end	
					
					mo.waveangle = $ + ANG1
					
					-- update sprite group:
					local ang = mo.waveangle
					for i = 1, #mo.speedwave
						local part = mo.speedwave[i]
						
						if not part or not part.valid
							continue
						end	
						
						local x = mo.x + mo.waveradius*cos(ang)
						local y = mo.y + mo.waveradius*sin(ang)
						local z = mo.z + FRACUNIT*24
						
						P_MoveOrigin(part, x, y, z)
						part.angle = ang + ANG1*90
						
						ang = $ + 360/NUMPARTS *ANG1
						
						if timer == TICRATE
							part.destscale = 8*FRACUNIT
							part.fuse = TICRATE*2
						end						
					end
				end
				
				local heal = 1
				if mo.weapon
					heal = mo.weapon.atk/10
				end
				
				for i = 1, #hittargets
					local t = timer - (i-1)*10
					
					if t == TICRATE*2
						damageObject(hittargets[i], -heal)
						buffStat(hittargets[i], "agi")
						
						BTL_logMessage(mo.battlen, "Agility up!")
						
					end
				end
				
				if timer == TICRATE*3 + TICRATE/2 + (#hittargets-1)*10
					return true
				end	
			end,
}

attackDefs["atr_auraburst"] = {
	name = "Aura Burst",
	type = ATK_ALMIGHTY,
	target = TGT_ALLENEMIES,
	costtype = CST_SPPERCENT,
	cost = 4,
	desc = "Almighty dmg to all enemies.\nRemoves barriers and stat buffs.\nRaise caster's defence by 2 stages.",
	aetherspell = true,
	aetherspellmult = FRACUNIT/5,
	power = 20,
	anim = 	function(mo, target, hittargets, timer)
				
				if timer == 10
					ANIM_set(mo, mo.anim_special1, false)
					playSound(mo.battlen, sfx_abchrg)
				
				elseif timer == 40
					ANIM_set(mo, mo.anim_special2, false)
					playSound(mo.battlen, sfx_bkpoof)
				elseif timer == 45
					
					for i = 1, 32
						local ang = 360/32*i*ANG1
						local a = P_SpawnMobj(mo.x, mo.y, mo.z, MT_SUPERSPARK)
						P_InstaThrust(a, ang, 48*FRACUNIT)
					end
					
					for i = 1, #hittargets
						local t = hittargets[i]
						
						damageObject(t)
						local buffs = {"atk", "mag", "def", "agi"}
						for j = 1, #buffs
							if t.buffs[buffs[j]][1] > 0
								t.buffs[buffs[j]][1] = 0
								t.buffs[buffs[j]][2] = 0
							end	
						end
						
					end
					
					BTL_logMessage(mo.battlen, "Stat bonuses reverted")
					
				
				elseif timer == 80
					-- buff def by 2 stages
					buffStat(mo, "def", BUFFSTEP*2)
					
					local s = P_SpawnMobj(mo.x, mo.y, mo.z, MT_DUMMY)
					s.state = S_AETHER_SHIELD1
					mo.ashield = s
					-- start the shield anim
				
				elseif timer == 100
					mo.ashield.state = S_AETHER_SHIELD3
					mo.ashield.health = 0	-- yeah trust me this works
					return true
				end
			end,
}

attackDefs["atr_spells"] = {
	name = "Spell Mastery",
	type = ATK_PASSIVE,
	desc = "Enable the user to cast\nSpells selected with C1 and C2.",
	passive = 0,
	
	implicit = "atr_spells.start",
}

attackDefs["atr_spells.start"] = {
	name = "Spell Mastery - Give Spell",
	type = ATK_PASSIVE,
	passive = PSV_STARTBATTLE,
	desc = "Gives the proper Spell Skills at the start of the battle",
	anim =	function(mo)
				mo.t_spell = $ or 1	-- just in case it bugs?
				mo.atr_neuterdmg = nil
				ATR_giveSpell(mo, spell_2_skill[mo.t_spell])
			end,
	
	implicit = "atr_spells.end",
}

-- NOTE: This makes the assumption that any weapon-scaled attack is a Spell
attackDefs["atr_spells.end"] = {
	name = "Spell Mastery - Remove Spell",
	type = ATK_PASSIVE,
	passive = PSV_ENDBATTLE,
	desc = "Removes the Spell Skills when battles end",
	anim =	function(mo)
				mo.atr_neuterdmg = nil
				ATR_removeSpells(mo)
			end,
			
	implicit = "atr_spells.boost",
}

attackDefs["atr_spells.boost"] = {
	name = "Spell Mastery - Boost",
	type = ATK_PASSIVE,
	passive = PSV_TRIGGERDAMAGE,
	desc = "Increases spell damage\nwith equipped weapon",
	
	anim =	function(mo, target, dmg, crit)
				local atk = mo.attack
				if atk.aetherspell
				and mo.weapon
					
					local watk = mo.weapon.atk*FRACUNIT/10
					
					-- characters always use starter weapons in challenge mode...
					-- so scale the damage with aether's *level* instead.
					if server.gamemode == GM_CHALLENGE
						watk = 35*FRACUNIT + FixedMul((mo.level-1)*FRACUNIT, 23*FRACUNIT/10)	
						--print("new atk: "..(watk/FRACUNIT))
					end
					
					local wpfac = FixedSqrt(watk)*4/10
					
					local spfac = atk.aetherspellmult or FRACUNIT
					--print(wpfac)
					--local wpfac = FixedDiv(mo.weapon.atk*FRACUNIT*4/10, 100*FRACUNIT)
					--print("DMG: "..dmg)
					dmg = FixedMul(spfac, FixedMul(dmg*FRACUNIT, wpfac))/FRACUNIT
					--print("NEW DMG: "..dmg)
					
					-- for shining swords specifically, reduce damage by 25% when attacking a singular target
					-- this is to avoid having a hassou tobi in the game...
					if atk.aetherswords
					and #mo.targets == 1
						dmg = $*75/100
					end
				end
				
				return dmg
			end,
	
}


attackDefs["atr_natk"] = {
	name = "Shine Spark",
	type = ATK_PIERCE,
	target = TGT_ENEMY,
	costtype = CST_SP,
	cost = 0,
	physical = true,
	
	anim = 	function(mo, targets, hittargets, timer)
				
				local target = hittargets[1]
				
				if timer == 10
					mo.spear_origins = {}
					playSound(mo.battlen, sfx_hama3)
				end
				
				if timer >= 10
				and timer <= 40
				and timer%4 == 0	
					
					local x = target.x + P_RandomRange(-192, 192)*FRACUNIT
					local y = target.y + P_RandomRange(-192, 192)*FRACUNIT
					local z = target.z + P_RandomRange(32, 255)*FRACUNIT
					
					local t = P_SpawnMobj(x, y, z, MT_DUMMY)
					t.scale = 1
					t.destscale = FRACUNIT
					t.tics = -1
					t.frame = A|FF_FULLBRIGHT
					t.color = SKINCOLOR_WHITE
					
					mo.spear_origins[#mo.spear_origins+1] = t
				end
				
				if timer == 70
					damageObject(target)
					
					for i = 1, #mo.spear_origins
						local o = mo.spear_origins[i]
						o.fuse = TICRATE/2
						o.destscale = 1
						
						for j = 1, 32
							local percent = FRACUNIT/32 *j
							
							local x = lerp(o.x, target.x, percent)
							local y = lerp(o.y, target.y, percent)
							local z = lerp(o.z, target.z, percent)
							
							local b = P_SpawnMobj(x, y, z, MT_DUMMY)
							b.scale = FRACUNIT/2
							b.destscale = 1
							b.tics = -1
							b.frame = A|FF_FULLBRIGHT
							b.fuse = TICRATE/2
							b.color = SKINCOLOR_WHITE
						end
					end
				end
				
				if timer == 90
					return true
				end
			end,
	
}


personaList["atr_kaguya"] = {
		-- GENERAL
		name = "Kaguya-Hime",
		arcana = ARCANA_MOON,

		-- STATS (at level 10. so we can have some precision, y'see)

		sprite = SPR_PSN0,
		anim_idle = 	{SPR_KGYA, A, B, C, B, 12},
		anim_atk = 		{SPR_KGYA, D, E, F, G, H, I, I, I, I, I, I, 2},
		startframe = 0,

		strength = 4,
		magic = 12,
		endurance = 5,
		luck = 9,
		agility = 8,
		weak = ATK_STRIKE|ATK_CURSE,
		resist = ATK_WIND,
		block = ATK_BLESS,
		skills = {
				
					{"atr_spells", 1}, {"kouha", 1}, {"dia", 4}, {"hama", 9},
					{"makakaja", 13}, {"counter", 18}, {"kouga", 23, "kouha"}, {"diarama", 28, "dia"},
					{"invigorate 1", 32}, {"hamaon", 36, "hama"}, {"mamakakaja", 41, "makakaja"}, {"counterstrike", 45, "counter"},
					{"kougaon", 51, "kouga"}, {"invigorate 2", 55, "invigorate 1"}, {"diarahan", 61, "diarama"}, {"high counter", 67, "counterstrike"},
					{"invigorate 3", 75, "invigorate 2"}
				},
}

charStats["aether"] = {

		-- GENERAL

		name = "Aether",			-- This is only for bots / enemies
		basehp = 45,			-- base HP for level 1
		basesp = 65,
		persona = "atr_kaguya",
		melee_natk = "atr_natk",	-- normal melee attack

		supercolor_start = SKINCOLOR_ROSEGOLD1,

		atk = 30,
		acc = 85,
		crit = 3,

		wep = "heels_01",
		weptype = WPT_HEELS,

		-- ANIMS
		-- Anims are built like this:
		-- {SPR_SPRITE, frame1, frame2, frame3, ... , duration between each frame}

		anim_stand = 		{SPR_PLAY, A, B, C, D, C, B, 8, "SPR2_STND"},		-- standing
		anim_stand_hurt =	{SPR_PLAY, A, B, C, D, C, B, 16, "SPR2_STND"},		-- standing (low HP)
		anim_stand_bored =	{SPR_PLAY, A, B, C, D, 8, "SPR2_WAIT"},
		anim_guard = 		{SPR_PLAY, A, 1, "SPR2_GARD"},		-- guarding
		anim_move =			{SPR_PLAY, A, B, C, D, 2, "SPR2_WALK"},		-- moving
		anim_run =			{SPR_PLAY, A, B, C, D, 2, "SPR2_RUN_"},	-- guess what
		anim_atk =			{SPR_PLAY, A, B, C, D, 1, "SPR2_ROLL"},	-- attacking
		anim_aoa_end =		{SPR_PLAY, D, C, B, A, 1, "SPR2_ROLL"},	-- jumping out of all-out attack
		anim_hurt =			{SPR_PLAY, A, 35, "SPR2_PAIN"},		-- taking damage
		anim_getdown =		{SPR_PLAY, A, 1, "SPR2_PAIN"},		-- knocked down from weakness / crit
		anim_downloop =		{SPR_PLAY, B, 1, "SPR2_BNCE"},		-- is down
		anim_getup =		{SPR_PLAY, A, B, C, D, 1, "SPR2_ROLL"},		-- gets up from down
		anim_death =		{SPR_PLAY, A, 30, "SPR2_PAIN"},		-- dies
		anim_revive =		{SPR_PLAY, A, B, C, D, 1, "SPR2_ROLL"},		-- gets revived
		anim_evoker =		{SPR_ATSH, A, 1},
		anim_evoker_shoot =	{SPR_ATSH, B, A, 4},
		
		anim_special1 =		{SPR_PLAY, B, 1, "SPR2_TRNS"},
		anim_special2 =		{SPR_PLAY, C, D, E, F, G, 4, "SPR2_TRNS"},
		
		vfx_summon = {"sfx_atrsu1", "sfx_atrsu2"},
        vfx_skill = {"sfx_atrsk1", "sfx_atrsk2", "sfx_atrsk3", "sfx_atrsk4"},
        vfx_item = {"sfx_atrit1", "sfx_atrit2"},
        vfx_hurt = {"sfx_atrhr1", "sfx_atrhr2", "sfx_atrhr3", "sfx_atrhr4", "sfx_atrhr5", "sfx_atrhr6", "sfx_atrhr7"},
        vfx_hurtx = {"sfx_atrhx1", "sfx_atrhx2", "sfx_atrhx3", "sfx_atrhx4", "sfx_atrhx5", "sfx_atrhx6", "sfx_atrhx7"},
        vfx_die = {"sfx_atrdi1", "sfx_atrdi2", "sfx_atrdi3", "sfx_atrdi4"},
        vfx_heal = {"sfx_atrhe1", "sfx_atrhe2", "sfx_atrhe3", "sfx_atrhe4", "sfx_atrhs3", "sfx_atrhs4", "sfx_atrhs5", "sfx_atrhs6"},
        vfx_healself = {"sfx_atrhs1", "sfx_atrhs2", "sfx_atrhs3", "sfx_atrhs4", "sfx_atrhs5", "sfx_atrhs6"},
        vfx_kill = {"sfx_atrkl1", "sfx_atrkl2", "sfx_atrkl3", "sfx_atrkl4", "sfx_atrkl5", "sfx_atrkl6"},
        vfx_1more = {"sfx_atr1m1", "sfx_atr1m2", "sfx_atr1m3", "sfx_atr1m4", "sfx_atr1m5"},
        vfx_crit = {"sfx_atrcr1", "sfx_atrcr2", "sfx_atrcr3"},
        vfx_aoaask = {"sfx_atrai1", "sfx_atrai2"},
        vfx_aoado = {"sfx_atrao1", "sfx_atrao2"},
        vfx_aoarelent = {"sfx_atrar1"},
        vfx_miss = {"sfx_atrmi1", "sfx_atrmi2", "sfx_atrmi3"},
        vfx_dodge = {"sfx_atrdg1", "sfx_atrdg2", "sfx_atrdg3", "sfx_atrdg4", "sfx_atrdg5", "sfx_atrdg6"},
        vfx_win = {"sfx_atrwi1", "sfx_atrwi2", "sfx_atrwi3", "sfx_atrwi4", "sfx_atrwi5"},
        vfx_levelup = {"sfx_atrlv1", "sfx_atrlv2"},

		icon = "ICO_ATR",		-- icon for net select and other menus
		hudbust = "ATR_H",			-- patch to draw on the stat hud
		hudbustlayer = "",		-- transparent patch to draw for status conditions
		hudaoa = "H_ATRAOA",	-- patch for all out attack hud
		hudcutin = "ATR_A",	-- cut in
		aoa_quote = "Face the Light\nof Justice!",

		//tip = "DISGUSTING BLACK CREATURES!",
		tip = "Specializes in Bless and Expel. She can also use unique Spells that get stronger depending on her equipped weapon's strength. Mind her low HP and her Spells' SP costs, and don't replace her Spell Mastery Skill by mistake!",

}

VFX_freeslotC("aether")

-- HUD shit
-- draw the main command selection for a player
local cmd_names = {
	"Arcane Rush",
	"Sword Rain",
	"Speedwave",
	"Aura Burst",
}

local function drawAetherCommands(v, mo)
	if mo.t_hidehud return end
	local battle = server.P_BattleStatus[mo.battlen]
	local timer = battle.hudtimer.start or battle.hudtimer.newturn
	-- center x y are of fixed position, don't bother:
	
	local tx, ty = -mo.t_atrhudtimer*24, -mo.t_atrhudtimer*24
	
	local cx, cy = 20<<FRACBITS + tx*FRACUNIT, 50<<FRACBITS + ty*FRACUNIT
	local s_ang = ANG1*90
	local b_ang = ANG1*65 + mo.t_spell*s_ang + ANG1*20*mo.t_spellangle-- base angle for drawing
	local radius = 48 + 32*timer
	
	-- render BG
	PDraw(v, 0 - 24*timer + tx, -10 - 24*timer + ty, v.cachePatch("H_ATRSB"), V_SNAPTOTOP|V_SNAPTOLEFT)

	local angle, x, y, scale = 0, 0, 0, 0

	for i = 1, 4

		angle = b_ang - (i)*s_ang - 16*timer*ANG1	-- start timer fancies
		local cosang = cos(angle)

		x = cx + radius*(cosang)*3/4 - abs(16*cosang)	-- adjust to ease shape on X coord
		y = cy - radius*(sin(angle)*5/6)	-- scale Y so that it looks more like an oval
		scale = max(FRACUNIT/6, FRACUNIT/4 + cos(angle-ANG1*65)/6 + timer*FRACUNIT/6)

		local pp = "H_ATRS"..i
		local index = i

		if x > -20*FRACUNIT and scale > 1
			--print(i..":"..x)
			v.drawScaled(x, y, scale, v.cachePatch(pp), V_SNAPTOTOP|V_SNAPTOLEFT)
		end
	end

	-- command name
	V_drawString(v, 40 + tx, 12 + ty, cmd_names[mo.t_spell]:upper(), "FPIMP", V_SNAPTOTOP|V_SNAPTOLEFT, nil, 0, 31)
end

-- add aether's hud
-- @TODO: move to the actual srb2p hud later
-- ^nevermind im retarded i forgot to rawset LOLOLOLOLOL
hud.add(function(v, p, c)
	if SYS_systemmenu.running return end	-- yeah no
	if not p.P_party return end
	
	local btl = server.P_BattleStatus[p.P_party]
	if btl.running
		local mo = btl.turnorder[1]
		if mo and mo.control and mo.control.P_party == p.P_party
		and btl.battlestate == BS_DOTURN	

			-- check if we have "atr_spells" in our skills list
			local spells
			for i = 1, #mo.passiveskills
				if mo.passiveskills[i] == "atr_spells"
					spells = true
					break
				end	
			end
		
			if spells and mo.t_spell
				drawAetherCommands(v, mo)
			end	
		end
	end
end)