/*
Aether the Gazelle
Spell thinkers
	by Tatsuru
	& Lat'
	& Frostiikin
	& Lach
	& SMS Alfredo
beta
*/

local Aether = Silverhorn.Aether

rawset(_G, "SPL_NONE", 0)
rawset(_G, "SPL_ARCANE", 1)
rawset(_G, "SPL_SWORDRAIN", 2)
rawset(_G, "SPL_BARRIER", 3)
rawset(_G, "SPL_AURA", 4)
rawset(_G, "SPL_TRUEMAX", 4) // Use instead of SPL_MAX to exclude Genesis

rawset(_G, "SPL_GENESIS", 5) // Battle-exclusive and Super-exclusive
rawset(_G, "SPL_MAX", SPL_GENESIS)

rawset(_G, "MAXSPELLTIME", {}) -- How much time each spell lasts
rawset(_G, "SPELLCOOLDOWNS", {}) -- The spell cooldowns
rawset(_G, "SPELLCOSTS", {})	-- Spell costs (for BattleMod)

MAXSPELLTIME[SPL_NONE] = 69	-- special handling in the spell thinker
MAXSPELLTIME[SPL_ARCANE] = states[S_AETHER_ARCANERUSH].tics
MAXSPELLTIME[SPL_SWORDRAIN] = TICRATE*3/2	-- 1.5
MAXSPELLTIME[SPL_BARRIER] = TICRATE*20
MAXSPELLTIME[SPL_AURA] = TICRATE*12
MAXSPELLTIME[SPL_GENESIS] = 50

SPELLCOOLDOWNS[SPL_NONE] = TICRATE/2
SPELLCOOLDOWNS[SPL_ARCANE] = TICRATE*5
SPELLCOOLDOWNS[SPL_SWORDRAIN] = TICRATE*2
SPELLCOOLDOWNS[SPL_BARRIER] = TICRATE*15
SPELLCOOLDOWNS[SPL_AURA] = TICRATE*10
SPELLCOOLDOWNS[SPL_GENESIS] = TICRATE*6

SPELLCOSTS[SPL_NONE] = 0
SPELLCOSTS[SPL_ARCANE] = 0
SPELLCOSTS[SPL_SWORDRAIN] = 0
SPELLCOSTS[SPL_BARRIER] = 0
SPELLCOSTS[SPL_AURA] = 0
SPELLCOSTS[SPL_GENESIS] = 0
-- ^ these will be set within battlemod

rawset(_G, "FLAGSPELLS", {}) //Which spells can we use when holding the flag?
FLAGSPELLS[SPL_NONE] = true
FLAGSPELLS[SPL_SWORDRAIN] = true

rawset(_G, "CanUseSpellWithFlag", function(p, spellnum) //Check the above
	if p.gotflag or p.gotcrystal
		return FLAGSPELLS[spellnum]
	else
		return true
	end
end)

rawset(_G, "AURA_RADIUS", 96)	-- speedwave radius

rawset(_G, "SpellThinker", {})
rawset(_G, "SpellAbort", {})
rawset(_G, "SpellSpin", {})

rawset(_G, "DeactivateSpell", function(p, spellnum, aborttype)
	local spell = p.aether.spells[spellnum]
	
	if not spell.active
		return
	end
	
	if p.powers[pw_super] and spellnum != SPL_GENESIS
		spell.cooldown = 0
	else
		spell.cooldown = SPELLCOOLDOWNS[spellnum]
		if p.gotflag or p.gotcrystal
			spell.cooldown = $*2
		end
		if aborttype == ABORT_TAGTEAM
			spell.cooldown = $ + MAXSPELLTIME[spellnum] - spell.timer // prevent cheesing speedwave/aura barrier cooldowns
		end
		
		-- apply debt
		if spell.debt
			local min_cooldown = TICRATE*2
			local cooldown_debt = (spell.cooldown * spell.debt) / 10
			spell.cooldown = max(min_cooldown, $ + cooldown_debt)
		end
	end
	
	spell.active = false
	spell.timer = 0
	spell.debt = 0	-- remove any lingering debt
	
	if SpellAbort[spellnum]
		SpellAbort[spellnum](p, aborttype)
	end
end)

local zonebuilderstoptrollingme = mobjinfo

// Spell abort types
rawset(_G, "ABORT_MAXSPELLTIME", 1) // spell timed out naturally
rawset(_G, "ABORT_RESET", 2) // the player died, switched skins, spectated, etc.
rawset(_G, "ABORT_TAGTEAM", 3) // player switched places with the bot player

// returns true if a spell from one player should benefit another
rawset(_G, "SpellSympathy", function(p1, p2)
	if gametyperules & GTR_FRIENDLY
		return true
	end
	
	if G_GametypeHasTeams()
	and p1.ctfteam == p2.ctfteam
		return true
	end
	
	if G_TagGametype()
	and (p1.pflags & PF_TAGIT) == (p2.pflags & PF_TAGIT)
		return true
	end
	
	return false
end)

--
-- Aura Burst
--

local AURABURST_RADIUS_SMALL = 512*FRACUNIT // base radius, used when the shield is spawned
local AURABURST_RADIUS_LARGE = 1536*FRACUNIT // armageddon shield radius, used when the shield is broken
local AURABURST_RADIUS_SOL = 1600*FRACUNIT // replaces others when sol

local function BustSpikesAndMonitorsAndFOFs(player, radius)
	local mo = player.mo
	local moz = mo.z + (mo.height >> 1)
	local sectordata = {
		{
			sector = mo.subsector.sector,
			coords = {mo.x, mo.y},
			dist = 0
		}
	}
	local function SectorSearch(mo, line)
		local x, y = P_ClosestPointOnLine(mo.x, mo.y, line)
		local v1 = line.v1
		local v2 = line.v2
		local minx = min(v1.x, v2.x)
		local maxx = max(v1.x, v2.x)
		local miny = min(v1.y, v2.y)
		local maxy = max(v1.y, v2.y)
		local dist
		
		x = min(max($, minx), maxx)
		y = min(max($, miny), maxy)
		dist = FixedHypot(mo.x - x, mo.y - y)
		
		if dist > radius
			return
		end
		
		for i, sector in ipairs({line.frontsector, line.backsector})
			local added = false
			for j, data in ipairs(sectordata)
				if sector == data.sector
					if dist < data.dist
						data.coords = {x = x, y = y}
					end
					added = true
					break
				end
			end
			if not added
				table.insert(sectordata, {
					sector = sector,
					coords = {x = x, y = y},
					dist = dist
				})
			end
		end
	end
	
	// find all sectors within the radius
	searchBlockmap("lines", SectorSearch, mo,
		mo.x - radius, mo.x + radius,
		mo.y - radius, mo.y + radius)
	
	// bust FOFs and monitors and spikes in those sectors
	for i, data in ipairs(sectordata)
		local sector = data.sector
		
		// fofs
		for fof in sector.ffloors()
			if (fof.flags & (FF_EXISTS|FF_BUSTUP)) ~= (FF_EXISTS|FF_BUSTUP)
				continue
			end
			
			local x = data.coords.x
			local y = data.coords.y
			local z
			local topz
			local bottomz
			
			if fof.t_slope
				topz = P_GetZAt(fof.t_slope, x, y)
			else
				topz = fof.topheight
			end
			
			if fof.b_slope
				bottomz = P_GetZAt(fof.b_slope, x, y)
			else
				bottomz = fof.bottomheight
			end
			
			z = min(max(moz, bottomz), topz)
			
			if FixedHypot(data.dist, z - moz) <= radius
				Silverhorn.MobjBustFOF(mo, fof)
			end
		end
		
		// spikes and monitors
		local hitlist = {}
		for thing in sector.thinglist()
			if not (thing.valid and thing.health > 0)
				continue
			end
			
			if not P_CheckSight(mo, thing)
				continue
			end
			
			local goodthing = thing.flags & MF_MONITOR
			or (thing.type == MT_SPIKE or thing.type == MT_WALLSPIKE)
				and thing.flags & MF_SOLID
			
			if not goodthing
				continue
			end
			
			local z = min(max(moz, thing.z), thing.z + thing.height)
			if FixedHypot(FixedHypot(thing.x - mo.x, thing.y - mo.y), z - moz) <= radius
				table.insert(hitlist, thing) // when spikes die, they spawn MORE SPIKES! so don't kill them here, otherwise thinglist gets fucked
			end
		end
		
		for i, thing in ipairs(hitlist)
			P_KillMobj(thing, mo, mo)
		end
	end
end

local auraburst_tics = TICRATE/2
local function burstcircle(mobj, mobjtype, ns, vertical, vertangle)
	local i = 0	while i < 20
		local fa = (i*(ANGLE_180/10))
		local mo = P_SpawnMobjFromMobj(mobj, 0, 0, 0, mobjtype) // your range was -5*FRACUNIT*FRACUNIT, which is a negligible -5 :V
		if mo and mo.valid
			mo.source = mobj
			mo.target = mobj
			mo.scale = mobj.scale
			mo.fuse = auraburst_tics
			mo.tics = max($, mo.fuse - 6)
			
			mo.reactiontime = mo.fuse/2
			local power = ns / mo.reactiontime
			
			if vertical
				mo.momz = FixedMul(sin(fa),power)
				P_InstaThrust(mo, vertangle+ANGLE_90,
				FixedMul(cos(fa),power)/2)
			else
				mo.momx = FixedMul(sin(fa),power)
				mo.momy = FixedMul(cos(fa),power)
			end
		end
		i = $+1
	end
end

local function AuraBurst(p)
	local a = p.aether
	local mo = p.mo
	local barrier = mo.aetherbarrier
	local radius
	local power = 0
	
	if barrier and barrier.valid and barrier.health > 0
		power = barrier.health
		P_KillMobj(barrier)
	end
	
	if Silverhorn.IsSol(p)
		radius = AURABURST_RADIUS_SOL
	elseif not power or CBW_Battle
		radius = AURABURST_RADIUS_SMALL
	else
		radius = AURABURST_RADIUS_LARGE
	end
	
	if power == 1
		radius = $ >> 1
	end
	radius = FixedMul($, mo.scale)
	
	a.barrierbursthit = {}	-- just so we dont hit the same thing twice
	a.barrierburstrange = radius
	a.barrierbursttics = auraburst_tics
	
	mo.frame = $ | FF_FRAMEMASK
	mo.state = S_PLAY_SUPER_TRANS1
	burstcircle(mo, MT_AETHER_BURSTPARTICLE, radius, false, 0)
	a.auraflash = 10
	S_StartSound(mo, sfx_abxp)
end

local function AuraBarrierHalt(p)
	local a = p.aether
	local mo = p.mo
	a.dashmode = 0
	a.flight = 0
	p.rmomx = 0
	p.rmomy = 0
	mo.state = S_PLAY_SUPER_TRANS2
	mo.tics = -1
end

addHook("MobjThinker", function(mo)
	if mo and mo.valid and mo.reactiontime and mo.fuse == mo.reactiontime
		mo.momx = -$
		mo.momy = -$
		mo.momz = -$
	end
end, MT_AETHER_BURSTPARTICLE)

SpellAbort[SPL_BARRIER] = function(p)
	local aether = p.aether
	
	aether.auraflash = 0
	aether.auracountdown = 0
end

SpellSpin[SPL_BARRIER] = function(p)
	local mo = p.mo
	local a = p.aether
	local barrier = p.mo.aetherbarrier
	
	if barrier and barrier.valid and barrier.health > 0
	and not a.auracountdown
		barrier.fuse = max($, 2) // barrier should not time out before the burst
		a.auracountdown = TICRATE
		a.spells[SPL_BARRIER].timer = min($, MAXSPELLTIME[SPL_BARRIER] - a.auracountdown - auraburst_tics) // arbitrary, but prevents the spell from running out before the barrier bursts
		S_StartSound(mo, sfx_abchrg)
		AuraBarrierHalt(p)
	end
end

local function GiveBarrier(mo, source)
	if not (mo and mo.valid) then return end
	
	if (source == nil) then source = mo end	-- failsafe?
	if not ((source and source.valid) and source.player) then return end
	if not (source.player.aether) then return end
	
	local shield = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_AETHER_BURSTSHIELD)
	shield.scale = FixedMul(5*source.scale/4, source.player.shieldscale)
	shield.target = mo
	if Silverhorn.IsSol(source.player)
		shield.fuse = 600*TICRATE
		shield.color = SKINCOLOR_BUBBLEGUM
		shield.colorized = true
	else
		local s = source.player.aether.spells[SPL_BARRIER]
		shield.fuse = MAXSPELLTIME[SPL_BARRIER] - s.timer
	end
	if (CBW_Battle)
		shield.fuse = $ - 5*TICRATE
	end
	S_StartSound(mo, shield.info.seesound)
	mo.aetherbarrier = shield
end

local function burst_blockmap(ref, found, range)
	if not (range) then return true end	-- no range given...
	
	if (ref == found) then return end	-- dont damage ourself!
	if not ((found and found.valid) or found.health) then return end	-- gotta be alive
	if not ((found.flags & (MF_SHOOTABLE|MF_ENEMY)) or found.player) then return end	-- not damageable
	if not (P_CheckSight(ref, found)) then return end	-- no LoS
	
	-- check if we hit them
	local hittable = ref.player.aether.barrierbursthit
	if (hittable[found]) then return end
	
	-- airdodge check
	if (found.player and found.player.airdodge) then return end
	
	-- damage nearby objects!
	if (gametyperules & (GTR_FRIENDLY|GTR_TEAMS))
	and (found.player)
	and not (found.player.playerstate or found.player.exiting)
	and not (found.aetherbarrier and found.aetherbarrier.valid)
	and SpellSympathy(ref.player, found.player)
		GiveBarrier(found, ref)
		hittable[found] = true	-- just so we don't hit them again
		return
	end
	
	local dist = FixedHypot(FixedHypot(found.x - ref.x, found.y - ref.y), found.z - ref.z)
	if not (dist <= range) then return end	-- too far away
	P_DamageMobj(found, ref, ref, 1, 0)
	hittable[found] = true	-- just so we don't hit them again
end

SpellThinker[SPL_BARRIER] = function(p)
	local a = p.aether
	local mo = p.mo
	local s = a.spells[SPL_BARRIER]
	local barrier = mo.aetherbarrier
	
	if a.auracountdown
		if mo.state ~= S_PLAY_SUPER_TRANS2
			a.auracountdown = 1
		elseif a.auracountdown == 1
			AuraBurst(p)
		end
		a.auracountdown = $ - 1
	end
	
	if s.timer == 1
		S_StartSound(mo, sfx_abchrg)
		AuraBarrierHalt(p)
		a.barrierbursttics = 0
		a.barrierburstrange = 0
	end
	
	if barrier and barrier.valid and barrier.health > 0
		barrier.fuse = max($, 2) // barrier should not time out before the burst
	end
	
	/*if s.timer < TICRATE*3/4	-- charge
		mo.momx = $*80/100 // ??? this doesn't seem like it does anything since she halts all momentum upon use of the spell anyway
		mo.momy = $*80/100
		mo.momz = $*80/100

		mo.sprite2 = SPR2_TRNS
		mo.frame = A*/
		
	local cancel_burst = false
	
	if mo.state < S_PLAY_SUPER_TRANS1 or mo.state > S_PLAY_SUPER_TRANS6
		cancel_burst = true
		S_StopSoundByID(mo, sfx_abchrg)
	elseif not CanUseSpells(p, true, true, true)
		mo.state = S_PLAY_WALK
		cancel_burst = true
		S_StopSoundByID(mo, sfx_abchrg)
	else
		p.panim = PA_DASH
		
		-- Make sure we actually get crushed if we get crushed
		if mo.ceilingz - mo.floorz < min(mo.height, P_GetPlayerSpinHeight(p))
		and P_DamageMobj(mo, nil, nil, 1, DMG_CRUSHED)
			cancel_burst = true
		end
	end
	
	if s.timer < TICRATE
		if cancel_burst
			DeactivateSpell(p, SPL_BARRIER)
			return
		end
	elseif s.timer == TICRATE		-- initial barrier burst
		AuraBurst(p)
		GiveBarrier(mo)	-- self barrier

		-- @TODO: visuals :noose:
		--not anymore, bitch. -frostii
	elseif s.timer > TICRATE
	and not (barrier and barrier.valid and barrier.health > 0)
	and (not a.barrierbursttics or cancel_burst)
		DeactivateSpell(p, SPL_BARRIER)
		return
	end
	
	if a.barrierbursttics
		local range_tics = (a.barrierbursttics * 180) / auraburst_tics
		local range_angle = (range_tics - 90) * ANG1
		local range = FixedMul(abs(cos(range_angle)), a.barrierburstrange)
		
		-- break bustables
		BustSpikesAndMonitorsAndFOFs(p, range)
		
		-- damage stuff around us! (or give players barriers!)
		searchBlockmap("objects", function(ref, found)
			return burst_blockmap(ref, found, range)
		end, mo, mo.x - range, mo.x + range, mo.y - range, mo.y + range)
			
		a.barrierbursttics = $-1
	end
	
	if a.auraflash
		a.auraflash = $ - 1
	end
end

addHook("MobjThinker", function(barrier)
	if not barrier.valid then return end
	
	local mo = barrier.target
	local player = mo and mo.valid and mo.player
	
	if not player
	or not mo.health
	or player.playerstate
	or player.exiting
	or mo.aetherbarrier != barrier
		P_RemoveMobj(barrier)
		return
	end
	
	local z
	local camangle
	
	barrier.flags2 = $ & ~MF2_OBJECTFLIP | (mo.flags2 & MF2_OBJECTFLIP)
	barrier.eflags = $ & ~MFE_VERTICALFLIP | (mo.eflags & MFE_VERTICALFLIP)
	barrier.scale = FixedMul(mo.scale*5/4, player.shieldscale)
	
	P_ColorTeamMissile(barrier, player)
	if (barrier.color)
		barrier.colorized = true
	end
	
	if mo.eflags & MFE_VERTICALFLIP
		z = mo.z + mo.height - barrier.height
	else
		z = mo.z
	end
	
	if splitscreen
		camangle = player.cmd.angleturn << 16
	else
		camangle = R_PointToAngle(mo.x, mo.y)
	end
	
	P_MoveOrigin(barrier,
		mo.x - P_ReturnThrustX(mo, camangle, mo.scale),
		mo.y - P_ReturnThrustY(mo, camangle, mo.scale),
		z)
end, MT_AETHER_BURSTSHIELD)

addHook("MobjFuse", function(barrier)
	if not (barrier.valid and barrier.health > 0) then return end
	
	P_KillMobj(barrier)
	return true
end, MT_AETHER_BURSTSHIELD)

addHook("MobjRemoved", function(barrier)
	if not barrier.valid then return end
	
	local mo = barrier.target
	local player = mo and mo.valid and mo.player
	
	if not player
		return
	end
	
	if player.aether
	and not player.aether.barrierbursttics
		DeactivateSpell(player, SPL_BARRIER)
	end
	mo.aetherbarrier = nil
	S_StartSound(mo, barrier.info.deathsound)
end, MT_AETHER_BURSTSHIELD)

hud.add(function(v, player, cam)
	local aether = player.aether
	if aether and aether.auraflash and not srb2p
		local af = aether.auraflash/2 // because it's supposed to be faint
		
		v.fadeScreen(148, af)
	end
end, "game")

addHook("ShouldDamage", function(mo, inf, src, damage, damagetype)
	local barrier = mo.aetherbarrier
	local player = mo.player
	
	if not (barrier and barrier.valid and player)
	or damagetype & DMG_DEATHMASK
		return
	end
	
	local player2 = src and src.player
	if (player2 == nil)
		player2 = inf and inf.player
	end
	if player2 and SpellSympathy(player, player2)
		return
	end
	
	if barrier.health > 0
	and not player.powers[pw_flashing]
		local state = barrier.info.painstate
		local flashtics = 0
		
		barrier.state = state
		barrier.health = $ - 1
		
		while states[state].tics > 0
			flashtics = $ + states[state].tics
			state = states[state].nextstate
		end
		player.powers[pw_flashing] = flashtics
		S_StartSound(mo, barrier.info.painsound)
		
		if CBW_Battle and not barrier.health
			local angle = mo.angle - ANGLE_180
			if inf and inf.valid
				angle = R_PointToAngle2(mo.x-mo.momx,mo.y-mo.momy,inf.x-inf.momx,inf.y-inf.momy)
			end
			
			local thrust = FRACUNIT*10
			if twodlevel then thrust = B.TwoDFactor($) end
			P_SetObjectMomZ(mo,thrust)
			
			CBW_Battle.DoPlayerTumble(player, 55, angle, mo.scale*3, true)
		end
		
		return false
	end
end, MT_PLAYER)

--
-- ARCANE RUSH

local ROCKET_STRENGTH = 24
local THOK_SPEED = 55
local ROCKET_SOL = 30
local THOK_SOL = 82
local ROCKET_TUMBLE = TICRATE*13/9
local THOK_TUMBLE = TICRATE*2/3

local arcaneboost = function(p, arcanetype, angle, sol, tumble)
	local mo = p.mo
	if arcanetype
		if (CBW_Battle)
		and (tumble)
			CBW_Battle.DoPlayerTumble(p, ROCKET_TUMBLE)
		end
		
		local strength = ROCKET_STRENGTH<<FRACBITS
		if CBW_Battle
			strength = $*2/3
		end
		if sol == true
		  strength = ROCKET_SOL<<FRACBITS
		end
		if mo.eflags & MFE_UNDERWATER
			strength = $*2/3
		end
		P_SetObjectMomZ(mo, strength)
		mo.z = $ + P_MobjFlip(mo) // puts player off the ground, which prevents them from jumping during the next tic
		//a.arcanedata = {mo.x, mo.y, mo.z, P_MobjFlip(mo)}
		S_StartSound(mo, sfx_arrshg)
	else
		if (CBW_Battle)
		and (tumble)
			CBW_Battle.DoPlayerTumble(p, THOK_TUMBLE)
		end
		
		-- otherwise, "thok"!
		if angle != nil
			mo.angle = angle
		end
		P_SetObjectMomZ(mo, 2<<FRACBITS)
		if sol == true
		  P_InstaThrust(mo, mo.angle, THOK_SOL*mo.scale)
		else
		  P_InstaThrust(mo, mo.angle, THOK_SPEED*mo.scale)
		end
		p.drawangle = mo.angle
		//a.arcanedata = nil	-- don't have anything on the floor obv
		S_StartSound(mo, sfx_arcrsh)
		-- boom
		/*for i = 1, 16
			local f = P_SpawnMobj(mo.x + P_RandomRange(-32, 32)*mo.scale, mo.y + P_RandomRange(-32, 32)*mo.scale, mo.z + P_RandomRange(0, 32)*mo.scale, MT_AETHER_ARCANEPARTICLE)
			f.scale = mo.scale
			f.momx = P_RandomRange(-32, 32)*mo.scale
			f.momy = P_RandomRange(-32, 32)*mo.scale
			f.momz = P_RandomRange(-32, 32)*mo.scale
		end*/
	end
	
	if not (tumble)
		if p.aether
			p.aether.jumped = 2
		end
		if not P_IsObjectOnGround(mo) and not p.powers[pw_carry]
			p.pflags = $ | P_GetJumpFlags(p)
		end
		p.arcanetype = arcanetype
		mo.state = S_AETHER_ARCANERUSH
		p.powers[pw_justsprung] = max($, mo.tics)
		p.panim = PA_ROLL
	end
end

SpellThinker[SPL_ARCANE] = function(p)
	local a = p.aether
	local mo = p.mo
	local s = a.spells[SPL_ARCANE]

	if s.timer == 1
	    if Silverhorn.IsSol(p)
			arcaneboost(p, P_IsObjectOnGround(mo), nil, true)
		else
			arcaneboost(p, P_IsObjectOnGround(mo), nil, false)
		end
		
		local booster = P_SpawnMobjFromMobj(mo, 0,0,0, MT_AETHER_ARCANEBOOSTER)
		booster.target = mo
		P_SetScale(booster, booster.scale/4)
		if Silverhorn.IsSol(p)
			booster.sol = true
			booster.color = SKINCOLOR_ICY
			booster.colorized = true
		end
		if p.arcanetype
			booster.flags2 = $|MF2_SPLAT
			booster.renderflags = $|RF_NOSPLATBILLBOARD|RF_SLOPESPLAT
			
			P_CreateFloorSpriteSlope(booster)
			booster.floorspriteslope.o = {
				x = booster.x,
				y = booster.y,
				z = booster.z + (booster.eflags & MFE_VERTICALFLIP and booster.height or 0)*4
			}
			
			//Gravity helper since having a large radius is stupid
			booster.tracer = P_SpawnMobjFromMobj(mo, 0,0,0, MT_GHOST)
			booster.tracer.state = S_INVISIBLE
			booster.tracer.tics = booster.tics
			booster.tracer.flags = $&~MF_NOGRAVITY&~MF_NOCLIP&~MF_NOCLIPHEIGHT|MF_NOCLIPTHING
		else
			booster.spriteyoffset = $ + 16*FRACUNIT
			booster.angle = mo.angle - ANGLE_90
			booster.renderflags = $|RF_PAPERSPRITE
			booster.flags = $|MF_PAPERCOLLISION
		end
	end
end

//Make the effect based on the state so other players can use it
addHook("PlayerThink", function(p)
	if not (p.mo and p.mo.valid and p.mo.state == S_AETHER_ARCANERUSH) return end
	local mo = p.mo
	local followmobj = p.followmobj
	if followmobj and followmobj.valid and followmobj.type == MT_TAILSOVERLAY
		p.followmobj.tics = 1
	end
	
	/*if a.arcanedata	-- fire on floor!
		local x = a.arcanedata[1]
		local y = a.arcanedata[2]
		local z = a.arcanedata[3]
		local fl = a.arcanedata[4]

		for i = 1, 6
			local f = P_SpawnMobj(x + P_RandomRange(-32, 32)*mo.scale, y + P_RandomRange(-32, 32)*mo.scale, z, MT_AETHER_ARCANEPARTICLE)
			f.scale = mo.scale*3
			f.momz = P_RandomRange(8, 30)*fl*mo.scale
		end
	end*/

	// orb particles
	for i = 1, 2
		local scale = FRACUNIT/2
		local info = zonebuilderstoptrollingme[MT_AETHER_ARCANEPARTICLE]
		local hangle = FixedAngle(P_RandomKey(360)*FRACUNIT)
		local zangle = FixedAngle((P_RandomKey(360) - 180)*FRACUNIT)
		
		local zspeed = FixedMul(sin(zangle), info.speed)
		local hspeed = FixedMul(cos(zangle), info.speed)
		local xspeed = P_ReturnThrustX(mo, hangle, hspeed)
		local yspeed = P_ReturnThrustY(mo, hangle, hspeed)
		
		local part = P_SpawnMobjFromMobj(mo, xspeed, yspeed,
			zspeed + p.height/2, MT_AETHER_ARCANEPARTICLE)
		part.target = mo
		part.aether_vectors = {
			pos = {
				x = part.x - mo.x,
				y = part.y - mo.y,
				z = part.z - mo.z  + (mo.eflags & MFE_VERTICALFLIP and (mo.height + (part.height >> 1)) or 0)
			},
			mom = {
				x = FixedMul(xspeed, mo.scale),
				y = FixedMul(yspeed, mo.scale),
				z = FixedMul(zspeed, mo.scale)
			}
		}
		part.scale = FixedMul($, scale)
		part.destscale = 0
		part.scalespeed = part.scale / part.tics
	end
	
	// streak particles
	local radius = 3*FixedDiv(mo.radius, mo.scale)
	local part = P_SpawnMobjFromMobj(mo,
		2*FixedMul(radius, P_RandomFixed()) - radius,
		2*FixedMul(radius, P_RandomFixed()) - radius,
		2*FixedMul(radius, P_RandomFixed()) - radius + p.height/2,
		MT_AETHER_ARCANEPARTICLE)
	
	part.angle = p.drawangle //pw_justsprung makes you strafe so there's no reason to check for momentum
	part.frame = $ & ~(FF_FRAMEMASK|FF_ANIMATE) | K
	
	if p.arcanetype
		local zangle = R_PointToAngle2(0, 0, FixedHypot(mo.momx, mo.momy), mo.momz)
		local slope
		
		part.frame = $ + 1
		part.flags2 = $ | MF2_AMBUSH
		if zangle == ANGLE_90 or zangle == ANGLE_270
			part.scalespeed = part.scale/12
			part.frame = $ | FF_PAPERSPRITE
			part.angle = $ - ANGLE_90
		else
			part.renderflags = $ | RF_FLOORSPRITE | RF_SLOPESPLAT | RF_NOSPLATBILLBOARD
			
			P_CreateFloorSpriteSlope(part)
			slope = part.floorspriteslope
			slope.o = {
				x = part.x,
				y = part.y,
				z = part.z + (part.eflags & MFE_VERTICALFLIP and part.height or 0)
			}
			slope.xydirection = part.angle
			slope.zangle = zangle
			part.spriteyscale = abs(cos(slope.zangle))
			part.scalespeed = part.spriteyscale/12
		end
	else
		part.scalespeed = part.scale/12
		part.frame = $ | FF_PAPERSPRITE
		P_Thrust(part, part.angle, FixedMul(part.scale, part.info.speed))
	end
end)

addHook("MobjThinker", function(part)
	if not part.valid
		return
	end
	
	local target = part.target
	local vectors = part.aether_vectors
	
	if vectors
		if not (part.target and part.target.valid)
			return
		end
	
		local mom = vectors.mom
		local pos = vectors.pos
		
		local vars = {"x", "y", "z"}
		
		for i, var in ipairs(vars)
			mom[var] = FixedMul($, part.info.painchance)
			pos[var] = $ + mom[var]
		end
		
		P_MoveOrigin(part,
			target.x + pos.x,
			target.y + pos.y,
			target.z + pos.z - (part.height >> 1))
	else
		if part.flags2 & MF2_AMBUSH
			part.spriteyscale = $ + part.scalespeed
		else
			part.spritexscale = $ + part.scalespeed
		end
	end
end, MT_AETHER_ARCANEPARTICLE)

addHook("MobjThinker", function(booster)
	if booster.tracer and booster.tracer.valid
		P_MoveOrigin(booster, booster.tracer.x, booster.tracer.y, booster.tracer.z)
	end
	
	if not (booster.tics % 3)
		local ghost = P_SpawnGhostMobj(booster)
		
		//ghost.fuse = min($, booster.tics)
		ghost.blendmode = AST_ADD
		ghost.scalespeed = ghost.scale/16
		ghost.destscale = ghost.scale*16
		ghost.spriteyoffset = booster.spriteyoffset
		ghost.flags2 = booster.flags2 & ~MF2_DONTDRAW
		ghost.renderflags = booster.renderflags
		
		if booster.flags & MF_PAPERCOLLISION
			P_InstaThrust(ghost, booster.angle + ANGLE_90, 8*booster.scale)
		else
			P_SetObjectMomZ(ghost, 8*FRACUNIT, false)
			
			if booster.eflags & MFE_VERTICALFLIP
				ghost.z = $ + booster.height
			end
		end
	end
	if booster.tics < TICRATE
		booster.flags2 = $ ^^ MF2_DONTDRAW
	end
end, MT_AETHER_ARCANEBOOSTER)

addHook("TouchSpecial", function(booster, mo)
	if booster and booster.valid and mo and mo.valid and mo.player and mo.health
	and booster.target and booster.target.valid
	and booster.target.player and not mo.player.playerstate
	and not mo.player.powers[pw_carry] and not mo.player.powers[pw_justsprung]
	and (not (booster.flags & MF_PAPERCOLLISION) or not mo.player.powers[pw_pushing])
	and (P_IsObjectOnGround(mo) or booster.flags & MF_PAPERCOLLISION)
		local friendly_booster = true
		if (CBW_Battle)
		and not (CBW_Battle.MyTeam(booster.target.player, mo.player))
			-- tumble them!
			friendly_booster = false
		end
		
	   if booster.color == SKINCOLOR_ICY
	     arcaneboost(mo.player, not (booster.flags & MF_PAPERCOLLISION), booster.angle + ANGLE_90, true, not friendly_booster)
	   else
		 arcaneboost(mo.player, not (booster.flags & MF_PAPERCOLLISION), booster.angle + ANGLE_90, false, not friendly_booster)
	   end	
	end
	return true
end, MT_AETHER_ARCANEBOOSTER)

--
-- GENESIS
--	@TODO: MOVE THIS TO BATTLE SOMEDAY...
// hello it has been moved to Battle
// it has been left as untouched as possible because I'm lazy
-- now this is used in battle AND while super!

SpellThinker[SPL_GENESIS] = function(p)
	local aether = p.aether
	local mo = p.mo
	local spell = aether.spells[SPL_GENESIS]

	local range = 960*mo.scale	-- take scale into account so aether can become small... or very big...!!!
	
	//Start Genesis
	if spell.timer == 1
		P_ResetPlayer(p)
		aether.flags = $ & ~Aether.FLAGS.FLEW
		aether.flight = 0
		mo.state = S_PLAY_NIGHTS_TRANS1
		p.panim = PA_DASH
	end
	
	//Player state during Genesis
	if spell.timer >= MAXSPELLTIME[SPL_GENESIS]
		mo.state = S_PLAY_FALL
	elseif mo.state < S_PLAY_NIGHTS_TRANS1 or mo.state > S_PLAY_NIGHTS_TRANS5
		DeactivateSpell(p, SPL_GENESIS) //Incorrect state! Abort!
		return
	else
		p.panim = PA_DASH
		if mo.state == S_PLAY_NIGHTS_TRANS5
			mo.frame = 6
			mo.tics = -1
		end
	end
	
	//Momentum and flags during Genesis
	p.pflags = $|PF_JUMPSTASIS&~PF_JUMPED&~PF_SHIELDABILITY&~PF_SPINNING
	if P_IsObjectOnGround(mo) 
		p.pflags = $|PF_STASIS
	else
		mo.momz = 0
		mo.momx = FixedMul($, FRACUNIT*95/100)
		mo.momy = FixedMul($, FRACUNIT*95/100)
	end

	//Genesis spawner
	if (leveltime % 3 == 0) and spell.timer < 12
		local target
		local potentialtargets = {}	-- Lat: Let's build a table of objects so we can determine which is the closest to us.
		-- this makes the ability feel more deliberate as it will wipe out everything around you

		searchBlockmap("objects", function(ref, found)
			if found.eth_genesis and found.eth_genesis.valid
				return
			end
			-- Lat: Do NOT use tracer, this can break so many stupid things because of how retarded srb2 is,
			-- many objects will have an undenified behaviour if their tracer is to become invalid, as well
			-- (this notably fixes CEZ3 weirdness)
			
			local p2 = found.player
			
			if (p2
			and not p2.spectator
			and not SpellSympathy(p, p2))
			// TODO: any other Battle conditions to consider?
			or ((found.flags & (MF_ENEMY|MF_BOSS))
			and (found.flags & MF_SHOOTABLE))
			--and abs(found.z - ref.z) <= 256*FRACUNIT
			--and P_AproxDistance(found.x - ref.x, found.y - ref.y) <= 960*FRACUNIT redundant, the blockmap already does this
				table.insert(potentialtargets, found)
				--return true
			end
		end, mo, mo.x - range, mo.x + range, mo.y - range, mo.y + range)

		-- sort found targets by distance, take the first one
		table.sort(potentialtargets, function(a, b)
			return R_PointToDist2(a.x, a.y, mo.x, mo.y) < R_PointToDist2(b.x, b.y, mo.x, mo.y)
		end)

		target = potentialtargets[1]	-- first element will be the closest

		local origin = target or mo
		local x, y = 0, 0

		if not target
			x = P_RandomRange(1, 64)*9*FRACUNIT * (P_RandomChance(FRACUNIT/2) and -1 or 1)
			y = P_RandomRange(1, 64)*9*FRACUNIT * (P_RandomChance(FRACUNIT/2) and -1 or 1)
		end

		if not R_PointInSubsectorOrNil(origin.x + FixedMul(x, mo.scale), origin.y + FixedMul(y, mo.scale)) return end

		local genesis = P_SpawnMobjFromMobj(origin, x, y, 192*FRACUNIT, MT_GENESISSPAWNPOINT)

		genesis.target = mo
		genesis.scale = mo.scale/64
		genesis.destscale = 3*mo.scale/2
		genesis.scalespeed = mo.scale/16
		genesis.threshold = spell.timer
		
		P_ColorTeamMissile(genesis, p)
		if genesis.color then genesis.colorized = true end

		if target
			target.eth_genesis = genesis
			genesis.found = target
		end
	end
end

-- Spawn the mess
addHook("MobjThinker", function(mo)
	if not mo.bands
		-- Lower band
		local t1 = CreatePaperspriteCircle(mo, 15, 57, MT_THOK, S_AETHER_RUNEBAND, 0, 1)
		-- Upper band
		local t2 = CreatePaperspriteCircle(mo, 12, 45, MT_THOK, S_AETHER_RUNEBAND, 16, 2)
		-- Orbs
		local t3 = {
			{mobj = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK), state = S_AETHER_ORB, offset = {x = 0, y = 0, z = 24}, angle = 0, rotation = 0, id = 3},
			{mobj = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK), state = S_AETHER_ORB, offset = {x = 0, y = 0, z = 24}, angle = 0, rotation = 0, id = 4},
			{mobj = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK), state = S_AETHER_ORB, offset = {x = 0, y = 0, z = 24}, angle = 0, rotation = 0, id = 5}
		}
		-- Lower rune circle
		local t4 = CreatePaperspriteCircle(mo, 15, 127, MT_THOK, S_AETHER_RUNES, 0, 6)
		-- Upper rune circle
		local t5 = CreatePaperspriteCircle(mo, 12, 99, MT_THOK, S_AETHER_RUNES, 16, 7)

		for _, t in ipairs({t4, t5})
			for _, m in ipairs(t)
				m.frame = P_RandomKey(7)
			end
		end

		t1 = MergeTables($, t2)
		t1 = MergeTables($, t3)
		t1 = MergeTables($, t4)
		t1 = MergeTables($, t5)

		mo.bands = t1
		P_BuildSpriteGroup(mo, mo.bands)

		S_StartSound(mo, sfx_gencha)
	end

	mo.threshold = $ + 1
	local timer = mo.threshold

	P_UpdateSpriteGroup(mo, function(seg)
		if mo.fuse and not seg.mobj.fuse
			seg.mobj.fuse = mo.fuse
		end

		if seg.frame
			seg.mobj.frame = seg.frame|FF_PAPERSPRITE
		end
		seg.mobj.frame = $|FF_FULLBRIGHT

		if seg.id == 2
			seg.mobj.scale = 3*mo.scale/2
		elseif seg.id == 3
			seg.mobj.scale = 6*mo.scale/3
			seg.mobj.color = SKINCOLOR_SUPERGOLD3
			seg.mobj.frame = $ | FF_TRANS50
		elseif seg.id == 4
			seg.mobj.scale = 5*mo.scale/3
			seg.mobj.color = SKINCOLOR_SUPERGOLD2
			seg.mobj.frame = $ | FF_TRANS30
			seg.mobj.shadowscale = FRACUNIT
		elseif seg.id == 5
			seg.mobj.scale = 3*mo.scale/2
			seg.mobj.color = SKINCOLOR_SUPERGOLD1
			seg.mobj.frame = $ | FF_TRANS10
		elseif seg.id == 6
			seg.mobj.scale = mo.scale/2
			seg.angle = $ + 1
			seg.rotation = $ + 1
		elseif seg.id == 7
			seg.mobj.scale = 3*mo.scale/4
			seg.angle = $ - 1
			seg.rotation = $ - 1
		end
		
		if mo.colorized
			if not seg.mobj.color
				seg.mobj.colorized = true
			end
			seg.mobj.color = mo.color
		end
	end)
	
	if mo.cusval or (timer <= 30
	and not (mo.target and mo.target.valid
	and mo.target.player and mo.target.player.aether
	and mo.target.player.aether.spells[SPL_GENESIS].active))
		if not mo.cusval
			mo.destscale = FRACUNIT/96
			mo.fuse = 15
			mo.cusval = 1
		end
		return
	end

	-- Fire!
	if timer == 30
		-- Ughhhhh
		local surface = (P_MobjFlip(mo) < 0) and mo.ceilingz or mo.floorz
		surface = $ - 16*mo.scale*P_MobjFlip(mo)

		local laser = P_SpawnMobjFromMobj(mo, 0, 0, 16*FRACUNIT, MT_THOK)
		laser.state = S_AETHER_GENESISBEAM
		laser.scale = 2*mo.scale/3
		laser.fuse = MAXSPELLTIME[SPL_GENESIS] - timer + 15
		
		if (mo.target.player)
			P_ColorTeamMissile(laser, mo.target.player)
			if laser.color then laser.colorized = true end
		end
		
		laser.height = laser.scale
		laser.eflags = $ ^^ MFE_VERTICALFLIP
		laser.spriteyscale = FixedDiv(abs(mo.z - surface), mo.scale*2/3)/64

		S_StartSound(mo, sfx_genblt)
	end

	if timer >= 30 and timer <= MAXSPELLTIME[SPL_GENESIS]
		local flip = P_MobjFlip(mo)
		-- Damage!
		searchBlockmap("objects", function(_, found)
			if found == mo.target
			or FixedHypot(found.x - mo.x, found.y - mo.y) > mo.radius
			or flip > 0 and found.z > mo.z + mo.height
			or flip < 0 and found.z + found.height < mo.z
				return
			end

			P_DamageMobj(found, mo, mo.target)
		end, mo, mo.x - 128*FRACUNIT, mo.x + 128*FRACUNIT, mo.y - 128*FRACUNIT, mo.y + 128*FRACUNIT)

		-- Do a quake!
		// not anymore
		/*for p in players.iterate
			if not p.mo continue end -- LOL

			local dist = P_AproxDistance(p.mo.x - mo.x, p.mo.y - mo.y)
			if dist > 512*FRACUNIT continue end

			P_StartQuake(FixedDiv(4096*FRACUNIT, dist or 1), 2)
		end*/
	end

	if timer >= MAXSPELLTIME[SPL_GENESIS] and not mo.fuse
		mo.destscale = FRACUNIT/96
		mo.fuse = 15
	end
end, MT_GENESISSPAWNPOINT)

--
-- SWORD RAIN
--

SpellAbort[SPL_SWORDRAIN] = function(p, aborttype)
	local mo = p.mo
	local aether = p.aether
	local spells = aether.spells
	
	p.thrustfactor = skins[mo.skin].thrustfactor
	if mo.state == S_AETHER_SWORDRAIN
		mo.state = S_PLAY_JUMP
	end
	
	if p.playerstate ~= PST_LIVE // don't adjust mobj flags/state while dead
		return
	end
end

SpellThinker[SPL_SWORDRAIN] = function(p)
	local aether = p.aether
	local mo = p.mo
	--local ang = aether.swordrain_aim

	local s = aether.spells[SPL_SWORDRAIN]
	local maxtime = MAXSPELLTIME[SPL_SWORDRAIN]

	if not CanUseSpells(p, true, true)
		DeactivateSpell(p, SPL_SWORDRAIN)
	else
		if s.timer == 1	-- on activation
			aether.swordraintargets = {}
		end
		
		-- no gravity, slide along walls, don't touch the floor
		if not P_IsObjectOnGround(mo) and not P_InQuicksand(mo)
			mo.momz = $ - P_GetMobjGravity(mo)
		end
		
		if mo.state != S_AETHER_SWORDRAIN
			mo.state = S_AETHER_SWORDRAIN
		end
		
		-- always be slightly over the floor
		if not P_InQuicksand(mo)
			if mo.eflags & MFE_VERTICALFLIP
				local fz = mo.ceilingz
				if mo.standingslope
					fz = P_GetZAt(mo.standingslope, mo.x, mo.y)
				end
				mo.z = min((fz-mo.height) - mo.scale*2, $)
			else
				local fz = mo.floorz
				if mo.standingslope
					fz = P_GetZAt(mo.standingslope, mo.x, mo.y)
				end
				mo.z = max(fz + mo.scale*2, $)
			end
		end
		
		if P_IsObjectOnGround(mo)
			p.thrustfactor = skins[mo.skin].thrustfactor*3/2
		else
			p.pflags = $|P_GetJumpFlags(p)
			p.thrustfactor = skins[mo.skin].thrustfactor*3
		end

		-- vertical momentum shenanigans
		local momz = mo.momz * P_MobjFlip(mo)
		if momz < 0 and not P_InQuicksand(mo)
			mo.momz = $*95/100	-- keep 95% of initial momentum
		end

		-- visuals:
		if leveltime & 1
			local g = P_SpawnGhostMobj(mo)
			g.colorized = true
			g.color = SKINCOLOR_GOLD	-- to match w/ sword rain
			g.tics = 4
		end
	end
	
	local battlemod_amount = 5
	local battlemod_rate = (maxtime / battlemod_amount)
	
    if (not CBW_Battle and Silverhorn.IsSol(p) and aether.timer % 3)
	or (not CBW_Battle and not Silverhorn.IsSol(p) and aether.timer % 5)
	or (CBW_Battle and s.timer % battlemod_rate)
		return
	end

	local dist = P_RandomRange(16, 128)
	local flip = P_RandomChance(FRACUNIT/2) and -1 or 1
	local ang = P_RandomRange(-50, 50)*ANG1
	
	local x = dist * flip * cos(mo.angle + ang)
	local y = dist * flip * sin(mo.angle + ang)
	local z = p.height*2/3
	
	if CBW_Battle
		dist = FRACUNIT*48
		x = FixedMul(dist, cos(mo.angle))
		y = FixedMul(dist, sin(mo.angle))
		z = p.height*7/9
		
		local battlemod_spawncount = s.timer / battlemod_rate
		if (battlemod_spawncount != 1)	-- first one is always accurate
			-- make it "circle" around the spawn location
			local angle_diff = 360 / (battlemod_amount - 1)
			local angle_offset = -45
			local vertangle = ((battlemod_spawncount - 1) * angle_diff + angle_offset) * ANG1
			local horzangle = mo.angle + ANGLE_90
			
			local circle_dist = FRACUNIT*24
			x = $ + FixedMul(FixedMul(circle_dist, cos(horzangle)), cos(vertangle))
			y = $ + FixedMul(FixedMul(circle_dist, sin(horzangle)), cos(vertangle))
			z = $ + FixedMul(circle_dist, sin(vertangle))
		end
	else
		local floordist = min(P_MobjFlip(mo) > 0 and (mo.z - mo.floorz) or (mo.ceilingz - (mo.z + mo.height)), -16*mo.scale)
		local ceildist = min(P_MobjFlip(mo) > 0 and (mo.ceilingz - (mo.z + mo.height)) or (mo.z - mo.floorz), 128*mo.scale)
		z = P_RandomRange(FixedInt(-floordist), FixedInt(ceildist)) * FRACUNIT
		z = FixedDiv($, mo.scale)
	end

	local portal = P_SpawnMobjFromMobj(mo, x + FixedDiv(mo.momx, mo.scale),
	y + FixedDiv(mo.momy, mo.scale), z + FixedDiv(mo.momz, mo.scale), MT_SWORDSPAWNPOINT)
	portal.target = mo
	portal.angle = mo.angle
	portal.fuse = 40
	P_ColorTeamMissile(portal, p)

	portal.scale = mo.scale/64
	portal.scalespeed = mo.scale/4
	portal.destscale = mo.scale
	
	S_StartSound(portal, sfx_srport)

	P_BuildSpriteGroup(portal, {
		{mobj = P_SpawnMobjFromMobj(portal, 0, 0, 0, MT_THOK), state = S_AETHER_PORTALRING, offset = {x = -16, y = 0, z = 0}, angle = 90, rotation = 0, id = 1},
		{mobj = P_SpawnMobjFromMobj(portal, 0, 0, 0, MT_THOK), state = S_AETHER_PORTALRING, offset = {x = 16, y = 0, z = 0}, angle = 90, rotation = 0, id = 2},
	})
end

-- GATE OF BABYLON
local function ItemInTable(t, item)
	for i = #t, 1, -1
		if t[i] == item
			return true
		end
	end
end

addHook("MobjThinker", function(mo)
	P_UpdateSpriteGroup(mo, function(seg)
		seg.mobj.fuse = $ or mo.fuse
		if (mo.color)
			seg.mobj.color = mo.color
		else
			seg.mobj.color = SKINCOLOR_SUPERGOLD2
		end

		if seg.mobj.fuse < 10
			seg.mobj.frame = (seg.mobj.fuse<<FRACBITS) | FF_PAPERSPRITE
		end

		seg.mobj.scale = FixedDiv(mo.scale, 5*seg.id*FRACUNIT/2)
	end)
	
	local target = mo.target
	local player = target and target.valid and target.player
	local aether = player and player.aether
	
	if not aether
		return
	end

	if not mo.shot and mo.scale == mo.destscale

		-- find an enemy
		local range = 1024*target.scale
		local span = ANGLE_45
		local x = mo.x + P_ReturnThrustX(mo, mo.angle, range >> 1)
		local y = mo.y + P_ReturnThrustY(mo, mo.angle, range >> 1)
		local boundx1 = mo.x - range
		local boundx2 = mo.x + range
		local boundy1 = mo.y - range
		local boundy2 = mo.y + range

		local closest = 0
		local foundt

		searchBlockmap("objects", function(ref, found)
			//Prevent targeting Battle's target dummy since we can target players anyways
			if (pcall(function() local v = _G["MT_TARGETDUMMY"] end, "MT_TARGETDUMMY")
			and found.type == _G["MT_TARGETDUMMY"]) then return end

			if found == ref or found == target
			or not found.valid or not found.health
			or (not P_CheckSight(ref, found) and not P_CheckSight(target, found))
			or found.flags & (MF_NOCLIP|MF_NOCLIPTHING) or found.flags2 & MF2_FRET
			or found.type == MT_EGGMAN_BOX or found.type == MT_EGGMAN_GOLDBOX
				return
			end
			
			//Only aim at objects at angles in the span range
			local diff = ref.angle - R_PointToAngle2(ref.x, ref.y, found.x, found.y)
			if abs(InvAngle(InvAngle(diff))) > span
				return
			end
			
			if found.player
				if found.player.spectator or found.player.powers[pw_flashing]
					return
				elseif gametyperules&GTR_FRIENDLY and not CV_FindVar("friendlyfire").value
					return
				elseif G_GametypeHasTeams() and found.player.ctfteam == player.ctfteam
					return
				elseif G_TagGametype() and (found.player.pflags & PF_TAGIT) == (player.pflags & PF_TAGIT)
					return
				end
			elseif found.flags & (MF_MONITOR) and (G_RingSlingerGametype() or (CBW_Battle and CBW_Battle.BattleGametype()))
				return
			elseif (!((found.flags & (MF_ENEMY|MF_BOSS|MF_MONITOR) and (found.flags & MF_SHOOTABLE))) == !(found.flags2 & MF2_INVERTAIMABLE)) // allows if it has the flags desired XOR it has the invert aimable flag
				return // not a valid target
			end
			
			local dist = R_PointToDist2(ref.x, ref.y, found.x, found.y)
			dist = R_PointToDist2(0, ref.z + ref.height/2, $, found.z + found.height/2)
			
			if (found.type != MT_METALSONIC_BATTLE
			and ItemInTable(aether.swordraintargets, found))
			or dist > (closest or range)
				return
			end
			
			/*local test = P_SpawnMobj(found.x, found.y, found.z, MT_THOK)
			test.frame = A
			test.tics = TICRATE*2
			test.color = SKINCOLOR_WHITE*/
			foundt = found
			closest = dist

		end, mo, boundx1, boundx2,boundy1, boundy2)

		local sword
		if foundt	-- found enemy, home in!
			sword = P_SpawnPointMissile(target, foundt.x, foundt.y, foundt.z + foundt.height >> 1, MT_AETHER_SWORD, mo.x, mo.y, mo.z)
			table.insert(aether.swordraintargets, foundt)
			if sword and sword.valid
				local angle = R_PointToAngle2(0,0, FixedHypot(sword.momx, sword.momy), sword.momz)
				if abs(angle) >= ANGLE_11hh
					local mul = angle*P_MobjFlip(target) > 0 and 1 or -1
					
					if abs(angle) >= ANGLE_22h
						sword.frame = $ + 2*mul
					else
						sword.frame = $ + 1*mul
					end
				end
			end
		else	-- keep old behaviour
			sword = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_AETHER_SWORD)
			if sword and sword.valid
				P_InstaThrust(sword, mo.angle, FixedMul(sword.info.speed, sword.scale))
				sword.angle = mo.angle
			end
		end
		
		if sword and sword.valid
			sword.target = target
			sword.fuse = TICRATE
			
			sword.blockable = 1
			sword.block_hthrust = 4
			sword.block_vthrust = 12
			sword.block_stun = 0
			sword.hit_sound = sfx_hit02
			
			if mo.color
				sword.color = mo.color
				sword.colorized = true
			end
		end
		
		S_StartSound(mo, sfx_srfire)
		mo.shot = true
	end
end, MT_SWORDSPAWNPOINT)

-- Sword Rain sword thinker
addHook("MobjThinker", function(mo)
	if not mo.threshold
		mo.blendmode = AST_ADD
		mo.threshold = $ + 1
	end
	
	if leveltime & 1
		local g = P_SpawnGhostMobj(mo)
		g.blendmode = mo.blendmode
		g.destscale = mo.scale*8
		g.scalespeed = mo.scale/12
	end
	
	if not (mo.flags & MF_MISSILE)
		if S_SoundPlaying(mo, sfx_srtral) 
			S_StopSoundByID(mo, sfx_srtral)
		end
		return
	end

	--print(AngleFixed(ang)/FRACUNIT)

	local x = P_RandomRange(-1, 1) * cos(mo.angle)
	local y = P_RandomRange(-1, 1) * sin(mo.angle)
	local z = P_RandomRange(-1, 1) * FRACUNIT

	local particle = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
	particle.momx = x
	particle.momx = y
	particle.momx = z
	particle.state = S_AETHER_PARTICLE2
	particle.frame = (P_RandomChance(FRACUNIT/2) and 1 or 0) | FF_TRANS20
	if mo.color
		particle.color = mo.color
	else
		particle.color = SKINCOLOR_SUPERGOLD2
	end
	
	if not S_SoundPlaying(mo, sfx_srtral)
		S_StartSound(mo, sfx_srtral)
	end
end, MT_AETHER_SWORD)

local function SwordDamage(mo, inf, flags)
	if not (inf and inf.valid and inf.type == MT_AETHER_SWORD)
	or not mo.valid
	or flags and not (mo.flags & flags)
		return
	end
	
	local info = zonebuilderstoptrollingme[MT_AETHER_SWORDPARTICLE]
	local z
	local pitchrange
	if P_IsObjectOnGround(mo)
		z = 0
		pitchrange = 90
	else
		z = FixedDiv(mo.height, mo.scale) >> 1
		pitchrange = 180
	end
	
	for i = 1, 24
		local hangle = FixedAngle(P_RandomKey(360)*FRACUNIT)
		local zangle = FixedAngle((P_RandomKey(pitchrange) - (pitchrange - 90))*FRACUNIT)
		local zspeed = FixedMul(sin(zangle), info.speed)
		local hspeed = FixedMul(cos(zangle), info.speed)
		local xspeed = P_ReturnThrustX(inf, hangle, hspeed)
		local yspeed = P_ReturnThrustY(inf, hangle, hspeed)
		
		local particle = P_SpawnMobjFromMobj(mo, xspeed, yspeed, z + zspeed, MT_AETHER_SWORDPARTICLE)
		particle.color = SKINCOLOR_GOLD
		particle.frame = $ + P_RandomKey(2)
		if P_RandomChance(FRACUNIT/3)
			particle.blendmode = AST_ADD
		end
		particle.momx = FixedMul(xspeed, particle.scale)
		particle.momy = FixedMul(yspeed, particle.scale)
		P_SetObjectMomZ(particle, zspeed)
		particle.scale = $ << 1
	end
end

addHook("MobjDamage", function(mo, inf) SwordDamage(mo, inf) end)
addHook("MobjDeath", function(mo, inf) SwordDamage(mo, inf, MF_MONITOR) end)

addHook("MobjThinker", function(mo)
	local factor = mo.info.painchance
	
	mo.momx = FixedMul($, factor)
	mo.momy = FixedMul($, factor)
	mo.momz = FixedMul($, factor)
	
	if states[mo.state].nextstate == S_NULL
	and mo.tics < 10
		mo.frame = $ | ((10 - mo.tics) * FF_TRANS10)
	end
end, MT_AETHER_SWORDPARTICLE)

-- Sword Rain vs. Bustable Blocks
addHook("MobjLineCollide", function(mo, line)
	for _, sec in ipairs({line.frontsector, line.backsector})
		for fof in sec.ffloors()
			if not HeightCheck(mo, fof) continue end
			if not (fof.flags & FF_EXISTS) continue end
			if not (fof.flags & FF_BUSTUP) continue end

			Silverhorn.MobjBustFOF(mo, fof)
			return false
		end
	end
end, MT_AETHER_SWORD)

--
-- Speedwave
--

-- Spawns an aura on the player mobj
local function spawnaura(mo, tics, sol)
	if not (mo.aura and mo.aura.valid)
		mo.aura = P_SpawnMobjFromMobj(mo, 0, 0, mo.height, MT_AETHER_AURACENTER)
		mo.aura.target = mo
		mo.aura.fuse = tics or MAXSPELLTIME[SPL_AURA] - 1	-- fallback
		if sol == true
		   mo.aura.color = SKINCOLOR_ORANGE
		   mo.aura.colorized = true
		end
		S_StartSound(mo, sfx_swon)
		
		local t = CreatePaperspriteCircle(mo, 15, AURA_RADIUS, MT_THOK, S_AETHER_RUNES, 0)
		if sol == true
		   t.color = SKINCOLOR_ORANGE
		   t.colorized = true
		end		

		for _, m in ipairs(t)
			m.frame = P_RandomKey(7)
			m.mobj.blendmode = AST_ADD
		end

		P_BuildSpriteGroup(mo.aura, t)

		mo.player.wasinaura = true	-- had an aura
		
		if mo.player.aether
			local spell = mo.player.aether.spells[SPL_AURA]
			spell.cooldown = 0
			spell.active = true
			spell.timer = MAXSPELLTIME[SPL_AURA] - mo.aura.fuse
		end
	end
end

SpellAbort[SPL_AURA] = function(p, aborttype)
	local mo = p.mo
	local aether = p.aether
	local spells = aether.spells
	
	if not p.powers[pw_carry]
	and mo.state == S_PLAY_SPINDASH
		if P_IsObjectOnGround(mo)
			mo.state = S_PLAY_STND
		else
			mo.state = S_PLAY_FALL
		end
	end
end


SpellThinker[SPL_AURA] = function(p)
	local aether = p.aether
	local mo = p.mo

	if aether.spells[SPL_AURA].timer == 1
		if Silverhorn.IsSol(p)
			spawnaura(mo, MAXSPELLTIME[SPL_AURA] - 1, true)
		else
			spawnaura(mo, MAXSPELLTIME[SPL_AURA] - 1, false)
		end
		mo.state = S_PLAY_SPINDASH
	end
	
	if mo.state == S_PLAY_SPINDASH
	and P_IsObjectOnGround(mo) and p.panim != PA_DASH
		p.panim = PA_IDLE
	end
end

addHook("MobjThinker", function(mo)
	if not mo.target or not mo.target.valid or not mo.target.player
	then P_RemoveMobj(mo) return end
	
	if not mo.target.health or mo.target.aura != mo
	or mo.target.player.playerstate or mo.target.player.exiting
	or (CBW_Battle and (P_PlayerInPain(mo.target.player) or mo.target.player.tumble))
		-- Lat: non aether players can now have auras
		if mo.target.aura == mo and mo.target.player.aether
			DeactivateSpell(mo.target.player, SPL_AURA)
		end
		P_RemoveMobj(mo)
		return
	end
	
	mo.eflags = $ & ~MFE_VERTICALFLIP | (mo.target.eflags & MFE_VERTICALFLIP)
	local z = mo.target.z - mo.target.height/6*P_MobjFlip(mo)
	if not (mo.eflags & MFE_VERTICALFLIP)
		z = $ + mo.target.height
	end
	P_MoveOrigin(mo, mo.target.x, mo.target.y, z)
	mo.angle = $ + ANG1 * 2

	if mo.fuse < TICRATE*3	-- Time is running out!
		if leveltime & 1
			mo.flags2 = $|MF2_DONTDRAW
		else
			mo.flags2 = $ & ~MF2_DONTDRAW
		end
	end

	P_UpdateSpriteGroup(mo, function(seg)
		if seg.frame
			seg.mobj.frame = seg.frame|FF_PAPERSPRITE
		end

		seg.mobj.flags2 = mo.flags2 & MF2_DONTDRAW	-- Update DONTDRAW for low duration remaining flashing
		seg.mobj.eflags = $ & ~MFE_VERTICALFLIP | (mo.eflags & MFE_VERTICALFLIP)
		
		if seg.mobj.eflags & MFE_VERTICALFLIP
			seg.mobj.z = $ - seg.mobj.height
		end
        if mo.color == SKINCOLOR_ORANGE
			seg.mobj.color = SKINCOLOR_AQUA
		else
			seg.mobj.color = Silverhorn.EmeraldColor
		end
		seg.mobj.colorized = true
		seg.mobj.fuse = $ or mo.fuse
	end)

	-- Aura check
	if not (gametyperules & (GTR_FRIENDLY|GTR_TEAMS)) // only give buffs to friendly players/teammates
		return
	end
	
	for p in players.iterate
		if not p.mo
		or not p.mo.health
		or p.playerstate
		or p.exiting
		or P_CheckDeathPitCollide(p.mo) -- prevents an infinite loop
		or (p.mo.aura and p.mo.aura.valid)
		or not SpellSympathy(mo.target.player, p)
		--or p.aether -- No cheese sorry
		-- Lat: not anymore, aether can benefit from her own aura!
			continue
		end

		local dist = AURA_RADIUS * mo.scale
		if R_PointToDist2(p.mo.x, p.mo.y, mo.x, mo.y) > dist continue end
		if p.mo.z > mo.z + 256*FRACUNIT continue end
		if p.mo.z + p.mo.height < mo.z - 256*FRACUNIT continue end

        if mo.color == SKINCOLOR_ORANGE
			spawnaura(p.mo, max(5*TICRATE, mo.fuse), true)	-- give me a sol aura!
		else
			spawnaura(p.mo, max(5*TICRATE, mo.fuse), false)	-- give me an aura!
		end
	end
end, MT_AETHER_AURACENTER)

addHook("MobjRemoved", function(mo)
	P_UpdateSpriteGroup(mo, function(seg)
		P_RemoveMobj(seg.mobj)
	end)
end, MT_AETHER_AURACENTER)

-- Aura Sharing
local function SpeedwaveStatReset(p)
	p.wasinaura = false
	-- Reset stats (Don't reset dashmode stats though!)
	if p.dashmode < TICRATE*3 //DASHMODE_THRESHOLD
		p.jumpfactor = skins[p.skin].jumpfactor
		if not (p.inazuma and p.inazuma.neuralshock)
			p.normalspeed = skins[p.skin].normalspeed
		end
	end
	p.fakenormalspeed = p.normalspeed
end

addHook("PlayerThink", function(p)
	local mo = p.mo
	if (not mo or p.playerstate ~= PST_LIVE)
		if p.wasinaura
			SpeedwaveStatReset(p)
		end
		return
	end

	-- aura shit
	if mo.aura and mo.aura.valid
		if mo.aura.color == SKINCOLOR_ORANGE
			if not p.isxmomentum
				p.normalspeed = max($,skins[p.skin].normalspeed*150/100) -- +50%
			else
				p.fakenormalspeed = max($,skins[p.mo.skin].normalspeed*150/100) -- +50%
			end
			p.jumpfactor = max($,skins[p.mo.skin].jumpfactor*150/100)	-- also 50%	  
		else
			if not p.isxmomentum
				p.normalspeed = max($,skins[p.skin].normalspeed*133/100) -- +33%
			else
				p.fakenormalspeed = max($,skins[p.mo.skin].normalspeed*133/100) -- +33%
			end
			p.jumpfactor = max($,skins[p.mo.skin].jumpfactor*133/100)	-- also 33%
		end
		
		//Dashmode charge
		if p.charflags & SF_DASHMODE and not p.gotflag
		and not p.powers[pw_carry] and not p.exiting
		and not (maptol & TOL_NIGHTS) and not metalrecording
			if p.dashmode < TICRATE*3 //DASHMODE_THRESHOLD
				S_StartSound(mo, (p.charflags & SF_MACHINE) and sfx_kc4d or sfx_cdfm40)
			end
			p.dashmode = TICRATE*3 + 3 //DASHMODE_MAX
			
			//Reset dash state
			if mo.state == S_PLAY_DASH and p.speed < FixedMul(p.runspeed, mo.scale)
				mo.state = S_PLAY_WALK
			end
		
		//Neural Shock charge
		elseif mo.skin == "inazuma" and p.inazuma and not p.gotflag
		and not p.gotcrystal and not p.gotflagdebuff
			if not p.inazuma.neuralshock
				S_StartSound(mo, sfx_boltbl)
				
				local fx = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
				fx.state = S_BOLTBLAST
				fx.color = mo.aura.color or AETHERHUD[SPL_AURA].color
			end
			p.inazuma.neuralshock = max($ or 0, TICRATE)
		
		//Burstmode charge (too janky to allow while standing)
		elseif mo.skin == "blaze"
			p.blazeburstcharge = max($, TICRATE*2 - 1)
			
		//Marine P-Speed
		elseif mo.skin == "marine" and p.marine and not (p.pflags & PF_THOKKED)
		and not P_PlayerInPain(p) and not p.powers[pw_carry]
		and not p.exiting and not p.gotflag and not p.gotcrystal
		and p.playerstate == PST_LIVE
			p.marine.run = TICRATE*2
		end
		
	elseif p.wasinaura
		-- Reset stats
		SpeedwaveStatReset(p)
		--also play sound he haw
		S_StartSound(mo, sfx_swoff)
	end
end)

addHook("PlayerSpawn", function(player)
	if player.wasinaura
		SpeedwaveStatReset(player)
	end
end)

--
-- Light Orb
--

local lightorb_startup = TICRATE/2

local function DeactivateAetherOrb(p)
	local aether = p.aether
	local mo = p.mo
	local orb = aether.lightorb
	
	if orb and orb.valid and orb.type == MT_AETHER_LIGHTORB
		S_StartSound(orb, orb.info.deathsound)
		orb.flags2 = ($&~MF2_OBJECTFLIP)|(mo.flags2&MF2_OBJECTFLIP)
		orb.eflags = ($&~MFE_VERTICALFLIP)|(mo.eflags&MFE_VERTICALFLIP)
		P_SetObjectMomZ(orb, 8*FRACUNIT, false)
		P_SpawnMobjFromMobj(orb, 0,0, -16*FRACUNIT, MT_EXPLODE)
		P_SetScale(orb, orb.scale/2)
		P_InstaThrust(orb, mo.angle, 2*mo.scale)
		aether.lightorb = nil
	end
end

SpellThinker[SPL_NONE] = function(p)
	local aether = p.aether
	local mo = p.mo
	local orb = aether.lightorb
	
	-- spawn the orb
	if aether.spells[SPL_NONE].timer == 1
		orb = P_SpawnMobjFromMobj(mo, FixedMul(16*FRACUNIT, cos(mo.angle)), FixedMul(16*FRACUNIT, sin(mo.angle)), p.height/2, MT_AETHER_LIGHTORB)
		P_SpawnMobjFromMobj(orb, 0,0, -16*FRACUNIT, MT_EXPLODE)
		aether.lightorb = orb
		orb.target = mo
		orb.tracer = mo
		orb.sprite = SPR_ALOB
		orb.reactiontime = lightorb_startup
		P_SetScale(orb, orb.scale/2)
		S_StartSound(orb, orb.info.seesound)
		AuraBarrierHalt(p)
		mo.state = S_PLAY_GLIDE_LANDING
		p.powers[pw_nocontrol] = max($, lightorb_startup)
		mo.tics = p.powers[pw_nocontrol]
	end
	
	if P_PlayerInPain(p)
		-- destroy the orb if we're damaged
		DeactivateAetherOrb(p)
		DeactivateSpell(p, SPL_NONE)
	elseif not (orb and orb.valid and orb.type == MT_AETHER_LIGHTORB and orb.health)
		-- orb isnt active anymore, end the spell
		DeactivateSpell(p, SPL_NONE)
	elseif aether.spells[SPL_NONE].timer > lightorb_startup
		-- keep the spell active
		aether.spells[SPL_NONE].timer = 57
	end
end

SpellSpin[SPL_NONE] = function(p)
	local aether = p.aether
	local mo = p.mo
	local orb = aether.lightorb
	
	if aether.spells[SPL_NONE].timer > lightorb_startup
	and orb and orb.valid and orb.type == MT_AETHER_LIGHTORB
		DeactivateAetherOrb(p)
		mo.state = S_PLAY_GASP
	end
end

SpellAbort[SPL_NONE] = function(p, aborttype)
	local mo = p.mo
	local aether = p.aether
	local orb = aether.lightorb
	
	if aborttype != ABORT_MAXSPELLTIME
		if orb and orb.valid
			P_SetObjectMomZ(orb, 8*FRACUNIT, false)
		end
		aether.lightorb = nil
	end
end

//Orb FOF Breaker Function
local orbfofbreak = function(orb, sectorthing)
	if not (sectorthing and sectorthing.valid) return end
	for fof in sectorthing.ffloors()
		if fof.valid and (fof.flags & FF_BUSTUP) and (fof.flags & FF_EXISTS)
			local foffloor = fof.topheight
			local fofceiling = fof.bottomheight
			if fof.b_slope and fof.b_slope.valid
				fofceiling = P_GetZAt(fof.b_slope, orb.x, orb.y)
			end
			if fof.t_slope and fof.t_slope.valid
				foffloor = P_GetZAt(fof.t_slope, orb.x, orb.y)
			end
			if foffloor >= min(orb.z, orb.z+orb.momz)
			and fofceiling <= max(orb.z+orb.height, orb.z+orb.height+orb.momz)
				Silverhorn.MobjBustFOF(orb, fof)
				if fof.master.flags & ML_EFFECT5
					P_LinedefExecute(P_AproxDistance(fof.master.dx, fof.master.dy)>>FRACBITS, orb, sectorthing)
				end
				return true
			end
		end
	end
end

//Orb Wall Breaker
addHook("MobjLineCollide", function(orb, line)
	if not (line and line.valid and orb and orb.valid
	and (not orb.health or orb.reactiontime != 1)) then return end
	
	local killone = orbfofbreak(orb, line.frontsector)
	local killtwo = orbfofbreak(orb, line.backsector)
	
	return killone or killtwo
end, MT_AETHER_LIGHTORB)

//Light Orb thinker
addHook("MobjThinker", function(orb)
	if not orb.health
		if not orb.fuse then orb.fuse = TICRATE*5 end
		
		if (orb.aether_bursted)
			A_ForceStop(orb)
		else
			-- regular death sequence
			orb.sprite = SPR_ALOB
			orb.frame = 0
			orb.flags = $&~MF_NOGRAVITY&~MF_NOCLIPHEIGHT&~MF_NOCLIP&~MF_NOCLIPTHING
			if P_IsObjectOnGround(orb) 
				if not orb.momx and not orb.momy and not orb.momz
					if orb.fuse > TICRATE/2
						orb.fuse = TICRATE/2
					else
						orb.momx = FixedMul($,orb.friction)
						orb.momy = FixedMul($,orb.friction)
					end
				end
				if orb.fuse > TICRATE/2 and orb.subsector and orb.subsector.valid
					orbfofbreak(orb, orb.subsector.sector)
				end
			end
		end
		
		return
	elseif not orb.reactiontime
		A_ForceStop(orb)
		return
	end
	
	local mo = orb.target
	if mo and mo.valid and mo.player and mo.skin == "aether"
	and mo.health and not mo.player.playerstate and mo.player.aether
	and mo.player.aether.lightorb == orb
		if orb.reactiontime > 1
			orb.reactiontime = $-1
			if orb.reactiontime == 1
				P_MoveOrigin(orb, mo.x, mo.y, mo.z+mo.height/2)
				if P_TryMove(orb, mo.x+FixedMul(16*mo.scale, cos(mo.angle)), mo.y+FixedMul(16*mo.scale, sin(mo.angle)), false)
					orb.momx = mo.momx
					orb.momy = mo.momy
				else
					P_TryMove(orb, mo.x+FixedMul(-16*mo.scale, cos(mo.angle)), mo.y+FixedMul(-16*mo.scale, sin(mo.angle)), false)
				end
				orb.momz = mo.momz
				orb.frame = 1
				P_SetScale(orb, orb.scale/2)
				S_StartSound(orb, orb.info.attacksound)
				P_SpawnMobjFromMobj(orb, 0,0, -16*FRACUNIT, MT_SPARK)
				
				if not (CBW_Battle and CBW_Battle.BattleGametype())
					orb.flags = $|MF_NOCLIPTHING
				else
					orb.flags = $|MF_SHOOTABLE|MF_BOSS
					orb.flags = $ & ~(MF_NOCLIPTHING)
				end
			else
				P_MoveOrigin(orb, mo.x+FixedMul(16*mo.scale, cos(mo.angle)), mo.y+FixedMul(16*mo.scale, sin(mo.angle)), mo.z+mo.height/2)
			end
		else
			local oldmomentum = {x = orb.momx, y = orb.momy, z = orb.momz}
			local playermomentum = {x = mo.momx, y = mo.momy, z = mo.momz}
			if mo.player.powers[pw_carry]
			and mo.player.powers[pw_carry] != CR_NIGHTSMODE
			and mo.player.powers[pw_carry] != CR_NIGHTSFALL
			and mo.player.powers[pw_carry] != CR_ZOOMTUBE
			and mo.tracer and mo.tracer.valid
				playermomentum = {x = mo.tracer.momx, y = mo.tracer.momy, z = mo.tracer.momz}
			end
			
			orb.destscale = mo.scale
			orb.flags2 = $&~MF2_OBJECTFLIP
			orb.eflags = $&~MFE_VERTICALFLIP
			if mo.eflags & MFE_VERTICALFLIP
				orb.frame = $ | FF_VERTICALFLIP
				orb.spriteyoffset = 8*FRACUNIT
			else
				orb.frame = $ &~ FF_VERTICALFLIP
				orb.spriteyoffset = 0
			end
			
			local distance = FixedHypot(R_PointToDist2(mo.x, mo.y, orb.x, orb.y), abs(mo.z - orb.z))
			if distance > orb.radius*INT8_MAX
				P_SetScale(orb, orb.scale/2)
				P_MoveOrigin(orb, mo.x, mo.y, mo.z+mo.height/2)
			else
				local factor = max(128-distance/orb.scale, 8)
				local factor2 = max(factor-1, 1)
				P_HomingAttack(orb, mo)
				orb.momx = $/factor + oldmomentum.x*factor2/factor + playermomentum.x/factor + P_SignedRandom()*orb.scale/128
				orb.momy = $/factor + oldmomentum.y*factor2/factor + playermomentum.y/factor + P_SignedRandom()*orb.scale/128
				orb.momz = $/factor + oldmomentum.z*factor2/factor + playermomentum.z/factor + P_SignedRandom()*orb.scale/128
			end
			
			if P_IsObjectOnGround(orb)
				if abs(orb.z-orb.floorz) < abs(orb.z-orb.ceilingz)
					orb.momz = abs($)
				else
					orb.momz = -abs($)
				end
			end
			
			if leveltime%2
				for i = 0, 8
					P_SPMAngle(orb, MT_AETHER_RAYCAST, orb.angle+ANGLE_22h*i, 0)
				end
			end
			
			local angle = R_PointToAngle(orb.x,orb.y)
			local thok = P_SpawnMobjFromMobj(orb, cos(angle), sin(angle), 0, MT_THOK)
			thok.scale = $/2 + FixedMul(orb.scale/8, sin(FixedAngle(leveltime*8*FRACUNIT)))
			thok.momx = orb.momx
			thok.momy = orb.momy
			thok.momz = orb.momz
			thok.blendmode = AST_ADD
			thok.sprite = SPR_OILF
			thok.tics = 2
		end
	elseif orb.health
		orb.aether_noburst = true	-- prevent burst
		P_KillMobj(orb)
	end
end, MT_AETHER_LIGHTORB)

//Destroy spikes via the orb
local orbspike = function(spike, orb)
	if spike and spike.valid and orb and orb.valid and orb.type == MT_AETHER_LIGHTORB
	and (not orb.health or orb.reactiontime != 1)
	and orb.z >= spike.z-orb.height and orb.z <= spike.z+spike.height+orb.height
		P_KillMobj(spike, orb)
		return true
	end
end
addHook("MobjCollide", function(thing, tmthing) orbspike(thing, tmthing) end, MT_SPIKE)
addHook("MobjCollide", function(thing, tmthing) orbspike(thing, tmthing) end, MT_WALLSPIKE)

addHook("MobjMoveCollide", function(orb, mo)
	-- no interactions with purely solid objects
	if mo and mo.valid and mo.flags & MF_SOLID
	and not (mo.player or mo.flags & MF_SHOOTABLE)
		return false
	end
	
	if (mo.player and orb.target != mo)
		if not (CBW_Battle) or not (orb.health)
			-- no burst interactions outside of battle
			return false
		end
			
		-- burst the orb on contact
		local topz = orb.z + orb.height
		local bottomz = orb.z
		
		if ((topz >= mo.z) and (bottomz <= mo.z + mo.height))	-- height check
		and (P_PlayerCanDamage(mo.player, orb) or mo.battle_atk)
		and not (CBW_Battle.MyTeam(orb.target.player, mo.player))
			P_KillMobj(orb)
		end
		
		return false
	end
end, MT_AETHER_LIGHTORB)

addHook("MobjMoveBlocked", function(orb)
	if not orb.health or orb.reactiontime != 1
	then S_StartSound(orb, orb.info.activesound) end
	P_BounceMove(orb)
	return true
end, MT_AETHER_LIGHTORB)

addHook("MobjFuse", function(orb)
	if (orb.aether_bursted) then return end	-- prevent this with burst functionality
	P_SpawnGhostMobj(orb)
end, MT_AETHER_LIGHTORB)

addHook("ShouldDamage", function(mo, orb)
	if orb and orb.valid and orb.type == MT_AETHER_LIGHTORB
		return false
	end
end, MT_PLAYER)

//Ray Cast thinker
addHook("MobjThinker", function(mobj)
	if not mobj.threshold
		mobj.threshold = 1
		mobj.fuse = TICRATE/2
	end
	if mobj.subsector and mobj.subsector.valid
	and mobj.subsector.sector and mobj.subsector.sector.valid
		if mobj.subsector.sector.lightlevel < 170
			P_SpawnLightningFlash(mobj.subsector.sector)
			mobj.subsector.sector.lightlevel = 170
		end
		for fof in mobj.subsector.sector.ffloors()
			if fof.valid and (fof.flags & FF_EXISTS)
			and (fof.bottomheight >= mobj.z
			or (mobj.z < fof.topheight
			and mobj.z+mobj.height > fof.bottomheight))
			and fof.sector.valid and fof.sector.lightlevel < 170
			P_SpawnLightningFlash(fof.sector)
			fof.sector.lightlevel = 170
			end
		end
	end
end, MT_AETHER_RAYCAST)