
-- =============================================================
-- This sequence is dedicated to Black Panther, and Chadwick Boseman- the Black Panther. Enjoy. [\'x'/]
-- =============================================================

-- Prevent loading this namespace again
if _G["CreditMaps"] then return end


rawset(_G, "CreditMaps", {endingmap=M_MapNumber(Silverhorn.SpMaps["Credits"])})


-- Handles warping to our credit map used for displaying a custom ending
function CreditMaps.WarpToMap()

	-- Get the map header
	local mapheader = mapheaderinfo[gamemap]

	-- Get the defined map for an ending map
	local endingmap = CreditMaps.endingmap

	-- Don't run on this type of map (custom credits), and get stuck in a recursive exit loop
	if (mapheader.silverhorn_endingmap) then return end

	for p in players.iterate do

		-- Custom credits warp: Runs when exiting time is almost over
		if (mapheader.nextlevel > 1100 and p.exiting and p.exiting < 2) then

			-- (Verifies the header exists before warping to custom Credits Map. This block will be skipped if clearall was used.)
			for i = 0, #mapheaderinfo do
				if mapheaderinfo[i] and mapheaderinfo[i].silverhorn_endingmap then

					-- Warp to custom credits if verified
					G_SetCustomExitVars(endingmap, 1)
					G_ExitLevel()
					p.exiting = 0
					break
				end
			end
		end

	end
end






-- =============================
-- .......
-- Credits
-- .......
-- =============================
freeslot("SPR_CRRS")
local End_33_Aftermath_Position = {x = -6976*FU, y = -21664*FU, z = 2296*FU}

EventSH.new("sh_credits_select", {
function(c,e)

	-- Disable Hud
	EvActionSH.SetInternalHudStatus({"time", "score", "rings", "lives"}, false)

	-- Patch and sound displays
	c.card = ""
	c.cardmusic = ""
	c.star = ""

	-- (Handle conditionals)

	local gotemeralds = All7Emeralds(emeralds)
	local visitedms = MindscapeSessionFlags & MS_VISITEDMINDSCAPE
	local beatms = false

	-- TODO: change to actuall beaten rather than getting horns
	if visitedms then
		EvActionSH.doforall(function(player)
			if player.hassilverhorns then
				beatms = true
				return
			end
		end)
	end

	-- (Ending Selection)
	c.endtype = ""

	-- good ending (got emeralds and beat mindscape)
	if gotemeralds and beatms then
		c.endtype = "good"
		c.card = "SH_ENDG1"
		c.cardmusic = "ENDGD"
		c.star = "SHSTAR4"
		Silverhorn.AchieveEnding("has_ending_g")

	-- true ending (simply visited mindscape)
	elseif visitedms then
		c.card = "SH_ENDT"
		c.cardmusic = "ENDTR"
		c.star = "SHSTAR3"
		Silverhorn.AchieveEnding("has_ending_t")

	-- neutral ending (got emeralds, but never visited mindscape)
	elseif gotemeralds then
		c.card = "SH_ENDN"
		c.cardmusic = "ENDNU"
		c.star = "SHSTAR2"
		Silverhorn.AchieveEnding("has_ending_n")

	-- bad ending (did none of the above)
	else
		c.card = "SH_ENDB"
		c.cardmusic = "ENDBD"
		c.star = "SHSTAR1"
		Silverhorn.AchieveEnding("has_ending_b")
	end
end,
function(c,e)
	if (MindscapeSessionFlags & MS_BEATMINDSCAPE
	and MindscapeSessionFlags & MS_SPAWNLEAVEPORTAL) then
		-- Go to Developer Credits, if you beat mindscape
		EventSH.start("sh_credits", c)
	else
		-- Go to endcard, if you didn't beat mindscape
		EventSH.start("sh_credits_endcard", c)
	end
end})





-- Retruns a table to reduce tediousness below
local function SH_CredSprite(spritename, spriteframe, sprwidth, sprheight)
	return {spr=spritename, frm=R_Char2Frame(spriteframe), w=sprwidth, h=sprheight}
end

local credit_list = {

	-- Names
	["Dirkthehusky"]    = SH_CredSprite(SPR_CCC1, "A", 258, 29),
	["DrStephen"]       = SH_CredSprite(SPR_CCC1, "B", 197, 29),
	["Inazuma"]         = SH_CredSprite(SPR_CCC1, "C", 163, 29),
	["Hyperchaotix"]    = SH_CredSprite(SPR_CCC1, "D", 265, 29),
	["Saneko"]          = SH_CredSprite(SPR_CCC1, "E", 148, 29),
	["Cystallinegazer"] = SH_CredSprite(SPR_CCC1, "F", 341, 29),
	["Flameeliwood"]    = SH_CredSprite(SPR_CCC1, "G", 276, 29),
	["Motorroach"]      = SH_CredSprite(SPR_CCC1, "H", 251, 29),
	["Machturne"]       = SH_CredSprite(SPR_CCC1, "I", 214, 29),
	["Telosturntable"]  = SH_CredSprite(SPR_CCC1, "J", 304, 29),
	["Starlight"]       = SH_CredSprite(SPR_CCC1, "K", 188, 29),
	["Lightdasher"]     = SH_CredSprite(SPR_CCC1, "L", 237, 29),
	["Ralphjeremy65"]   = SH_CredSprite(SPR_CCC1, "M", 290, 29),
	["Blueblur"]        = SH_CredSprite(SPR_CCC1, "N", 171, 28),
	["Soapsurfin"]      = SH_CredSprite(SPR_CCC1, "O", 236, 29),
	["Hattyboyo"]       = SH_CredSprite(SPR_CCC1, "P", 205, 29),
	["Spinslash165"]    = SH_CredSprite(SPR_CCC1, "Q", 242, 29),
	["Alacroix"]        = SH_CredSprite(SPR_CCC1, "R", 180, 29),
	["Delpho"]          = SH_CredSprite(SPR_CCC1, "S", 138, 29),
	["Vadapega"]        = SH_CredSprite(SPR_CCC1, "T", 189, 29),
	["Krabs"]           = SH_CredSprite(SPR_CCC1, "U", 111, 29),
	["Megabaz"]         = SH_CredSprite(SPR_CCC1, "V", 169, 29),
	["Kameasinger"]     = SH_CredSprite(SPR_CCC1, "W", 252, 29),
	["Awesomedubs"]     = SH_CredSprite(SPR_CCC1, "X", 485, 31),
	["Othius"]          = SH_CredSprite(SPR_CCC1, "Y", 127, 29),
	["Kanna"]           = SH_CredSprite(SPR_CCC1, "Z", 125, 29),
	["So2ro"]           = SH_CredSprite(SPR_CCC1, "0", 125, 29),
	["Spectorious"]     = SH_CredSprite(SPR_CCC1, "1", 243, 29),
	["Amperbee"]        = SH_CredSprite(SPR_CCC1, "2", 181, 28),
	["Cyanknight"]      = SH_CredSprite(SPR_CCC1, "3", 225, 29),
	["Lune_Wulf"]       = SH_CredSprite(SPR_CCC1, "4", 199, 30),
	["Frostiikin"]      = SH_CredSprite(SPR_CCC1, "5", 187, 29),
	["Logan8r"]         = SH_CredSprite(SPR_CCC1, "6", 169, 29),
	["Anonymouscoder"]  = SH_CredSprite(SPR_CCC1, "7", 368, 29),
	["Tatsuru"]         = SH_CredSprite(SPR_CCC1, "8", 149, 29),
	["Lat"]             = SH_CredSprite(SPR_CCC1, "9", 67, 28),
	["Smsalfredo"]      = SH_CredSprite(SPR_CCC1, "a", 243, 29),
	["Lach"]            = SH_CredSprite(SPR_CCC1, "b", 89, 29),
	["OrdoMandalore"]   = SH_CredSprite(SPR_CCC1, "c", 329, 29),
	-- 10-25-2023
	["Furless"]        = SH_CredSprite(SPR_CCC1, "d", 144, 29),
	["Snu"]            = SH_CredSprite(SPR_CCC1, "e", 69, 29),
	["FlyingNosaj"]    = SH_CredSprite(SPR_CCC1, "f", 238, 29),
	["Metalwario64"]   = SH_CredSprite(SPR_CCC1, "g", 265, 29),

	-- Roles
	["TitleLead"]   = SH_CredSprite(SPR_CCCT, "A", 576, 84), -- Lead designer
	["TitleLore"]   = SH_CredSprite(SPR_CCCT, "B", 576, 84), -- Story
	["TitleSpr"]    = SH_CredSprite(SPR_CCCT, "C", 231, 25), -- Main Sprites
	["TitleAdSpr"]  = SH_CredSprite(SPR_CCCT, "D", 339, 25), -- Additional Sprites
	["TitleAdArt"]  = SH_CredSprite(SPR_CCCT, "E", 264, 25), -- Additional Art
	["TitleSnd"]    = SH_CredSprite(SPR_CCCT, "F", 239, 25), -- Sound
	["TitleVox"]    = SH_CredSprite(SPR_CCCT, "G", 231, 25), -- Voice
	["TitleMap"]    = SH_CredSprite(SPR_CCCT, "H", 194, 25), -- Map
	["TitleTx"]     = SH_CredSprite(SPR_CCCT, "I", 249, 25), -- Texture
	["TitleBeta"]   = SH_CredSprite(SPR_CCCT, "J", 213, 25), -- Beta Testing
	["TitleCinema"] = SH_CredSprite(SPR_CCCT, "K", 396, 25), -- Cinematics
	["TitleProg"]   = SH_CredSprite(SPR_CCCT, "L", 249, 25), -- Programming
	["TitleExec"]   = SH_CredSprite(SPR_CCCT, "M", 576, 26), -- "Exec Ideas"
	["Presents"]    = SH_CredSprite(SPR_CCCT, "N", 275, 51), -- Silverhorn Dev Presents
	["Logo"]        = SH_CredSprite(SPR_CCCT, "O", 438, 146), -- Silverhorn Logo
	["TAG"]         = SH_CredSprite(SPR_CCCT, "P", 381, 186), -- TAG Logo
	["Combi"]       = SH_CredSprite(SPR_CCCT, "Q", 381, 332), -- TAG Logo
	["And"]         = SH_CredSprite(SPR_CCCT, "R", 53, 20), -- With inclusion
	["With"]        = SH_CredSprite(SPR_CCCT, "S", 58, 20), -- And inclusion

	-- ["SequenceD"] = {spr=SPR_CCCT, frame=R_Char2Frame("L")},
	-- ["Lune_Wulf"] = {spr=SPR_CCC2, frame=R_Char2Frame("K")},
}
-- rawset(_G, "SHCRDSPR", function(str)
local function SHCRDSPR(str)
    if credit_list[str] then
        return credit_list[str]
    else
    	print("Credit Sprite" + str + " not found!")
    end
end

function CreditMaps.floatToFixed(decimalString)
    local int, frac = decimalString:match("(%d+)%.(%d+)")

    if int and frac then
        local int = tonumber(int)
        local fracval = tonumber(frac)

        local fixedValue = int * FRACUNIT + (fracval * FRACUNIT / 10 ^ #frac)

        return fixedValue
    else
        print("Invalid decimal string: " .. decimalString)
        return 0
    end
end

-- TODO: split
local function find_lineref_mobj(mobjtype, angle) --stringarg, stringargnum)
	for mt in mapthings.iterate do
		-- create fake object until stringargs works...
		if (mt.type == mobjinfo[mobjtype].doomednum and mt.angle == angle)-- and mt.stringargs[stringargnum or 0] == stringarg)
		and (mt and mt.valid) then

			local lineref = {}
			local line_angle = 0

			-- Refer to special line for object settings because of limited options
			lineref = lines[P_FindSpecialLineFromTag(2022, mt.angle)]
			if lineref then
				line_angle = R_PointToAngle2(0, 0, lineref.dx, lineref.dy) or 0
			end

			return {mt=mt, mobj=mt.mobj, angle=line_angle}
		end
	end
end

-- Gets all objects with a stringarg in them on from the udmf enabled map
local function get_udmf_string_objects()
	local udmf_objects = {}
	for mt in mapthings.iterate do
		if (mt.stringargs and mt.stringargs[0] and string.len(mt.stringargs[0]) > 0) and (mt and mt.valid) then
			-- print("Added:" .. mt.stringargs[0])
			udmf_objects[mt.stringargs[0]] = {
				x = mt.mobj.x,
				y = mt.mobj.y,
				z = mt.mobj.z,
				angle = mt.mobj.angle,
				tag = mt.tag
			}
		end
	end
	return udmf_objects
end

-- TODO: replace
local function set_camera_angle(_camera, angle)
	_camera.angle = angle
end

local function set_object_sprite(mobj, spritenum, frame)
	mobj.sprite = spritenum
	mobj.frame = frame or $1
end

local function move_object_position(mobj, setposition, interpolate)
	local movefunction = interpolate and P_MoveOrigin or P_SetOrigin -- or P_TeleportMove
	movefunction(mobj, setposition.x, setposition.y, setposition.z)
end

local function move_object_relative(mobj, relmobj, offx, offy, offz, offang, interpolate)

	local warporigin = interpolate and P_MoveOrigin or P_SetOrigin -- or P_TeleportMove

	warporigin(mobj,
		relmobj.x + P_ReturnThrustX(relmobj.angle+(offang or 0), offy) + offx,
		relmobj.y + P_ReturnThrustY(relmobj.angle+(offang or 0), offy),
		relmobj.z + offz)

	-- old (For reference)
	-- 	c.camera_ref.x + P_ReturnThrustX(c.camera_ref.angle, c.offy)+c.offx,
	-- 	c.camera_ref.y + P_ReturnThrustY(c.camera_ref.angle, c.offy),
	-- 	c.camera_ref.z + c.offz)
end
rawset(_G, "find_lineref_mobj", find_lineref_mobj)
rawset(_G, "set_camera_angle", set_camera_angle)

EventSH.new("crd_emitter", {
function(c,e)
	c.emitter_time = 1
	c.seedx = {16, -64, -96, 32, -16, 96, 48, 64, 8, -32, -48, -8}
	c.seedy = {-32, -64, 32, 64, -48, 16, 96, -96, 48, -16, -8, 8}
	c.seedz = {-32, 8, 64, -64, -96, -8, 32, -16, -48, 48, 96, 16}
end,
function(c,e)
	for multiplier=1, 3 do
		-- TODO: effect that is the same every time it's seen
		-- local rx = c.seedx[c.emitter_time]*FU
		-- local ry = c.seedy[c.emitter_time]*FU
		-- local rz = c.seedz[c.emitter_time]*FU
		local rx = EventSH.RANDOMCHOICE({-96, -64, -48, -32, -16, -8, 8, 16, 32, 48, 64, 96}, true)*FU --P_RandomRange(-128, 128)*FU
		local ry = EventSH.RANDOMCHOICE({-96, -64, -48, -32, -16, -8, 8, 16, 32, 48, 64, 96}, true)*FU --P_RandomRange(-128, 128)*FU
		local rz = EventSH.RANDOMCHOICE({-96, -64, -48, -32, -16, -8, 8, 16, 32, 48, 64, 96}, true)*FU --P_RandomRange(-128, 128)*FU
		local f = P_SpawnMobj(c.emitter_mobj.x+rx, c.emitter_mobj.y+ry, c.emitter_mobj.z+rz, MT_INAZUMA_PARTICLE)
		f.tics = 15
		f.fuse = 15
		f.blendmode = AST_ADD
		f.color = SKINCOLOR_CERULEAN
	end
	if c.emitter_time >= #c.seedx then c.emitter_time = 1 else c.emitter_time = $1 + 1 end

	EventSH.wait(e, 6*FU)
end})








EventSH.new("sh_credit_title", {
function(c,e)
	------------------
	-- Setup title
	------------------
	c.fadein_counter = 10
	c.fadeout_counter = 0
	c.title_timer = 0
	EventSH.wait(e, c.startwaittime or 1, true)
end,
function(c,e)
	if (c.sticky) then move_object_relative(c.titlemo, c.camera_ref, c.offx, c.offy, c.offz, 0, false) end -- (2.2.11 interpolation patch)
end,
function(c,e)
	------------------
	-- Transition in
	------------------
	if not (c.title_timer) then
		c.titlemo.flags2 = $ &~ MF2_DONTDRAW
	end

	if (c.transition_in) then
		if c.no_fade_in then
			-- Instant fade-in
			c.titlemo.frame = c.titlemo_frame|(0<<FF_TRANSSHIFT)
		else
			-- Fade in
			if (c.title_timer % c.transition_in == 0) then
				c.fadein_counter = max(0, $1-1)
			end
			c.titlemo.frame = c.titlemo_frame | (c.fadein_counter << FF_TRANSSHIFT)
			EventSH.waitUntil(e, c.fadein_counter <= 0)
		end
	end

	c.title_timer = $1 + 1
	if (c.sticky) then move_object_relative(c.titlemo, c.camera_ref, c.offx, c.offy, c.offz, nil, true) end -- (2.2.11 interpolation patch)
end,
function(c,e)
	------------------
	-- Wait time between transition
	------------------
	if (c.sticky) then move_object_relative(c.titlemo, c.camera_ref, c.offx, c.offy, c.offz, nil, true) end -- (2.2.11 interpolation patch)
	EventSH.wait(e, c.delay or 2*TICRATE)
end,
function(c,e)
	------------------
	-- Transition out
	------------------
	if (c.transition_out) then
		-- Fade out
		if (c.title_timer % c.transition_out == 0) then
			c.c.fadeout_counter = min($1+1, 10)
		end
		c.titlemo.frame = c.titlemo_frame | (c.fadeout_counter << FF_TRANSSHIFT)

		EventSH.waitUntil(e, c.fadeout_counter >= 10)
	end
	c.title_timer = $1 + 1
	if (c.sticky) then move_object_relative(c.titlemo, c.camera_ref, c.offx, c.offy, c.offz, 0, true) end
end,
function(c,e)
	-- Remove title
	c.titlemo.fuse = 1
	c.titlemo.tics = 1
	c.titlemo.state = S_NULL -- (2.2.11 interpolation patch)
end,})




local function add_camera_title_offset(mocam, offx, offy, offz, spritedata, args)

	-- Container contents
	local con = {}

	con["camera_ref"] = mocam
	con["offx"] = offx -- Offset x
	con["offy"] = offy -- Offset y
	con["offz"] = offz -- Offset z

	-- Custom parameters
	con["startwaittime"] = args and args.startwait -- Time to wait before showing
	-- TODO: if you just disable transition in...does this still need to be kept?
	con["no_fade_in"] = args and args.nofadein -- Don't fade in for transition inward
	con["transition_in"] = args and args.intran -- Transition in time
	con["delay"] = args and args.delay -- Delay waiting time after transition inward
	con["transition_out"] = args and args.outtran -- Transition out time
	con["sticky"] = (args and args.sticky) and true or false -- Relatively follows the reference it's tied to

	-- Get sprite and frame
	local sprite = spritedata and spritedata.spr or SPR_UNKN
	local frame = spritedata and spritedata.frm or 0

	-- Full brightness for the title sprite
	local fullbright = (args and args.fullbright) and RF_FULLBRIGHT or 0

	-- Sprite offset alignment
	local ax = ((args and args.sprx) and not nil) and args.sprx or 0 --args and args.sprx or 0
	local ay = ((args and args.spry) and not nil) and args.spry or 0 --args and args.spry or 0

	-- Create object
	local titlemobj = P_SpawnMobj(
		mocam.x + P_ReturnThrustX(mocam.angle, offy) + offx,
		mocam.y + P_ReturnThrustY(mocam.angle, offy),
		mocam.z + offz,
		MT_THOK)

	titlemobj.scale = (args and args.scale) or FU
	titlemobj.sprite = sprite
	titlemobj.tics = FRACUNIT
	titlemobj.flags = $1|MF_NOCLIP|MF_NOCLIPHEIGHT
	titlemobj.flags2 = $ | MF2_DONTDRAW -- (2.2.11 interpolation patch)
	titlemobj.color = SKINCOLOR_WHITE

	-- Setup frame
	titlemobj.frame = frame|(con.transition_in and (10<<FF_TRANSSHIFT) or 0)
	titlemobj.spritexoffset = ax*FU
	titlemobj.spriteyoffset = ay*FU
	titlemobj.renderflags = $1|(fullbright)

	con["titlemo"] = titlemobj -- Title object
	con["titlemo_frame"] = frame -- Title object frame

	EventSH.start("sh_credit_title", con)
	return titlemobj
end

local function titleParam(starttime, nofadein, transitionin, delay, transitionout, sticky, fullbright, sprxoff, spryoff, scale)
	return {startwait=starttime, nofadein=nofadein, intran=transitionin, delay=delay, outtran=transitionout, sticky=sticky, fullbright=fullbright, sprx=sprxoff, spry=spryoff, scale=scale}
end


local function add_camtitle(container, name, mocam, offx, offy, offz, spritedata, args)
	container["credit_" .. name] = add_camera_title_offset(mocam, offx, offy, offz, spritedata, args)
end

-- Creates a preset parameter table for huditems
local function hudtitleParam(hname, crdsprname, hx, hy, hs)
	return {order=70, name=hname, x=hx, y=hy, scale=hs, sprite={SHCRDSPR(crdsprname).spr, SHCRDSPR(crdsprname).frm, 0, 0}}
end

-- Returns a center aligned patch accounting for scale
local function alignpatch(intwidth, scale, align)
	if align == "right" then
		return FixedDiv(FixedMul(FU*intwidth, scale) + FixedMul(FU*intwidth, scale), 2*FU)
	else
		return FixedDiv(FixedMul(FU*intwidth, scale), 2*FU)
	end
end

local function music_check_fast_forward(position, checkonly)
	if checkonly then
		print(S_GetMusicPosition())
	else
		-- if (S_GetMusicPosition() < position) then
			S_SetMusicPosition(position)
		-- end
	end
end


EventSH.new("crd_animtimer", {
function(c,e)
	EvActionSH.togglerender(c)
	c.displaytime = 0
end,
function(c,e)
	c.displaytime = $1 + 1
	local Hours = string.format("%02d", G_TicsToHours(c.displaytime))
	local Minutes = string.format("%02d", G_TicsToMinutes(c.displaytime))
	local Seconds = string.format("%02d", G_TicsToSeconds(c.displaytime))
	local Centiseconds = string.format("%02d", G_TicsToCentiseconds(c.displaytime))
	local Milliseconds = string.format("%02d", G_TicsToMilliseconds(c.displaytime))
	EvActionSH.rendertext(c, 1, "time", 320/2, 6, Hours..":"..Minutes..":"..Seconds..":"..Centiseconds, V_ALLOWLOWERCASE|V_MONOSPACE, "center")
	EventSH.wait(e, 99999999999)
end,
function(c,e)
end})

local function start_anim_timer()
	EventSH.start("crd_animtimer")
end

local function pause_anim_timer()
	EventSH.pause("crd_animtimer")
end

-- For simple credit showing
EventSH.new("crd_hudshow_sequence", {
function(c,e)
	EvActionSH.togglerender(c)
	EventSH.wait(e, c.waitingtime, true)
end,
function(c,e)
	for i, ttl in ipairs(c.titles) do
		EvActionSH.renderscaledpatch(c, ttl.order, ttl.name, ttl.x, ttl.y, ttl.scale, ttl.sprite, 0, SKINCOLOR_WHITE)
	end
end,
function(c,e)
	EventSH.wait(e, c.delay)
end})

local function show_hud_credit(titles, waitingtime, delay, hud_id, remove_ttl)
	if remove_ttl then
		if hud_id then
			EventSH.destroy(hud_id, {hasvariable=true})
		else
			EventSH.destroy("crd_hudshow_sequence")
		end
	else
		EventSH.start("crd_hudshow_sequence", {titles=titles, waitingtime=waitingtime, delay=delay})
	end
end


EventSH.new("sh_credits", {
function(c,e)

	--Check if we've seen the credits
	-- if MindscapeSessionFlags & MS_SEENCREDITS then
	-- 	Event.destroyself(e)
	-- 	Event.start("sh_credits_endcard")
	-- 	-- EventSH.start("sh_mscredits_skip")
	-- end

	-- Make camera
	c.camera = P_SpawnMobj(0, 0, 0, MT_PUSH)

	-- Check game things before continuing
	c.pos_cr_start = find_lineref_mobj(MT_TELEPORTMAN, 1)
	c.pos_cr_inright = find_lineref_mobj(MT_TELEPORTMAN, 2)
	c.pos_cr_aeleft = find_lineref_mobj(MT_TELEPORTMAN, 3)
	c.pos_cr_inaecenter = find_lineref_mobj(MT_TELEPORTMAN, 4)
	c.pos_cr_additart = find_lineref_mobj(MT_TELEPORTMAN, 5)
	-- c.pos_cr_moresound = find_lineref_mobj(MT_TELEPORTMAN, 6)
	c.pos_cr_maptex = find_lineref_mobj(MT_TELEPORTMAN, 7)
	c.pos_cr_TAG_move_a = find_lineref_mobj(MT_TELEPORTMAN, 8)
	c.pos_cr_TAG_move_b = find_lineref_mobj(MT_TELEPORTMAN, 9)
	c.pos_cr_TAG_trms_move_a = find_lineref_mobj(MT_TELEPORTMAN, 10)
	c.pos_cr_TAG_trms_move_b = find_lineref_mobj(MT_TELEPORTMAN, 11)
	c.pos_cr_TAG_msz_move_a = find_lineref_mobj(MT_TELEPORTMAN, 12)
	c.pos_cr_TAG_msz_move_b = find_lineref_mobj(MT_TELEPORTMAN, 13)
	c.pos_cr_inshad = find_lineref_mobj(MT_TELEPORTMAN, 14)
	c.pos_cr_aeshad = find_lineref_mobj(MT_TELEPORTMAN, 15)
	c.pos_cr_faccshad = find_lineref_mobj(MT_TELEPORTMAN, 16)
	c.pos_cr_fb_1 = find_lineref_mobj(MT_TELEPORTMAN, 17)
	c.pos_cr_fb_2 = find_lineref_mobj(MT_TELEPORTMAN, 18)
	c.pos_cr_fb_3 = find_lineref_mobj(MT_TELEPORTMAN, 19)
	c.pos_cr_fb_4 = find_lineref_mobj(MT_TELEPORTMAN, 20)
	c.pos_cr_ulti1 = find_lineref_mobj(MT_TELEPORTMAN, 22)
	c.pos_cr_ulti2 = find_lineref_mobj(MT_TELEPORTMAN, 23)
	c.pos_cr_title = find_lineref_mobj(MT_TELEPORTMAN, 24)
	c.pos_cr_title_f = find_lineref_mobj(MT_TELEPORTMAN, 21)

	c.pos_orbpath_start = find_lineref_mobj(MT_TELEPORTMAN, 390).mobj
	c.pos_orbpath_end = find_lineref_mobj(MT_TELEPORTMAN, 391).mobj
	c.pos_bigtitle = find_lineref_mobj(MT_TELEPORTMAN, 400).mobj


	c.pos_cr_char1_run_a = find_lineref_mobj(MT_TELEPORTMAN, 250).mobj
	c.pos_cr_char2_run_a = find_lineref_mobj(MT_TELEPORTMAN, 251).mobj
	c.pos_cr_char1_run_b = find_lineref_mobj(MT_TELEPORTMAN, 252).mobj
	c.pos_cr_char2_run_b = find_lineref_mobj(MT_TELEPORTMAN, 253).mobj

	c.pos_cr_char1_dance = find_lineref_mobj(MT_TELEPORTMAN, 260).mobj
	c.pos_cr_char2_dance = find_lineref_mobj(MT_TELEPORTMAN, 261).mobj

	c.pos_cr_char1_fb = find_lineref_mobj(MT_TELEPORTMAN, 340).mobj
	c.pos_cr_char2_fb = find_lineref_mobj(MT_TELEPORTMAN, 341).mobj
	c.pos_cr_char3_fb = find_lineref_mobj(MT_TELEPORTMAN, 342).mobj

	-- 10-27-2023: Cache udmf objects instead
	c.UDMFcache = get_udmf_string_objects()

	c.character_in = P_SpawnMobj(0, 0, 0, MT_PUSH)
	c.character_in.sprite = SPR_PLAY
	c.character_in.skin = "inazuma"
	c.character_in.radius = 4*FU
	c.character_in.shadowscale = 6*FU
	c.character_in.color = SKINCOLOR_CERULEAN
	c.character_ae = P_SpawnMobj(0, 0, 0, MT_PUSH)
	c.character_ae.sprite = SPR_PLAY
	c.character_ae.skin = "aether"
	c.character_ae.radius = 4*FU
	c.character_ae.shadowscale = 4*FU
	c.character_ae.color = SKINCOLOR_AETHER
	c.character_fc = P_SpawnMobj(0, 0, 0, MT_PUSH)
	c.character_fc.sprite = SPR_CCC2
	c.character_fc.radius = 4*FU
	c.character_fc.shadowscale = 4*FU

	c.sil_spr = SPR_CRRS
	EvActionSH.Sq_CreateMobjActor(c, "ae_silh", MT_PUSH, {x=0,y=0,z=0}, function(ma)
		ma.sprite = c.sil_spr
		ma.frame = R_Char2Frame("A")
	end)
	EvActionSH.Sq_CreateMobjActor(c, "in_silh", MT_PUSH, {x=0,y=0,z=0}, function(ma)
		ma.sprite = c.sil_spr
		ma.frame = R_Char2Frame("B")
	end)
	EvActionSH.Sq_CreateMobjActor(c, "fx_silh", MT_PUSH, {x=0,y=0,z=0}, function(ma)
		ma.sprite = c.sil_spr
		ma.frame = R_Char2Frame("C")
	end)
	EvActionSH.movemobjto(c.m_ae_silh, c.UDMFcache["pos_aether_silh"])
	EvActionSH.movemobjto(c.m_in_silh, c.UDMFcache["pos_ina_silh"])
	EvActionSH.movemobjto(c.m_fx_silh, c.UDMFcache["pos_aether_light"])
end,
function(c,e)
	EvActionSH.setcamera(true, c.camera, c.pos_cr_start.mobj, c.pos_cr_start.angle, 0, true)

	S_ChangeMusic("INACRD", false)
end,
-- ======================================
-- Starting Point
-- ======================================
function(c,e)
	-- PROJECT BY:
	-- ---------------------------
	add_camtitle(c, "titlelead", c.camera, 0, 256*FU, 0, SHCRDSPR("TitleLead"), titleParam(0,false,nil,3*TICRATE+32,nil,false,true,SHCRDSPR("TitleLead").w/2,SHCRDSPR("TitleLead").h,FU/2))
	MobjMoverSH.moveto("move_title", c.credit_titlelead, c.camera, {speed=FU/4, noangle=true})

end,
function(c,e)
	-- EventSH.wait(e, 99*TICRATE + 20) -- Testing
	EventSH.wait(e, 2*TICRATE + 20)
end,
-- ======================================
-- Follow Path
-- ======================================
function(c,e)
	-- set_object_sprite(c.crd_bigtitle, SPR_NULL)

	c.orb = P_SpawnMobj(c.pos_orbpath_start.x, c.pos_orbpath_start.y, c.pos_orbpath_start.z, MT_THOK)
	c.orb.sprite = SPR_NULL
	c.orb.tics = FRACUNIT
	EventSH.start("crd_emitter", {emitter_mobj=c.orb})

	MobjMoverSH.moveto("move_orb", c.orb, c.pos_orbpath_end, {speed=20*FU, ease="inoutsine", harc=270*FU, varc=-96*FU})
end,
function(c,e)
	EventSH.wait(e, 1*TICRATE + 15)
end,
function(c,e)
	-- MAIN SPRITEWORK:
	-- ---------------------------
	add_camtitle(c, "lore", c.camera, 0, 256*FU, 0, SHCRDSPR("TitleLore"), titleParam(0,false,nil,4*TICRATE,nil,false,true,SHCRDSPR("TitleLore").w/2,SHCRDSPR("TitleLore").h,FU/2))
	MobjMoverSH.moveto("move_title", c.credit_lore, c.camera, {speed=FU/6, noangle=true})
end,
function(c,e)
	EventSH.wait(e, 2*TICRATE + 20)
	-- EventSH.wait(e, 3*TICRATE + 20)
end,
-- ======================================
-- Running Right
-- ======================================
function(c,e)
	MobjMoverSH.stop("move_title")
	music_check_fast_forward(6673)
end,
function(c,e)
	move_object_position(c.character_in, c.pos_cr_char1_run_a)
	move_object_position(c.character_ae, c.pos_cr_char2_run_a)

	c.character_in.angle = FixedAngle(225*FU)
	c.character_in.angle = $
	FrameAnimSH.addspr2("in", c.character_in, SPR2_RUN_, "A", nil, {loop=32*TICRATE, delay=2}, true)
end,
function(c,e)
	EvActionSH.setcamera(true, c.camera, c.pos_cr_inright.mobj, c.pos_cr_inright.angle, 0, true)

	-- MAIN SPRITEWORK:
	-- ---------------------------
	add_camtitle(c, "mainsprite", c.camera, -220*FU, 256*FU, 96*FU, SHCRDSPR("TitleSpr"), titleParam(1*TICRATE,true,1,10*TICRATE,nil,false,true,0,21,FU/2 - FU/16)) -- FU/2 + FU/8
	-- add_camtitle(c, "mainsprite", c.camera, -150*FU, 256*FU, 96*FU, SHCRDSPR("TitleSpr"), titleParam(1*TICRATE,true,1,10*TICRATE,nil,false,true,115,21,FU/3 + FU/5)) -- FU/2 + FU/8

	add_camtitle(c, "credit1", c.camera, -220*FU, 256*FU, 96*FU-(FU*25), SHCRDSPR("Inazuma"), titleParam(2*TICRATE,false,1,10*TICRATE,nil,false,true,0,0,FU/2))
	add_camtitle(c, "credit2", c.camera, -220*FU, 256*FU, 96*FU-(FU*25)*2, SHCRDSPR("DrStephen"), titleParam(3*TICRATE,false,1,10*TICRATE,nil,false,true,0,0,FU/2))
	add_camtitle(c, "credit3", c.camera, -220*FU, 256*FU, 96*FU-(FU*25)*3, SHCRDSPR("Dirkthehusky"), titleParam(4*TICRATE,false,1,10*TICRATE,nil,false,true,0,0,FU/2))

	P_LinedefExecute(8000)
end,
function(c,e)
	EventSH.wait(e, 9*TICRATE)
end,
-- ----------------------
-- Running Left
-- ----------------------
function(c,e)
	MobjMoverSH.moveto("move_camera", c.camera, c.pos_cr_aeleft.mobj, {speed=32*FU, ease="inoutquad", noangle=true})

	c.character_ae.angle = FixedAngle(225*FU)+FixedAngle(90*FU)
	FrameAnimSH.addspr2("ae", c.character_ae, SPR2_RUN_, "A", nil, {loop=32*TICRATE, delay=2}, true)

end,
function(c,e)
	MobjMoverSH.waitmovedone(e, "move_camera")
end,
function(c,e)

	-- ADDITIONAL SPRITEWORK:
	-- ---------------------------
	add_camtitle(c, "addsprite", c.camera, 128*FU, 256*FU, 96*FU, SHCRDSPR("TitleAdSpr"), titleParam(1*TICRATE,true,1,10*TICRATE,nil,false,true,SHCRDSPR("TitleAdSpr").w-230,0,FU/2 - FU/10)) --FU/2 + FU/8

	add_camtitle(c, "sprcredit1", c.camera, 128*FU, 256*FU, 96*FU-(FU*16), SHCRDSPR("Lightdasher"), titleParam(2*TICRATE,false,1,10*TICRATE,nil,false,true,SHCRDSPR("Lightdasher").w-220,0,FU/3 + FU/9))
	add_camtitle(c, "sprcredit2", c.camera, 128*FU, 256*FU, 96*FU-(FU*16)*2, SHCRDSPR("Starlight"), titleParam(2*TICRATE+10,false,1,10*TICRATE,nil,false,true,SHCRDSPR("Starlight").w-220,0,FU/3 + FU/9))
	add_camtitle(c, "sprcredit3", c.camera, 128*FU, 256*FU, 96*FU-(FU*16)*3, SHCRDSPR("Telosturntable"), titleParam(2*TICRATE+20,false,1,10*TICRATE,nil,false,true,SHCRDSPR("Telosturntable").w-220,0,FU/3 + FU/9))
	add_camtitle(c, "sprcredit4", c.camera, 128*FU, 256*FU, 96*FU-(FU*16)*4, SHCRDSPR("Machturne"), titleParam(2*TICRATE+30,false,1,10*TICRATE,nil,false,true,SHCRDSPR("Machturne").w-220,0,FU/3 + FU/9))
	add_camtitle(c, "sprcredit5", c.camera, 128*FU, 256*FU, 96*FU-(FU*16)*5, SHCRDSPR("Motorroach"), titleParam(2*TICRATE+40,false,1,10*TICRATE,nil,false,true,SHCRDSPR("Motorroach").w-220,0,FU/3 + FU/9))
	add_camtitle(c, "sprcredit6", c.camera, 128*FU, 256*FU, 96*FU-(FU*16)*6, SHCRDSPR("Flameeliwood"), titleParam(2*TICRATE+50,false,1,10*TICRATE,nil,false,true,SHCRDSPR("Flameeliwood").w-220,0,FU/3 + FU/9))
	add_camtitle(c, "sprcredit7", c.camera, 128*FU, 256*FU, 96*FU-(FU*16)*7, SHCRDSPR("Cystallinegazer"), titleParam(2*TICRATE+60,false,1,10*TICRATE,nil,false,true,SHCRDSPR("Cystallinegazer").w-220,0,FU/3 + FU/9))
	add_camtitle(c, "sprcredit8", c.camera, 128*FU, 256*FU, 96*FU-(FU*16)*8, SHCRDSPR("Saneko"), titleParam(2*TICRATE+70,false,1,10*TICRATE,nil,false,true,SHCRDSPR("Saneko").w-220,0,FU/3 + FU/9))
	add_camtitle(c, "sprcredit9", c.camera, 128*FU, 256*FU, 96*FU-(FU*16)*9, SHCRDSPR("Hyperchaotix"), titleParam(2*TICRATE+80,false,1,10*TICRATE,nil,false,true,SHCRDSPR("Hyperchaotix").w-220,0,FU/3 + FU/9))
	add_camtitle(c, "sprcredit9_1", c.camera, 128*FU, 256*FU, 96*FU-(FU*16)*10, SHCRDSPR("Delpho"), titleParam(2*TICRATE+90,false,1,10*TICRATE,nil,false,true,SHCRDSPR("Delpho").w-220,0,FU/3 + FU/9))
	add_camtitle(c, "sprcredit9_2", c.camera, 128*FU, 256*FU, 96*FU-(FU*16)*11, SHCRDSPR("FlyingNosaj"), titleParam(2*TICRATE+100,false,1,10*TICRATE,nil,false,true,SHCRDSPR("FlyingNosaj").w-220,0,FU/3 + FU/9))

	P_LinedefExecute(8002)
end,
function(c,e)
	EventSH.wait(e, 8*TICRATE)
end,
-- ----------------------
-- Running Center
-- ----------------------
function(c,e)
	move_object_position(c.character_ae, c.pos_cr_char1_run_b)
	move_object_position(c.character_in, c.pos_cr_char2_run_b)
	c.character_in.angle = FixedAngle(330*FU)
	c.character_in.angle = $
	c.character_ae.angle = FixedAngle(190*FU)
	c.character_ae.angle = $

	EvActionSH.setcamera(true, c.camera, c.pos_cr_inaecenter.mobj, c.pos_cr_inaecenter.angle, 0, true)

	c.useogzoom = false

	-- zoom
	if c.useogzoom then
		MobjMoverSH.moveto("move_camera", c.camera, {
			x=c.pos_cr_inaecenter.mobj.x,
			y=c.pos_cr_inaecenter.mobj.y+512*FU,
			z=c.pos_cr_inaecenter.mobj.z}, {speed=FU, ease="inquad", noangle=true})
	end
	P_LinedefExecute(7990)
	c.momstart = 0
	c.momend = 0
	c.intime = 0
end,
function(c,e)
	if not c.useogzoom then
		c.intime = $1 + 1
		if c.intime < 4*TICRATE+15 then
			c.camera.momy = min($1 + CreditMaps.floatToFixed("0.45"), FU)
		else
			c.momend = $1 + CreditMaps.floatToFixed("0.06")
			c.camera.momy = $1 + c.momend
		end
	end
	EventSH.wait(e, 6*TICRATE + 5)
end,
-- ======================================
-- Addit Artwork
-- ======================================
function(c,e)
	MobjMoverSH.stop("move_camera")

	EvActionSH.setcamera(true, c.camera, c.pos_cr_additart.mobj, c.pos_cr_additart.angle, 0, true)
	c.off = 0
	c.offz = 0
	-- MobjMoverSH.moveto("move_camera", c.camera, {
	-- 	x=c.pos_cr_additart.mobj.x+800*FU,
	-- 	y=c.pos_cr_additart.mobj.y+800*FU,
	-- 	z=c.pos_cr_additart.mobj.z}, {speed=1*FU, arc={horz=1500*FU}, noangle=true})
end,
function(c,e)
	-- ADDITIONAL ARTWORK:
	-- ---------------------------
	local tt_artwork_1_start = 1*TICRATE+15
	local tt_artwork_1_length = 4*TICRATE
	local tt_artwork_x = (FU*320/2)+128*FU
	local tt_artwork_y = (FU*4)
	local tt_artwork_scl = CreditMaps.floatToFixed("0.25")
	local tt_artwork_subscl = CreditMaps.floatToFixed("0.30")
	local tt_artwork_space = (FU*8)

	local tt_artwork_2_start = 6*TICRATE
	local tt_artwork_2_length = 4*TICRATE+15
	show_hud_credit({
		hudtitleParam("hud0", "TitleAdArt", tt_artwork_x - alignpatch(SHCRDSPR("TitleAdArt").w, tt_artwork_scl, "right"), (tt_artwork_y + tt_artwork_space)*1, tt_artwork_scl),
	}, tt_artwork_1_start, tt_artwork_1_length+tt_artwork_2_length+18)

	show_hud_credit({
		hudtitleParam("hud1", "Inazuma", tt_artwork_x - alignpatch(SHCRDSPR("Inazuma").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*2, tt_artwork_subscl),
		hudtitleParam("hud2", "Delpho", tt_artwork_x - alignpatch(SHCRDSPR("Delpho").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*3, tt_artwork_subscl),
		hudtitleParam("hud3", "Alacroix", tt_artwork_x - alignpatch(SHCRDSPR("Alacroix").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*4, tt_artwork_subscl),
		hudtitleParam("hud4", "Flameeliwood", tt_artwork_x - alignpatch(SHCRDSPR("Flameeliwood").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*5, tt_artwork_subscl),
		hudtitleParam("hud5", "DrStephen", tt_artwork_x - alignpatch(SHCRDSPR("DrStephen").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*6, tt_artwork_subscl),
		hudtitleParam("hud6", "Spinslash165", tt_artwork_x - alignpatch(SHCRDSPR("Spinslash165").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*7, tt_artwork_subscl),
		hudtitleParam("hud7", "Hattyboyo", tt_artwork_x - alignpatch(SHCRDSPR("Hattyboyo").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*8, tt_artwork_subscl),

	}, tt_artwork_1_start, tt_artwork_1_length)

	show_hud_credit({
		hudtitleParam("hud1", "Saneko", tt_artwork_x - alignpatch(SHCRDSPR("Saneko").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*2, tt_artwork_subscl),
		hudtitleParam("hud2", "Soapsurfin", tt_artwork_x - alignpatch(SHCRDSPR("Soapsurfin").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*3, tt_artwork_subscl),
		hudtitleParam("hud3", "Starlight", tt_artwork_x - alignpatch(SHCRDSPR("Starlight").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*4, tt_artwork_subscl),
		hudtitleParam("hud4", "Blueblur", tt_artwork_x - alignpatch(SHCRDSPR("Blueblur").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*5, tt_artwork_subscl),
		hudtitleParam("hud5", "Lightdasher", tt_artwork_x - alignpatch(SHCRDSPR("Lightdasher").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*6, tt_artwork_subscl),
		hudtitleParam("hud6", "Dirkthehusky", tt_artwork_x - alignpatch(SHCRDSPR("Dirkthehusky").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*7, tt_artwork_subscl),
		hudtitleParam("hud7", "Ralphjeremy65", tt_artwork_x - alignpatch(SHCRDSPR("Ralphjeremy65").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*8, tt_artwork_subscl),
		hudtitleParam("hud8", "Metalwario64", tt_artwork_x - alignpatch(SHCRDSPR("Metalwario64").w, tt_artwork_subscl, "right"), (tt_artwork_y + tt_artwork_space)*9, tt_artwork_subscl),

	}, tt_artwork_2_start, tt_artwork_2_length)

	P_LinedefExecute(8004)
	P_LinedefExecute(5050)
end,
function(c,e)
	c.off = $1+((FU)-(FU/3)-(FU/10)) -- floats would really be cool too
	c.offz = $1-FU/2
	move_object_relative(c.camera, {x=4096*FU,y=-2944*FU,z=256*FU,angle=0}, 0, 720*FU, c.offz, FixedAngle(225*FU)+FixedAngle(c.off))
	c.camera.angle = R_PointToAngle2(c.camera.x, c.camera.y, 4096*FU, -2944*FU)+FixedAngle(-30*FU)
	EventSH.wait(e, 11*TICRATE + 8)
end,
-- ======================================
-- More Sounds
-- ======================================
function(c,e)
	music_check_fast_forward(42369)

	move_object_position(c.character_in, c.pos_cr_char1_dance)
	move_object_position(c.character_ae, c.pos_cr_char2_dance)
	FrameAnimSH.addspr2("in", c.character_in, SPR2_DNCE, "A", "H", {loop=15*TICRATE, delay=4})
	FrameAnimSH.addspr2("ae", c.character_ae, SPR2_DNCE, "A", nil, {loop=15*TICRATE, delay=4})

	EvActionSH.setcamera(true, c.camera, c.UDMFcache["pos_snd_view1"], c.UDMFcache["pos_snd_view1"].angle, 0, true)
	-- EvActionSH.setcamera(true, c.camera, c.pos_cr_moresound.mobj, c.pos_cr_moresound.angle, 0, true)

	P_LinedefExecute(7990)
end,
function(c,e)
	-- SOUND TEST & VOICE:
	-- ---------------------------
	local tt_sound_start = TICRATE/2
	local tt_sound_length = 4*TICRATE
	local tt_sound_x = (FU*320/2)
	local tt_sound_y = (FU*4)
	local tt_sound_scl = FU/3
	local tt_sound_subscl = FU/3
	local tt_sound_space = (FU*8)

	local tt_vox_start = tt_sound_length+25
	local tt_vox_length = 3*TICRATE+15
	show_hud_credit({
		hudtitleParam("hud0", "TitleSnd", tt_sound_x - alignpatch(SHCRDSPR("TitleSnd").w, tt_sound_scl), (tt_sound_y + tt_sound_space)*1, tt_sound_scl),
		hudtitleParam("hud1", "Megabaz", tt_sound_x - alignpatch(SHCRDSPR("Megabaz").w, tt_sound_subscl), (tt_sound_y + tt_sound_space)*2, tt_sound_subscl),
		hudtitleParam("hud2", "Krabs", tt_sound_x - alignpatch(SHCRDSPR("Krabs").w, tt_sound_subscl), (tt_sound_y + tt_sound_space)*3, tt_sound_subscl),
		hudtitleParam("hud3", "Vadapega", tt_sound_x - alignpatch(SHCRDSPR("Vadapega").w, tt_sound_subscl), (tt_sound_y + tt_sound_space)*4, tt_sound_subscl)
	}, tt_sound_start, tt_sound_length)

	show_hud_credit({
		hudtitleParam("hud0", "TitleVox", tt_sound_x - alignpatch(SHCRDSPR("TitleVox").w, FU/3), (tt_sound_y + tt_sound_space)*1, FU/3),
		hudtitleParam("hud1", "Awesomedubs", tt_sound_x - alignpatch(SHCRDSPR("Awesomedubs").w, FU/3), (tt_sound_y + tt_sound_space)*2, FU/3),
		hudtitleParam("hud2", "Kameasinger", tt_sound_x - alignpatch(SHCRDSPR("Kameasinger").w, FU/3), (tt_sound_y + tt_sound_space)*3, FU/3)
	}, tt_vox_start, tt_vox_length)
end,
function(c,e)
	EventSH.wait(e, 1*TICRATE)
end,
function(c,e)
	-- 10 25 2023
	MobjMoverSH.moveto("move_camera", c.camera, {
		x=c.UDMFcache["pos_snd_view2"].x,
		y=c.UDMFcache["pos_snd_view2"].y,
		z=c.UDMFcache["pos_snd_view2"].z}, {speed=4*FU+FU/4, noangle=true, harc=-512*FU, locked=true, ease="inoutquart"})

end,
function(c,e)
	local viewcenter = {x=c.character_ae.x-96*FU, y=c.character_ae.y, z=c.character_ae.z}
	-- EvActionSH.cameratrack(true, c.camera, viewcenter)
	EvActionSH.setmobjeangle(c.camera, EvActionSH.getlookangle(c.camera, viewcenter))
	EventSH.wait(e, 7*TICRATE+10)
end,
-- ======================================
-- Mapping Textures
-- ======================================
function(c,e)
	MobjMoverSH.stop("move_camera")
	EvActionSH.setcamera(true, c.camera, c.pos_cr_maptex.mobj, c.pos_cr_maptex.angle, 0, true)
end,
function(c,e)
	-- MAP:
	-- ---------------------------
	local maptitle_z = 12*FU
	local mapcredit_z = 6*FU
	local credoffset = 4*FU
	add_camtitle(c, "map", c.camera, -75*FU, 79*FU, maptitle_z, SHCRDSPR("TitleMap"), titleParam(TICRATE/2,true,1,3*TICRATE-14,nil,true,true,0,0, FU/8))

	add_camtitle(c, "mpcredit1", c.camera, -75*FU, 79*FU, mapcredit_z-(credoffset*1), SHCRDSPR("Inazuma"), titleParam(TICRATE/2,true,1,3*TICRATE-14,nil,true,true,0,0, FU/7))
	add_camtitle(c, "mpcredit2", c.camera, -75*FU, 79*FU, mapcredit_z-(credoffset*2), SHCRDSPR("So2ro"), titleParam(TICRATE/2,true,1,3*TICRATE-14,nil,true,true,0,0, FU/7))
	add_camtitle(c, "mpcredit3", c.camera, -75*FU, 79*FU, mapcredit_z-(credoffset*3), SHCRDSPR("Kanna"), titleParam(TICRATE/2,true,1,3*TICRATE-14,nil,true,true,0,0, FU/7))
	add_camtitle(c, "mpcredit3", c.camera, -75*FU, 79*FU, mapcredit_z-(credoffset*4), SHCRDSPR("Othius"), titleParam(TICRATE/2,true,1,3*TICRATE-14,nil,true,true,0,0, FU/7))
	add_camtitle(c, "mpcredit3", c.camera, -75*FU, 79*FU, mapcredit_z-(credoffset*5), SHCRDSPR("OrdoMandalore"), titleParam(TICRATE/2,true,1,3*TICRATE-14,nil,true,true,0,0, FU/7))

	-- TEXTURE:
	-- ---------------------------
	add_camtitle(c, "map", c.camera, -75*FU, 79*FU, mapcredit_z-(credoffset*4), SHCRDSPR("TitleTx"), titleParam(3*TICRATE+10,true,1,13*TICRATE,nil,false,true,0,0, FU/8))

	add_camtitle(c, "texcredit", c.camera, -75*FU, 79*FU, mapcredit_z-(credoffset*5), SHCRDSPR("Spectorious"), titleParam(3*TICRATE+10,true,1,13*TICRATE,nil,false,true,0,0, FU/7))

	P_LinedefExecute(8006)
end,
function(c,e)
	EventSH.wait(e, 7*TICRATE + 10)
end,
-- ======================================
-- Mindscape Portal Discovery
-- ======================================
function(c,e)
	music_check_fast_forward(58166)

	EvActionSH.setcamera(true, c.camera, c.pos_cr_TAG_move_a.mobj, c.pos_cr_TAG_move_a.angle, 0, true)
end,
function(c,e)
	c.portal = P_SpawnMobj(c.pos_cr_TAG_move_b.mobj.x, c.pos_cr_TAG_move_b.mobj.y, c.pos_cr_TAG_move_b.mobj.z, MT_FACCIOLO2_SPACEPORTAL)
	c.portal.scale = 0
	c.portal.colorized = true
	c.portal.color = SKINCOLOR_LIME
	c.portal.facc = 1
	-- c.portal.angle = ANGLE_90
	-- c.portal.renderflags = RF_PAPERSPRITE
	c.portal.spriteyoffset = -45*FU
	c.portal.scalespeed = FU/128
	c.portal.destscale = 1*FU
end,
function(c,e)
	-- BETA:
	-- ---------------------------
	add_camtitle(c, "beta", c.camera, 60*FU, 64*FU, 55*FU, SHCRDSPR("TitleBeta"), titleParam(TICRATE/2,true,1,4*TICRATE-10,nil,true,true,SHCRDSPR("TitleBeta").w,0, FU/10))

	add_camtitle(c, "betaTAG", c.camera, 60*FU, 64*FU, 55*FU-(FU*4)*1, SHCRDSPR("TAG"), titleParam(TICRATE/2,true,1,4*TICRATE-10,nil,true,true,SHCRDSPR("TAG").w,0, FU/12))
	add_camtitle(c, "betaTAG", c.camera, 60*FU, 64*FU, 55*FU-(FU*4)*5, SHCRDSPR("Combi"), titleParam(TICRATE/2,true,1,4*TICRATE-10,nil,true,true,SHCRDSPR("Combi").w,0, FU/12))
	add_camtitle(c, "betaTAG", c.camera, 60*FU, 64*FU, 55*FU-(FU*7)*7, SHCRDSPR("Cyanknight"), titleParam(TICRATE/2,true,1,4*TICRATE-10,nil,true,true,SHCRDSPR("Cyanknight").w,0, FU/9))
	add_camtitle(c, "betaTAG", c.camera, 60*FU, 64*FU, 55*FU-(FU*7)*8, SHCRDSPR("Furless"), titleParam(TICRATE/2,true,1,4*TICRATE-10,nil,true,true,SHCRDSPR("Furless").w,0, FU/9))

	EvActionSH.fadescreen("out", 0, 2, 4*TICRATE+10)
end,
function(c,e)
	MobjMoverSH.moveto("move_camera", c.camera, c.pos_cr_TAG_move_b.mobj, {speed=5*FU, ease="inoutquad", noangle=true})
	EventSH.wait(e, 1*TICRATE, true)
end,
function(c,e)
	EventSH.wait(e, 4*TICRATE)
end,
-- ----------------------
-- Warp tunnel
-- ----------------------
function(c,e)
	MobjMoverSH.stop("move_camera")

	EvActionSH.setcamera(true, c.camera, c.pos_cr_TAG_trms_move_a.mobj, c.pos_cr_TAG_trms_move_a.angle, 0, true)
	for p in players.iterate do
		p.viewrollangle = ANGLE_90
	end
	c.camera.angle = ANGLE_90
end,
function(c,e)
	-- SEQUENCE:
	-- ---------------------------
	local tt_cinema_x = FU*(320/2)
	local tt_cinema_y = FU*(200/2) + 25*FU
	local tt_cinema_scl = FU/4
	local tt_cinema_sub_scl = FU/3
	local tt_cinema_space = (FU*12)
	show_hud_credit({
		hudtitleParam("hud0", "TitleCinema", tt_cinema_x - alignpatch(SHCRDSPR("TitleCinema").w, tt_cinema_scl, "center"), (tt_cinema_y + tt_cinema_space), tt_cinema_scl),
		hudtitleParam("hud1", "Lune_Wulf", tt_cinema_x - alignpatch(SHCRDSPR("Lune_Wulf").w, tt_cinema_sub_scl, "center"), (tt_cinema_y) + tt_cinema_space*2, tt_cinema_sub_scl),
		hudtitleParam("hud2", "Inazuma", tt_cinema_x - alignpatch(SHCRDSPR("Inazuma").w, tt_cinema_sub_scl, "center"), (tt_cinema_y) + tt_cinema_space*3, tt_cinema_sub_scl),
	}, 2*TICRATE, 3*TICRATE)

	EvActionSH.fadescreen("in", 0, 2, 28)
end,
function(c,e)
	MobjMoverSH.moveto("move_camera", c.camera, c.pos_cr_TAG_trms_move_b.mobj, {speed=12*FU, ease="outquad", noangle=true})
end,
function(c,e)
	EventSH.wait(e, 3*TICRATE)
end,
function(c,e)
	EvActionSH.fadescreen("out", 0, 1, 3*TICRATE)
end,
function(c,e)
	if (AngleFixed(c.camera.angle) <= 180*FU) then c.camera.angle = $1+FixedAngle(FU/3) end

	for p in players.iterate do
		p.viewrollangle = max(1, $1-FixedAngle(FU/3))
	end
	-- EventSH.wait(e, 5*TICRATE)
	EventSH.wait(e, 4*TICRATE)
end,
function(c,e)
-- ----------------------
-- Arena
-- ----------------------
	MobjMoverSH.stop("move_camera")
	for p in players.iterate do
		p.viewrollangle = 0
	end

	EvActionSH.setcamera(true, c.camera, c.pos_cr_TAG_msz_move_a.mobj, c.pos_cr_TAG_msz_move_a.angle, 0, true)
	EvActionSH.fadescreen("in", 0, 1, 15)
end,
function(c,e)
	MobjMoverSH.moveto("move_camera", c.camera, c.pos_cr_TAG_msz_move_b.mobj, {speed=15*FU, ease="inoutquad", varc=-32*FU})
end,
function(c,e)
	MobjMoverSH.waitmovedone(e, "move_camera")
end,
function(c,e)
	EventSH.wait(e, 1*TICRATE-10)
end,
-- ----------------------
-- Shadow
-- ----------------------
function(c,e)
	music_check_fast_forward(74809)
	-- start_anim_timer()
	EvActionSH.setcamera(true, c.camera, c.pos_cr_inshad.mobj, c.pos_cr_inshad.angle, 0, true)
end,
function(c,e)
	-- PROGRAM 1:
	-- ---------------------------
-- 	add_camtitle(c, "program", c.camera, 25*FU, 64*FU, 16*FU, SHCRDSPR("TitleProg"), titleParam(2,true,1,47,nil,true,true,0,0, FU/9))

-- 	add_camtitle(c, "progcredit1", c.camera, 25*FU, 64*FU, 16*FU-(FU*4)*2, SHCRDSPR("Lat"), titleParam(2,true,1,47,nil,true,true,0,0, FU/7))
-- 	add_camtitle(c, "progcredit2", c.camera, 25*FU, 64*FU, 16*FU-(FU*4)*3, SHCRDSPR("Krabs"), titleParam(2,true,1,47,nil,true,true,0,0, FU/7))
-- 	add_camtitle(c, "progcredit3", c.camera, 25*FU, 64*FU, 16*FU-(FU*4)*4, SHCRDSPR("Tatsuru"), titleParam(2,true,1,47,nil,true,true,0,0, FU/7))
-- 	add_camtitle(c, "progcredit3_1", c.camera, 25*FU, 64*FU, 16*FU-(FU*4)*5, SHCRDSPR("Amperbee"), titleParam(2,true,1,47,nil,true,true,0,0, FU/7))

	-- Redo
	local tt_prog_x = FU*(320/2)
	local tt_prog_y = FU*(200/2) + 15*FU
	local tt_prog_scl = FU/4
	local tt_prog_sub_scl = FU/3
	local tt_prog_space = (FU*10)
	show_hud_credit({
		hudtitleParam("hud0", "TitleProg", tt_prog_x + 64*FU, (tt_prog_y + tt_prog_space), tt_prog_scl),
		hudtitleParam("hud1", "Lat", tt_prog_x + 64*FU, (tt_prog_y) + tt_prog_space*2, tt_prog_sub_scl),
		hudtitleParam("hud2", "Krabs", tt_prog_x + 64*FU, (tt_prog_y) + tt_prog_space*3, tt_prog_sub_scl),
		hudtitleParam("hud3", "Tatsuru", tt_prog_x + 64*FU, (tt_prog_y) + tt_prog_space*4, tt_prog_sub_scl),
		hudtitleParam("hud3_1", "Amperbee", tt_prog_x + 64*FU, (tt_prog_y) + tt_prog_space*5, tt_prog_sub_scl),
	}, 2, 52)

	-- PROGRAM 2:
	-- ---------------------------
-- 	add_camtitle(c, "progcredit4", c.camera, -60*FU, 64*FU, 16*FU-(FU*4)*2, SHCRDSPR("Anonymouscoder"), titleParam(2*TICRATE,true,1,60,nil,true,true,0,0, FU/7))
-- 	add_camtitle(c, "progcredit5", c.camera, -60*FU, 64*FU, 16*FU-(FU*4)*3, SHCRDSPR("Logan8r"), titleParam(2*TICRATE,true,1,60,nil,true,true,0,0, FU/7))
-- 	add_camtitle(c, "progcredit6", c.camera, -60*FU, 64*FU, 16*FU-(FU*4)*4, SHCRDSPR("Frostiikin"), titleParam(2*TICRATE,true,1,60,nil,true,true,0,0, FU/7))
-- 	add_camtitle(c, "progcredit6_1", c.camera, -60*FU, 64*FU, 16*FU-(FU*4)*5, SHCRDSPR("Snu"), titleParam(2*TICRATE,true,1,60,nil,true,true,0,0, FU/7))

	show_hud_credit({
		hudtitleParam("hud0", "Anonymouscoder", tt_prog_x - 140*FU, (tt_prog_y + tt_prog_space), tt_prog_sub_scl),
		hudtitleParam("hud1", "Logan8r", tt_prog_x - 140*FU, (tt_prog_y) + tt_prog_space*2, tt_prog_sub_scl),
		hudtitleParam("hud2", "Frostiikin", tt_prog_x - 140*FU, (tt_prog_y) + tt_prog_space*3, tt_prog_sub_scl),
		hudtitleParam("hud3", "Snu", tt_prog_x - 140*FU, (tt_prog_y) + tt_prog_space*4, tt_prog_sub_scl),
	}, 2*TICRATE+12, 60)

end,
function(c,e)
	c.currentTime = 0
	c.momentumstop = 0
	c.camera.momx = CreditMaps.floatToFixed("0.3")
end,
function(c,e)
	local timeframe_1 = 1*TICRATE + 6
	local timeframe_2 = timeframe_1 + 9
	local timeframe_3 = timeframe_2 + 12
	local remainder_frame_time = (2*TICRATE + 6)
	EvActionSH.timedexec(function()
		-- Ease
		c.camera.momx = max(0, $1-CreditMaps.floatToFixed("0.006"))
	end, 0, timeframe_1-2, c.currentTime)
	EvActionSH.timedexec(function()
		-- Bounce
		c.camera.momx = $ - CreditMaps.floatToFixed("0.25")
	end, (timeframe_1), timeframe_2, c.currentTime)
	EvActionSH.timedexec(function()
		-- Bounce Leave
		c.momentumstop = $1 + CreditMaps.floatToFixed("0.630") -- 0.388 -- 0.7612 -- 0.630
		c.camera.momx = max(CreditMaps.floatToFixed("0.3"), $1 + (CreditMaps.floatToFixed("7.4") - c.momentumstop)) --5.0 --8.0 --7.5
	end, timeframe_2, timeframe_3 + remainder_frame_time, c.currentTime)
	-- EvActionSH.timedexec(function()
	-- 	-- Ease 2
	-- 	c.camera.momx = max(0, $1-CreditMaps.floatToFixed("0.006"))
	-- end, timeframe_3-10, timeframe_3 + remainder_frame_time, c.currentTime)

	c.currentTime = $1 + 1
	EventSH.wait(e, 4*TICRATE + 7)
end,
-- ----------------------
-- Shadow 2
-- ----------------------
function(c,e)
	-- pause_anim_timer()
	c.camera.momx = 0
	c.camera.momy = 0
	c.camera.momz = 0
	MobjMoverSH.stop("move_camera")

	EvActionSH.setcamera(true, c.camera, c.pos_cr_faccshad.mobj, c.pos_cr_faccshad.angle, 0, true)
	MobjMoverSH.moveto("move_camera", c.camera, {
		x=c.pos_cr_faccshad.mobj.x,
		y=c.pos_cr_faccshad.mobj.y+128*FU,
		z=c.pos_cr_faccshad.mobj.z}, {speed=FU/8, ease="linear", noangle=true})
	P_LinedefExecute(8008)
end,
function(c,e)
	-- PROGRAM 3:
	-- ---------------------------
	add_camtitle(c, "withinclusion", c.camera, -55*FU, 64*FU, 50*FU-(FU*4)*1, SHCRDSPR("With"), titleParam(5,true,1,80,nil,true,true,0,0, FU/7))
	add_camtitle(c, "progcredit7", c.camera, -55*FU, 64*FU, 50*FU-(FU*4)*2, SHCRDSPR("Smsalfredo"), titleParam(5,true,1,80,nil,true,true,0,0, FU/7))
end,
function(c,e)
	EventSH.wait(e, 3*TICRATE)
end,
-- ----------------------
-- The Final
-- ----------------------
function(c,e)
	-- Character
	set_object_sprite(c.character_in, SPR_CCC2, 0)
	c.character_in.frame = 0
	set_object_sprite(c.character_ae, SPR_CCC2, 1)
	set_object_sprite(c.character_fc, SPR_CCC2, 2)
	c.character_in.scale = FU/2
	c.character_ae.scale = FU/2
	c.character_fc.scale = FU/2
	move_object_position(c.character_in, c.pos_cr_char1_fb)
	move_object_position(c.character_ae, c.pos_cr_char2_fb)
	move_object_position(c.character_fc, c.pos_cr_char3_fb)

	-- Camera
	MobjMoverSH.stop("move_camera")

	EvActionSH.setcamera(true, c.camera, c.pos_cr_fb_1.mobj, c.pos_cr_fb_1.angle, 0, true)

	-- Pan
	MobjMoverSH.moveto("move_camera", c.camera, {
		x=c.pos_cr_fb_1.mobj.x+512*FU,
		y=c.pos_cr_fb_1.mobj.y,
		z=c.pos_cr_fb_1.mobj.z}, {speed=FU/4, ease="linear", noangle=true})
end,
function(c,e)
	EventSH.wait(e, 25)
end,
function(c,e)
	-- Character A
	MobjMoverSH.stop("move_camera")

	EvActionSH.setcamera(true, c.camera, c.pos_cr_fb_2.mobj, c.pos_cr_fb_2.angle, 0, true)

	MobjMoverSH.moveto("move_camera", c.camera, {
		x=c.pos_cr_fb_2.mobj.x+512*FU,
		y=c.pos_cr_fb_2.mobj.y,
		z=c.pos_cr_fb_2.mobj.z}, {speed=FU/4, ease="linear", noangle=true})
end,
function(c,e)
	EventSH.wait(e, 25)
end,
function(c,e)
	-- Character B
	MobjMoverSH.stop("move_camera")

	EvActionSH.setcamera(true, c.camera, c.pos_cr_fb_3.mobj, c.pos_cr_fb_3.angle, 0, true)

	MobjMoverSH.moveto("move_camera", c.camera, {
		x=c.pos_cr_fb_3.mobj.x+512*FU,
		y=c.pos_cr_fb_3.mobj.y,
		z=c.pos_cr_fb_3.mobj.z}, {speed=FU/4, ease="linear", noangle=true})
end,
function(c,e)
	EventSH.wait(e, 25)
end,
function(c,e)
	-- Character C-1
	MobjMoverSH.stop("move_camera")

	EvActionSH.setcamera(true, c.camera, c.pos_cr_fb_4.mobj, c.pos_cr_fb_4.angle, 0, true)

	MobjMoverSH.moveto("move_camera", c.camera, {
		x=c.pos_cr_fb_4.mobj.x,
		y=c.pos_cr_fb_4.mobj.y+512*FU,
		z=c.pos_cr_fb_4.mobj.z}, {speed=FU/6, ease="linear", noangle=true})
end,
function(c,e)
	EventSH.wait(e, 40)
end,
function(c,e)
	-- Character C-2
	MobjMoverSH.stop("move_camera")

	EvActionSH.setcamera(true, c.camera, {x=c.pos_cr_fb_4.mobj.x, y=c.pos_cr_fb_4.mobj.y-512*FU, z=c.pos_cr_fb_4.mobj.z}, c.pos_cr_fb_4.angle, 0, true)
	move_object_position(c.character_fc, {x=c.pos_cr_char3_fb.x, y=c.pos_cr_char3_fb.y, z=c.pos_cr_char3_fb.z+140*FU})

	MobjMoverSH.moveto("move_camera", c.camera, {
		x=c.pos_cr_fb_4.mobj.x,
		y=c.pos_cr_fb_4.mobj.y-96*FU,
		z=c.pos_cr_fb_4.mobj.z}, {speed=FU/6, ease="linear", noangle=true})
end,
function(c,e)
	EventSH.wait(e, 30)
end,
-- ----------------------
-- Ulti 1
-- ----------------------
function(c,e)
	MobjMoverSH.stop("move_camera")

	EvActionSH.setcamera(true, c.camera, c.pos_cr_ulti1.mobj, c.pos_cr_ulti1.angle, 0, true)
	MobjMoverSH.moveto("move_camera", c.camera, {
		x=c.pos_cr_ulti1.mobj.x,
		y=c.pos_cr_ulti1.mobj.y-96*FU,
		z=c.pos_cr_ulti1.mobj.z}, {speed=FU/8, ease="linear", noangle=true})
end,
function(c,e)
	-- PROGRAM 4:
	-- ---------------------------
	add_camtitle(c, "withinclusion", c.camera, -55*FU, 64*FU, 55*FU-(FU*4)*1, SHCRDSPR("With"), titleParam(5,true,1,50,nil,true,true,0,0, FU/7))
	add_camtitle(c, "progcredit7", c.camera, -55*FU, 64*FU, 55*FU-(FU*4)*2, SHCRDSPR("Lach"), titleParam(5,true,1,50,nil,true,true,0,0, FU/7))

end,
function(c,e)
	EventSH.wait(e, 2*TICRATE)
end,
function(c,e)
	MobjMoverSH.stop("move_camera")
	MobjMoverSH.moveto("move_camera", c.camera, {
		x=c.pos_cr_ulti2.mobj.x,
		y=c.pos_cr_ulti2.mobj.y,
		z=c.pos_cr_ulti2.mobj.z}, {speed=32*FU, ease="inoutquad", noangle=true})
end,
function(c,e)
	-- PROGRAM 5:
	-- ---------------------------
	add_camtitle(c, "andinclusion", c.camera, -55*FU, 64*FU, 55*FU-(FU*4)*1, SHCRDSPR("And"), titleParam(5,true,1,50,nil,true,true,0,0, FU/7))
	add_camtitle(c, "progcredit7", c.camera, -55*FU, 64*FU, 55*FU-(FU*4)*2, SHCRDSPR("Lune_Wulf"), titleParam(5,true,1,50,nil,true,true,0,0, FU/7))

end,
function(c,e)
	EventSH.wait(e, 2*TICRATE)
end,
-- ----------------------
-- Title
-- ----------------------
function(c,e)
	EvActionSH.setcamera(true, c.camera, c.pos_cr_title.mobj, c.pos_cr_title.angle, 0, true)

	MobjMoverSH.moveto("move_camera", c.camera, {
		x=c.pos_cr_title_f.mobj.x,
		y=c.pos_cr_title_f.mobj.y-64*FU,
		z=c.pos_cr_title_f.mobj.z}, {speed=20*FU, ease="inoutquad", noangle=true})
end,
function(c,e)
	-- FUN:
	-- ---------------------------
	add_camtitle(c, "exec", c.camera, 0, 64*FU, 20*FU, SHCRDSPR("TitleExec"), titleParam(5,true,1,2*TICRATE,nil,true,true,SHCRDSPR("TitleExec").w/2,0, FU/7))
	add_camtitle(c, "execperson", c.camera, 0, 64*FU, 20*FU-(FU*4)*2, SHCRDSPR("Inazuma"), titleParam(5,true,1,2*TICRATE,nil,true,true,SHCRDSPR("Inazuma").w/2,0, FU/8))

	-- PRESENTER:
	-- ---------------------------
	add_camtitle(c, "presenter", c.camera, 0, 64*FU, 20*FU, SHCRDSPR("Presents"), titleParam(3*TICRATE-8,true,1,1*TICRATE+10,nil,true,true,SHCRDSPR("Presents").w/2,SHCRDSPR("Presents").h/2, FU/4))

	-- TITLE:
	-- ---------------------------
	add_camtitle(c, "silverhornlogo", c.pos_cr_title_f.mobj, 0, 0*FU, 16*FU, SHCRDSPR("Logo"), titleParam(3*TICRATE,true,1,12*TICRATE,nil,false,true,SHCRDSPR("Logo").w/2,SHCRDSPR("Logo").h/2, FU/5))

end,
function(c,e)
	MobjMoverSH.waitmovedone(e, "move_camera")
end,
function(c,e)
	P_LinedefExecute(50) -- (leftover linedef)
	MobjMoverSH.moveto("move_camera", c.camera, {
		x=c.pos_cr_title_f.mobj.x,
		y=c.pos_cr_title_f.mobj.y-256*FU,
		z=c.pos_cr_title_f.mobj.z}, {speed=FU/16, ease="linear", noangle=true})
end,
function(c,e)
	EventSH.wait(e, 3*TICRATE+10)
end,
function(c,e)
	MobjMoverSH.stop("move_camera")
	EvActionSH.setcamera(true, c.camera, {x=c.pos_cr_title_f.mobj.x, y=c.pos_cr_title_f.mobj.y-512*FU, z=c.pos_cr_title_f.mobj.z}, ANGLE_90, 0, true)
	-- final (:
	EvActionSH.fadescreen("in", 0, 7)

	--  Check off credits
	MindscapeSessionFlags = $1 | MS_SEENCREDITS
end,
function(c,e)
	EventSH.wait(e, 4*TICRATE)
end,
function(c,e)
	-- Play ending
	if (c.endtype and c.endtype == "good") then
		EventSH.start("sh_credits_endcard", c)
	else
		-- Return home
		G_SetCustomExitVars(Silverhorn.MS_RETURNMAP, 1)
		G_ExitLevel()
	end
end,})





-- =============================
-- .......
-- Ending
-- .......
-- =============================
EventSH.new("sh_credits_endcard", {
function(c,e)
	-- Enable render
	EvActionSH.togglerender(c)
end,
function(c,e)
	EvActionSH.renderstretchedpatch(c, 1, "backdrop", 0, 0, EvActionSH.vidwidth*FU, EvActionSH.vidheight*FU, "PAL_31G", V_SNAPTOTOP|V_SNAPTOLEFT, 0)
	EventSH.wait(e, 3*TICRATE, true)
end,
function(c,e)
	-- EvActionSH.renderfill(c, 0, "back", 0, 0, 320, 200, 31)
	EvActionSH.fadescreen("in", 31, 6)
	EvActionSH.renderscaledpatch(c, 70, "Ending_Card", (320*FU)/2, (200*FU)/2, FU/4, c.card, 0, 0)
	EvActionSH.renderscaledpatch(c, 71, "Star_Rating", 246*FU, 6*FU, FU/4, c.star, V_SNAPTORIGHT|V_SNAPTOTOP, 0)
	S_ChangeMusic(c.cardmusic, true)

	EventSH.wait(e, 3*TICRATE, true)
end,
function(c,e)
	-- EvActionSH.renderscaledpatch(c, 71, "Star_Rating", 246*FU, 6*FU, FU/4, c.star, V_SNAPTORIGHT|V_SNAPTOTOP, 0)
	EvActionSH.Sq_SetQTimer(c, 22*TICRATE)
end,
function(c,e)
	-- Wait for input from host
	EvActionSH.waitForInput(e, server, BT_JUMP)
	-- Wait for next phase
	EvActionSH.Sq_AdvanceQTimer(e, c)
end,
function(c,e)
	EvActionSH.fadescreen("out", 31, 6)

	-- non-fade exclusive to good ending
	if not (c.endtype and c.endtype == "good") then
		S_FadeOutStopMusic(5*MUSICRATE)
	end

	EventSH.wait(e, 5*TICRATE, true)
end,
function(c,e)
	EvActionSH.renderitemvisibility(c, "Ending_Card", -1)
	EvActionSH.renderitemvisibility(c, "Star_Rating", -1)
	EventSH.wait(e, 2*TICRATE, true)
end,
-- (Exclusive to good ending)
-- ========================================================
function(c,e)
	if not (c.endtype and c.endtype == "good") then return end

	EvActionSH.rendertext(c, 80, "Credit_Final_Text", 320/2, 200/2, "INAZUMA will return.", 0, "center")
	EvActionSH.fadescreen("in", 31, 2)
end,
function(c,e)
	if not (c.endtype and c.endtype == "good") then return end
	S_FadeOutStopMusic(6*MUSICRATE)
	EventSH.wait(e, 5*TICRATE, true)
end,
function(c,e)
	if not (c.endtype and c.endtype == "good") then return end
	EvActionSH.fadescreen("out", 31, 2)
	EventSH.wait(e, 4*TICRATE, true)
end,
-- ========================================================
function(c,e)
	if (c.endtype and c.endtype == "good") then
		-- Return home
		G_SetCustomExitVars(Silverhorn.MS_RETURNMAP, 1)
		G_ExitLevel()
	else
		-- Warp to credits (SRB2)
		G_SetCustomExitVars(1102, 1)
		G_ExitLevel()
	end
end,
function(c,e)
end})





-- =============================
-- .......
-- Gallery
-- .......
-- =============================
Silverhorn.MapAC_GalleryWallLines = {}

local galleryInfo = {
	[8510] = {
		ending = "has_ending_g",
		texture = "SHG_PICG"
	},
	[8511] = {
		ending = "has_ending_t",
		texture = "SHG_PICT"
	},
	[8512] = {
		ending = "has_ending_n",
		texture = "SHG_PICN"
	},
	[8513] = {
		ending = "has_ending_b",
		texture = "SHG_PICB"
	}
}

-- Gets linedefs on mapload for easier optimized iteration
function Silverhorn.LoadGalleryWallLines()
	-- Only run on AC
	if not (mapheaderinfo[gamemap].aetherscastle) then return end

	Silverhorn.MapAC_GalleryWallLines = {}
	-- if mapheaderinfo[mapnum].mindscape then
	for line in lines.iterate do
		if galleryInfo[line.tag] then
			table.insert(Silverhorn.MapAC_GalleryWallLines, line)
		end
	end
	-- end
end

-- Loads correct unlocked gallery images for the current player on mapload
function Silverhorn.LoadGallery(mapnum)
	-- Only run on AC
	if not (mapheaderinfo[gamemap].aetherscastle) then return end

	if Silverhorn.MapAC_GalleryWallLines == nil then
		return
	end

	for i = #Silverhorn.MapAC_GalleryWallLines, 1, -1 do

		local line = Silverhorn.MapAC_GalleryWallLines[i]
		local info = galleryInfo[line.tag]

		if not info then
			table.remove(Silverhorn.MapAC_GalleryWallLines, i)
			continue
		end

		if Silverhorn.CheckGameValue(info.ending, true, true) then
			line.backside.midtexture = R_TextureNumForName(info.texture)
		else
			line.backside.midtexture = R_TextureNumForName("SHG_UNKN")
		end
	end
end

-- Disables player viewpoint when player is in gallery view
function Silverhorn.GalleryViewpoint(player, viewedPlayer)
	-- Only run on AC
	if not (mapheaderinfo[gamemap].aetherscastle) then return end

	if player == viewedPlayer then
		return
	end

	-- Skip any players tagged with this
	if (viewedPlayer.mo.sh_c_cantbeviewed) then
		return false
	end
end

-- Executes gallery image popup from a linedef
function Silverhorn.ShowGalleryImage(line, trigger, sector)

	-- @ Flag Effects:
	-- none
	-- @ Default: runs EV 'silverhorn_endcard_gallery' with textureoffset being the number id to reference inside of it
	if not (trigger and trigger.player and trigger.player.valid) then return end

	local player = trigger.player

	if (player.bot > 0) then return end -- emergency eject because bots shouldn't be able to trigger this

	local cardnum = (line.frontside.textureoffset >> 16)

	-- TODO: snap to center

	if (cardnum == 1 and not Silverhorn.PlayerHasEnding(player, "has_ending_g")) then
		return
	elseif (cardnum == 2 and not Silverhorn.PlayerHasEnding(player, "has_ending_t")) then
		return
	elseif (cardnum == 3 and not Silverhorn.PlayerHasEnding(player, "has_ending_n")) then
		return
	elseif (cardnum == 4 and not Silverhorn.PlayerHasEnding(player, "has_ending_b")) then
		return
	-- elseif (cardnum == 4 and not Silverhorn.GetUnlockPlayerData(player, "has_ending_u")) then
		-- return
	end

	EventSH.start("silverhorn_endcard_gallery", {line_cardnum=cardnum, linedef=line, execmobj=trigger, sector=sector, execplayer=player})
end

function Silverhorn.get33()
	local msg = ""

	for i,v in ipairs({79,110,101,32,100,111,119,110,46,32,84,119,111,32,116,111,32,103,111,46}) do
		msg = $1 .. string.char(v)
	end
	return msg

end





EventSH.new("silverhorn_endcard_gallery", {
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	// Lach: cancel spectating this player
	if c.execplayer == displayplayer
	and displayplayer ~= consoleplayer then
		displayplayer = consoleplayer
	end

	c.selectsounds = {
		sfx_hoop1, -- prompt (hoop)
		sfx_none, -- decline (antiring)
		sfx_3db16, -- view (hoop)
		sfx_3db06, -- leave (flicky)
	}

	-- Assign triggerer player to this
	EvActionSH.filterforplayers(c, {c.execplayer})

	-- Prompt Display
	EvActionSH.togglerender(c)
	c.originx = 160
	c.originy = 0
	EvActionSH.renderstretchedpatch(c, 29, "promptback", ((c.originx-50)-4)*FU, ((c.originy+32)-4)*FU, 108*FU, 58*FU, "PAL_31G", V_40TRANS, 0)
	EvActionSH.rendertext(c, 30, "Prompt_Text0", c.originx, c.originy+32, "View Gallery?", 0, "center")
	EvActionSH.rendertext(c, 30, "Prompt_Text1", c.originx, c.originy+42, "\x82[JUMP] - View", 0, "thin-center")
	EvActionSH.rendertext(c, 30, "Prompt_Text2", c.originx, c.originy+52, "\x82[SPIN] - Quit", 0, "thin-center")
	EvActionSH.rendertext(c, 30, "Prompt_Text3", c.originx, c.originy+65, "(While Viewing)", 0, "thin-center")
	EvActionSH.rendertext(c, 30, "Prompt_Text4", c.originx, c.originy+75, "\x82[SPIN] - Quit", 0, "thin-center")

	-- Player settings
	c.execplayer.mo.sh_c_cantbeviewed = true
	c.execplayer.powers[pw_nocontrol] = 99*TICRATE
	S_StartSound(c.execplayer.mo, c.selectsounds[1])
end,
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	-- Wait for input (Cancel/Accept)
	if (c.execplayer.cmd.buttons & BT_SPIN) then
		c.execplayer.mo.sh_c_cantbeviewed = false
		S_StartSound(c.execplayer.mo, c.selectsounds[2], c.execplayer)
		EventSH.destroyself(e)
	else
		c.execplayer.mo.momx = 0
		c.execplayer.mo.momy = 0
		c.execplayer.mo.momz = 0
		EventSH.waitUntil(e, c.execplayer.cmd.buttons & BT_JUMP)
	end
	c.execplayer.powers[pw_nocontrol] = 1
end,
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	-- Clear prompt
	c.execplayer.powers[pw_nocontrol] = 99*TICRATE
	EvActionSH.renderitemvisibility(c, "promptback", -1)
	EvActionSH.renderitemvisibility(c, "Prompt_Text0", -1)
	EvActionSH.renderitemvisibility(c, "Prompt_Text1", -1)
	EvActionSH.renderitemvisibility(c, "Prompt_Text2", -1)
	EvActionSH.renderitemvisibility(c, "Prompt_Text3", -1)
	EvActionSH.renderitemvisibility(c, "Prompt_Text4", -1)

	S_StartSound(c.execplayer.mo, c.selectsounds[3])

	-- Debug display
	-- print("Explayer: "..#c.execplayer)
	-- print("Displayer: "..#displayplayer)
	-- print("Consplayer: "..#consoleplayer)

	-- Setup card (default fallback graphic)
	c.card = "SH_UNKN"
	-- (Good)
	if (c.line_cardnum == 1) then
		-- (Transcended)
		if (c.execplayer.mo.skin == "fang") then
			EventSH.start("silverhorn_endcard_33", c)
			EventSH.destroyself(e)
			return
		else
			c.card = "SH_ENDG1"
		end
	-- (True)
	elseif (c.line_cardnum == 2) then
		c.card = "SH_ENDT"
	-- (Neutral)
	elseif (c.line_cardnum == 3) then
		c.card = "SH_ENDN"
	-- (Bad)
	elseif (c.line_cardnum == 4) then
		c.card = "SH_ENDB"
	-- (Ultimate)
	elseif (c.line_cardnum == 5) then
		c.card = "SH_ENDUL"
	end
end,
function(c,e)
	EvActionSH.renderstretchedpatch(c, 1, "back", 0, 0, EvActionSH.vidwidth*FU, EvActionSH.vidheight*FU, "PAL_31G", V_SNAPTOTOP|V_SNAPTOLEFT, 0)
end,
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	-- Show card!
	EvActionSH.fadescreen("in", 1, 1, nil, {c.execplayer}, tostring(#c.execplayer))
	EvActionSH.renderscaledpatch(c, 70, "Endcard", (320*FU)/2, (200*FU)/2, FU/4, c.card, 0, 0)
end,
function(c,e)
	EventSH.wait(e, 3*TICRATE, true)
end,
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	-- Ask the player if they're ready to leave
	c.execplayer.powers[pw_nocontrol] = 4*TICRATE
	EventSH.waitUntil(e, c.execplayer.cmd.buttons & BT_SPIN)
end,
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	S_StartSound(c.execplayer.mo, c.selectsounds[4])
	EvActionSH.fadescreen("out", 1, 1, nil, {c.execplayer}, tostring(#c.execplayer))
end,
function(c,e)
	EventSH.wait(e, 15, true)
end,
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	EvActionSH.fadescreen("in", 1, 1, nil, {c.execplayer}, tostring(#c.execplayer))
	-- Reset player settings
	c.execplayer.powers[pw_nocontrol] = 1
	c.execplayer.mo.sh_c_cantbeviewed = false
end})




EventSH.new("silverhorn_endcard_33", {
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	-- startmindtext(-1111)
	EvActionSH.togglerender(c)
	-- EvActionSH.renderfill(c, 0, "back", 0, 0, 319, 199, 31)
	EvActionSH.renderstretchedpatch(c, 1, "back", 0, 0, 320*FU, 200*FU, "PAL_31G", 0, 0)

	-- EvActionSH.fadescreen("full", 31, 1)
	c.alpha = 9
	c.timer = 0
	-- EventSH.wait(e, 1*TICRATE, true)

	-- Player settings
	c.execplayer.powers[pw_nocontrol] = 99*TICRATE
	c.execplayer.mo.sh_c_cantbeviewed = true
end,
function(c,e)
	EvActionSH.fadescreen("in", 1, 1, nil, {c.execplayer}, tostring(#c.execplayer))
	EvActionSH.renderscaledpatch(c, 70, "Endcard", (320*FU)/2, (200*FU)/2, FU/4, "SH_ENDG1", 0, 0)
end,
function(c,e)
	EvActionSH.renderscaledpatch(c, -1, "Endcard2", (320*FU)/2, (200*FU)/2, FU/4, "SH_ENDG2", (9 << V_ALPHASHIFT), 0)
	EvActionSH.renderscaledpatch(c, -1, "Endcard_F", (320*FU)/2, (200*FU)/2, FU/4, "SH_ENDG3", (9 << V_ALPHASHIFT), 0)

	-- EvActionSH.togglerender(c)
	-- EvActionSH.fadescreen("in", 31, 12)
	EventSH.wait(e, 7*TICRATE, true)
end,
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	EvActionSH.editrender(c, "Endcard2").order = 71
	S_StartSound(c.execplayer.mo, sfx_msrg10, c.execplayer)
	S_StartSound(c.execplayer.mo, sfx_msrg10, c.execplayer)
	S_StartSound(c.execplayer.mo, sfx_msrg10, c.execplayer)
	S_StartSound(c.execplayer.mo, sfx_msrg10, c.execplayer)
	S_FadeOutStopMusic(2*MUSICRATE, c.execplayer)
end,
function(c,e)
	c.timer = $1+1

	if (c.timer % 6 == 0) then
		c.alpha = max(0, $1-1)
		EvActionSH.editrender(c, "Endcard2").flags = (c.alpha << V_ALPHASHIFT)
	end
	EventSH.wait(e, 5*TICRATE)
end,
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	S_ChangeMusic("ENDFN", true, c.execplayer)
	EventSH.wait(e, 3*TICRATE, true)
end,
function(c,e)
	EvActionSH.editrender(c, "Endcard_F").order = 72
	c.timer = 0
	c.alpha = 10
	-- EvActionSH.togglerender(c)
	-- EvActionSH.fadescreen("in", 31, 12)
end,
function(c,e)
	c.timer = $1+1

	if (c.timer % 3 == 0) then
		c.alpha = max(0, $1-1)
		EvActionSH.editrender(c, "Endcard_F").flags = (c.alpha << V_ALPHASHIFT)
	end
	EvActionSH.editrender(c, "Endcard_F").scale = $1+FU/8096
	EventSH.wait(e, 4*TICRATE)
end,
function(c,e)
	EvActionSH.rendertext(c, 80, "Textf", 160, 200-25, Silverhorn.get33(), 0, "center")
end,
function(c,e)
	EvActionSH.editrender(c, "Endcard_F").scale = $1+FU/8096
	EventSH.wait(e, 9*TICRATE)
end,
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	EvActionSH.fadescreen("out", 31, 6, nil, {c.execplayer}, tostring(#c.execplayer))
	S_FadeOutStopMusic(6*MUSICRATE, c.execplayer)
end,
function(c,e)
	EvActionSH.editrender(c, "Endcard_F").scale = $1+FU/8096
	EventSH.wait(e, 5*TICRATE)
end,
function(c,e)
	EventSH.wait(e, 3*TICRATE)
end,
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	EvActionSH.togglerender(c, false)
	move_object_position(c.execplayer.mo, End_33_Aftermath_Position) -- (2.2.11 interpolation patch)
	c.execplayer.mo.angle = 0 --FixedAngle(180*FU)
end,
function(c,e)
	if not EvActionSH.isplayervalid(c.execplayer) then return end
	EvActionSH.EVS_validateplayer(e, c.execplayer)

	EvActionSH.fadescreen("in", 1, 1, nil, {c.execplayer}, tostring(#c.execplayer))
	-- EvActionSH.fadescreen("clear")

	S_ChangeMusic("EVOCSL", true, c.execplayer)

	-- Reset player settings
	c.execplayer.mo.sh_c_cantbeviewed = false
	c.execplayer.powers[pw_nocontrol] = 1
end})

