
rawset(_G, "EvActionSH", {})

-- ------------------------------------------------------
-- @region SYSTEM - System functions
-- ------------------------------------------------------

-- Allows zero values or an alternative if value is nil (prev: notnil)
function EvActionSH.defvalue(value, altvalue)
	return (value and not nil) and value or altvalue
end

-- A wrapper for players.iterate
function EvActionSH.doforall(func)
	for player in players.iterate do
		-- if not (player.mo and player.mo.valid) then continue end
		do func(player) end
	end
end

-- Wrapper for checking if a player is valid
function EvActionSH.isplayervalid(checkedplayer)
	if (checkedplayer and checkedplayer.valid) then
		return true
	else
		return false
	end
end

-- Validates if a player exists if used in an EV, and shuts it down if they don't exist
function EvActionSH.EVS_validateplayer(ev, checkedplayer)
	if not (checkedplayer and checkedplayer.valid) then
		EventSH.destroyself(ev)
	end
end

-- Iterates players and adds them to a list that is returned
function EvActionSH.createplayerlist()
	local t = {}

	for player in players.iterate do
		if not (player and player.valid) then continue end
		table.insert(t, player)
	end
	return t -- return table
end

-- Iterates players and adds them to a list that is returned (on conditional)
function EvActionSH.createplayerlistfromcond(fn)
	local t = {}

	for player in players.iterate do
		if not (player and player.valid) then continue end

		-- do thing because of cond fn
		if fn(player) then
			table.insert(t, player)
		end
	end
	return t -- return table
end

-- Selectively assigns players to an EV State Sequence
-- Includes only specified players and filters out the rest
function EvActionSH.filterforplayers(container, playerlist)
	-- Create assigned players list
	if not container["assignedplayers"] then container["assignedplayers"] = {} end

	-- Check for valid players before assigning (because otherwise the game might throw yet another fit!)
	for _, play in ipairs(playerlist) do
		(function()-- not valid, skip
			if not (play and play.mo.valid) then return end
			-- Insert player
			table.insert(container["assignedplayers"], play)
		end)()
	end
end

function EvActionSH.timedexec(fn, tstart, tend, timer)
	if timer >= tstart and timer <= tend then
		fn()
	end
end

-- Converts a fixed distance (factored with speed and a scale) to tics.
function EvActionSH.DistToTics(distance, speed, scale)
	return FixedCeil(FixedDiv(FixedDiv(distance, speed), scale)) / FRACUNIT
end

-- Gets distance of two objects in 3D space
function EvActionSH.GetDist3D(p1, p2)
    return FixedHypot(FixedHypot(p1.x-p2.x, p1.y-p2.y), p1.z-p2.z)
end

-- Gets distance of two objects in 2D space
function EvActionSH.GetDist2D(p1, p2)
    return FixedHypot(p1.x-p2.x, p1.y-p2.y)
end

-- Custom Random function that uses leveltime in case the random seed desynched
local function RandomRange_P(min, max)
	if min == max then return min end -- What's the point if there's no range?

	-- Well this won't work if min isn't actually min!
	if min > max then
		local oldmax = max
		max = min
		min = oldmax
	end

	-- Get a mapthing's positioning based off of the leveltime
	local position = 0
	local thing = mapthings[leveltime % (#mapthings + 1)]
	if thing then
		if check == 0 then
			position = thing.x
		elseif check == 1 then
			position = thing.y
		else
			position = thing.z
		end

		--Add each player's lives and skin
		for player in players.iterate do
			position = $1 * player.skin * player.lives
		end

		if position < 0 then
			position = abs($1 + INT32_MAX)
		end
	end

	-- Get our value
	local factor = position or leveltime
	local value = (factor % (max - min))
	return value + min
end

-- Random Choice port
function EvActionSH.RandomChoice(choices, vanillarandom)
	local RandomKey = 0
	if vanillarandom then
		RandomKey = P_RandomRange(1, #choices)
	else
		RandomKey = RandomRange_P(1, #choices)
	end
    if type(choices[RandomKey]) == "function" then
        choices[RandomKey]()
    else
        return choices[RandomKey]
    end
end

-- Creates a value in the container
function EvActionSH.Sq_CreateValue(container, valuename, value)
	container[valuename] = value
end

-- ------------------------------------------------------
-- @region INPUT - Effect functions
-- ------------------------------------------------------

-- Waits for an input by a selected player. Returns true if the decline button is pressed
function EvActionSH.waitForInput(ev, selectplayer, button, fn)
	if not selectplayer and selectplayer.valid then return end

	EventSH.waitUntil(ev, selectplayer.cmd.buttons & button)
end

-- Sets a timer to pass for automatically advancing.
function EvActionSH.Sq_SetQTimer(container, time)
	container["qt_autoadvancetime"] = 0
	container["qt_autoadvancetime_max"] = time
end

-- Waits for a timer to pass before automatically advancing. Not the same as waiting.
function EvActionSH.Sq_AdvanceQTimer(ev, container, triggerpause)
	-- if not e then return end

	if (container["qt_autoadvancetime_max"]) then

		container["qt_autoadvancetime"] = $1+1

		if (container["qt_autoadvancetime"] >= container["qt_autoadvancetime_max"]) then
			container["qt_autoadvancetime"] = 0
			container["qt_autoadvancetime_max"] = 0
			EventSH.resume(ev)
			return
		else
			if triggerpause then EventSH.pause(ev) end
		end
	end
end

-- ------------------------------------------------------
-- @region GAME - Gameplay functions
-- ------------------------------------------------------

-- Sets the checkpoint number manually for all players
function EvActionSH.setcheckpoint(num, position, chkangle)
	for player in players.iterate do
		player.starpostnum = num
		player.starposttime = player.realtime
		player.starpostx = position.x or 0*FRACUNIT
		player.starposty = position.y or 0*FRACUNIT
		player.starpostz = position.z or 0*FRACUNIT
		player.starpostangle = chkangle
	end
end

-- Forces a player to be re-born if they are dead in a netgame
function EvActionSH.forceplayerreborn(player)
	if not (player and player.valid) then return end -- no player exists

	if (netgame and (player.playerstate == PST_DEAD and player.lives > 0)) then
		player.playerstate = PST_REBORN
	end
end

-- Teleports all players to a location
function EvActionSH.relocateplayers(position, angle, interpolate)
	local warporigin = interpolate and P_MoveOrigin or P_SetOrigin -- or P_TeleportMove

	for player in players.iterate do
		warporigin(player.mo, position.x, position.y, position.z) -- (2.2.11 interpolation patch)
		-- P_MoveOrigin(player.mo, position.x, position.y, position.z)
		player.mo.angle = angle or 0
	end
end

EventSH.new("a_relocate_players", {
function(c,e)
	-- Capture players that should be re-located
end,
function(c,e)
	EvActionSH.relocateplayers(c.position, c.angle, c.interpolate)
	EventSH.wait(e, c.timing or 5*TICRATE)
end})

-- Teleports all players to a location constantly
function EvActionSH.relocateplayers_start(timing, position, angle, interpolate)
	EventSH.start("a_relocate_players", {timing = timing, position = position, angle = angle, interpolate = interpolate})
end


-- Stops teleporting all players to a location constantly
function EvActionSH.relocateplayers_end()
	EventSH.destroy("a_relocate_players")
end

-- @notice (8-31-2023): Was prone to a random error before
-- Creates a suspended position for suspending a player in place
local function Evie_CreateSuspendedPosition(player)
	player.mo._evie_suspendedxyz = {}
	player.mo._evie_suspendedxyz.x = player.mo.x
	player.mo._evie_suspendedxyz.y = player.mo.y
	player.mo._evie_suspendedxyz.z = player.mo.z
	player.mo._evie_suspendedxyz.drawangle = player.drawangle
	player.mo._evie_suspendedxyz.angle = player.mo.angle
end

-- Creates a sequence to suspend a player in place
EventSH.new("a_freeze_player", {
function(c,e)
	-- Get player original position
	for _, player in ipairs(c.playerlist) do

		if (player and player.valid) then
			Evie_CreateSuspendedPosition(player)
		end
	end
end,
function(c,e)
	-- Freeze player in place
	for _, player in ipairs(c.playerlist) do
		if (player and player.valid) then

			if not player.mo._evie_suspendedxyz then Evie_CreateSuspendedPosition(player) end

			-- Warp in place
			local mo = player.mo
			P_SetOrigin(mo, mo._evie_suspendedxyz.x, mo._evie_suspendedxyz.y, mo._evie_suspendedxyz.z) -- (2.2.11 interpolation patch)
			-- P_MoveOrigin(mo, mo._evie_suspendedxyz.x, mo._evie_suspendedxyz.y, mo._evie_suspendedxyz.z)

			-- Set angle
			mo.angle = mo._evie_suspendedxyz.angle
			player.drawangle = mo._evie_suspendedxyz.drawangle

			-- TODO: pw_nocontrol option for players to be allowed to jump
			-- Disable controls if enabled
			if (c.nocontrol) then
				player.powers[pw_nocontrol] = 1
			end
		end
	end
	EventSH.waitUntil(e, false)
end})

-- Freezes players in place, and also allows control restricting
function EvActionSH.FreezePlayers(frozenplayerlist, restrictcontrols)
	EventSH.start("a_freeze_player", {playerlist = frozenplayerlist, nocontrol = restrictcontrols})
end

-- Unfreezes players
function EvActionSH.UnfreezePlayers()
	EventSH.destroy("a_freeze_player")
end

-- ------------------------------------------------------
-- @region CAMERA - Camera functions
-- ------------------------------------------------------

-- Creates an object to act as a camera inside of the container
function EvActionSH.Sq_MakeCamera(container, ox, oy, oz, custommobjtype)
	local x = (ox and not nil) and ox or 0
	local y = (oy and not nil) and oy or 0
	local z = (oz and not nil) and oz or 0

	container["camera"] = P_SpawnMobj(x, y, z, custommobjtype or MT_PUSH)
	container["camera"].isevcamera = true -- To let iterators know it's a camera, it will apply this variable to itself
end


-- Resets the awayviewmobj
local function P_ResetAwayviewCamera(play, viewmobj)
	play.awayviewmobj = nil
	play.awayviewmobj = viewmobj
end

-- Sets an awayviewmobj to act as a camera indefinitely
function EvActionSH.setactivecamera(player, awaycam, awayaiming)

	-- player is userdata but not valid
	if (type(player) == "userdata" and not player.valid) or (type(player) == "nil") then return end

	if not (awaycam and awaycam.valid) then print("Camera does not exist.") return end

	-- Triggers for all players, or only the one specified
	if (type(player) == "boolean" and player == true) then
		EvActionSH.doforall(function(play)
			P_ResetAwayviewCamera(play, awaycam)
			play.awayviewmobj = awaycam
			play.awayviewtics = INT32_MAX
			play.awayviewaiming = (awayaiming and not nil) and awayaiming or 0
		end)
	elseif (type(player) == "userdata") then
		P_ResetAwayviewCamera(player, awaycam)
		player.awayviewmobj = awaycam
		player.awayviewtics = INT32_MAX
		player.awayviewaiming = (awayaiming and not nil) and awayaiming or 0
	end

end

-- Moves the camera to a position, and updates player viewpoints unless specified
function EvActionSH.setcamera(player, awaycam, position, angle, aiming, useang, interpolate)

	-- player is userdata but not valid
	if (type(player) == "userdata" and not player.valid) or (type(player) == "nil") then return end

	if not (awaycam and awaycam.valid) then print("Camera does not exist.") return end

	-- Unused: position angle for backwards compat
	-- local pangle = EvActionSH.defvalue(position and position.angle, 0)
	local fangle = EvActionSH.defvalue(angle, 0)
	local faiming = EvActionSH.defvalue(aiming, 0)

	-- Determine whether to target all players, or just the one given
	EvActionSH.setactivecamera(player, awaycam, useang and faiming or FixedAngle(faiming))
	-- EvActionSH.setactivecamera(player, awaycam, FixedAngle(aiming))

	awaycam.angle = ( useang and fangle or FixedAngle(fangle) )
	-- awaycam.angle = FixedAngle(angle or position.angle)

	local warporigin = interpolate and P_MoveOrigin or P_SetOrigin -- or P_TeleportMove

	warporigin(awaycam, position.x, position.y, position.z) -- (2.2.11 interpolation patch)
	-- P_ResetAwayviewCamera(player, awaycam)
	-- P_MoveOrigin(awaycam, position.x, position.y, position.z)
end

-- Moves the camera relative to where it was last
-- TODO: replace with camera orientation
function EvActionSH.setcamera_relative(player, awaycam, rx, ry, rz, angle, aiming)

	local ax = (awaycam.x and not nil) and awaycam.x or 0
	local ay = (awaycam.y and not nil) and awaycam.y or 0
	local az = (awaycam.z and not nil) and awaycam.z or 0
	local aim = (aiming and not nil) and aiming or 0

	local position = {x=ax + rx, y=ay + ry, z=az + rz}

	EvActionSH.setcamera(player, awaycam, position, angle or awaycam.angle, aim)
end

-- Make the camera track a point in space (prev: lookAtZ)
function EvActionSH.cameratrack(player, awaycam, destination)

	-- player is userdata but not valid
	if (type(player) == "userdata" and not player.valid) or (type(player) == "nil") then return end

	if not (awaycam and awaycam.valid) then print("Camera does not exist.") return end

	local x1, y1, z1 = awaycam.x, awaycam.y, awaycam.z
	local x2, y2, z2 = destination.x, destination.y, destination.z

	-- Point to track destination, and z height of the destination
	awaycam.angle = R_PointToAngle2(x1, y1, x2, y2)
	-- player.awayviewaiming = R_PointToAngle2(0, 0, FixedHypot(x2 - x1, y2 - y1), z2 - z1)
	local aimdirection = R_PointToAngle2(0, 0, FixedHypot(x2 - x1, y2 - y1), z2 - z1)

	-- Determine whether to target all players, or just the one given
	EvActionSH.setactivecamera(player, awaycam, aimdirection)
end

-- Changes the viewroll for the specified player
function EvActionSH.setplayerviewroll(player, rollangle, addangle)

	-- player is userdata but not valid
	if (type(player) == "userdata" and not player.valid) or (type(player) == "nil") then return end

	-- Triggers for all players, or only the one specified
	if (type(player) == "boolean") then
		EvActionSH.doforall(function(play)
			play.viewrollangle = (addangle==true and $1+rollangle or rollangle)
		end)
	else
		player.viewrollangle = (addangle==true and $1+rollangle or rollangle)
	end
end

-- TODO: may disrupt anything else that uses awayviewmobj
-- Ceases all awayviewmobj activity
function EvActionSH.stopcameraview(player)

	-- player is userdata but not valid
	if (type(player) == "userdata" and not player.valid) or (type(player) == "nil") then return end

	-- Triggers for all players, or only the one specified
	if (type(player) == "boolean") then
		EvActionSH.doforall(function(play)
			play.awayviewmobj = nil
			play.awayviewtics = 0
			play.awayviewaiming = 0
		end)
	else
		player.awayviewmobj = nil
		player.awayviewtics = 0
		player.awayviewaiming = 0
	end
end

-- ------------------------------------------------------
-- @region SOUND - Sound functions
-- ------------------------------------------------------

-- local MUSIC_RELOADRESET = MUSIC_RELOADRESET or 0x8000

-- Plays a musicname with S_ChangeMusic, with the ability to set it as the main map playback track
function EvActionSH.SetCurrentMusic(mainmus, musname, loop, player, flags, position, prefade, fadein)
	S_ChangeMusic(musname, loop, player, flags, position, prefade, fadein)

	-- Sets main playback track
	if mainmus then	mapmusname = musname end

	if multiplayer then
		mapmusflags = 0 -- Don't reload music on death
	else
		mapmusflags = 32768 -- 0x8000 Reload music on death
	end
end

-- TODO: More study on how this works is needed
-- Sets music flags to mapmusflags
-- Flag Types: (reload [MUSIC_RELOADRESET] (Reload on death) | force [MUSIC_FORCERESET] (???))
function EvActionSH.SetMapMusicFlags(flagtype)
	local flag = 0

	if flagtype == "reload" then
		flag = 32768 --0x8000
	elseif (flagtype == "force") then
		flag = 16384 --0x4000
	end

	mapmusflags = flag
end

-- Event Action wrapper to play a sound
function EvActionSH.startsound(mobj, soundnum, volume, player)
	if (volume) then
		S_StartSoundAtVolume(mobj, soundnum, volume, player)
	else
		S_StartSound(mobj, soundnum, player)
	end
end

-- Plays a sound at a timed rate
function EvActionSH.PlaySoundRate(source, soundnum, time, volume, player)
	if (leveltime % time == 0) then
		local sound = (volume) and S_StartSoundAtVolume(source, soundnum, volume or 255, player) or S_StartSound(source, soundnum, player)
	end
end

-- ------------------------------------------------------
-- @region ACTOR - Actor functions
-- ------------------------------------------------------

-- Creates a character that can be modified by the sequence (prefixed under ac_)
function EvActionSH.Sq_CreateActor(container, charname, skin, sprite, radius, shadow, color, attr)

	container["ac_" .. charname] = P_SpawnMobj(0, 0, 0, MT_PUSH) -- (Change when needed)
	container["ac_" .. charname].sprite = sprite or SPR_PLAY
	if skin then container["ac_" .. charname].skin = skin or "Sonic" end
	container["ac_" .. charname].radius = radius
	container["ac_" .. charname].shadowscale = shadow
	container["ac_" .. charname].color = color or SKINCOLOR_BLUE

	-- Sets custom attributes in function format if wanted
	if attr then
		do attr(container["ac_" .. charname]) end
	end
end

-- Creates an object that can be modified by the sequence (prefixed under m_)
function EvActionSH.Sq_CreateMobjActor(container, mobjname, mobjtype, position, attr)

	local x = (position.x and not nil) and position.x or 0
	local y = (position.y and not nil) and position.y or 0
	local z = (position.z and not nil) and position.z or 0

	container["m_" .. mobjname] = P_SpawnMobj(x, y, z, mobjtype)

	-- Sets custom attributes in function format if wanted
	if attr then
		do attr(container["m_" .. mobjname]) end
	end
end

-- ------------------------------------------------------
-- @region OBJECT - Object/Mobj functions
-- ------------------------------------------------------

-- Gets distance between two mobj/point
function EvActionSH.getmobjdistance(mobja, mobjb)

	if not ((mobja and mobja.valid) or (mobjb and mobjb.valid)) then return end -- no mobj

	return FixedHypot(FixedHypot(mobja.x-mobjb.x, mobja.y-mobjb.y), mobja.z-mobjb.z)
end

-- A wrapper for P_TeleportMove in the eventaction table with extras
function EvActionSH.movemobjto(mobj, position, interpolate)

	position = EvActionSH.defvalue($1, {x=0,y=0,z=0})

	if not (mobj and mobj.valid) then return end -- no mobj

	local px = EvActionSH.defvalue(position.x, 0)
	local py = EvActionSH.defvalue(position.y, 0)
	local pz = EvActionSH.defvalue(position.z, 0)
	-- local px = (position.x and not nil) and position.x or 0
	-- local py = (position.y and not nil) and position.y or 0
	-- local pz = (position.z and not nil) and position.z or 0

	local warporigin = interpolate and P_MoveOrigin or P_SetOrigin -- or P_TeleportMove
	warporigin(mobj, px, py, pz) -- (2.2.11 interpolation patch)
	-- P_MoveOrigin(mobj, px, py, pz)
end

-- Moves a mobj relatively from its current position
function EvActionSH.movemobjrelative(mobj, rx, ry, rz, interpolate)

	if not (mobj and mobj.valid) then return end -- no mobj

	local ax = (mobj.x and not nil) and mobj.x or 0
	local ay = (mobj.y and not nil) and mobj.y or 0
	local az = (mobj.z and not nil) and mobj.z or 0
	-- local position = {x=ax + rx, y=ay + ry, z=az + rz}

	local warporigin = interpolate and P_MoveOrigin or P_SetOrigin -- or P_TeleportMove
	warporigin(mobj, ax + rx, ay + ry, az + rz) -- (2.2.11 interpolation patch)
	-- P_MoveOrigin(mobj, ax + rx, ay + ry, az + rz)
end

-- Event action wrapper for changing the mobj sprite (+frame)
function EvActionSH.setmobjsprite(mobj, newsprite, newframe)

	if not (mobj and mobj.valid) then return end -- no mobj

	mobj.sprite = newsprite or SPR_PLAY -- fallback to error sprite (use play, so it's not confusing)
	if (newframe) then mobj.frame = newframe end
end

-- Event action wrapper for changing the mobj frame
function EvActionSH.setmobjframe(mobj, newframe)

	if not (mobj and mobj.valid) then return end -- no mobj

	mobj.frame = newframe or 0
end

-- Sets a mobj scale (or destscale)
function EvActionSH.setmobjscale(mobj, scale, speed)

	if not (mobj and mobj.valid) then return end -- no mobj

	if (speed) then
		mobj.destscale = scale
		mobj.scalespeed = speed
	else
		mobj.scale = scale
	end
end

-- Sets multiple mobj scale (wrapper)
function EvActionSH.setmultiscale(mobjlist, scale, speed)
	for i=1,#mobjlist do
		-- if not (mobjlist[i] and mobjlist[i].valid) then return end -- no mobj
		EvActionSH.setmobjscale(mobjlist[i], scale, speed)
	end
end

-- Wrapper for setting the dontdraw flag
function EvActionSH.setmobjvisible(mobj, isvisible)
	if not (mobj and mobj.valid) then return end -- no mobj

	if (isvisible) then
		mobj.flags2 = $1 &~ MF2_DONTDRAW
	else
		mobj.flags2 = $1|MF2_DONTDRAW
	end
end

-- Make the mobj look at a direction (in angles) (prev: mobjchangeangle)
function EvActionSH.setmobjeangle(mobj, directionangle, addangle)
	if not (mobj and mobj.valid) then return end -- no mobj
	mobj.angle = (addangle == true) and $1+directionangle or directionangle
end

-- Make the mobj look at another mobj or point (prev: mobjlookat)
function EvActionSH.setmobjlookat(mobj, point)
	if not (mobj and mobj.valid) then return end -- no mobj
	mobj.angle = R_PointToAngle2(mobj.x, mobj.y, point.x, point.y)
end

-- Get the direction of where to look at a mobj or point
function EvActionSH.getlookangle(mobj, point)
	if not (mobj and mobj.valid) then return end -- no mobj
	return R_PointToAngle2(mobj.x, mobj.y, point.x, point.y)
end

-- Set the mobj roll/rotation angle (or continuously add)
function EvActionSH.setspriteroll(mobj, angle, addangle)
	if not (mobj and mobj.valid) then return end -- no mobj

	if (addangle) then
		mobj.rollangle = $1+angle
	else
		mobj.rollangle = angle
	end
end

-- Stops the mobj's momentum
function EvActionSH.stopmobjmomentum(mobj)
	if not (mobj and mobj.valid) then return end -- no mobj

	mobj.momx = 0
	mobj.momy = 0
	mobj.momz = 0
end

-- ------------------------------------------------------
-- @region EFFECT - Effect functions
-- ------------------------------------------------------

-- ------------------------------------------------------
-- @region GAMEMAP - Game Map functions
-- ------------------------------------------------------

-- For coordinate placement (binary/non-udmf), set an origin to reference positions from (prefixed with mo_ 'map origin')
function EvActionSH.coordinateanchor(container, placement_name, mapx, mapy, mapz, mangle)
	container["mo_" .. placement_name] = {x=mapx, y=mapy, z=mapz}
end

-- Uses the coordinate anchor given to add positions to a made-up grid
function EvActionSH.coordinategetoffset(container, placement_name, offsetx, offsety, offsetz)
	if not container["mo_" .. placement_name] then
		print(string.format("Coordinate anchor %s does not exist!", placement_name))
		return
	end

	local anchor = container["mo_" .. placement_name]

	return {x=anchor.x + offsetx, y=anchor.y + offsety, z=anchor.z + offsetz}
end

-- ------------------------------------------------------
-- @region VIDEO - Video functions
-- ------------------------------------------------------

EvActionSH.vidwidth = 0
EvActionSH.vidheight = 0
EvActionSH.viddup = 0

-- Toggles default hud visibility
function EvActionSH.SetInternalHudStatus(namelist, huditemvisible)
	for i=1,#namelist do
		if (huditemvisible) then
			hud.enable(namelist[i])
		else
			hud.disable(namelist[i])
		end
	end
end

-- Sets the container of the sequence to use HUD drawing
function EvActionSH.togglerender(container, setting)
	container.sendtohud = (setting == nil or setting == true) and true or false
	if not container["huditems"] then container["huditems"] = {} end
end

-- TODO: Edit another render from another event
-- function EvActionSH.updaterenderfrom(evname, itemname)
-- end

-- Edits a hud item in the container
function EvActionSH.editrender(container, name)
	return container["huditems"][name]
end

function EvActionSH.renderitemvisibility(container, name, visible)
	container["huditems"][name].order = (visible and not nil) and visible or -1
end

-- Adds drawfill to the hud item list
function EvActionSH.renderfill(container, order, name, x, y, width, height, color)
	-- TODO: priority sort
	-- local tempname = string.format("%s_%s", order, name)
	if not container["huditems"] then print("Please use togglerender() first to enable the screen renderer!") return end
	container["huditems"][name] = {}
	container["huditems"][name].huditem = true
	container["huditems"][name].order = order
	container["huditems"][name].type = "fill"
	container["huditems"][name].x = x or (x and not nil) and x or 0
	container["huditems"][name].y = y or (y and not nil) and y or 0
	container["huditems"][name].width = width or 0
	container["huditems"][name].height = height or 0
	container["huditems"][name].color = color or 0
end

-- Adds drawstretched to the hud item list
function EvActionSH.renderstretchedpatch(container, order, name, x, y, hscale, vscale, patch, flags, colormap)
	-- TODO: priority sort
	-- local tempname = string.format("%s_%s", order, name)
	if not container["huditems"] then print("Please use togglerender() first to enable the screen renderer!") return end
	container["huditems"][name] = {}
	container["huditems"][name].huditem = true
	container["huditems"][name].order = order
	container["huditems"][name].type = "stretchedgraphic"
	container["huditems"][name].x = x or (x and not nil) and x or 0
	container["huditems"][name].y = y or (y and not nil) and y or 0
	container["huditems"][name].hscale = hscale or FU
	container["huditems"][name].vscale = vscale or FU
	container["huditems"][name].patch = patch or "UNKNOWN"
	container["huditems"][name].flags = flags or (flags and not nil) and flags or 0
	container["huditems"][name].colormap = colormap or 0
end

-- Adds drawscaled to the hud item list
function EvActionSH.renderscaledpatch(container, order, name, x, y, scale, patch, flags, colormap)
	-- TODO: priority sort
	-- local tempname = string.format("%s_%s", order, name)
	if not container["huditems"] then print("Please use togglerender() first to enable the screen renderer!") return end
	container["huditems"][name] = {}
	container["huditems"][name].huditem = true
	container["huditems"][name].order = order
	container["huditems"][name].type = "scaledgraphic"
	container["huditems"][name].x = x or (x and not nil) and x or 0
	container["huditems"][name].y = y or (y and not nil) and y or 0
	container["huditems"][name].scale = scale or FU
	-- Assign sprite patch or a normal patch
	if (type(patch) == "table") then
		container["huditems"][name].type = "scaledsprite" -- change the name
		container["huditems"][name].patch = patch[1] or SPR_PLAY
		container["huditems"][name].spr_frame = patch[2] or 0
		container["huditems"][name].spr_rotate = patch[3] or 0
		container["huditems"][name].spr_roll = patch[4] or 0
	else
		container["huditems"][name].patch = patch or "UNKNOWN"
	end
	container["huditems"][name].flags = flags or (flags and not nil) and flags or 0
	container["huditems"][name].colormap = colormap or 0
end

-- Adds drawstring to the hud item list
function EvActionSH.rendertext(container, order, name, x, y, text, flags, alignment)
	-- local tempname = string.format("%s_%s", order, name)
	if not container["huditems"] then print("Please use togglerender() first to enable the screen renderer!") return end
	container["huditems"][name] = {}
	container["huditems"][name].huditem = true
	container["huditems"][name].order = order
	container["huditems"][name].type = "string"
	container["huditems"][name].x = x or (x and not nil) and x or 0
	container["huditems"][name].y = y or (y and not nil) and y or 0
	container["huditems"][name].text = text or ""
	container["huditems"][name].textflags = flags or 0
	container["huditems"][name].textalign = alignment or "left"
end


-- Sequence version of fadescreen
EventSH.new("h_fadescreen", {
function(c,e)
	 -- frame 1 of the initialization
	EventSH.wait(e, c.waitingtime or 0, true)
end,
function(c,e)
	-- c.showfade = true

    -- Options (full, in, out)
	if (c.fadetype == "full") then
		c.time = c.maxstrength
	else
		if (leveltime % c.delay == 0) and not paused then
			if c.fadetype == "out" then
				c.time = min(c.maxstrength, $+1 * 1)
			elseif c.fadetype == "in" then
				c.time = max(0, $-1 * 1)
			end
		end
	end
	EventSH.waitUntil(e, e.status == "dead")
end,
function(c,e)
end})

-- The return of fadescreen()!
function EvActionSH.fadescreen(fadetype, color, delay, waitingtime, fadedplayers, fadetag)
	local t = {}

	t.fadetype = fadetype or "clear"
	-- t.fadetype = (fadetype == "in" or fadetype == "out" or fadetype=="clear") and fadetype or "clear"

	-- t.color = color
	-- Control colortypes
	local colortypes = {
		level = 0xFF00,
		title = 0xFA00,
		special = 0xFB00,
    }
    for k,v in pairs(colortypes) do
        if (color and color == k) then
            t.color = v
            break
        else
            t.color = color or 0 -- color
        end
    end
    -- Max strength is different for those other fade colors
	t.maxstrength = ((t.color == 0xFF00 or t.color == 0xFA00 or t.color == 0xFB00) and 31 or 10)
	-- t.maxstrength = 10
	t.time = 0
	t.delay = delay or 1

	if t.fadetype == "in" then t.time = t.maxstrength end
	if t.fadetype == "full" then t.time = t.maxstrength end

	-- Fade Tag ID stores what used the function to avoid overwrite conflicts.
	if (fadetag) then
		-- print("Set TID:" .. fadetag)
		t.fadetag = fadetag
	end

	if t.fadetype == "clear" then
		-- print("Removing TID:" .. fadetag)
		EventSH.destroy("h_fadescreen", fadetag and {haskey={"fadetag", fadetag}} or nil)
	else
		-- Filter players from the passed table to only show fading for included players only, otherwise show for everyone
		if (fadedplayers and #fadedplayers > 0) then
			EvActionSH.filterforplayers(t, fadedplayers)
			-- t.fadedplayers = fadedplayers -- (currently unused)
		end
		-- Set up graphics toggle and others, then start
		t.waitingtime = waitingtime or 0
		t.showfade = true
		EventSH.destroy("h_fadescreen", fadetag and {haskey={"fadetag", fadetag}} or nil)
		EventSH.start("h_fadescreen", t)
	end
end

function EvActionSH.fadescreenopt(prefs)
	local fadetype = EvActionSH.defvalue(prefs.fade, "clear")
	local color = EvActionSH.defvalue(prefs.color, nil)
	local delay = EvActionSH.defvalue(prefs.delay, nil)
	local waitingtime = EvActionSH.defvalue(prefs.waiting, nil)
	local fadedplayers = EvActionSH.defvalue(prefs.playerfilter, nil)
	local fadetag = EvActionSH.defvalue(prefs.fadetag, nil)
	EvActionSH.fadescreen(fadetype, color, delay, waitingtime, fadedplayers, fadetag)
end

-- Sequence version of drawborder
EventSH.new("a_cineborder", {
function(c,e)
	EvActionSH.togglerender(c)
end,
function(c,e)
	EvActionSH.renderstretchedpatch(c, c.order, "cineborder_top", 0, 0, EvActionSH.vidwidth*FU,c.width*FU, "~031G", V_SNAPTOTOP, 0)
	EvActionSH.renderstretchedpatch(c, c.order, "cineborder_bottom", 0, (200-c.width)*FU, EvActionSH.vidwidth*FU, c.width*FU, "~031G", V_SNAPTOBOTTOM, 0)
	EventSH.waitUntil(e, false)
end,
function(c,e)
end})

function EvActionSH.drawcineborders(c, order, borderwidth)
	local t = {}

	t.order = order or 50 -- drawing order
	t.width = borderwidth or 8 -- border width

	-- EventSH.destroy("a_cineborder")
	-- EventSH.start("a_cineborder", t)

	-- temporary until sequence can appear under other graphics drawn (+animations)
	EvActionSH.renderstretchedpatch(c, t.order, "cineborder_top", 0, 0, EvActionSH.vidwidth*FU, t.width*FU, "PAL_31G", V_SNAPTOTOP|V_SNAPTOLEFT, 0)
	EvActionSH.renderstretchedpatch(c, t.order, "cineborder_bottom", 0, (200-t.width)*FU, EvActionSH.vidwidth*FU, t.width*FU, "PAL_31G", V_SNAPTOBOTTOM|V_SNAPTOLEFT, 0)
end

function EvActionSH.removecineborders(container)
	container["huditems"]["cineborder_top"] = nil
	container["huditems"]["cineborder_bottom"] = nil
end

-- Checks if the graphics EVS hud control has any assigned players
-- to it if the table exists, then returns true if there is a display mismatch.
-- Used to stop showing graphics to all players, and only each hudplayer individually
function EvActionSH.EVassignedplayer_isnt_hudplayer(rc, hudplayer)

	if rc.assignedplayers then
		for i, pl in ipairs(rc.assignedplayers) do
			if (EvActionSH.isplayervalid(consoleplayer) and hudplayer ~= pl) then return true else return false end
		end
	end
end

-- Reads sequence data for hud drawing content
function EvActionSH.hudcontrol(v, stplyr, cam)

	-- Hacks to get vid information that should be non-exclusive to hud :|
	EvActionSH.vidwidth = v.width()
	EvActionSH.vidheight = v.height()
	EvActionSH.viddup = v.dupx()

	EventSH.read("any", function(rc)

		-- Graphics should only show to the players assigned to this unless specified (ie. dont assign players)
		if EvActionSH.EVassignedplayer_isnt_hudplayer(rc, stplyr) then return end

		-- replacement for render() and renderplayer()
		if not (rc.sendtohud == true) then return end -- is not drawing to the HUD

		for keyname, rend in EventSH.spairs(rc.huditems, function(t,a,b) return t[b].order > t[a].order end) do

			if (type(rend) == "table" and rend.huditem) then

				if (rend.order == -1) then continue end -- skip negative ordered

				if (rend.type == "fill") then
					v.drawFill(rend.x, rend.y, rend.width, rend.height, rend.color)
				elseif (rend.type == "stretchedgraphic") then
					v.drawStretched(rend.x, rend.y, rend.hscale, rend.vscale, v.cachePatch(rend.patch), rend.flags or 0, v.getColormap(1, rend.colormap))
				elseif (rend.type == "scaledgraphic") then
					v.drawScaled(rend.x, rend.y, rend.scale, v.cachePatch(rend.patch), rend.flags or 0, v.getColormap(1, rend.colormap))
				elseif (rend.type == "scaledsprite") then
					local spr = v.getSpritePatch(rend.patch, rend.spr_frame, rend.spr_rotate, rend.spr_roll)
					v.drawScaled(rend.x, rend.y, rend.scale, spr, rend.flags or 0, v.getColormap(1, rend.colormap))
				elseif (rend.type == "string") then
					v.drawString(rend.x, rend.y, rend.text, rend.textflags, rend.textalign)
				end
			end
		end
	end)
	EventSH.read("h_fadescreen", function(rc)

		-- Graphics should only show to the players assigned to this unless specified (ie. dont assign players)
		if EvActionSH.EVassignedplayer_isnt_hudplayer(rc, stplyr) then return end

		-- Handle the hud entry options (full, clear, or default)
		if rc.showfade then
			v.fadeScreen(rc.color, rc.time)
		end
	end)
end

-- Detect if the player is viewing through a custom EV-made camera (allows seenames through any EV that doesn't use a camera)
function EvActionSH.seenamesoff(player, seenplayer)
	-- :|
	if not (player and player.mo and player.mo.valid) then return end

	if (((player.awayviewmobj and player.awayviewmobj.valid) and player.awayviewmobj.isevcamera))
	or (((seenplayer.awayviewmobj and seenplayer.awayviewmobj.valid) and seenplayer.awayviewmobj.isevcamera)) then
		return false
	end
end

