/*
ForceBot
	by Lach

Lua.forcebot = <skinname>
	- Sets the name of the skin to force the 2P bot to use. If the 2P bot does not exist, a new bot is spawned.
Lua.forcebot_persistent = 1
	- By default, any bot spawned by this script will be removed when switching to a non-forcebot map. Set this parameter to instead allow the bot to be carried into other levels.
	- For vanilla 2P bots, this parameter lets the bot keep its skin when entering other levels, otherwise it will switch back to its old skin.

Reusability clause:
	- You may distribute copies of this script as-is in any add-on released on the SRB2 Message Board.
	- You may edit this script for private use or to experiment with/learn about SRB2 Lua scripting.
	- You may not distribute copies of this script that are edited by an author not listed in this header.
	- You may reuse code from this script in any add-on released on the SRB2 Message Board on the condition that it does not conflict with this script when both are loaded.
*/

local version = 0
local firstInstance = true

if ForceBot
	if ForceBot.version >= version
		return
	end
	
	firstInstance = false
else
	rawset(_G, "ForceBot", {})
end

local FB = ForceBot

FB.version = version

// ---------
// variables
// ---------

if firstInstance == true
	FB.persistent = false
end

// ---------
// functions
// ---------

function FB.RemoveBot()
	local player = FB.bot
	FB.bot = nil
	
	if not (player and player.valid)
		return
	end
	
	if FB.oldSkin
		R_SetPlayerSkin(player, FB.oldSkin)
		FB.oldSkin = nil
		FB.newSkin = nil
	else
		player.forceBot_hideMe = true
		G_RemovePlayer(#player)
	end
end

function FB.MapChange(mapNum)
	// never use this in multiplayer
	if multiplayer
		FB.RemoveBot()
		return
	end
	
	local map = mapheaderinfo[mapNum]
	
	if not map
	or not map.forcebot
		if not FB.persistent
			FB.RemoveBot()
		end
		return
	end
	
	local skinName = map.forcebot:lower()
	FB.newSkin = skinName
	
	// check whether this skin exists at all
	if not skins[skinName]
		if not FB.persistent
			FB.RemoveBot()
			return
		end
	end
	
	local player = players[1]
	
	if player
		if not FB.oldSkin
			FB.oldSkin = skins[player.skin].name
		end
	else
		player = G_AddPlayer(skinName, skins[skinName].prefcolor, nil, BOT_2PAI)
	end
	FB.bot = player
	
	FB.persistent = (map.forcebot_persistent ~= nil)
end

function FB.PlayerSpawn(player)
	if player == FB.bot
		R_SetPlayerSkin(player, FB.newSkin)
	end
	
	if not player.forceBot_hideMe
		return
	end
	
	local mo = player.mo
	
	if not mo
		return
	end
	
	// TO THE SHADOW REALM WITH YOU
	(P_SetOrigin or P_TeleportMove)(mo, INT32_MIN, INT32_MIN, INT32_MIN)
	mo.state = S_INVISIBLE
end

function FB.PlayerQuit(player)
	if player == FB.bot
		FB.bot = nil
		FB.oldSkin = nil
		FB.newSkin = nil
	end
end

function FB.GameQuit()
	FB.RemoveBot()
	FB.oldSkin = nil
	FB.newSkin = nil
	FB.persistent = false
end

// -----
// hooks
// -----

if not firstInstance
	return
end

addHook("MapChange", function(...) return FB.MapChange(...) end)
addHook("PlayerSpawn", function(...) return FB.PlayerSpawn(...) end)
addHook("PlayerQuit", function(...) return FB.PlayerQuit(...) end)
addHook("GameQuit", function (...) return FB.GameQuit(...) end)