--[[
* frameanimate.lua
* (Author: sprki_zard)
* (June 6, 2020 23:29)
* Desc: n/a
*
* Notes: n/a
]]
rawset(_G, "FrameAnimSH", {})

FrameAnimSH.playing = {}


-- Adds a new animation to the playing animation list
function FrameAnimSH.add(animname, animator, nsprite, startf, endf, args, use_dupe)

	-- Duplicate animation? Set it to be replaced immediately
	if not (use_dupe) then
		for i,anim in ipairs(FrameAnimSH.playing) do
			if (anim.name == animname) then
				anim.deleted = true
			end
		end
	end

	-- Create Animation Table
	local anim = {
		name=animname,
		mobj=animator,
		sprite=nsprite,
		timer=0,
		frame=R_Char2Frame(startf), --startf,
		startframe=(args and args.loopframe and R_Char2Frame(args.loopframe)) or R_Char2Frame(startf), --startf,
		endframe=R_Char2Frame(endf or startf), --endf,
		loop=(args and args.loop) or ((R_Char2Frame(endf)-R_Char2Frame(startf)) * ((args and args.delay) or 1)), -- if no loop, set to end frame (ef*delay for slower animations)
		forever=(args and args.loopalways),
		frameskip=(args and args.frameskip) or 1, -- TODO: implement frameskip
		delay=(args and args.delay) or 1,
		flag=(args and args.flag) or 0, -- TODO: sprite flag that applies to separate frames. currently does not work.
		-- spritelist = {}, -- TODO: separate flags per frame?
		paused=false,
		deleted=false,
	}

	-- Throw it into the playing list
	table.insert(FrameAnimSH.playing, anim)

	--Set object to these frames
	animator.sprite = nsprite
	animator.frame = anim.frame | anim.flag

	-- print("Inserted Animation: "..animname)
end

-- add wrapper for spr2 animations
function FrameAnimSH.addspr2(spr2animname, animator, spr2, framestart, frameend, animargs, use_dupe)

	local skin = false -- animargs and animargs.skin

	-- Set skin if given one (used for objects, not a good idea for players)
	if skin then animator.skin = skin end

	-- Set spr2 frame
	animator.sprite2 = spr2

	-- Super frame swapper
	if animargs and animargs.super then
		animator.sprite2 = $1 | FF_SPR2SUPER
	end

	-- TODO: report to issues: SPR2_STND cannot be accessed with skin.sprites[]
	local maxframe = 0
	if (spr2 == SPR2_STND) then
		maxframe = frameend
	else
		maxframe = R_Frame2Char(skins[animator.skin].sprites[spr2].numframes-1)
	end

	-- FrameAnimSH.add("anim_spr2_play", player.mo, SPR_PLAY, framestart or "A", frameend or maxframe, animargs)
	FrameAnimSH.add("anim_spr2_" .. spr2animname, animator, SPR_PLAY, framestart or "A", frameend or maxframe, animargs, use_dupe)
end

-- Seeks an animation and runs a function when found (also returns true if exists, false if not)
function FrameAnimSH.seek(animname, callback)
	for i,anim in ipairs(FrameAnimSH.playing) do
		if (anim.name == animname) then
			if (callback) then
				do callback(anim) end
			end
			return true
		end
	end
	return false
end

-- Removes one or all animations
function FrameAnimSH.remove(animname, removeall)
	if (removeall) then
		for i=1, #FrameAnimSH.playing do
			FrameAnimSH.playing[i].deleted = true
		end
	else
		FrameAnimSH.seek(animname, function(an) an.deleted = true end)
	end
end

-- Pauses an animation
function FrameAnimSH.pause(animname, ispaused)
	FrameAnimSH.seek(animname, function(an) an.paused = ispaused end)
end

-- Checks if an animation just finished
function FrameAnimSH.checkexisting(animname)
	return FrameAnimSH.seek(animname)
end

-- ================
-- Event Extension
-- Checks if a certain script is loaded to extend some of it's features
function FrameAnimSH.EventExists()
	if not _G["EventSH"] then print("(!)\x82 This function requires 'Event' to work!") return false else return true end
end

-- Add a custom wrapper event + others when the companion script exists
if FrameAnimSH.EventExists() then

	-- Suspends an event until the animation ends
	function FrameAnimSH.waitframedone(event, animname)
		if (FrameAnimSH.checkexisting(animname)) then EventSH.pause(event) else EventSH.resume(event) end
	end
-- else
-- 	local handler_str = "(!)\x82 Function [FrameAnimSH.waitframedone] requires 'Event' to work, is it available?"

-- 	function FrameAnimSH.waitframedone()
-- 		print(handler_str)
-- 	end
end

-- ================


-- The thinker function that handles animations
function FrameAnimSH.Thinker()

	for i=1, #FrameAnimSH.playing do

		local anim = FrameAnimSH.playing[i]

		-- Remove if animation is set to be deleted
		if (anim and anim.deleted) then
			table.remove(FrameAnimSH.playing, i)
		end

		-- no mobj means to not continue
		if anim and not (anim.mobj and anim.mobj.valid) then
			anim.deleted = true
		end

		-- Continue next iteration if animation is set to be deleted, otherwise continue
		(function()
			if (anim and not anim.deleted) then

				-- Set sprite
				anim.mobj.sprite = anim.sprite
				anim.mobj.frame = anim.frame | anim.flag

				-- the animation is paused
				if (anim.paused) then return end

				-- increase animation timer
				anim.timer = $ + 1

				-- loop on time, or loop forever
				if (anim.loop > 0 or anim.forever) then

					-- Play the sprite's frames by set speed and delay
					if (anim.timer % anim.delay == 0) then
						anim.frame = ($1+1 * anim.frameskip)
					end

					-- Reset the framecount to the beginning
					if (anim.frame > anim.endframe) then
						anim.frame = anim.startframe
					end

					anim.loop = anim.forever and 1 or $1-1
				else
					-- When our sprite loop time has ended, set the animation to be removed
					anim.deleted = true
					-- print(string.format("animation [%s] finished", anim.name))
				end

				--[[-- Play the sprite's frames by set speed and delay
				if (anim.timer % anim.delay == 0) then
					anim.frame = $1+1 * anim.speed
				end

				-- Reset the framecount to the beginning
				if (anim.frame > anim.endframe) then
					anim.frame = anim.startframe
				end

				-- Set sprite again
				anim.mobj.frame = anim.frame | anim.flag

				-- When our sprite loop time has ended, set the animation to be removed
				if (anim.loop <= 0 and not anim.forever) then
					anim.deleted = true
					-- print("animation finished")
					return
				else
					anim.loop = $1-1
				end--]]
			end
		end)()
	end

end

function FrameAnimSH.netvars(n)
	FrameAnimSH.playing = n($)

	--[[local a = #FrameAnimSH.playing
	a = n(a)
	for i = 1, a do
		FrameAnimSH.playing[i] = n($)
	end--]]
end



-- Uncomment for Example
--[[addHook("NetVars", function(network)
	FrameAnimSH.netvars(network)
end)

addHook("ThinkFrame", function()
	FrameAnimSH.Thinker()
end)--]]
--[[addHook("ThinkFrame", function()

	-- Checks for existing
	print(FrameAnimSH.checkexisting("existing"))

	if leveltime == 2*TICRATE then
		local s = P_SpawnMobj(-64*FU, -96*FU, 32*FU, MT_PULL)
		-- FrameAnimSH.add("name", s, SPR_EGGM, A, F)
		-- FrameAnimSH.add("name", s, SPR_EGGM, V, W, {loop=2*TICRATE})
		-- FrameAnimSH.add("name", s, SPR_PIKE, A, P)
		FrameAnimSH.add("name", s, SPR_GFZD, R_Frame2Char(0), R_Frame2Char(31), {loop=10*TICRATE})
	end

	if leveltime == 7*TICRATE then
		-- FrameAnimSH.pause("name", true)
	end

	if leveltime == 10*TICRATE then
		-- FrameAnimSH.remove("name", true)
	end

	if leveltime == 8*TICRATE then
		local s = P_SpawnMobj(-64*FU, -96*FU, 64*FU, MT_PULL)
		FrameAnimSH.add("existing", s, SPR_EGGM, "V", "W", {loop=8*TICRATE, delay=15})
	end
end)
--]]