/*
Inazuma the Deer
Battle Support
	by Lach
	& Krabs
*/

local B
local S
local Act
local G

local Inazuma = Silverhorn.Inazuma
local HITLAG_TIME = TICRATE/3 -- Bolt Stream Hit Lag Time
local TUMBLE_TIME = 48 -- Bolt Stream Tumble Time

local STATE_CHARGE = 1
local STATE_RELEASE = 2

local CHARGE_MIN_TIME = 5
local CHARGE_MAX_TIME = TICRATE
local CHARGE_STALL_TIME = CHARGE_MAX_TIME/2 + CHARGE_MAX_TIME
local CHARGE_END_TIME = TICRATE/3

local CHARGE_COST = 10
local CHARGE_COOLDOWN = TICRATE*11/4
local CANCEL_COOLDOWN = TICRATE*3/2

local FLASH_MIN_RANGE, FLASH_MAX_RANGE = FRACUNIT*64, FRACUNIT*256
local FLASH_MIN_THRUST = FRACUNIT*19/4
local FLASH_MIN_LAUNCH = FRACUNIT*8
local FLASH_MIN_TUMBLE = TICRATE*3/2
local FLASH_MAX_MODIFIER = FRACUNIT*3/2	-- affects FLASH_MIN_THRUST, FLASH_MIN_LAUNCH and FLASH_MIN_TUMBLE based on charge
local FLASH_COOLDOWN = TICRATE*7/2

local DEBUFF_TICS = TICRATE*12
local DEBUFF_MIN_FLASHBANG = TICRATE*7/10	-- also affected by FLASH_MAX_MODIFIER
local DEBUFF_FLASHBANG_BLIND = FRACUNIT*2/3	-- how long of the flashbang duration can we not see ANYTHING?
local DEBUFF_SOUND_MAX = 80	-- the max volume (remember it'll overlay)
local DEBUFF_SOUND_FADE = TICRATE*4	-- how long do our ears ring

local ScriptLoaded = false

//Shock thinker
addHook("MobjThinker", function(mo)
	//Don't allow this to slide up slopes
	if mo.prevz == nil
		mo.prevz = mo.z
	end
	if mo.prevz != mo.z
		P_KillMobj(mo)
	end
	mo.prevz = mo.z
	
	//Match the scale and color of the target object
	local target = mo.target
	if not mo.color and target and target.valid
	and target.inazuma_shockcolor
		mo.color = target.inazuma_shockcolor
		if mo.scale != target.scale
			mo.scale = target.scale
			mo.momx = FixedMul($, mo.scale)
			mo.momy = FixedMul($, mo.scale)
			mo.momz = FixedMul($, mo.scale)
		end
	end
	
	//Facciolo behavior
	if mo.target.type == MT_FACCIOLO_BOSS
		mo.blendmode = AST_ADD
		mo.renderflags = $|RF_FULLBRIGHT

		if leveltime & 1
			local trail = P_SpawnGhostMobj(mo)
			trail.blendmode = AST_ADD
		end
	end
end, MT_ZUMASPREAD)

//Electric hit
addHook("MobjDamage", function(target, mo)
	if mo and mo.valid and mo.type == MT_ZUMASPREAD
		S_StartSound(target, sfx_elcbmp)
	end
end)

local InaCollide = function(n1,n2,plr,mo,atk,def,weight,hurt,pain,ground,angle,thrust,thrust2,collisiontype)
	if not (plr[n1] and plr[n1].valid and plr[n1].playerstate == PST_LIVE)
		or not plr[n1].inazuma
		or not mo[n1].health
		or not (mo[n1].state == S_INAZUMA_BOLTSTREAM)
		or plr[n1].pflags & PF_SPINNING
		return false
	end

	local inazuma = plr[n1].inazuma
	
	if (hurt == 1 and n1 == 1) or (hurt == -1 and n1 == 2)
		Inazuma.ResetBoltStream(plr[n1])
		mo[n1].momx = $ / 2
		mo[n1].momy = $ / 2
		mo[n1].momz = P_MobjFlip(mo[n1])*FixedMul(mo[n1].scale, 2*plr[n1].normalspeed/5)
		
		//Gleem
		local gleam = P_SpawnMobj(
			mo[n2].x/2 + mo[n1].x/2,
			mo[n2].y/2 + mo[n1].y/2,
			((mo[n1].z + mo[n1].height >> 1) / 2) + ((mo[n2].z + mo[n2].height >> 1) / 2),
			MT_GHOST)
		gleam.sprite = SPR_IZFX
		gleam.scale = mo[n1].scale * 3
		gleam.blendmode = AST_SUBTRACT
		gleam.spritexscale = FRACUNIT >> 6
		gleam.spriteyscale = FRACUNIT << 6
		gleam.fuse = 12
		inazuma.gleam = gleam
		
		//Spike
		local spiked
		if P_MobjFlip(mo[n2]) == P_MobjFlip(mo[n1]) and not ground[n2]
			S_StartSound(mo[n1], sfx_boltsk)
			mo[n2].momz = -mo[n1].momz * 3/2
			spiked = true
		end
		
		plr[n1].pflags = $ & ~PF_STARTJUMP | PF_NOJUMPDAMAGE | PF_THOKKED
		inazuma.flags = $ | Inazuma.FLAGS.LIGHTNINGJUMP
		
		S_StartSound(mo[n1], sfx_ljmp)
		S_StartSound(mo[n1], sfx_hit01)
		
		local strike = Inazuma.SpawnMobj(mo[n1].x, mo[n1].y, mo[n1].z, MT_THOK, plr[n1])
		strike.scale = mo[n2].scale << 1
		strike.state = S_INAZUMA_LIGHTNINGSTRIKE
		strike.eflags = mo[n2].eflags
		if strike.eflags & MFE_VERTICALFLIP
			strike.z = $ + mo[n1].height - strike.height
		end
		
		//Set hitlag
		plr[n1].powers[pw_nocontrol] = HITLAG_TIME
		mo[n1].inazuma_stop = {tics = HITLAG_TIME, pushed = mo[n2],
		x = mo[n1].momx, y = mo[n1].momy, z = mo[n1].momz,
		angle = plr[n1].drawangle}
		if plr[n2]
			plr[n2].powers[pw_nocontrol] = HITLAG_TIME
			mo[n2].inazuma_stop = {tics = HITLAG_TIME, pushed = mo[n1],
			x = mo[n2].momx, y = mo[n2].momy, z = mo[n2].momz,
			angle = plr[n2].drawangle, shock = true, tumble = spiked}
		end
		
		return true
	elseif hurt == 0
		Inazuma.ResetBoltStream(plr[n1])
		mo[n1].state = S_PLAY_FALL
		plr[n1].pflags = $ | PF_THOKKED | PF_NOJUMPDAMAGE
		return false
	end
end

local InaMaxExhaustTime = 5*TICRATE
local InaExhaust = function(player)
	if player.inazuma and player.mo and player.mo.valid
		local inazuma = player.inazuma
		local mo = player.mo
		
		if inazuma.flags & Inazuma.FLAGS.BARRIER
			inazuma.exhaust_cooldown = 3 * TICRATE
			if not (inazuma.prevflags & Inazuma.FLAGS.BARRIER)
				player.exhaustmeter = max(0,$-10*FRACUNIT/InaMaxExhaustTime)//Penalty for tapping the shield
			end
			if player.gotflagdebuff
				player.exhaustmeter = max(0,$-(3/2*FRACUNIT)/InaMaxExhaustTime)
			else
				player.exhaustmeter = max(0,$-FRACUNIT/InaMaxExhaustTime)
			end
			if player.exhaustmeter == 0
				S_StartSound(mo,  sfx_s3k72)
			end
		else
			if inazuma.exhaust_cooldown
				inazuma.exhaust_cooldown = $ - 1
				if inazuma.exhaust_cooldown <= 0
					player.exhaustmeter = FRACUNIT
					S_StartSound(mo,  sfx_s3k5c)
					inazuma.exhaustcolor2 = 3
					mo.exhaustcolor = false
					mo.colorized = true
					mo.color = SKINCOLOR_SKY
				end
			end
		end
		if inazuma.exhaustcolor2
			inazuma.exhaustcolor2 = $ - 1
			if inazuma.exhaustcolor2 <= 0
				mo.colorized = false
				mo.color = player.skincolor
			end
		end
		inazuma.prevflags = inazuma.flags
	end
	return true//override the standard exhaustion refill condition
end

//Spawn the shockwave (rawset it so we can use it for Ultimate's Lightning Jump
rawset(_G, "InazumaProjectileStream", function(mo, color, charge)
	//Set the color on the spawner so we can give it to the shockwaves later
	if not color
		if mo.player
			color = Inazuma.GetLightningColor(mo.player)
		else
			color = SKINCOLOR_INAFX
		end
	end
	mo.renderflags = $ | RF_NOCOLORMAPS
	mo.inazuma_shockcolor = color
	
	local info = mobjinfo[MT_ZUMASPREAD]
	local oldfuse = info.painchance
	if charge
		info.painchance = $ * charge / 100
		A_Shockwave(mo, MT_ZUMASPREAD, 24) //Spawn the shockwave
		info.painchance = oldfuse
	else
		A_Shockwave(mo, MT_ZUMASPREAD, 24) //Spawn the shockwave
	end
end)

addHook("ThinkFrame", function()
	if not(ScriptLoaded) and CBW_Battle
		ScriptLoaded = true
		B = CBW_Battle
		S = B.SkinVars
		Act = B.Action
		G = B.GuardFunc
		
		mobjinfo[MT_ZUMASPREAD].name = "Overload shockwave"
		mobjinfo[MT_INAZUMA_LIGHTNINGBLAST].name = "Lightning Blast"
		
		//Bypassing Battle's underwater missile speed function
		B.UnderwaterMissileSuper = B.UnderwaterMissile
		B.UnderwaterMissile = function(mo)
			if mo.type != MT_ZUMASPREAD
				B.UnderwaterMissileSuper(mo)
			end
		end
		
		B.AddPriorityFunction("ebar", function(mo,other)
			local player = mo and mo.valid and mo.player
			if player and player.exhaustmeter
				player.exhaustmeter = max(0,$-20*FRACUNIT/InaMaxExhaustTime)
				if player.exhaustmeter == 0
					S_StartSound(mo, sfx_s3k72)
				end
			end
		end)
		
		B.Action.ElectricRing_Priority = function(player)
			if player.actionstate == STATE_CHARGE then
				B.SetPriority(player,1,1,nil,1,1,"Overload charge")
			else
				local inazuma = player.mo.inazuma
				if inazuma and inazuma.laser
					B.SetPriority(player,0,3,nil,0,3,"Ultra Lightning Blast")
				elseif inazuma and (inazuma.flags & Inazuma.FLAGS.BARRIER)
					B.SetPriority(player,1,2,"ebar",1,2,"Electric Barrier")
				elseif player.mo.state == S_INAZUMA_BOLTSTREAM
					B.SetPriority(player,1,1,nil,1,1,"Bolt Stream")
				end
			end
		end
		
		B.Action.RadialFlashBlockmap = function(ref, found, flashinfo)	
			if not (flashinfo) then return true end	-- no info given...
	
			if (ref == found) then return end	-- dont damage ourself!
			if not ((found and found.valid) or found.health) then return end	-- gotta be alive
			if not ((found.flags & (MF_SHOOTABLE|MF_ENEMY)) or found.player) then return end	-- not damageable
			
			-- airdodge check
			if (found.player and found.player.airdodge) then return end
			
			if not (P_CheckSight(ref, found)) then return end	-- no LoS
			local dist = FixedHypot(FixedHypot(found.x - ref.x, found.y - ref.y), found.z - ref.z)
			if not (dist <= flashinfo["range"]) then return end	-- too far away
			
			if (found.player)	-- flashbang them!
				if (B.MyTeam(ref.player, found.player)) then return end -- same team
				if (found.player.airdodge) then return end	-- airdodge check
				
				if (found.player.guard > 0)	-- they guarded it, protect from effects!
					if (B.SkinVars[found.player.skinvars].func_guard_trigger)
						B.SkinVars[found.player.skinvars].func_guard_trigger(found, ref, ref)
					elseif (B.SkinVars[-1].func_guard_trigger)
						B.SkinVars[-1].func_guard_trigger(found, ref, ref)
					end
				else
					-- tumble them!
					found.z = $ + P_MobjFlip(found)	-- GOD i hate SRB2
					found.momz = flashinfo["launch"] * P_MobjFlip(ref)
					local angle = R_PointToAngle2(ref.x, ref.y, found.x, found.y)
					B.DoPlayerTumble(found.player, flashinfo["tumble"], angle, flashinfo["thrust"], true)
					if (ref.player) then B.PlayerCreditPusher(found.player, ref) end
					
					-- slow them down
					found.inazuma_flashed = DEBUFF_TICS
					found.inazuma_flashed_colour = Inazuma.GetLightningColor(ref.player)
					
					-- think fast chucklenuts!
					found.inazuma_flashed_flashbang = flashinfo["flashbang"]
					found.inazuma_flashed_flashbang_max = flashinfo["flashbang"]	-- also save this
					found.inazuma_flashed_flashbang_blind = FixedMul(DEBUFF_FLASHBANG_BLIND, flashinfo["flashbang"])	-- useful cache
					S_StartSound(found, sfx_kc59)
				end
			else	-- just damage them
				P_DamageMobj(found, ref, ref, 1, 0)
			end
		end
		
		-- hijack B.GotFlagStats to debuff us!
		local oldFunc = B.GotFlagStats
		B.GotFlagStats = function(p)
			-- just call the old function, we dont need to hijack it
			local mo = p.mo
			if not ((mo and mo.valid) and mo.inazuma_flashed)
				oldFunc(p)
				return
			end
			
			-- hijack the function
			local oldGotFlag = p.gotflag
			p.gotflag = 1
			oldFunc(p)
			p.gotflag = oldGotFlag
		end
		
		B.Action.ElectricRing = function(mo,doaction,pressedaction)
			local player = mo.player
			
			if player.actionstate == STATE_CHARGE
				player.actiontext = "Charging " + min(player.actiontime * 100 / CHARGE_MAX_TIME, 100) + "%"
				player.action2text = "\128Radial Flash"
			else
				player.actiontext = "Overload"
				player.actionrings = CHARGE_COST
			end
			
			if not(B.CanDoAction(player))
				player.actionstate = 0 //Default action state
				return
			end
			
			if player.actionstate == 0
				if doaction == 1
					player.actionstate = STATE_CHARGE //enter state 1: Startup lag
					player.actiontime = 0 //state timer
					B.PayRings(player)
					S_StartSoundAtVolume(mo,sfx_s3ka1, 190)
					S_StartSound(mo,sfx_s3k9f)
				end
			end
			
			if player.actionstate == STATE_CHARGE //This state will serve as our startup lag
				player.actiontime = $ + 1 //Tic this upwards to keep track of how long we've been in this state.
				
				//Halt our momentum
				local an = R_PointToAngle2(0,0,mo.momx,mo.momy)
				local spd = R_PointToDist2(0,0,mo.momx,mo.momy)
				local momz = player.actiontime * FRACUNIT / CHARGE_STALL_TIME
				momz = max((FRACUNIT - $)*3/2, 0)
				P_Thrust(mo,an,-spd/6)
				P_SetObjectMomZ(mo, momz, false)
				player.pflags = $&~(PF_SPINNING|PF_JUMPED)
				
				//State
				if mo.state != S_TAILSOVERLAY_EDGE
					mo.state = S_TAILSOVERLAY_EDGE
					mo.tics = CHARGE_MAX_TIME/2
				end
				
				//Flash
				if player.actiontime == CHARGE_MAX_TIME
					S_StartSound(mo, sfx_s3k45)
					mo.state = S_TAILSOVERLAY_EDGE
					mo.frame = ($ &~ FF_FRAMEMASK) | FF_ADD
					mo.tics = 4
				end
				
				//Sparks
				Inazuma.SpawnParticle(mo, mo.scale, MT_INAZUMA_FX)
				if player.actiontime >= CHARGE_MAX_TIME
					Inazuma.SpawnParticle(mo, mo.scale)
				end
				
				if (player.inazuma and player.inazuma.buttons[BT_TOSSFLAG] == 1)
					-- cancel charge
					mo.state = S_PLAY_FALL
					player.actionstate = 0
					B.ApplyCooldown(player, CANCEL_COOLDOWN) //Waittime before player can use this action again
					S_StartSound(mo, sfx_kc65)
				else 
					//End charge
					player.actioninaspin = player.inazuma
					and (player.inazuma.buttons[BT_JUMP] == 1
					or player.inazuma.buttons[BT_SPIN] == 1)
					
					if (player.actiontime >= CHARGE_STALL_TIME
					or not doaction or player.actioninaspin)
					and player.actiontime >= CHARGE_MIN_TIME
						player.actionstate = STATE_RELEASE
						player.actioninacharge = min(player.actiontime * 100 / CHARGE_MAX_TIME, 100)
						player.actiontime = 0
						
						local cooldown = player.actioninaspin and FLASH_COOLDOWN or CHARGE_COOLDOWN
						mo.frame = ($ &~ FF_FRAMEMASK) | 2
						B.ApplyCooldown(player, cooldown) //Waittime before player can use this action again
					end
				end
			end
			
			if player.actionstate == STATE_RELEASE
				player.actiontime = $ + 1
				//Creating our attack
				if player.actiontime == 1 then
					S_StopSoundByID(mo, sfx_s3ka1)
					
					-- radial flash
					if player.actioninaspin
						-- calculate flashinfo then flashbang in range! FLASH_MAX_MODIFIER
						local multiplier = (player.actioninacharge * FRACUNIT) / 100
						
						local modifier_range = FLASH_MAX_MODIFIER - FRACUNIT
						local modifier = FRACUNIT + FixedMul(multiplier, modifier_range)
						
						local range_range = FLASH_MAX_RANGE - FLASH_MIN_RANGE
						
						local flashinfo = {
							range = FixedMul(mo.scale, FLASH_MIN_RANGE + FixedMul(multiplier, range_range)),
							thrust = FixedMul(mo.scale, FixedMul(modifier, FLASH_MIN_THRUST)),
							launch = FixedMul(mo.scale, FixedMul(modifier, FLASH_MIN_LAUNCH)),
							tumble = FixedMul(modifier, FLASH_MIN_TUMBLE),
							flashbang = FixedMul(modifier, DEBUFF_MIN_FLASHBANG),
						}
						searchBlockmap("objects", function(ref, found)
							return B.Action.RadialFlashBlockmap(ref, found, flashinfo)
						end, mo, mo.x - flashinfo["range"], mo.x + flashinfo["range"], mo.y - flashinfo["range"], mo.y + flashinfo["range"])
						
						-- effects
						local blast_offset = -(FRACUNIT*32)
						local blast = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_THOK)
						blast.state = S_INAZUMA_LIGHTNINGBLAST_FAST
						blast.scale = FixedDiv(flashinfo["range"], FRACUNIT*70)	-- SPR_RADF radius
						P_SetOrigin(blast, blast.x, blast.y, blast.z + FixedMul(blast.scale, blast_offset))	-- try to align it more with the player/radius
						Inazuma.ColorMobj(blast, player)
						S_StartSound(mo, sfx_boltbl)
						
						-- debug visualiser
						/*
						local angles = 24
						local angle_thoks = 16
						
						for angle_i = 1, angles
							local angle = ((360 / angles) * angle_i) * ANG1
							
							for thok_i = 1, angle_thoks
								local vertangle = ((360 / angle_thoks) * thok_i) * ANG1
								
								local xoff = FixedMul(cos(angle), FixedMul(cos(vertangle), flashinfo["range"]))
								local yoff = FixedMul(sin(angle), FixedMul(cos(vertangle), flashinfo["range"]))
								local zoff = FixedMul(sin(vertangle), flashinfo["range"])
								
								local thok = P_SpawnMobjFromMobj(mo, xoff, yoff, zoff, MT_THOK)
								thok.frame = ($ & ~(FF_TRANSMASK)) | (FF_FULLBRIGHT|FF_TRANS40)
								thok.scale = FixedMul(FRACUNIT/3, $)
								thok.tics = blast.tics
								Inazuma.ColorMobj(thok, player)
							end
						end*/
					//Stream
					else
						InazumaProjectileStream(mo, nil, player.actioninacharge)
						S_StartSound(mo, mobjinfo[MT_ZUMASPREAD].seesound)
						S_StartSoundAtVolume(mo,sfx_s3k4e, 100)
					end	
					
					//Set player state
					mo.state = S_TAILSOVERLAY_EDGE
				end
				
				//Do some recoil
				player.pflags = $&~(PF_SPINNING)
				player.drawangle = mo.angle
					
				//Reset player state
				if player.actiontime >= CHARGE_END_TIME
					mo.state = S_PLAY_FALL
					player.actionstate = 0
				else
					-- nvm, halt z momentum
					mo.momz = 0
				end
			end
		end //End of action function.
		
		S["inazuma"] = {
			flags = SKINVARS_GUARD|SKINVARS_NOSPINSHIELD,
			weight = 100,
			shields = 1,
			special = Act.ElectricRing,
			guard_frame = 1,
			func_priority_ext = Act.ElectricRing_Priority,
			func_collide = InaCollide,
			func_exhaust = InaExhaust
		}
		
		print("\136Inazuma\128 asserts his striking vigor!")
		
	//Lightning Jump hitlag
	elseif CBW_Battle
		for player in players.iterate
			local mo = player.mo
			if not (mo and mo.valid) then continue end
			
			local stop = mo and mo.inazuma_stop
			if not stop then continue end
			if stop.state == nil then stop.state = mo.state end
			
			//Keep pushed info intact
			mo.pushed = stop.pushed
			mo.pushed_last = stop.pushed
			mo.pushtics = 1
			
			//Decrease timer
			stop.tics = $ - 1
			
			//Stop it if we change to another state or the tics run out
			if mo.state != stop.state or not stop.tics
				mo.momx = stop.x
				mo.momy = stop.y
				mo.momz = stop.z
				
				P_XYMovement(mo)
				P_PlayerZMovement(mo)
				
				if mo.state == S_INAZUMA_BOLTSTREAM
					mo.state = S_PLAY_SPRING
				end
				mo.frame = $ &~ FF_FULLDARK
				
				if stop.tumble
					B.DoPlayerTumble(player, TUMBLE_TIME)
					mo.state = S_PLAY_FALL
					mo.sprite2 = SPR2_PAIN
					player.panim = PA_PAIN
					mo.tics = player.tumble
				end
				
				mo.spritexoffset = 0
				mo.spriteyoffset = 0
				mo.inazuma_stop = nil
				
			//Don't move!
			else
				A_ForceStop(mo)
				player.drawangle = stop.angle
				
				if stop.shock
					local angle = P_RandomRange(1, 360)*ANG1
					mo.spritexoffset = stop.tics*cos(angle)*2
					mo.spriteyoffset = stop.tics*sin(angle)*2
					
					Inazuma.SpawnParticle(mo, mo.scale, MT_WATERZAP, true)
					mo.frame = $ ^^ FF_FULLDARK
				end
			end
			
			local mobj = player.followmobj
			if mobj and mobj.valid
				mobj.spritexoffset = mo.spritexoffset
				mobj.spriteyoffset = mo.spriteyoffset
			end
		end
	end
end)

-- redefine these here
local NEURAL_COLORS = {
	SKINCOLOR_ICY,
	SKINCOLOR_VAPOR,
	SKINCOLOR_CORNFLOWER
}
local function valid(mo)
	return mo and mo.valid
end

-- handle Radial Flash debuff
addHook("ThinkFrame", do
	for p in players.iterate
		local mo = p.mo
		if not (mo and mo.valid) then continue end
		
		-- flashbang handling
		if (mo.inazuma_flashed_flashbang)
			mo.inazuma_flashed_flashbang = $ and $-1 or 0
		end
		
		-- actual flash debuff handling
		mo.inazuma_flashed_ghosts = $ or {}
		if not (mo.inazuma_flashed)
			-- might as well just remove this
			mo.inazuma_flashed = nil
			for i = #mo.inazuma_flashed_ghosts, 1, -1
				local ghost = mo.inazuma_flashed_ghosts[i]
				if valid (ghost)
					P_RemoveMobj(ghost)
				end
				table.remove(mo.inazuma_flashed_ghosts, i)
			end
			
			continue 
		end
		mo.inazuma_flashed = $ and $-1 or 0
		
		-- neural shock ghost for this!
		local ghost = P_SpawnGhostMobj(mo)
		ghost.fuse = 5
		if valid(ghost.tracer)
			P_RemoveMobj(ghost.tracer)
		end
			
		if mo.inazuma_flashed % 5 == 0
			ghost.color = mo.color
			ghost.colorized = true
		else
			table.insert(mo.inazuma_flashed_ghosts, ghost)
		end
			
		for i = #mo.inazuma_flashed_ghosts, 1, -1
			local ghost = mo.inazuma_flashed_ghosts[i]
			if not valid(ghost)
				table.remove(mo.inazuma_flashed_ghosts, i)
			else
				ghost.height = mo.height
				ghost.angle = p.drawangle
				ghost.color = NEURAL_COLORS[P_RandomKey(#NEURAL_COLORS << 1)] or mo.color
				P_MoveOrigin(ghost, mo.x + FixedMul(3*mo.scale, cos(FixedAngle(P_RandomKey(360)*FRACUNIT))), mo.y + FixedMul(3*mo.scale, sin(FixedAngle(P_RandomKey(360)*FRACUNIT))), mo.z)
			end
		end
		
		-- particles
		local flashbang = mo.inazuma_flashed_flashbang
		local flashbang_blind = mo.inazuma_flashed_flashbang_blind
		local isBlind = flashbang and ((flashbang > flashbang_blind) and 2 or 1) or 0
		
		if (isBlind)	-- faster frequency when flashbanged
		or (mo.inazuma_flashed % 2)
			local scale = mo.scale/2
			if (isBlind) then scale = (isBlind == 2) and mo.scale or mo.scale*3/4 end
			local spark = Inazuma.SpawnParticle(mo, scale)
			spark.color = mo.inazuma_flashed_colour or $
		end
		
		-- ear ringing sound
		local volume = 0
		if (flashbang and (flashbang > flashbang_blind))
			-- max volume
			volume = DEBUFF_SOUND_MAX
		else
			-- scale down
			local sound_tics = mo.inazuma_flashed - (DEBUFF_TICS - DEBUFF_SOUND_FADE) + (mo.inazuma_flashed_flashbang_max - flashbang_blind)
			volume = max(0, (sound_tics * DEBUFF_SOUND_MAX) / DEBUFF_SOUND_FADE)
		end
		if (volume) then S_StartSoundAtVolume(nil, sfx_rflsh, volume, p) end
	end
end)

hud.add(function(v, p, c)
	local mo = p.realmo
	if not (mo and mo.valid) then return end
	
	if not (mo.inazuma_flashed_flashbang) then return end
	local flashbang = mo.inazuma_flashed_flashbang
	local flashbang_max = mo.inazuma_flashed_flashbang_max
	local flashbang_blind = mo.inazuma_flashed_flashbang_blind
	-- ^ in retrospect, maybe all of these couldve been stored in a table...
	
	-- calculate strength
	local max_strength = 12
	local strength = 0
	if (flashbang > flashbang_blind)
		strength = max_strength
	else
		strength = (flashbang * max_strength) / flashbang_blind
	end
	
	-- flashbang them!
	if (strength) then v.fadeScreen(0xFB00, strength) end
end, "game")