/*
Changeable effect colors
	by Lach
	
v2
*/

local MAXSPLITSCREENPLAYERS = 2
local COMMANDNAME = "lightningcolor"
local COMMANDDESCRIPTION = "Sets the color of Inazuma's electric effects. Can be the name or number of a skincolor!"
local VARNAME = "inazuma_lightningcolor"
local SAVENAME = "lightningcolor"

// return a skincolor based on the name read from the file
local function ReadColor(playernum)
	local val = Silverhorn.LocalData["setting"][SAVENAME + playernum]
	if val != nil
		val = R_GetColorByName($)
	end
	return val or 0
end

// save the skincolor names to the file
local function WriteColor(val, playernum)
	local color = val and skincolors[val].name or "None"
	Silverhorn.SaveValue("setting", SAVENAME + playernum, color)
end

// set this player's effect color to the given skincolor, and save
local function SetColor(player, n)
	player[VARNAME] = n
	if not netgame or player == consoleplayer
		WriteColor(player[VARNAME], player == secondarydisplayplayer and 2 or 1)
	end
end

local function ChangeColor(player, ...)
	local silent
	local args = {...}
	local arg = ""
	
	for i, word in ipairs(args)
		word = $:upper()
		if word == "-SILENT"
			silent = true
		else
			arg = $ .. word .. " "
		end
	end
	
	arg = $:sub(1, $:len() - 1)
	
	if not arg:len()
		if not silent
			local playernum = player == secondarydisplayplayer and 2 or 1
			CONS_Printf(splitscreen and consoleplayer or player,
				COMMANDDESCRIPTION..
				"\nCurrently set to "..(Silverhorn.LocalData["setting"][SAVENAME + playernum])..".")
		end
		return
	end
	
	if arg == "DEFAULT"
	or arg == "OFF"
		SetColor(player, 0)
		return
	end
	
	local n = tonumber(arg)
	if n ~= nil
	and n >= 0 and n <= #skincolors
		SetColor(player, n)
		return
	end
	
	local exists, color = pcall(function(name) return _G["SKINCOLOR_"..name] end, arg)
	if exists
		SetColor(player, color)
		return
	end
	
	color = R_GetColorByName(arg)
	if color
		SetColor(player, color)
		return
	end
	
	if not silent
		CONS_Printf(splitscreen and consoleplayer or player,
			"Not a valid color!")
	end
end

COM_AddCommand(COMMANDNAME..2, ChangeColor, COM_SPLITSCREEN)
COM_AddCommand(COMMANDNAME, ChangeColor)

Silverhorn.AddExecCallback(function(player, playerNum)
	if player[VARNAME] ~= nil
		return
	end
	
	local text = COMMANDNAME
	if playerNum == 2
		text = $..2
	end
	text = $ .. " -silent " .. ReadColor(playerNum)
	CommandQueue_Add(text)
end)

Silverhorn.AddPlayerConnectCallback(function(player)
	player[VARNAME] = nil
end)
