/*
Mindscape
General objects and functions
	by SMS Alfredo
	& Lach
	& Tatsuru
*/

local Inazuma = Silverhorn.Inazuma

-- local returnablemap = 1 TODO: deleteme
Silverhorn.MS_RETURNMAP = 1 -- The last gamemap you visited
local exittimer = 0
local playerlives
local mindscapemap

local ultmusic = false

Silverhorn.FINALSTARPOSTID = 6 // The last starpost's parameter value in Mindscape plus one
local bloblist = {} //Lists all the blobs in mindscape with a parameter!

//Set up flags for this Mindscape/Game session
//For the different endings, check MS_VISITEDMINDSCAPE, the collected emeralds,
//and the player's Silverhorn status
rawset(_G, "MindscapeSessionFlags", 0)
rawset(_G, "MS_TOUCHEDPORTAL", 1 << 0) //If we've touched the portal in a level
rawset(_G, "MS_INMINDSCAPE", 1 << 1) //If we're currently in Mindscape
rawset(_G, "MS_VISITEDMINDSCAPE", 1 << 2) //If we've visited Mindscape this Game session
rawset(_G, "MS_SEENINTRO", 1 << 3) //If we've seen the intro this Game session
rawset(_G, "MS_BEATPHASEONE", 1 << 4) //If we've beaten Phase 1 this Mindscape session
rawset(_G, "MS_ONEUPTRANS", 1 << 5) //Flag for handling transforming 1ups into Eggmen
rawset(_G, "MS_SEENCREDITS", 1 << 6) //If we've seen the Credits this Game session
rawset(_G, "MS_OPENEDDOOR", 1 << 7) -- Opened the door before the boss fight
rawset(_G, "MS_BEATMINDSCAPE", 1 << 8) -- The player has beaten mindscape in the game session
rawset(_G, "MS_ULTIMATEHORNS", 1 << 9) -- The player has beaten ULTIMATE mindscape in the game session
rawset(_G, "MS_FAILEDENTRY", 1 << 10) -- If we attempted to enter mindscape via the map command
rawset(_G, "MS_SPAWNLEAVEPORTAL", 1 << 11) -- If we should spawn a leaving portal


local function GetMindscapeMap()
	if mindscapemap ~= nil
		return mindscapemap
	end
	for i = 0, #mapheaderinfo
		if mapheaderinfo[i] and mapheaderinfo[i].mindscape
			mindscapemap = i
			return mindscapemap
		end
	end
end

addHook("NetVars", function(network)
	Silverhorn.MS_RETURNMAP = network($)
	exittimer = network($)
	playerlives = network($)
	mindscapemap = network($)
	bloblist = network($)
	MindscapeSessionFlags = network($) // the $ thing works in NetVars?! aw man
end)

//Makes it so that when you exit Mindscape Zone you return to the regular game
//Same for when you get a game over
//Also handle transforming 1ups into Eggman monitors
addHook("PreThinkFrame", do
	if mapheaderinfo[gamemap].mindscape
		G_SetCustomExitVars(Silverhorn.MS_RETURNMAP, 1)
		if exittimer
			exittimer = $ - 1
			return
		end
		
		if MindscapeSessionFlags & MS_FAILEDENTRY
			exittimer = TICRATE
			G_ExitLevel(Silverhorn.MS_RETURNMAP, 1)
		end
		
		if not G_GametypeUsesLives() return end
		
		local allplayersdead = true
		for player in players.iterate
			--Check if a player has infinite lives
			if not (MindscapeSessionFlags & MS_ONEUPTRANS)
			and (player.lives == INFLIVES
			or (multiplayer and not CV_FindVar("cooplives").value))
				MindscapeSessionFlags = $ | MS_ONEUPTRANS
				
				--Transform all 1ups into Eggmen
				for mobj in mobjs.iterate()
					if mobj.type == MT_1UP_BOX and mobj.health
						mobj.type = MT_EGGMAN_GOLDBOX
						mobj.state = mobj.info.spawnstate
						if mobj.tracer and mobj.tracer.valid
							P_RemoveMobj(mobj.tracer)
							mobj.tracer = nil
						end
					end
				end
			end
			
			--Check if a player is alive
			if not player.bot
			and (player.lives or player.deadtimer < TICRATE*3/2)
				allplayersdead = false
			end
		end
		
		--Everyone's dead? Exit Mindscape
		if allplayersdead
			exittimer = TICRATE
			G_ExitLevel(Silverhorn.MS_RETURNMAP, 1)
		end
	elseif not (G_IsSpecialStage(gamemap) or mapheaderinfo[gamemap].silverhorn_endingmap) then
		Silverhorn.MS_RETURNMAP = mapheaderinfo[gamemap].nextlevel
	end
end)

//Stuff on map load
addHook("MapLoad", function(mapnum)
	
	//Starpost / Spawnpoint portals that go to the boss fight
	if multiplayer and mapheaderinfo[gamemap].mindscape and GetMindscapeMap()
		for mobj in mobjs.iterate()
			if mobj and mobj.valid and mobj.type == MT_STARPOST
			and mobj.health < Silverhorn.FINALSTARPOSTID
				local portal = P_SpawnMobjFromMobj(mobj,0,0,160*FRACUNIT,MT_MINDSCAPE_PORTAL)
				if P_MobjFlip(mobj) < 0
					portal.z = $ + portal.height
				end
				
				portal.flags2 = $|MF2_DONTDRAW
				portal.scale = mobj.scale/16
				portal.scalespeed = $/8
				portal.reactiontime = mobj.scale
			end
		end
		for player in players.iterate
			if player.mo and player.mo.valid
				local portal = P_SpawnMobjFromMobj(player.mo,0,0,80*FRACUNIT,MT_MINDSCAPE_PORTAL)
				portal.flags2 = $|MF2_DONTDRAW
				portal.scale = player.mo.scale/16
				portal.scalespeed = $/8
				portal.reactiontime = player.mo.scale
				break
			end
		end
	end
	
	//Handle entering Mindscape and playing the cutscene/intro text
	if mapheaderinfo[gamemap].mindscape
		--Already in Mindscape, spawn normally
		if (MindscapeSessionFlags & MS_INMINDSCAPE)
			for player in players.iterate
				if player.mo
					player.mo.state = S_PLAY_FALL
				end
			end
			
		--Set mindscape status
		else
			MindscapeSessionFlags = $ | (MS_INMINDSCAPE | MS_VISITEDMINDSCAPE)
			if not (MindscapeSessionFlags & MS_FAILEDENTRY)
				MindscapeSessionFlags = $ | MS_SPAWNLEAVEPORTAL
			end
		end
		
		--Reset life monitor transformation
		MindscapeSessionFlags = $ &~ MS_ONEUPTRANS
		
	--Not in Mindscape
	else
		--Erase Mindscape session flags
		if MindscapeSessionFlags & (MS_INMINDSCAPE | MS_BEATPHASEONE)
			MindscapeSessionFlags = $ &~ (MS_INMINDSCAPE | MS_BEATPHASEONE | MS_SEENINTRO | MS_ONEUPTRANS)
		end
		
		--Spawn portal when leaving
		if MindscapeSessionFlags & MS_SPAWNLEAVEPORTAL
		and not mapheaderinfo[gamemap].silverhorn_endingmap
			if not G_IsSpecialStage()
				for player in players.iterate
					if player.mo and player.mo.valid
						Silverhorn.SpawnMindscapePlayer(player)
					end
				end
			end
			MindscapeSessionFlags = $ &~ MS_SPAWNLEAVEPORTAL
		end
	end
	
	//Make it so that you have 3 lives in Minescape Zone, and that your
	//original live count is restored on exit
	if not G_GametypeUsesLives()
		playerlives = nil
	elseif mapheaderinfo[gamemap].mindscape
		if not playerlives
			playerlives = {}
			for player in players.iterate
				if player.bot
					continue
				elseif playerlives[#player] == nil
					playerlives[#player] = player.lives
					player.lives = 3
				end
			end
		end
	elseif playerlives
		for player in players.iterate
			if player.bot
				continue
			elseif playerlives[#player] != nil
				player.lives = playerlives[#player]
			end
		end
		playerlives = nil
	end
end)

--At the final checkpoint? Let's open that door!
addHook("PlayerSpawn", function(player)
	if not multiplayer and mapheaderinfo[gamemap].mindscape
	and player.starpostnum >= Silverhorn.FINALSTARPOSTID
	and not (MindscapeSessionFlags & MS_OPENEDDOOR)
		for sector in sectors.iterate
			if sector.tag == 148 then
				sector.tag = 0
			elseif sector.tag == 149 then
				sector.floorheight = $ - 552*FRACUNIT
				sector.tag = 0
				MindscapeSessionFlags = $ | MS_OPENEDDOOR
			end
		end
	end
end)

//Reset this
addHook("MapChange", function(mapnum)
	if mapheaderinfo[mapnum].mindscape and not mapheaderinfo[gamemap].mindscape
	and not (MindscapeSessionFlags & MS_TOUCHEDPORTAL)
	and not (server and Inazuma.CheckSilverhorns(server))
		MindscapeSessionFlags = $ | MS_FAILEDENTRY
	else
		MindscapeSessionFlags = $ &~ MS_FAILEDENTRY
	end
	MindscapeSessionFlags = $ &~ MS_TOUCHEDPORTAL &~ MS_OPENEDDOOR
	mindscapemap = nil
	bloblist = {}
end)

//Spawn in the portal above all signposts when the level starts
addHook("MobjSpawn", function(mobj)
	if (gametyperules & GTR_SPECIALSTAGES or Silverhorn.UltimateMode())
	and not mapheaderinfo[gamemap].mindscape
	and not G_IsSpecialStage(gamemap)
	and GetMindscapeMap()
		local portal = P_SpawnMobjFromMobj(mobj,0,0,128*FRACUNIT,MT_MINDSCAPE_PORTAL)
		portal.scale = mobj.scale/16
		portal.scalespeed = $/8
		portal.reactiontime = mobj.scale
	end
end, MT_SIGN)

//Portal Thinker
addHook("MobjThinker", function(mobj)
	if not mobj.reactiontime
	or (not mapheaderinfo[gamemap].mindscape
	and netgame and not (server and server.valid))
		mobj.flags2 = $|MF2_DONTDRAW
		return
	end
	
	local hasinazuma = false
	local fiftyrings = false
	local hasbeigehorns = false
	if mobj.isintroportal
		//sorry nothing
	elseif mapheaderinfo[gamemap].mindscape
		for player in players.iterate
			if player.starpostnum >= Silverhorn.FINALSTARPOSTID
				hasinazuma = true
				fiftyrings = true
				hasbeigehorns = true
			end
		end
	elseif All7Emeralds(emeralds)
		for player in players.iterate
			if skins[player.skin].name == "inazuma"
			or (player.mo and player.mo.valid and player.mo.skin == "inazuma")
				hasinazuma = true
			end
		
			if player.bot then continue end
			
			if player.rings >= 50 or Silverhorn.UltimateMode()
				fiftyrings = true
			end
			
			if player == server and not Inazuma.CheckSilverhorns(player)
				hasbeigehorns = true
			end
		end
	end
	
	if mobj.fuse == (TICRATE*4)-1 and mobj.target and mobj.target.valid
	and mobj.target.player
		S_StartSound(portal, sfx_s1c3, mobj.target.player)
	end
	
	if mobj.isintroportal and leveltime < TICRATE*5
	and mobj.fuse == (TICRATE*2) and mobj.target and mobj.target.valid
	and mobj.target.player and mobj.target.player == consoleplayer
	and mapheaderinfo[gamemap].mindscape
		startmindtext(7) //intro text
	end
	
	if hasinazuma and fiftyrings and hasbeigehorns
	and not (MindscapeSessionFlags & MS_TOUCHEDPORTAL)
		if S_SoundPlaying(mobj, sfx_s3kbds)
			S_StopSoundByID(mobj, sfx_s3kbds)
		end
		if not S_SoundPlaying(mobj, sfx_s3kbdl)
			S_StartSound(mobj, sfx_s3kbdl)
		end
		mobj.destscale = mobj.reactiontime
		mobj.flags = $|MF_SPECIAL
		mobj.flags2 = $&~MF2_DONTDRAW
	elseif mobj.scale <= mobj.reactiontime/16
		mobj.flags = $&~MF_SPECIAL
		mobj.flags2 = $|MF2_DONTDRAW
	else
		if S_SoundPlaying(mobj, sfx_s3kbdl)
			S_StopSoundByID(mobj, sfx_s3kbdl)
			S_StartSound(mobj, sfx_s3kbds)
		end
		mobj.flags = $&~MF_SPECIAL
		mobj.destscale = mobj.reactiontime/16
	end
end, MT_MINDSCAPE_PORTAL)

//Make it so the portal takes you to Mindscape Zone/to the final checkpoint
addHook("TouchSpecial", function(special, toucher)
	if special.isintroportal
		return true
	elseif mapheaderinfo[gamemap].mindscape
	and toucher.player and toucher.player.pflags&PF_JUMPED
		if toucher.player != displayplayer
			S_StartSound(special, sfx_s1c3)
		end
		S_StartSound(toucher, sfx_s1c3)
		for mobj in mobjs.iterate()
			if mobj and mobj.valid and mobj.type == MT_STARPOST
			and mobj.health == Silverhorn.FINALSTARPOSTID
				P_MoveOrigin(toucher, mobj.x, mobj.y, mobj.z+160*mobj.scale*P_MobjFlip(mobj))
				P_ResetPlayer(toucher.player)
				toucher.flags2 = ($&~MF2_OBJECTFLIP)|(mobj.flags2&MF2_OBJECTFLIP)
				toucher.eflags = ($&~MFE_VERTICALFLIP)|(mobj.eflags&MFE_VERTICALFLIP)
				toucher.scale = mobj.scale
				toucher.angle = mobj.angle
				toucher.player.drawangle = toucher.angle
				toucher.state = S_PLAY_FALL
				
				local portal = P_SpawnMobjFromMobj(toucher,0,0,0,MT_MINDSCAPE_PORTAL)
				portal.isintroportal = true
				portal.scale = toucher.scale
				portal.scalespeed = $/8
				portal.reactiontime = toucher.scale
				portal.fuse = TICRATE*4
				portal.target = toucher
				S_StartSound(portal, sfx_s3kbds)
				P_SetObjectMomZ(toucher,8*FRACUNIT,false)
				P_InstaThrust(toucher,toucher.angle,8*toucher.scale)
				toucher.state = S_PLAY_SPRING
				toucher.flags2 = $|MF2_DONTDRAW
				
				for p in players.iterate
					if p == toucher.player
					or (p.mo and P_CheckSight(p.mo, toucher))
						P_FlashPal(p, 1, 10)
					end
				end
				
				break
			end
		end
		return true
	elseif MindscapeSessionFlags & MS_TOUCHEDPORTAL
	or not (toucher.player and toucher.player.pflags&PF_FINISHED)
		return true
	end
	S_StartSound(special, sfx_s1c3)
	if multiplayer or netgame
		COM_BufInsertText(consoleplayer, "cecho You feel a dark energy pulling you...")
		G_SetCustomExitVars(GetMindscapeMap())
	else
		toucher.minescapedisappear = true
		S_FadeMusic(0,MUSICRATE)
		G_SetCustomExitVars(GetMindscapeMap(),1)
	end
	-- You have entered the portal
	MindscapeSessionFlags = $ | MS_TOUCHEDPORTAL
	Silverhorn.SaveValue("game", "ms_accessed", true)
	return true
end, MT_MINDSCAPE_PORTAL)

//Disappear when going into the Minescape Portal
//Also get silver horns if the variable is set to true
addHook("PlayerThink", function(player)
	if player.mo and player.mo.valid and player.exiting
	and MindscapeSessionFlags & MS_TOUCHEDPORTAL and player.mo.minescapedisappear
		player.mo.flags2 = $|MF2_DONTDRAW
	end
end)

//Mindscape Blob Object
addHook("MobjSpawn", function(mobj)
	mobj.flags = $ | MF_NOCLIPTHING
	mobj.flags2 = $ | MF2_INVERTAIMABLE | MF2_DONTDRAW
	
	//Extra objects
	local blob = P_SpawnMobjFromMobj(mobj,0,0,0,MT_MINDSCAPE_BLOB2)
	blob.tracer = mobj
	blob.flags2 = $|MF2_LINKDRAW
	mobj.target = blob
	
	local corona = P_SpawnMobjFromMobj(mobj, 0, 0, 20<<FRACBITS, MT_OVERLAY)
	corona.target = blob
	corona.tracer = blob --anonymous coder was here
	corona.state = S_MINDSCAPE_BLOB_CORONA
	mobj.hnext = corona
end, MT_MINDSCAPE_BLOB1)

//Spawn settings
addHook("MapThingSpawn", function(mobj, mapthing)
	//Don't give the blob a shadow if the extra flag is ticked
	if not (mapthing.options & MTF_EXTRA)
		mobj.shadowscale = FRACUNIT/2
	end
		
	//The reactiontime is set to the parameter value, which determines what
	//linedef offset will cause the blob to appear
	//Also add it to the blob list so we can make it show up later
	if mapthing.extrainfo
		mobj.reactiontime = mapthing.extrainfo
		table.insert(bloblist, mobj)
	end
end, MT_MINDSCAPE_BLOB1)

//The inner circle
//Also does thinking for the outer blob
addHook("MobjThinker", function(mobj)
	local tracer = mobj.tracer
	if not tracer then return end
	if mobj.state == S_INVISIBLE
		if mobj.sprite ~= SPR_SLSO
			mobj.sprite = SPR_SLSO
		end
		mobj.frame = FF_FULLBRIGHT | min(M, tracer.frame & FF_FRAMEMASK)
	end
	
	mobj.reactiontime = $ + 1
	
	//Set the fade in ID for mace blobs!
	local macespawn = tracer.hprev and tracer.tracer and tracer.tracer.spawnpoint
	if mobj.reactiontime == 1 and macespawn
	and not (tracer.flags2 & MF2_BOSSNOTRAP)
		tracer.reactiontime = macespawn.angle
		table.insert(bloblist, tracer)
		
	//Fade in blobs that don't normally fade in after a while
	elseif mobj.reactiontime == TICRATE and not tracer.reactiontime
		tracer.fuse = NUMTRANSMAPS*2
	end
	
	//Handle fade
	if tracer.fuse
		local trans = min((tracer.fuse - 1)/2, NUMTRANSMAPS - 1)
		trans = $ << FF_TRANSSHIFT
		
		mobj.frame = ($ &~ FF_TRANSMASK) | trans
		if states[tracer.state].frame & FF_TRANSMASK <= trans
			tracer.frame = ($ &~ FF_TRANSMASK) | trans
		end
		
		//Resolidify
		if tracer.flags & MF_NOCLIPTHING
			tracer.flags = $ &~ MF_NOCLIPTHING
			tracer.flags2 = $ &~ MF2_DONTDRAW
		end
	end
	
	//Make blob Ultimate if spawnpoint or mace spawnpoint have ambush ticked
	//or if we're simply in Ultimate Mode
	if not (tracer.flags2 & MF2_STRONGBOX) and (Silverhorn.UltimateMode()
	or (tracer.spawnpoint and tracer.spawnpoint.options & MTF_AMBUSH)
	or (macespawn and macespawn.options & MTF_AMBUSH))
		tracer.flags2 = $ | MF2_STRONGBOX
		tracer.state = S_SEESTATE
		if tracer.hnext
			tracer.hnext.frame = ($ &~ FF_TRANSMASK) | FF_TRANS50
		end
	end
	
	//Handle Ultimate spikes
	if tracer.flags2 & MF2_STRONGBOX and not (leveltime % mobj.info.activesound)
		tracer.state = S_MISSILESTATE
	end
	
	A_CapeChase(mobj,1,0)
end, MT_MINDSCAPE_BLOB2)

//No fuse deletion
addHook("MobjFuse", function(mobj) return true end, MT_MINDSCAPE_BLOB1)

//Bouncing on the blob
addHook("TouchSpecial", function(special, toucher)
	if not (special.valid and toucher.valid) return true end
	if toucher.eflags & MFE_SPRUNG return true end
	
	if special.state == special.info.xdeathstate
	and P_DamageMobj(toucher, special, special)
		toucher.eflags = $ | MFE_SPRUNG
		return true
	end
	
	local player = toucher.player
	
	if player and player.inazuma
	and player.inazuma.flags & Inazuma.FLAGS.LIGHTNINGJUMP
		return true
	end
	
	local prevmomx = toucher.momx
	local prevmomy = toucher.momy
	local candamage = false
	local homing = false
	local doLightningJump = false
	local pflags
	if player
		homing = player.homing
		doLightningJump = Inazuma.CheckStatus(player) and player.inazuma.homing
		candamage = (P_PlayerCanDamage(player, special) or homing)
		pflags = player.pflags
	end
	
	P_DoSpring(special,toucher)
	if special.flags2 & MF2_STRONGBOX
		special.state = S_MELEESTATE
	end
	
	if (homing or doLightningJump or (toucher.eflags & MFE_VERTICALFLIP and toucher.z+toucher.height < special.z+special.height)
	or (not (toucher.eflags & MFE_VERTICALFLIP) and toucher.z > special.z))
	and player and not P_IsObjectOnGround(toucher)
		if (toucher.eflags & MFE_VERTICALFLIP)
			toucher.z = special.z - toucher.height - 1
		else
			toucher.z = special.z + special.height + 1
		end
		toucher.momx = prevmomx
		toucher.momy = prevmomy
		toucher.eflags = $ | MFE_SPRUNG
		
		if doLightningJump
			Inazuma.LightningJump(player, special)
		elseif candamage
			if homing and pflags
				if (player.charability == CA_HOMINGTHOK
				and pflags&PF_THOKKED and not (pflags&PF_SHIELDABILITY)
				and not (player.charflags&SF_MULTIABILITY))
				or ((player.powers[pw_shield] & SH_NOSTACK) == SH_ATTRACT
				and pflags&PF_THOKKED and not (pflags&PF_SHIELDABILITY))
				and pflags&PF_JUMPED
					player.pflags = $|PF_JUMPED&~PF_THOKKED&~PF_SHIELDABILITY
				end
				toucher.momx = prevmomx/2
				toucher.momy = prevmomy/2
			end
			
			if (player.charability == CA_TWINSPIN and player.panim == PA_ABILITY)
			or (player.charability2 == CA2_MELEE and player.panim == PA_ABILITY2)
			or pflags & PF_BOUNCING
				P_SetObjectMomZ(toucher, 14*6*FRACUNIT/5, false)
			else
				P_SetObjectMomZ(toucher, 14*FRACUNIT, false)
			end
			
			toucher.state = S_PLAY_JUMP
			player.pflags = $ | P_GetJumpFlags(player) &~ PF_SHIELDABILITY
			if pflags & PF_BOUNCING
				toucher.state = S_PLAY_BOUNCE_LANDING
				S_StartSound(toucher, sfx_boingf)
			elseif toucher.skin != "inazuma"
				player.pflags = $ &~ PF_THOKKED
			end
		else
			P_SetObjectMomZ(toucher, 6*FRACUNIT, false)
			toucher.state = S_PLAY_SPRING
			player.pflags = $|PF_JUMPED|PF_NOJUMPDAMAGE
		end
	end
	return true
end,MT_MINDSCAPE_BLOB1)

//Hitting the blob with a projectile
addHook("MobjCollide", function(mobj, missile)
	if mobj and missile and missile.flags & MF_MISSILE
	and mobj.z - mobj.scale <= missile.z + missile.height
	and mobj.z + mobj.height + mobj.scale >= missile.z
		mobj.state = (mobj.flags2 & MF2_STRONGBOX) and S_MELEESTATE or S_RAISESTATE
		P_ExplodeMissile(missile)
	end
end,MT_MINDSCAPE_BLOB1)

//Make a blob fade in via linedef execution
//If the line's front X offset matches the blob's parameter, make it fade in!
//For maces, use the angle of the mace instead!
addHook("LinedefExecute", function(line)
	for i, mobj in ipairs(bloblist)
		if mobj and mobj.valid and mobj.reactiontime
		and mobj.reactiontime == line.frontside.textureoffset/FRACUNIT
			mobj.fuse = NUMTRANSMAPS*2
		end
	end
end, "MINDBLOB")

// do not load the following objects if Hollow Hill is already loaded
if pcall(do return _G["MT_MSTREE"] end)
	print("\131NOTICE:\128 Mind Tree and Mind Root objects already loaded.")
	return
end

//Big Tree in the Arena room

freeslot("MT_MSTREE","MT_MSTREEBRANCH", "MT_MSROOT",
"S_MSTREE","S_MSTREEBRANCH", "S_MSROOT",
"SPR_MSTT","SPR_MSTB", "SPR_MSRT", "SPR_RTSP")

mobjinfo[MT_MSTREE] = {
	--$Name "Mind Tree"
	--$Sprite MSTTA0
	--$Category "Mindscape"
	doomednum = 4094,
	spawnstate = S_MSTREE,
	radius = 20*FRACUNIT,
	height = 256*FRACUNIT,
	flags = MF_SOLID|MF_SCENERY
}
mobjinfo[MT_MSTREEBRANCH] = {
	spawnstate = S_MSTREEBRANCH,
	radius = 20*FRACUNIT,
	height = 256*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_SCENERY
}
states[S_MSTREE].sprite = SPR_MSTT
states[S_MSTREEBRANCH] = {
	sprite = SPR_MSTB,
	frame = FF_PAPERSPRITE|A
}

//spawning the big tree 

addHook("MapThingSpawn", function(mo, mt)
	mo.fuse = 1
end, MT_MSTREE)

addHook("MobjFuse", function(mo)
	local ang = ANG1 * 72
	for i = 0, 5
		local b = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_MSTREEBRANCH)
		b.angle = mo.angle + (ang * i)
	end
	return true
end, MT_MSTREE)

// Root scenery

mobjinfo[MT_MSROOT] = {
	--$Name "Mind Roots"
	--$Sprite MSRTA0
	--$Category "Mindscape"
	doomednum = 4095,
	spawnstate = S_MSROOT,
	radius = 21*FRACUNIT,
	height = 48*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_SCENERY
}

states[S_MSROOT] = {
	sprite = SPR_MSRT,
	frame = FF_FULLBRIGHT
}

addHook("MapThingSpawn", function(mo, mt)
	mo.frame = mt.angle % 6
	
	local sparkleradii = {20, 26, 30, 89, 42, 65}
	local heights = {69, 108, 87, 108, 158, 158}
	
	mo.sparkleradius = sparkleradii[(mo.frame & FF_FRAMEMASK) + 1]
	mo.height = heights[(mo.frame & FF_FRAMEMASK) + 1] * FRACUNIT
end, MT_MSROOT)

addHook("MobjThinker", function(mo)
	local rate = FixedInt(mo.sparkleradius / 72)
	if leveltime % ((5 - rate) * 2) return end
	
	local ang = P_RandomKey(180) * ANG2
	local dist = P_RandomRange(0, mo.sparkleradius)
	local height = P_RandomRange(0, FixedInt(mo.height/2))
	
	local sparkle = P_SpawnMobjFromMobj(mo, dist*cos(ang), dist*sin(ang), 0, MT_THOK)
	sparkle.sprite = SPR_RTSP
	sparkle.frame = P_RandomRange(0, 1) | (P_RandomRange(0, 3) * FRACUNIT)
	sparkle.tics = -1
	sparkle.color = P_RandomChance(FRACUNIT/2) and SKINCOLOR_AETHER or SKINCOLOR_BLUEBELL
	sparkle.blendmode = AST_ADD
	sparkle.momz = P_MobjFlip(mo) * FRACUNIT
	sparkle.fuse = FixedInt(mo.height) + P_RandomRange(-8, 8)
	sparkle.destscale = sparkle.scale / 8
	sparkle.scalespeed = $ / (16 - rate)
end, MT_MSROOT)

addHook("MobjCollide", function(mo, tmo)
	if tmo.type == mo.type return false end
end, MT_MSROOT)
