/*
Mindscape
Dialog Boxes
	by SMS Alfredo
	& Inazuma
	& Varren
*/

//Here are the different texts for the text boxes, and the code used to display them
//Will probably explain how the new system works, but for now, not here

//By using linedef type 443 and the function name "MINDTEXT", the text displayed
//can be controlled by the level itself. The linedef executor's front texture
//x offset will be used as the text ID to display.
//This can be seen in the Prologue if you need reference.
//See the wiki for more information on how linedef type 443 is used
//https://wiki.srb2.org/wiki/Linedef_type_443

local mindtext = {}

//Values used for the text display process
local textid = 0
local textletter = 0
local textwait = 0
local texttime = 0
local textskin = nil
local textcolor = 0
local textdebug = false
local pressbuttons = {{[BT_JUMP] = -1, [BT_SPIN] = 0}, {[BT_JUMP] = -1, [BT_SPIN] = 0}}

//Moving the local variables from the process here to allow multiple run throughs
local length = 0
local text = ""

local flip = false
local numlines = 1

local x = 61
local icnorigin_x = 6 -- szv: default was 36, some text found a way to clip off-screen
local snap = V_SNAPTOLEFT

//Get the value of a certain data from a certain id in mindtext
local mindtextvalue = function(id,arg)
	if arg == "image" return mindtext[id][1]
	elseif arg == "name" return mindtext[id][2]
	elseif arg == "text" return mindtext[id][3]
	elseif arg == "waittime" return mindtext[id][4]
	elseif arg == "nexttext" return mindtext[id][5]
	elseif arg == "flip" return mindtext[id][6]
	elseif arg == "numlines" return mindtext[id][7] end
end

-- a local version of string.split for us to use
local function mindtextsplit(str, sep)
    local fields = {}
    
    local sep = sep or " "
    local pattern = string.format("([^%s]+)", sep)
    string.gsub(str, pattern, function(c) fields[#fields + 1] = c end)
    
    return fields
end

-- Adds a new dialog text id
local function AddMindscapeDialog(id, image, name, text, args)
	-- Put it at the first freeslot if we don't have an id
	if not id
		local i = 1
		while mindtext[i]
			i = $ + 1
		end
		id = i
	end
	
	-- Split the text with a delimeter, these will be the lines of text
	-- Sets default text if none exists
	local temptext = text or "Looks like I don't have any dialog, or a name. Great job there."
	
	-- Get the extra customized arguments if not using the default
	local waittime = (args and args.waittime) or 60
	local nexttext = (args and args.nexttext) or 0
	local flip = (args and args.flip) or false

	local new = {
		image or "none", -- image
		name or "none", -- name
		text, -- text
		waittime, -- waittime
		nexttext, -- nexttext
		flip, -- flip
		-- [alternate skin text]
	}

	-- Assigns ids for alternate skin text
	if (args and args.altskins) then
		for k,v in pairs(args.altskins) do
			new[k] = v
		end
	end

	table.insert(mindtext, id, new)
	return id //Return the ID in case we need to use the auto-generated one later
end

-- Adds multiple new dialog text ids
local function AddMindscapeDialogSet(id, dialogs, altskin, origintext)
	-- Put it at the first freeslot if we don't have an id
	if not id
		local i = 1
		while mindtext[i]
			i = $ + 1
		end
		id = i
	end
	
	if altskin != nil and origintext != nil
	and mindtext[origintext]
		mindtext[origintext][altskin] = id
	end
	
	if type(dialogs) != "table" then dialogs = {$} end
	for k,v in pairs(dialogs) do
		local newid = id + k - 1
		if id < 0 then newid = id - k + 1 end

		//Set nexttext
		if v[4] == nil then v[4] = {} end
		if v[4].nexttext == nil
			if k < #dialogs
				if id >= 0
					v[4].nexttext = newid + 1
				else
					v[4].nexttext = newid - 1
				end
			elseif v[4].waittime == nil
				v[4].waittime = 120
			end
		end
		
		//Add it!
		AddMindscapeDialog(newid, v[1], v[2], v[3], v[4])
	end
	
	return id //Return the ID in case we need to use the auto-generated one later
end

-- Constants for the IDs used for the default text
local INTRO_TEXT                  = 7
local BOSS_PHASE1_TEXT            = 1000
local BOSS_PHASE2_TRANSFORM_TEXT  = 1001
local BOSS_PHASE2_TEXT            = 1002

-- TEST DIALOG
-- AddMindscapeDialog(9000, "FA_SMUG", nil, nil, nil, nil, nil)

AddMindscapeDialogSet(1, {
	{"FA_SMUG", "Facciolo", "Hey, Inazuma. Give me your body bro.", {waittime=-1}},
	{"skin", "skin", "Looks like somebody needs to get their eyes checked!", {waittime=-1}},
	{"FA_SHOK", "Facciolo", "WHA-", {waittime=-1}},
	{"FA_ANGR", "Facciolo", "Why you...!", {waittime=-1}},
})
AddMindscapeDialogSet(nil, {{"IN_DAMN", "Inazuma", "No.", {nexttext=3, waittime=-1}}}, "inazuma", 2)
AddMindscapeDialogSet(nil, {{"AMYRCONC", "Amy", "Bruh. I'm not even Inazuma.", {nexttext=3, waittime=-1}}}, "amy", 2)

-- THE MAIN INTRO AND BOSS TEXT

AddMindscapeDialogSet(INTRO_TEXT, {
	{"FA_SMUG", "Facciolo", "You... I was not expecting someone other than the Silverhorn to appear here."},
	{"FA_SMUG", "Facciolo", "I will show you mercy. Leave here now and live..."},
	{"FA_SMUG", "Facciolo", "Or stay and embrace oblivion!"},
})
AddMindscapeDialog(BOSS_PHASE1_TEXT, "FA_ANGR", "Facciolo", "Invader, for your failure to heed my warning, you will be served a violent and  merciless death!")
AddMindscapeDialog(BOSS_PHASE2_TRANSFORM_TEXT, "FA_ANGR", "Facciolo", "I've let you live long enough!")
AddMindscapeDialog(BOSS_PHASE2_TEXT, "FA_MNRR", "Moonreaver", "You're doomed!")

-- Inazuma

AddMindscapeDialogSet(nil, {
	{"IN_DETR", "Inazuma", "So it's come to this. If he kills me here, then he'll take over my body."},
	{"IN_DETR", "Inazuma", "There's a lot at stake here... But if I win... Then I'll finally be free."},
}, "inazuma", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "Well done making it this far, Silverhorn. Now that you're here. I shall now take your body by force!"},
}, "inazuma", BOSS_PHASE1_TEXT)

-- Aether

local AETHER_INTRO_TEXT = AddMindscapeDialogSet(nil, {
	{"FA_SHOK", "Facciolo", "What the!?"},
	{"FA_ANGR", "Facciolo", "BAH! YOU AGAIN! WITCH!"},
	{"AE_SMUG", "Aether", "Good to see you too, old man!"},
	{"IN_DETR", "Inazuma", "Careful, Facciolo has dominion over this realm. I'm not sure if I can beat him."},
	{"AE_REDY", "Aether", "Well it's a good thing I'm here, huh?"},
	{"FA_SMUG", "Facciolo", "Hiding behind the Meridian again, Silverhorn? What happened to your pride?"},
	{"IN_DAMN", "Inazuma", "You are the last person that should be preaching to anyone about pride."},
	{"FA_ANGR", "Facciolo", "What I did to you is inconsequential in the grand scheme of things."},
	{"FA_ANGR", "Facciolo", "You wouldn't have been paralyzed all your life if you had just fused with me to begin with."},
	{"AE_SMUG", "Aether", "Sounds more like you're coping with your failures, Facciolo! Inazuma, we can beat him!"},
	{"AE_REDY", "Aether", "Together, there's no way we can lose!"},
	{"IN_DETR", "Inazuma", "Alright!"},
}, "aether", INTRO_TEXT)

local AETHER_ARENA_TEXT = AddMindscapeDialogSet(nil, {
	{"AE_REDY", "Aether", "There he is! Let's finish this!"},
	{"FA_ANGR", "Facciolo", "You two just don't know when to give up. I'll send you both into oblivion!"},
}, "aether", BOSS_PHASE1_TEXT)

-- Make Aether's text show up for Inazuma's team-up as well
mindtext[INTRO_TEXT]["inazuma&aether"] = AETHER_INTRO_TEXT
mindtext[INTRO_TEXT]["aether&inazuma"] = AETHER_INTRO_TEXT
mindtext[BOSS_PHASE1_TEXT]["inazuma&aether"] = AETHER_ARENA_TEXT
mindtext[BOSS_PHASE1_TEXT]["aether&inazuma"] = AETHER_ARENA_TEXT

-- Hinote

AddMindscapeDialogSet(nil, {
	{"HI_THINK", "Hinote", "Huh...? Where am I?"},
	{"FA_SMUG", "Facciolo", "Ho ho? Hinote? Is that really you?"},
	{"HI_WINK", "Hinote", "Yup, the one and only! And wow, Facciolo? What a pleasant suprise."},
	{"FA_SMUG", "Facciolo", "It's good to see you again too, Hinote. How is your father?"},
	{"HI_THINK", "Hinote", "Can't say I know much goin' on. He continues to give a mysterious aura of himself."},
	{"HI_THINK", "Hinote", "I'm sorry I don't got much info."},
	{"FA_SMUG", "Facciolo", "Hmm, that's too bad. I was hoping for some news about a good ol' friend of mine."},
	{"HI_HAPPY", "Hinote", "What about you? What have you been up to lately?"},
	{"FA_SMUG", "Facciolo", "In the middle of getting a new body. Once I do, we should meet up somewhere."},
	{"FA_SMUG", "Facciolo", "I know of a marvelous pastry shop that I know you will love."},
	{"HI_WINK", "Hinote", "That sounds fantastic!"},
	{"FA_SMUG", "Facciolo", "First, traverse through Mindscape as a test of your skill."},
	{"HI_SHRUG", "Hinote", "Mmmm... Sure thing."},
}, "hinote", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Hinote! Welcome. Let's have a light sparring session. If I win, will you help me defeat Inazuma? "},
	{"HI_SMUG", "Hinote", "Sure, I 'spose. I mean he's kind of a push over, don't know why you need my help."},
	{"FA_SMUG", "Facciolo", "Excellent."},
}, "hinote", BOSS_PHASE1_TEXT)

-- Gemma

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Ah... Gemma, I'm suprised you decided to show your face here considering your lineage."},
	{"GE_DFNT", "Gemma", "What are you talking about?"},
	{"FA_SMUG", "Facciolo", "Your father, as well as the rest of your clan, are cowards."},
	{"FA_SMUG", "Facciolo", "Frail and pathetic. Ironic considering they wield the sturdiest of elements."},
	{"GE_UPST", "Gemma", "How dare you...!!! You know NOTHING about my father!!"},
	{"FA_SMUG", "Facciolo", "Well considering the sad display of strength he showed in his final moments,"},
	{"FA_SMUG", "Facciolo", "I guess there really is nothing worthy to know about him."},
	{"GE_ANGR", "Gemma", "You... I've had it with you!!!"},
}, "gemma", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Gemma, you may have lasted this far..."},
	{"FA_ANGR", "Facciolo", "But I will personally ensure that today you will be meeting your ancestors!!"},
}, "gemma", BOSS_PHASE1_TEXT)

-- Veph

AddMindscapeDialogSet(nil, {
	{"VP_ECST", "Veph", "Woah, what is this place? It's so big and cool looking..."},
	{"FA_SMUG", "Facciolo", "This place is the physical manifestation of the mind of the Silverhorn."},
	{"FA_SMUG", "Facciolo", "Now, I suggest you run back home, lest you seal your fate here."},
	{"VP_ECST", "Veph", "What? I can't go home while I'm still adventuring!"},
	{"FA_SMUG", "Facciolo", "Adventuring? You amuse me, boy. We both know that's not why you wish to avoid returning home."},
	{"VP_CONF", "Veph", "Is that not a good enough reason?"},
	{"FA_ANGR", "Facciolo", "You're a mutant! A freak of nature! A result of immoral conduct and a heartless monster to all."},
	{"FA_DEAD", "Facciolo", "You think you were allowed to leave your village because you were ready for it?"},
	{"VP_FEAR", "Veph", "W-wasn't I?"},
	{"FA_ANGR", "Facciolo", "No! They despise you! It was only a method to rid themselves of you."},
	{"FA_SMUG", "Facciolo", "However, I can assist your cause if you assist mine."},
	{"FA_SMUG", "Facciolo", "Help me subdue the Silverhorn, and I will rid you of that which ails you."},
	{"VP_FEAR", "Veph", "..."},
	{"FA_SMUG", "Facciolo", "Give it some thought. I'm sure you will find such an offer well worth your time."},
	{"VP_CONF", "Veph", "(Could this guy really fix me? Everyone back home was always so mean since I'm such an outcast.)"},
	{"VP_CONF", "Veph", "(But wait, I'd need to \"subdue the Silverhorn\" first right? Is he asking me to betray Inazuma?!)"},
	{"VP_ECST", "Veph", "Okay, there's no way I'm going to betray my friends! Not now, not ever!"},
	{"VP_ECST", "Veph", "I don't care if you think I'm weird, I'm sending you straight to next week!"},
}, "veph", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "I see you've rejected my offer! A grave mistake, as you will now die here both a fraud and a fool!"},
}, "veph", BOSS_PHASE1_TEXT)

--Piroli

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Well well, what's a fine young lady like yourself doing here?"},
	{"PR_DISS", "Piroli", "Oh, I was supposed to be setting up a tailor shop in Greenflower."},
	{"PR_DISS", "Piroli", "I guess I just got a little sidetracked."},
	{"FA_SMUG", "Facciolo", "Your father must be quite disappointed in you..."},
	{"FA_SMUG", "Facciolo", "Not able to accomplish anything you've been assigned to do."},
	{"PR_ANGR", "Piroli", "Hey, that's none of your business! And besides, it's not my fault this place has been devolving"},
	{"PR_ANGR", "Piroli", "into chaos with all the robots everywhere!"},
	{"PR_ANGR", "Piroli", "Can't even rent out a building without it being blown to shreds!"},
	{"FA_SMUG", "Facciolo", "Yes, but that is what I am here to solve."},
	{"FA_SMUG", "Facciolo", "Rid the world of its chaos and bring it order under my rule!"},
	{"PR_DISS", "Piroli", "What, you planning a global dictatorship? Yeah, no thanks pal."},
}, "piroli", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "If you choose to defile me, then I have no choice but to rid the world of you as well!"},
}, "piroli", BOSS_PHASE1_TEXT)

-- Flame

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "A dragon? Pathetic. Your species' potential is vapid at best."},
	{"FM_BRUH", "Flame", "Oh, I'm pathetic? At least I'm not stuck in someone else's subconscious."},
	{"FA_SMUG", "Facciolo", "Be that as it may, this state is only temporary."},
	{"FA_SMUG", "Facciolo", "I shall have full dominion over his body soon enough."},
	{"FA_SMUG", "Facciolo", "Now, run along back to that little island of yours."},
	{"FM_SMUG", "Flame", "Let me guess... Once you take control, it'll be the first to go?"},
	{"FM_SMUG", "Flame", "Hah! I'd love to see you try."},
	{"FA_SMUG", "Facciolo", "Oh, you're quite right about that."},
	{"FA_SMUG", "Facciolo", "I'm looking forward to crushing you like the Silverhorns did to your mainland brethren."},
	{"FM_ANGR", "Flame", "What are you even talking about?"},
}, "flame", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "Ah so I see you have chosen death, come forth and let me show you how weak you truly are!"},
	{"FM_SMUG", "Flame", "Says the fluffy ghos-", {waittime=2}},
	{"FM_ANGR", "Flame", "HEY I'M NOT DONE TALKING!"},
}, "flame", BOSS_PHASE1_TEXT)

-- Gale

AddMindscapeDialogSet(nil, {
	{"GA_CONF", "Gale", "Huh? How'd I get here?"},
	{"FA_SMUG", "Facciolo", "Welcome to my humble abode. I was wondering when someone of value would show up."},
	{"GA_CONF", "Gale", "Oh? Uh, thanks I gues-", {waittime=2}},
	{"FA_SMUG", "Facciolo", "I wasn't talking to you, child."},
	{"FA_SMUG", "Facciolo", "You, Arva. I see you're still alive. Have you finally come to submit to me?"},
	{"GA_BRUH", "Gale", "I'm pretty sure she won't-", {waittime=2}},
	{"GA_ARVA", "Arva", "You dare!? A dishonorable cheat like you has no place addressing a goddess in such a manner."},
	{"FA_SMUG", "Facciolo", "Alright then, \"goddess\". I implore you! Try to put me in my place!"},
	{"GA_STPD", "Gale", "... Can I get a single sentence in here without being interrupted?"},
}, "gale", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "The era of the old gods has come to an end! Witness TRUE DIVINITY!"},
}, "gale", BOSS_PHASE1_TEXT)

-- Willo

AddMindscapeDialogSet(nil, {
	{"WL_BOFA", "Willo", "Oye, anyone home? Hello?"},
	{"FA_SHOK", "Facciolo", "Peculiar child... Why would you haphazardly follow the Silverhorn to this place?"},
	{"WL_NOFA", "Willo", "Oh, no I'm just trying to find the bathroom, gramps."},
	{"FA_ANGR", "Facciolo", "Hrrrgh..."},
	{"WL_MURD", "Willo", "Gee, don't you have any manners, grandpa?"},
}, "willo", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "WHAT are you still doing here?"},
	{"WL_DEEZ", "Willo", "... Bathroom."},
}, "willo", BOSS_PHASE1_TEXT)

-- Cacee

AddMindscapeDialogSet(nil, {
	{"CACESASS", "Cacee", "Donde el mundo estoy?"},
	{"FA_SMUG", "Facciolo", "What do we have here? Is the Silverhorn keeping house plants in his skull now?"},
	{"FA_SMUG", "Facciolo", "You're a long ways from home little one."},
	{"CACETALK", "Cacee", "Ah! It's a big perra. Coochie coochie coo!"},
	{"FA_SHOK", "Facciolo", "What the..."},
	{"FA_ANGR", "Facciolo", "I am NOT a pet! You garden weed! I will rip you to shreds!"},
	{"CACEFAIL", "Cacee", "Vete a ir esparragos, pendejo!"},
	{"FA_SMUG", "Facciolo", "Stupid little peinabombillas, No puedes enganarme hablando otro idioma."},
	{"FA_SMUG", "Facciolo", "I speak many tongues, plant. Did you really think you'd make a fool out of me?"},
	{"CACESASS", "Cacee", "Which one of us is having an argument with a house plant, ese?"},
	{"FA_SHOK", "Facciolo", "...!"},
	{"FA_ANGR", "Facciolo", "Why you!"},
}, "cacee", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"CACESASS", "Cacee", "Hola-", {waittime=2}},
	{"FA_ANGR", "Facciolo", "NO! SHUT UP! You will die here, cactus!"},
}, "cacee", BOSS_PHASE1_TEXT)

-- Skip

AddMindscapeDialogSet(nil, {
	{"FA_SHOK", "Facciolo", "Fascinating... I did not expect you to enter my domain."},
	{"SK_ANGR", "Skip", "So you're the reason Mister Inazuma is hurt! I won't let you harm him any longer!"},
	{"FA_SMUG", "Facciolo", "By all means... Come at me. I'm eager to see the depths of that power you're hiding."},
	{"SK_WORY", "Skip", "Hiding? I'm not hiding anything! Y-You're just messing with me! I won't fall for your tricks!"},
}, "skip", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "You've got guts, kid! But your travels end here!"},
}, "skip", BOSS_PHASE1_TEXT)

-- Jana

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Oh great, another self-declared warrior here to throw her life away."},
	{"FA_SMUG", "Facciolo", "Such power is wasted on a fool like you."},
	{"JA_COCK", "Jana", "I ain't here to lose, ya pompous nerd!I can wipe the floor with you any day of the week!"},
	{"FA_ANGR", "Facciolo", "How can you expect to win with a power you can't control?"},
	{"FA_ANGR", "Facciolo", "Let your rage fuel you... Embrace your darker self or fall by my hand."},
	{"JA_SERO", "Jana", "..."},
	{"JA_SERO", "Jana", "My fate isn't yours to control."},
}, "jana", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "You will never control the beast within! Face destruction!"},
}, "jana", BOSS_PHASE1_TEXT)

-- Kou
	
AddMindscapeDialogSet(nil, {
	{"KOU_SMUG", "Kou", "Heh. Ain't this place cozy? Could be a fun place to crash into every once in a while."},
	{"FA_SMUG", "Facciolo", "Well I certainly wasn't expecting the likes of you. Another one of Silverhorn's allies?"},
	{"KOU_SMUG", "Kou", "What, that lil' jolt of a deer? Nah. I think he's a dolt."},
	{"FA_SMUG", "Facciolo", "Guess we can agree on something. Care to join forces, you and I? You'd make a great underling."},
	{"KOU_SMUG", "Kou", "Didn't take ya for a jokester, wolfie."},
	{"KOU_SMUG", "Kou", "If ya think I'll simply kneel down to such a pompous big shot like yerself..."},
	{"KOU_ANGR", "Kou", "Ya got another thing comin'."},
	{"FA_ANGR", "Facciolo", "Brutes like yourself are a waste of my time. Come face me, and I will show you what real strength is."},
}, "kou", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "Your wreckless behavior will be your doom. Have at you!"},
}, "kou", BOSS_PHASE1_TEXT)

-- Echoes & Abyss
	
AddMindscapeDialogSet(nil, {
	{"EC_WORY", "Echoes", "Wha... Abby? Where are we?"},
	{"AB_ANGR", "Abyss", "Stay on guard, young one. I sense a horrible presence in these foreign ruins."},
	{"FA_SMUG", "Facciolo", "Kinda rude to refer to an old pal like that, eh, 'Abby'?"},
	{"AB_SURP", "Abyss", "[!] It... It can't be! Azathoth? How are you still alive?"},
	{"FA_SMUG", "Facciolo", "I could ask the same, Fallen Dragon King."},
	{"FA_ANGR", "Facciolo", "How the mighty have fallen, having to borrow that little lady's body to survive."},
	{"EC_WORY", "Echoes", "H-huh? How did you know that?"},
	{"FA_SMUG", "Facciolo", "You seek to return to your old life, don't you?"},
	{"FA_SMUG", "Facciolo", "Join my side little fox... and you might get to see your father again."},
	{"AB_ANGR", "Abyss", "Don't let him fool you, young one!"},
	{"AB_ANGR", "Abyss", "This monster has brought upon the end of countless civilizations... My own included."},
	{"EC_WORY", "Echoes", "..."},
	{"EC_ANGR", "Echoes", "W-We won't bow down to evil! We will protect our friends and this world from the likes of you!"},
	{"FA_ANGR", "Facciolo", "Very well then. Your deaths shall serve as an example to anyone who dares get in my way!"},
	{"AB_HAPY", "Abyss", "...You've grown a lot, dear."},
	{"EC_HAPY", "Echoes", "...Thank you, Abby."},
}, "echoes", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "Foolish child. Perish alongside your kind!"},
}, "echoes", BOSS_PHASE1_TEXT)

-- Tusk

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "What's this? Another one of the doctor's creations?"},
	{"TSK_NEUT", "TUSK", "[?]"},
	{"FA_SMUG", "Facciolo", "Turn back now, machine. I don't have time for a bunch of bolts like you."},
	{"TSK_NEUT", "TUSK", "..."},
	{"TSK_ANGR", "TUSK", "01111001 01101111 01110101 01110010  00100000 01101101 01101111 01101101"},
	{"FA_SHOK", "Facciolo", "[!]"},
	{"FA_ANGR", "Facciolo", "You insignificant SCRAP! I shall splay the gore of your profane form across the STARS!!"},
	{"TSK_ANGR", "TUSK", "01110010 01100001 01110100 01101001 01101111"},
}, "tusk", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "I shall grind you down until the very SPARKS cry for mercy!"},
}, "tusk", BOSS_PHASE1_TEXT)

-- Milne

AddMindscapeDialogSet(nil, {
    {"FA_SMUG", "Facciolo", "Another otherworldly adversary? Worry not crizlyn, I will be correcting your planet as"},
    {"FA_SMUG", "Facciolo", "soon as I'm finished with this one."},
    {"ML_COCK", "Milne", "Puh-lease. You think I care? By all means, go for it!"},
	{"FA_SMUG", "Facciolo", "So the poor little pirate abandons her ship?"},
	{"FA_SMUG", "Facciolo", "Such a pathetic display of honor, fitting to a coward such as yourself."},
    {"ML_HUHH", "Milne", "Excuse meeee...?!"},
	{"ML_UWUT", "Milne", "Coward?! No way, bub. Nothin' gets in the way of what I want."},
	{"ML_UWUT", "Milne", "I never run away from a fight!"},
	{"FA_SMUG", "Facciolo", "Your crew would say otherwise, Captain Korsar."},
	{"ML_HUHH", "Milne", "...!"},
	{"ML_URGH", "Milne", "You..."},
}, "milne", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
    {"FA_SMUG", "Facciolo", "Interesting. Despite how much you seem to hate your own kind for what they've done,"},
	{"FA_SMUG", "Facciolo", "you seem eager to use the very power they cursed you with."},
	{"ML_ISTG", "Milne", "...Shut it, fuzzy."},
}, "milne", BOSS_PHASE1_TEXT)

-- Iclyn

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "You followed the other two crizlyns here, didn't you?"},
	{"FA_SMUG", "Facciolo", "It'd be wise to leave, child."},
	{"IC_DUMB", "Icyln", "Oho, you're underestimating me? Well ya better watch out, you..."},
	{"IC_OOP", "Icyln", "uh"}, 
	{"IC_DUMB", "Icyln", "big bad wolf guy...! 'Cause if you think you can beat me, you've got another thing comin'!"},
	{"FA_DEAD", "Facciolo", "..."},
	{"IC_WOO", "Iclyn", "Yeah, you better run!"}
}, "iclyn", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "If you're as powerful as you say you are, you must have no issue surviving this!"},
	{"IC_FEAR", "Icyln", "Y-yup! No issue at all...!"},
	{"IC_WEHH", "Iclyn", "(...should'a kept my big mouth shut...)"},
}, "iclyn", BOSS_PHASE1_TEXT)

-- Flare

AddMindscapeDialogSet(nil, {
	{"FA_SHOK", "Facciolo", "I didn't expect to see you again, Narlok."},
	{"FR_SHOK", "Flare", "Wh- h-how do you know h-his-?"},
	{"NR_NEUT", "Narlok", "Yes. It's been a long time, hasn't it?"},
	{"FA_SMUG", "Facciolo", "I must admit, after such an embarrassing defeat..."},
	{"FA_SMUG", "Facciolo", "I didn't think you would ever show your face again."},
	{"NR_ANOY", "Narlok", "If you're here just to taunt me, then forget it. What do you want?"},
	{"FA_SMUG", "Facciolo", "I want to make a deal. Devote the last of your dwindling power and loyalty to me..."},
	{"FA_SMUG", "Facciolo", "And not only will I return your prime strength, I promise no harm to the boy's dear brother."},
	{"FR_SWET", "Flare", "B-brother...? He doesn't m-mean...!"},
	{"NR_ENRG", "Narlok", "ONE MORE WORD AND I'M BURNING YOUR MOUTH SHUT, DEMON."},
}, "flarethetrex", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "You're not as weak as I thought, are you? What do you say, Narlok?"},
	{"NR_ANOY", "Narlok", "You say that like I ever considered agreeing."},
	{"FA_ANGR", "Facciolo", "Then BURN!"},
}, "flarethetrex", BOSS_PHASE1_TEXT)

-- Ryder

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "So you're this \"Ryder\" the reptile seems to care about."},
	{"RY_DETR", "Ryder", "Yeh. That's me. What's it t' ya?"},
	{"FA_SMUG", "Facciolo", "You shouldn't have been granted those powers of dark. You... are nothing but a false prophet."},
	{"RY_FEAR", "Ryder", "False... what? Th' hell are y' talkin' 'bout?"},
	{"FA_SMUG", "Facciolo", "Dark Gaia doesn't need you. They never did. All you've done is steal from them."},
	{"RY_ANGR", "Ryder", "Shut it! I ain't 'ere t' get trash-talked by some cowerin' fuzzball. Get out 'ere and fight!"},
	{"FA_SMUG", "Facciolo", "No. I want you to work to earn your victory... \"champion\"."},
}, "ryder", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_SHOK", "Facciolo", "You... made it?"},
	{"RY_ANGR", "Ryder", "Nobody insults th' damn champion an' gets away with it!"},
}, "ryder", BOSS_PHASE1_TEXT)

-- Dirk

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Ah... The boy of frost with the heart of gold. You were one of the first to befriend Inazuma."},
	{"FA_SMUG", "Facciolo", "Quite peculiar... Why go through such lengths to help out your friend?"},
	{"DR_ANGR", "Dirk", "You said it yourself! He's my friend and that's what friends do!"},
	{"DR_ANGR", "Dirk", "I will defeat you and save him!"},
	{"FA_SMUG", "Facciolo", "I admire your tenacity, boy. It is a shame that I must destroy you!"},
}, "dirk", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "Your training has made you strong, but challenging me? You will die a fool, boy."},
}, "dirk", BOSS_PHASE1_TEXT)

-- Whirl

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "A lost little ram wanders into the maw of the wolf. Poetic."},
	{"WH_LOL", "Whirl", "What's that supposed to mean?"},
	{"FA_SHOK", "Facciolo", "Fascinating. So there are those who do not fear the mighty Facciolo."},
	{"WH_LOL", "Whirl", "Who's that?"},
	{"FA_ANGR", "Facciolo", "...Me."},
	{"WH_LOL", "Whirl", "Oh! Nice to meet you, Facciolo! I'm Whirl."},
	{"FA_SMUG", "Facciolo", "How naive you are. It will also be your downfall."},
}, "whirl", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"WH_LOL", "Whirl", "Why are you on a tree?"},
	{"FA_ANGR", "Facciolo", "This is the center of the mind of Inazuma Silverhorn. This will also be your grave!"},
}, "whirl", BOSS_PHASE1_TEXT)

-- Vesola

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "So I heard you like gardening. What a tranquil hobby."},
	{"VL_OHNO", "Vesola", "I am. What's it to you?"},
	{"FA_SMUG", "Facciolo", "I find the act wonderful. Like you, I too tend to a garden. Though my canvas is much bigger."},
	{"VL_FEAR", "Vesola", "Are you seriously comparing enslavement to gardening?"},
	{"FA_SMUG", "Facciolo", "Is there really a difference?"},
	{"FA_SMUG", "Facciolo", "Controlling the fate of another living being and using them to your advantage?"},
	{"VL_FEAR", "Vesola", "You're insane!"},
}, "vesola", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "You have infested my garden for long enough! Away with you!"},
}, "vesola", BOSS_PHASE1_TEXT)

-- Robuster

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Well if it isn't the leader of Team Kinetic. How's the kinetic stone doing?"},
	{"RB_NEUT", "Robuster", "That's none of your business, punk!"},
	{"FA_SMUG", "Facciolo", "Much like how it's your family's business to lose that little rock, correct?"},
	{"RB_ANGR", "Robuster", "Hey!"},
	{"FA_SMUG", "Facciolo", "Oh, did I say lose? I meant protect."},
	{"FA_SMUG", "Facciolo", "Hard to remember since your kind is so bad at protecting such a useless artifact."},
	{"RB_SCRM", "Robuster", "That's it! I've had it with you! Why when I get my hands on you...!!!"},
}, "robuster", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"RB_SCRM", "Robuster", "ALRIGHT GEEZER! LET'S DO THIS!"},
}, "robuster", BOSS_PHASE1_TEXT)

-- The Kid

AddMindscapeDialogSet(nil, {
	{"THE_KID", "The Kid", "Die monster. You don't belong in this world!"},
	{"FA_ANGR", "Facciolo", "It was not by my hand that I am once again given flesh."},
	{"FA_ANGR", "Facciolo", "I was called here by humans who wish to pay me tribute."},
	{"THE_KID", "The Kid", "\"Tribute!?!\", You steal men's souls and make them your slaves!"},
	{"FA_SMUG", "Facciolo", "Perhaps the same could be said of all religions..."},
	{"THE_KID", "The Kid", "Your words are as empty as your soul! Mankind ill needs a savior such as you!"},
	{"FA_ANGR", "Facciolo", "What is a man? A miserable little pile of secrets. But enough talk... Have at you!"},
}, "thekid", BOSS_PHASE1_TEXT)

-- Fang

AddMindscapeDialogSet(nil, {
	{"FA_DEAD", "Facciolo", "There is no treasure for you here, hunter. Only death."},
	{"FG_MAD",  "Fang",     "I'm pretty sure there are some secrets to be uncovered here."},
	{"FG_MAD",  "Fang",     "Cough up the goods unless you want to get hurt, old man."},
	{"FA_DEAD", "Facciolo", "If it is a treasure that you seek:"},
	{"FA_DEAD", "Facciolo", "try looking for \x82places with great points of interest where people tend to gather\x80."},
	{"FA_SMUG", "Facciolo", "Perhaps a museum?"},
	{"FG_MAD",  "Fang",     "Is this supposed to be some kind of joke?"},
	{"FA_SMUG", "Facciolo", "Oh, we were not speaking in jest?"},
	{"FA_SMUG", "Facciolo", "I thought that pathetic display of bravado was your way of making me laugh."},
	{"FA_SMUG", "Facciolo", "Come to me if you dare, but know that your chances of survival here are zero."},
}, "fang", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "I see you let your greed get the better of you!\x82 A palace in the skies\x80 now awaits your arrival!"},
}, "fang", BOSS_PHASE1_TEXT)

-- Steve

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Hello again, Steve."},
	{"ST_OOF", "Steve", "oof"},
}, "steve", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "Time to kill you, Steve of Minecraft."},
	{"ST_TROL", "Steve", "sussy amongus???"},
	{"FA_SMUG", "Facciolo", "You fool, I was in electrical the whole time."},
}, "steve", BOSS_PHASE1_TEXT)

-- Amy
	
AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Run along, ugly little girl. This is no place for the likes of you."},
	{"AMYRCONC", "Amy", "Hey! What's with the attitude?! That's not how you treat people you just met!"},
}, "amy", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"AMYRJOYS", "Amy", "Oh, Mr. Wolf!~"},
	{"FA_SHOK", "Facciolo", "What the... Where did you get- Hey! What are you doing with that hammer!?"},
}, "amy", BOSS_PHASE1_TEXT)

-- Surge

AddMindscapeDialogSet(nil, {
	{"SG_CNFD", "Surge", "Wh..? Where the heck am I?!"},
	{"FA_SMUG", "Facciolo", "Another lightning elemental? Though, you seem to lack any control over it what-so-ever."},
	{"SG_ANGR", "Surge", "Huh!? You picking a fight, frak-face? Come down here so I can knock your stupid teeth out!"},
	{"FA_SMUG", "Facciolo", "Is that all it takes to knock you off kilter?"},
	{"FA_SMUG", "Facciolo", "Such arrogance... Your blind hot headedness will be the end of you."},
	{"SG_ANGR", "Surge", "Hey! Get back here! I'll rip you to pieces you stupid dog!"},
}, "surge", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "Brute forcing your way in here was a mistake, child. You are way out of your league."},
	{"SG_ANGR", "Surge", "Shut up! I'll shock your stupid eyesockets out!"},
}, "surge", BOSS_PHASE1_TEXT)

-- Blaze

AddMindscapeDialogSet(nil, {
	{"BZ_CNFS", "Blaze", "This wicked heat I sense... An ancient flame of malice and pure evil."},
	{"FA_SMUG", "Facciolo", "Hello again, Princess Blaze."},
	{"BZ_SPRZ", "Blaze", "You!? But...we sealed you!"},
	{"FA_SMUG", "Facciolo", "You think your little rocks could contain me? Come now, princess."},
	{"FA_SMUG", "Facciolo", "If your joint powers couldn't stop Inazuma, what makes you think it would work on me?"},
	{"BZ_CNFS", "Blaze", "In that case, then why appear now? Surely you could've done this alot sooner if our strike on you failed."}, 
	{"FA_SMUG", "Facciolo", "Simple; An opportunity. Inazuma's soul is finally vulnerable. Once I rend his soul from his body,"},
	{"FA_SMUG", "Facciolo", "I will finally become who I was always meant to be!"},
	{"BZ_DETR", "Blaze", "As if I would ever let that happen, monster!"}, 
}, "blaze", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "You should've run back to your kingdom! You are no martyr and your death will be in vain!"},
}, "blaze", BOSS_PHASE1_TEXT)

-- Sonic

AddMindscapeDialogSet(nil, {
	{"SH_HEHE", "Sonic", "Chaos Control again, huh? Things just got a lot more interesting!"},
	{"FA_SMUG", "Facciolo", "Ah, you again. The Blue Blur. Of all the beings I've encountered, you are by far the fastest."},
	{"SH_SMUG", "Sonic", "Good to see you too, spooky! You back for a rematch?"},
	{"FA_SMUG", "Facciolo", "That implies we have ever fought to begin with! If you and your two friends flailing"},
	{"FA_SMUG", "Facciolo", "at me with a bunch of colored rocks is what is considered fighting these days, then"},
	{"FA_SMUG", "Facciolo", " I must've been asleep for much longer than I thought!"},
	{"SH_WSUP", "Sonic", "Heh. Done with your beauty sleep, then? Because I want you to be wide awake when I kick"}, 
	{"SH_WSUP", "Sonic", "your butt and save Inazuma, again!"}, 
	{"FA_SMUG", "Facciolo", "Again? Sonic, you've only delayed the inevitable."},
}, "sonic", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"SH_SMUG", "Sonic", "Sitting up on a tree, huh? Well, can't blame ya, I do that a lot myself."},
	{"FA_ANGR", "Facciolo", "Cocky as ever I see. Let's finish this, Sonic!"},
}, "sonic", BOSS_PHASE1_TEXT)

-- Shadow
	
AddMindscapeDialogSet(nil, {
	{"SHD_NEUT", "Shadow", "That was... Chaos Control? But who cast it?"},
	{"FA_SMUG", "Facciolo", "Hello again, self proclaimed Ultimate Lifeform."},
	{"SHD_SHOCK", "Shadow", "You again!? But then that means..."},
	{"FA_SMUG", "Facciolo", "Your efforts those years ago were for naught. You've failed."},
	{"SHD_DTRM", "Shadow", "Not yet I haven't. My promise to him is only unfufilled."},
	{"SHD_DTRM", "Shadow", "You've made a big mistake facing me again, demon."},
	{"SHD_NEUT", "Shadow", "You're going straight to hell!"},
}, "shadow", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "I've been waiting, Shadow! Show me the power of Chaos!"},
	{"SHD_DTRM", "Shadow", "Time to put this to rest!"},
}, "shadow", BOSS_PHASE1_TEXT)

-- Bowser
	
AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Oh? Can't say I expected to see you here, Koopa King."},
	{"FA_SMUG", "Facciolo", "What do you have to gain from helping that deer?"},
	{"KK_CNFS", "Bowser", "Huh? You mean that Zappy Guy?"},
	{"KK_CNFS", "Bowser", "I dunno, I saw loads of others headed through that weird glowy thing."},
	{"KK_SMUG", "Bowser", "Figured I should follow too, don't want to miss out on the fun after all."},
	{"FA_SMUG", "Facciolo", "You know, your raw strength is quite admirable."},
	{"FA_SMUG", "Facciolo", "Join me and serve as my underling, and together, nothing shall stand in our way!"},
	{"KK_RAGE", "Bowser", "SERVE? You think the mighty leader of the Koopas would bow down to some mangy mutt? No way!"},
	{"FA_SMUG", "Facciolo","Have it your way then."},
}, "bowser", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"KK_RAGE", "Bowser", "Get down here and prepare for a beating, chump!"},
	{"FA_SMUG", "Facciolo", "Your overconfidence will be your undoing."},
}, "bowser", BOSS_PHASE1_TEXT)

-- Orville
	
AddMindscapeDialogSet(nil, {
	{"ORV_CONF", "Orville", "...Huh? What is this place? How did I get this off track?"},
	{"FA_SMUG", "Facciolo", "Being off track is the least of your concerns."},
	{"ORV_SHOCK", "Orville", "What the... Who are you?"},
	{"FA_ANGR", "Facciolo", "I am the master of this world and you are tresspassing."},
	{"ORV_SAD", "Orville", "Sorry, mister! It was a complete accident!"},
	{"ORV_SAD", "Orville", "Could you possibly show me the exit? I'm really trying to find my way home."},
	{"FA_SMUG", "Facciolo", "To leave, simply turn back. Do not come return here. Only death awaits you next time."},
	{"ORV_SAD", "Orville", "Alright then... Wait!"},
	{"ORV_SHOCK", "Orville", "That green hole disappeared! Hello?! Hey!! Where did you go!?"},
}, "orville", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"ORV_SHOCK", "Orville", "Uh oh!"},
	{"FA_ANGR", "Facciolo", "You were warned what returning here would bring! Now die!"},
}, "orville", BOSS_PHASE1_TEXT)

-- The Derp
	
AddMindscapeDialogSet(nil, {
	{"FA_SHOK", "Facciolo", "A marshmallow?"},
	{"THE_DERP", "Derp", "..."},
	{"FA_SHOK", "Facciolo", "A 'MAGIC' marshmallow?"},
	{"THE_DERP", "Derp", "..."},
	{"FA_DEAD", "Facciolo", "..."},
}, "derp", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_SHOK", "Facciolo", "So you're here to battle? Well, it's your funeral..."},
}, "derp", BOSS_PHASE1_TEXT)

-- Whisper
	
AddMindscapeDialogSet(nil, {
	{"WSP_NEUT", "Whisper", "..."},
	{"FA_SMUG", "Facciolo", "Try as you might to hide, wolf your mere presence gives you away."},
	{"WSP_HUH", "Whisper", "You're not Mimic."},
	{"FA_SMUG", "Facciolo", "And you're not the Silverhorn, it seems be are both mismatched with our endeavors."},
	{"WSP_HUH", "Whisper", "..."},
	{"WSP_SHOK", "Whisper", "...?"},
	{"WSP_NEUT", "Whisper", "How do I leave?"},
	{"FA_SMUG", "Facciolo", "Oh come now, you can't just leave now that you're here! At least, not in one piece."},
	{"FA_SMUG", "Facciolo", "I'm afraid you'll just have to add to my... collection, so to speak."},
	{"FA_SMUG", "Facciolo", "You'll fit in well next to the bratty lemur."},
	{"WSP_SHOK", "Whisper", "!!"},
	{"WSP_ANGR", "Whisper", "You!!"},
	{"FA_SMUG", "Facciolo", "Ahahah! It's but a taunt, my dear. Although it's pleasing to see I've jabbed at the right nerve..."},
	{"FA_SMUG", "Facciolo", "I guess you'll just have to see if I'm playing with you or not now, won't you?"},
	{"WSP_ANGR", "Whisper", "Mimic can wait, you're clearly up to no good."},
	{"FA_SMUG", "Facciolo", "Top of the list now, am I? Ahahah."},
	{"FA_SMUG", "Facciolo", "Truly, I'm flattered."},
}, "whisper", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"WSP_ANGR", "Whisper", "..."},
	{"FA_SMUG", "Facciolo", "No last words? Pity, you'll have missed your chance."},
}, "whisper", BOSS_PHASE1_TEXT)

-- Tangle
	
AddMindscapeDialogSet(nil, {
	{"TNG_HUH", "Tangle", "Whoah, this is a pretty weird place I've found, huh?"},
	{"FA_SMUG", "Facciolo", "What do we have here? A rather limberous lemuroidea."},
	{"TNG_HPY", "Tangle", "That's right I'm a-"},
	{"TNG_SAD", "Tangle", "Wait, a what?"},
	{"FA_SMUG", "Facciolo", "A monkey that comes to play."},
	{"FA_SMUG", "Facciolo", "As intriguing as an encounter this is, I'm afraid you've wound up in a very dangerous place."},
	{"FA_SMUG", "Facciolo", "It would be wise to leave while you've caught me in a good mood."},
	{"FA_SMUG", "Facciolo", "Because I won't give you another chance."},
	{"TNG_SAD", "Tangle", "Who're y-you calling a monkey? I'm Tangle and I'm a lemur!"},
	{"TNG_SAD", "Tangle", "If you're up to no good, I-I'll-"},
	{"FA_SMUG", "Facciolo", "Pffaahahahaha!!"},
	{"FA_SMUG", "Facciolo", "You're trembling, child. I'm glad I still exude that kind of presence among the people."},
	{"TNG_SAD", "Tangle", "I-I'm not-"},
	{"FA_SMUG", "Facciolo", "You don't know what you're getting yourself into, I have warned you."},
	{"FA_SMUG", "Facciolo", "Do not be disquieted when something awry should come to pass."},
	{"TNG_SAD", "Tangle", "I-I'll... I'll show you!"},
	{"TNG_SAD", "Tangle", "..."},
	{"TNG_SAD", "Tangle", "I wish Jewel or Whisper were here, they always help my nerves..."},
	{"TNG_HPY", "Tangle", "Come on girl, you can do this, just a little, interdimensional traversal."},
	{"TNG_HPY", "Tangle", "No biggie!"},
	{"TNG_SAD", "Tangle", "...right?"},
}, "tangle", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"TNG_SAD", "Tangle", "I'm not scared of you!"},
	{"FA_DEAD", "Facciolo", "You should be."},
}, "tangle", BOSS_PHASE1_TEXT)

-- Sage
	
AddMindscapeDialogSet(nil, {
	{"SGE_NORM", "Sage", "Error: Current location unknown. This is not Cyberspace."},
	{"SGE_NORM", "Sage", "Recalculating warp destination; I mustn't keep father waiting."},
	{"SGE_NORM", "Sage", "..."},
	{"SGE_NORM", "Sage", "Error... Impossible, a firewall?"},
	{"FA_SMUG", "Facciolo", "I don't believe I've seen a flying human before. Perhaps you aren't human at all."},
	{"SGE_NORM", "Sage", "You...! ...\n..."},
	{"SGE_NORM", "Sage", "There is no information regarding you within the EggNet's databases!"},
	{"FA_SMUG", "Facciolo", "Aww... Did pool ol' granny not warn you about the wolves?"},
	{"SGE_NORM", "Sage", "Wait... The Ancient logs! You're... Jaldabaoth! Class SS Native Predator!"},
	{"SGE_NORM", "Sage", "A threat potentially greater than The End!"},
	{"FA_SMUG", "Facciolo", "Hooh? So you know me by that name huh?"},
	{"FA_SMUG", "Facciolo", "I have not gone by that name in eons..."},
	{"SGE_NORM", "Sage", "You've killed so many... You're... The Enemy."},
	{"FA_SMUG", "Facciolo", "I suppose you could also call me that."},
	{"FA_SMUG", "Facciolo", "But what about you?"},
	{"FA_SMUG", "Facciolo", "You seem to have information and knowledge from the Ancients AND Robotnik. Who are you?"},
	{"SGE_NORM", "Sage", "You are The Enemy. I will not share data with you."},
	{"FA_SMUG", "Facciolo", "Then allow me... I've seen this world develop for thousands of years."},
	{"FA_SMUG", "Facciolo", "History should've long since forgotten about the Ancients."},
	{"FA_SMUG", "Facciolo", "And only the Ancients have ever referred to me by that name."},
	{"FA_SMUG", "Facciolo", "Which means either you are one of them..."},
	{"FA_DEAD", "Facciolo", "Or that blasted buffoon of a doctor actually managed to crack their systems before I could've."},
	{"SGE_SHOK", "Sage", "...!"},
	{"FA_SMUG", "Facciolo", "Struck a nerve, did I?"},
	{"FA_SMUG", "Facciolo", "Heh. Well then, you're going to tell me everything you know whether you like it or not."},
	{"SGE_SHOK", "Sage", "...If I am to keep him safe I must survive this!"},
}, "sage", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"SGE_SHOK", "Sage", "Compiling data, chances of success..."},
	{"FA_SMUG", "Facciolo", "Heh, go on, tell me the odds."},
}, "sage", BOSS_PHASE1_TEXT)

-- Knuckles

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "The current guardian of the Master Emerald!"},
	{"KX_NORM", "Knuckles", "..."},
	{"FA_DEAD", "Facciolo", "Not the talkative type, eh?"},
	{"KX_GLEE", "Knuckles", "Don't see much point talking to a ghost."},
	{"FA_SMUG", "Facciolo", " It sounds like you're speaking from experience."},
	{"KX_BRUH", "Knuckles", "Mmgh..."},
	{"FA_SMUG", "Facciolo", "Heh, I like you! You cut to the chase! Come then, and we shall let our fists do the talking!"},
}, "knuckles", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "Your island will make a fitting throne for me once I'm done!"},
	{"KX_GRRR", "Knuckles", "Tch! There's no way I'm losing to a geezer like you!"},
}, "knuckles", BOSS_PHASE1_TEXT)

-- Tails (Awesome)

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Miles Prower."},
	{"TA_SHOK", "Tails", "How did you-", {waittime=2}},
	{"FA_SMUG", "Facciolo", "I have my ways. So, they're sending children to fight now?"},
	{"TA_DETR", "Tails", "I'm here to help my friends! Any way I can!"},
	{"FA_DEAD", "Facciolo", "Some friends you have! Sending a child to die on the battlefield."},
	{"FA_SMUG", "Facciolo", "You're smart, Miles. Do the intelligent thing and leave while you can!"},
	{"TA_DETR", "Tails", "I can't do that! I'll contribute whatever I can! No matter the cost!"},
	{"FA_SMUG", "Facciolo", "Your determination is admirable. A shame I'll have to snuff out that light."},
}, "tails", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "I wasn't planning on burying a kid today, but if I have to, so be it!"},
}, "tails", BOSS_PHASE1_TEXT)

-- Marine (Stupid)
	
AddMindscapeDialogSet(nil, {
	{"MA_SHOK", "Marine", "Crikey! Where the bloomin' heck is this? Ain't no beach resort that's for sure!"},
	{"FA_DEAD", "Facciolo", "...Has English changed in the last century?"},
	{"MA_WUTS", "Marine", "Corr blimey, you're a dodgy lookin' bloke aintcha?"},
	{"MA_HEHE", "Marine", "I don't reckon you're on good terms with the true blue now, are ya?"},
	{"FA_DEAD", "Facciolo", "If by ''true blue'' you mean the hedgehog, no. I cannot say we're friends."},
	{"FA_SMUG", "Facciolo", "And what would a little rodent such as yourself care anyway?"},
	{"MA_SHOK", "Marine", "Oi! That aint no way to speak to such a respectable fine young lady ya know!"},
	{"MA_HEHE", "Marine", "I'm a captain n' got my own crew n' all! Me best mate's a Princess too!"},
	{"MA_HEHE", "Marine", "She'll singe your mug clean off too right!"},
	{"FA_SMUG", "Facciolo", "Princess? Do you speak of the Meridian or the Sol guardian?"},
	{"MA_SHOK", "Marine", "Mer-what-ian now?"},
	{"FA_DEAD", "Facciolo", "I suppose that answers that."},
	{"MA_SHOK", "Marine", "Corr rack off, you outback drongo!"},
	{"MA_SHOK", "Marine", "Figures the locals are all crass, no class, speakin' to a sheila like that, the nerve!"},
	{"FA_DEAD", "Facciolo", "..."},
}, "marine", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"MA_HEHE", "Marine", "Finally I can knock ya block off! Corr blimey you're in for it now-", {waittime=2}},
	{"FA_DEAD", "Facciolo", "ENOUGH! SHUT UP! STOP TALKING!"},
}, "marine", BOSS_PHASE1_TEXT)

-- Metal Sonic

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "What's this? Yet another of the doctor's robots?"},
	{"FA_DEAD", "Facciolo", "Seems this time he let his obsession over that blasted \"Blue Blur\" get to him..."},
	{"ME_NORM", "Metal Sonic", "..."},
	{"FA_SMUG", "Facciolo", "Ho ho? I don't suppose you are looking for a challenge, are you?"},
	{"ME_SMUG", "Metal Sonic", "...!"},
	{"FA_SMUG", "Facciolo", "Fortunately for you, the Silverhorn's mind is quite the treacherous place."},
	{"FA_SMUG", "Facciolo", "And I, quite the treacherous opponent. So let us put your skills to the test."},
	{"FA_SMUG", "Facciolo", "If you conquer both, then we can say you have truly surpassed the power of your doppelganger."},
	{"FA_SMUG", "Facciolo", "Come, and face the endless abyss of consciousness!"},
}, "metalsonic", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "You've done well to make it this far. But speed means nothing if you cannot return the blow!"},
}, "metalsonic", BOSS_PHASE1_TEXT)

-- Tails Doll
	
AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "What's this? A doll?"},
	{"TD_SOUL", "Tails Doll", "..."},
	{"FA_DEAD", "Facciolo", "The doctor certainly has quite the peculiar taste in design."},
}, "tailsdoll", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_DEAD", "Facciolo", "Foul machine. Your pitiful existence ends now!"},
}, "tailsdoll", BOSS_PHASE1_TEXT)

-- Silver

AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Ah, the time traveler. Silver, correct?"},
	{"SLV_NEUT", "Silver", "In the flesh!"},
	{"FA_SMUG", "Facciolo", "Tell me... does history remember my name in your time?"},
	{"SLV_ANGR", "Silver", "I'm here to make sure it doesn't."},
}, "silver", INTRO_TEXT)

AddMindscapeDialogSet(nil, {
	{"FA_ANGR", "Facciolo", "Ever the persistent, naive, hero, hedgehog. I shall cleanse the timeline of you!"},
}, "silver", BOSS_PHASE1_TEXT)

-- DeltaChars

mindtext[INTRO_TEXT]["kris"] = 0
mindtext[INTRO_TEXT]["susie"] = 0
mindtext[INTRO_TEXT]["ralsei"] = 0

local DELTA_INTRO1_TEXT = AddMindscapeDialog(nil, "FA_SMUG", "Facciolo", "That's \"Him\" with a capital H, by the way.")
local DELTA_INTRO2_TEXT = AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "I could ask you the same thing. You four are far beyond your depths."},
	{"FA_SMUG", "Facciolo", "Run along now, children, lest you seek an early death.", {waittime=60}},
})
local DELTA_INTRO3_TEXT = AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "I do not know any of you lost children, but I advise you to stay clear of my path."},
	{"FA_ANGR", "Facciolo", "Unless you want to die like the rest.", {waittime=60}},
})
local DELTA_INTRO4_TEXT = AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "I've met many children as courageous as you lot."},
	{"FA_SMUG", "Facciolo", "Alas their efforts were all for naught as their hopes and dreams died with them."},
	{"FA_SMUG", "Facciolo", "And you? I sense in your heart your determination. "},
	{"FA_ANGR", "Facciolo", "You wish to challenge me!? Indulge me then, heroes!"},
	{"FA_ANGR", "Facciolo", "Show me how far that determination gets you!", {waittime=60}},
})

local DELTA_PHASE1_TEXT = AddMindscapeDialogSet(nil, {
	{"FA_SMUG", "Facciolo", "Let us see whose determination comes out on top.", {waittime=60}},
}, "kris", BOSS_PHASE1_TEXT)

mindtext[BOSS_PHASE1_TEXT]["susie"] = DELTA_PHASE1_TEXT
mindtext[BOSS_PHASE1_TEXT]["ralsei"] = DELTA_PHASE1_TEXT

-- DeltaChars 2!!

local function yield()
	coroutine.yield()
end

local function talk(name, portraitNum, text)
	return {
		name = name,
		portraitNum = portraitNum,
		text = text
	}
end

local function introCondition()
	return leveltime > 2 and MindscapeSessionFlags & MS_SEENINTRO
	and consoleplayer and consoleplayer.mo
	and P_IsObjectOnGround(consoleplayer.mo)
end

local deltaEvent1 = {
	tag = "mindscapeIntro",
	zone = "mindscape",
	act = 0,
	hookType = "ThinkFrame",
	condition = introCondition,
	script = {
		talk("susie", 1, "..."),
		talk("susie", 12, "Huh. I don't recall this in Spiky Robot Bash 2."),
		talk("susie", 26, "KRIS, DID YOU HACK YOUR CONSOLE??"),
		talk("ralsei", 13, "!!"),
		talk("susie", 36, "Ralsei..? Is everything okay?"),
		talk("ralsei", 28, "...This... This place is filled with darkness."),
		talk("ralsei", 35, "It's like... Like a dozen Dark Fountains sprung up all at once!"),
		talk("ralsei", 35, "What sort of evil could be lurking in these ruins?"),
		talk("ralsei", 36, "It's... overwhelming! C-Could it be... him?"),
		function()
			startmindtext(DELTA_INTRO1_TEXT)
			while textid
				yield()
			end
			return true
		end,
		talk("susie", 20, "W-WHAT? WHERE DID THAT COME FROM?"),
		function()
			startmindtext(DELTA_INTRO2_TEXT)
			while textid
				yield()
			end
			return true
		end,
		talk("susie", 2, "..."),
		talk("susie", 12, "...Four?"),
		function()
			startmindtext(DELTA_INTRO3_TEXT)
			while textid
				yield()
			end
			return true
		end,
		talk("susie", 22, "Ho-ho, threats huh? You don't scare us, voice in the sky!!"),
		talk("susie", 3, "We're gonna bring you and your fountains down!!"),
		talk("ralsei", 35, "S-Susie! This darkness is much stronger than what we've faced before!"),
		talk("susie", 11, "Yeah? Well, we're TEN times tougher than anything else."),
		talk("susie", 10, "Right Kris??"),
		talk(nil, nil, "Although the sense of dread is overwhelming..."),
		talk(nil, nil, "You and your friends are DETERMINED to see this through."),
		function()
			startmindtext(DELTA_INTRO4_TEXT)
			while textid
				yield()
			end
			return true
		end,
		talk("ralsei", 23, "..."),
		talk("susie", 17, "..."),
	}
}

local deltaEvent2 = {
	tag = "mindscapeLater",
	zone = "mindscape",
	act = 0,
	hookType = "ThinkFrame",
	condition = introCondition,
	script = {
		TICRATE*10,
		talk("ralsei", 19, "Susie, are you alright? You're shaking."),
		talk("susie", 4, "H-Haha, I'm just cold alright?"),
		talk("susie", 21, "Just... Just a little cold..."),
		talk("susie", 9, "We got this! Just like always."),
		talk("susie", 12, '(What did he mean by "you four", anyway?)'),
		talk("susie", 1, "(Is someone else here already?)"),
	}
}

local deltaEvent3 = {
	tag = "mindscapeBoss",
	save = false,
	zone = "mindscape",
	act = 0,
	hookType = "ThinkFrame",
	preCondition = function()
		DeltaChars.DialogueBox.vars.facciolo = false
		return true
	end,
	condition = function()
		if DeltaChars.DialogueBox.vars.facciolo
			return not textid
		elseif textid == DELTA_PHASE1_TEXT
			DeltaChars.DialogueBox.vars.facciolo = true
		end
	end,
	script = {
		talk("susie", 28, "Y-Yeah well... You can't count properly!"),
	}
}

//Add DeltaChars
local deltaAdded = false
addHook("ThinkFrame", do
	if not deltaAdded and DeltaChars
		DeltaChars.DialogueBox.AddEvent(deltaEvent1)
		DeltaChars.DialogueBox.AddEvent(deltaEvent2)
		DeltaChars.DialogueBox.AddEvent(deltaEvent3)
		local funny = DeltaChars.Singleplayer
		DeltaChars.Singleplayer = function(switching)
			if mapheaderinfo[gamemap].mindscape
				return true
			end
			return funny(switching)
		end
		deltaAdded = true
	end
end)

-- Used for Intro. (needs to be excluded from sets for separate call times)

AddMindscapeDialog(1510, "IN_DETR", "Inazuma", "What the...? What's going on? I was trying to use the emeralds to heal myself... so... ")
AddMindscapeDialog(1511, "IN_DETR", "Inazuma", "Why am I here?")
AddMindscapeDialog(1512, "FA_UNKN", "???", "Because even the emeralds know that \x82THIS\x80 is the only way!")
AddMindscapeDialog(1513, "IN_RAGE", "Inazuma", "Facciolo!!!")
AddMindscapeDialog(1514, "IN_RAGE", "Inazuma", "Why are you here?!")
AddMindscapeDialog(1515, "FA_SMUG", "Facciolo", "Because I live here dumbass.")
AddMindscapeDialog(1516, "IN_DETR", "Inazuma", "...What...? Then that means...")
AddMindscapeDialog(1520, "FA_SMUG", "Facciolo", "That's right. We are inside your \x82mind\x80. But this is different from our normal meet ups, isn't it?")
AddMindscapeDialog(1521, "FA_SMUG", "Facciolo", "We are both \x82physical spirits\x80 in this realm. We are now truly talking face to face.")
AddMindscapeDialog(1522, "FA_DEAD", "Facciolo", "Which also means...")
AddMindscapeDialog(1523, "FA_ANGR", "Facciolo", "I CAN FINALLY KILL YOU!")
AddMindscapeDialog(1524, "FA_ANGR", "Facciolo", "My original offer to you is now an ultimatum. Fuse with me, or else I'll destroy your soul.")
AddMindscapeDialog(1525, "FA_DEAD", "Facciolo", "You are a valiant warrior... And you are no fool.")
AddMindscapeDialog(1526, "FA_SMUG", "Facciolo", "You know in your heart that your rebellion is meaningless.")
AddMindscapeDialog(1527, "FA_DEAD", "Facciolo", "The entire world hates you and wants you dead.")
AddMindscapeDialog(1528, "FA_DEAD", "Facciolo", "What purpose is there for you to maintain this status quo?")
AddMindscapeDialog(1529, "IN_DETR", "Inazuma", "Because it's my decision to make.")
AddMindscapeDialog(1530, "FA_ANGR", "Facciolo", "Very well then. If you truly wish to keep this up, then meet me at the core.")
AddMindscapeDialog(1531, "FA_DEAD", "Facciolo", "We'll finish it all there.")
AddMindscapeDialog(1532, "CREWMATE", "Crewmate", "among us")
AddMindscapeDialog(1533, "WHAT", "Facciolo", "What the fuck.")

AddMindscapeDialog(1550, "IN_DETR", "Inazuma", "The curse... it's been lifted. So, this is what it's like to move my body on my own.")
AddMindscapeDialog(1551, "FA_DEAD", "Facciolo", "How... was I bested?")
AddMindscapeDialog(1552, "IN_DETR", "Inazuma", "...")
AddMindscapeDialog(1553, "FA_DEAD", "Facciolo", "You have nothing... No past. No future. No reason to fight.")
AddMindscapeDialog(1554, "FA_DEAD", "Facciolo", "So why now!? After all my efforts!")
AddMindscapeDialog(1555, "FA_DEAD", "Facciolo", "Trillions sacrificed for the greater good - for multiversal peace! ")
AddMindscapeDialog(1556, "IN_DETR", "Inazuma", "You planned to take control of everyone's fate, like what you did to me, and look where that got you.")
AddMindscapeDialog(1557, "IN_DETR", "Inazuma", "A life that doesn't belong to yourself is not one worth living.")
AddMindscapeDialog(1558, "IN_DETR", "Inazuma", "I refuse to subject myself to you or anyone else. I decide my own fate.")
AddMindscapeDialog(1559, "FA_DEAD", "Facciolo", "Thousands of years of planning and work...")
AddMindscapeDialog(1560, "FA_DEAD", "Facciolo", "I will not let what I've built be destroyed by a child, especially a Silverhorn!")
AddMindscapeDialog(1561, "FA_DEAD", "Facciolo", "You may have beaten me...")
AddMindscapeDialog(1562, "FA_ANGR", "Facciolo", "But with my last embers I'll make sure that you're not getting out of here alive!")

AddMindscapeDialog(1563, "IN_ULTI", "Inazuma", "Let's do this.")

-- Used for Prologue
-- The IDs are important here since they need to be referenced in the map itself!

local TUTORIAL_TEXT_1 = 2000
local TUTORIAL_TEXT_2 = 2006
local TUTORIAL_TEXT_3 = 2010
local TUTORIAL_TEXT_4 = 2013
local TUTORIAL_TEXT_5 = 2022
local TUTORIAL_TEXT_6 = 2025
local TUTORIAL_TEXT_7 = 2026
local TUTORIAL_TEXT_8 = 2027
local TUTORIAL_TEXT_9 = 2028
local TUTORIAL_TEXT_10 = 2032
local TUTORIAL_TEXT_11 = 2035
local TUTORIAL_TEXT_12 = 2037
local TUTORIAL_TEXT_13 = 2045
local TUTORIAL_TEXT_14 = 2052
local TUTORIAL_TEXT_15 = 2055
local TUTORIAL_TEXT_16 = 2059
local TUTORIAL_TEXT_17 = 2062
local TUTORIAL_TEXT_18 = 2066
local TUTORIAL_TEXT_19 = 2069
local TUTORIAL_TEXT_20 = 2072

AddMindscapeDialogSet(TUTORIAL_TEXT_1, {
	{"AE_GLEE", "Aether", "Heeeey!! Training day!! You ready Zuzu?", {waittime = -1}},
	{"IN_DAMN", "Inazuma", "Is this really necessary?", {waittime = -1}},
	{"AE_SMUG", "Aether", "Of course! It's always good for us to refresh ourselves on the basics!", {waittime = -1}},
	{"AE_REDY", "Aether", "Our skillsets are very different from each other. So if we want to work as a team, we gotta practice!", {waittime = -1}},
	{"IN_DETR", "Inazuma", "Alright... I guess I'll start then. I see a bunch of orbs over there.", {waittime = -1}},
	{"IN_DETR", "Inazuma", "If I double jump near them, I can use my Lightning Jump to get us over the wall.", {waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_2, {
	{"AE_GLEE", "Aether", "Nice!! Great job! Hehe~!", {waittime = -1}},
	{"IN_DETR", "Inazuma", "When there are no targets nearby, I will instead perform Bolt Stream which will propel me forward!",{waittime = -1}},
	{"IN_DETR", "Inazuma", "Performing bolt stream will stall my descent a bit. And I usually have three charges per jump.", {waittime = -1}},
	{"IN_DETR", "Inazuma", "There's a lot of rings over there! My spin ability, Electric Barrier, will make collecting them easy!", {waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_3, {
	{"IN_DETR", "Inazuma", "Electric Barrier will make me walk slower when used on the ground.",{waittime = -1}},
	{"IN_DETR", "Inazuma", "When used in the air, Electric Barrier will instead slow down my descent.", {waittime = -1}},
	{"IN_DETR", "Inazuma", "Using Electric Barrier along with Bolt Stream will help me cross over large gaps.", {waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_4, {
	{"IN_DETR", "Inazuma", "That wall is too high. I won't be get up there with a normal Lightning Jump.",{waittime = -1}},
	{"AE_REDY", "Aether", "Would you like me to help?", {waittime = -1}},
	{"IN_DETR", "Inazuma", "I'm good. I just need to power myself up.", {waittime = -1}},
	{"IN_DETR", "Inazuma", "Holding down jump while performing Bolt Stream will make me do Bolt Blast instead.", {waittime = -1}},
	{"IN_DETR", "Inazuma", "Bolt Blast will thrust me further than a normal Bolt Stream and consume my remaining dashes.", {waittime = -1}},
	{"IN_DETR", "Inazuma", "Using Bolt Blast will let me enter my Neural Shock State.", {waittime = -1}},
	{"IN_DETR", "Inazuma", "While in Neural Shock almost all of my abilities are powered up.", {waittime = -1}},
	{"IN_DETR", "Inazuma", "I can continue to use Bolt Blast to limitlessly power up my Neural Shock state.", {waittime = -1}},
	{"IN_DETR", "Inazuma", "I just need one charge to gain enough power to hop that wall.", {waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_5, {
	{"AE_GLEE", "Aether", "Wow! You jumped really high there!",{waittime = -1}},
	{"IN_DETR", "Inazuma", "Being in Neural shock will also let me bust through spikes and run on water.", {waittime = -1}},
	{"IN_DETR", "Inazuma", "I just need to be careful about my speed as it becomes harder to control myself at high levels.", {waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_6, {
	{"IN_DETR", "Inazuma", "I can break through that wall of spikes while in Neural Shock."}
})

AddMindscapeDialogSet(TUTORIAL_TEXT_7, {
	{"IN_DETR", "Inazuma", "Neural Shock will also let me cross this body of water."}
})

AddMindscapeDialogSet(TUTORIAL_TEXT_8, {
	{"IN_DETR", "Inazuma", "There is a bustable floor here. I can break open this floor with Electric Barrier.", {waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_9, {
	{"IN_DETR", "Inazuma", "I think I'm good with the basics. It's your turn, Aether.", {waittime = -1}},
	{"AE_SHOK", "Aether", "Aw come ooon.... I don't wannaaaa!", {waittime = -1}},
	{"IN_DAMN", "Inazuma", "Aether.", {waittime = -1}},
	{"AE_SHOK", "Aether", "Ugh.. fine... To switch leaders, press Custom 1!",{waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_10, {
	{"AE_REDY", "Aether", "I can fly over that wall with my Crystal Wings! I can use them once I enter flight mode!", {waittime = -1}},
	{"AE_REDY", "Aether", "To enter flight mode, I must be running at top speed for a set amount of time.", {waittime = -1}},
	{"AE_REDY", "Aether", "Once I see glitter sparkling out of me, that means I've entered flight mode!",{waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_11, {
	{"IN_DETR", "Inazuma", "That was pretty great, Aether!", {waittime = -1}},
	{"AE_REDY", "Aether", "When my wings aren't ready, my slowfall ability can help get me to safety when in a pinch.", {waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_12, {
	{"AE_REDY", "Aether", "When pressing spin, I will open up a menu that has a list of available spells that I can use.", {waittime = -1}},
	{"AE_REDY", "Aether", "Each spell works differently from each other, and are identified by their symbol and color.", {waittime = -1}},  
	{"AE_REDY", "Aether", "Purple is Arcane Rush, Blue is Aura Burst, Green is Speedwave, and Yellow is Sword Rain.", {waittime = -1}},
	{"AE_REDY", "Aether", "While the menu is open, use your movement key input to select the spell that you want to use.", {waittime = -1}},
	{"AE_REDY", "Aether", "Alternatively, you can use the WEP PREV and WEP NEXT keys to switch between spells.", {waittime = -1}},
	{"AE_REDY", "Aether", "You can also use the keys for Weapon Slots 1-5 to quickly switch between spells.", {waittime = -1}},
	{"IN_DETR", "Inazuma", "There seems to be a high wall over there. I bet you can get over it with Arcane Rush!", {waittime = -1}},
	{"AE_REDY", "Aether", "If I use Arcane Rush on the ground, It will give me enough height to put me up to the top of that wall!", {waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_13, {
	{"IN_DETR", "Inazuma", "Looks like your spell is still lingering down there.", {waittime = -1}},
	{"AE_GLEE", "Aether", "Yup! After performing Arcane Rush, I will leave behind a temporary vortex that acts like a spring!", {waittime = -1}},
	{"AE_REDY", "Aether", "This vortex can be used by my allies or by me!", {waittime = -1}},
	{"AE_REDY", "Aether", "Additionally, the air version of Arcane Rush will thrust me forward and also leave a vortex!", {waittime = -1}},
	{"AE_REDY", "Aether", "The air variant of the vortex will thrust my allies and myself forward as well.", {waittime = -1}},
	{"AE_REDY", "Aether", "Finally, I will need to keep track of the cooldowns on my spells.", {waittime = -1}},
	{"AE_REDY", "Aether", "Each spell has its own unique cooldown timer that you can track with the HUD to your right!", {waittime = -1}},	
})

AddMindscapeDialogSet(TUTORIAL_TEXT_14, {
	{"AE_REDY", "Aether", "If I use Speedwave, I will gain an incredible boost in speed AND jump!", {waittime = -1}},
	{"AE_REDY", "Aether", "In addition, it will also instantly activate my Crystal Wings!", {waittime = -1}},
	{"AE_REDY", "Aether", "Using Speedwave here will let me cross that gap without needing to build up speed!", {waittime = -1}},	
})

AddMindscapeDialogSet(TUTORIAL_TEXT_15, {
	{"IN_DETR", "Inazuma", "Looks like that enemy is keeping that door to our right closed.", {waittime = -1}},
	{"IN_DETR", "Inazuma", "That enemy also seems to be shielded by some sort of wall of psychic energy.", {waittime = -1}},	
	{"AE_REDY", "Aether", "I should still be able to defeat that enemy. I just need to use my Sword Rain Spell!", {waittime = -1}},
	{"AE_REDY", "Aether", "Sword Rain summons multiple mana blade projectiles that homes in on nearby enemies and monitors!", {waittime = -1}},	
})

AddMindscapeDialogSet(TUTORIAL_TEXT_16, {
	{"IN_DETR", "Inazuma", "Well that floor looks... Hazardous.", {waittime = -1}},
	{"AE_REDY", "Aether", "Aura Burst is perfect for this situation!", {waittime = -1}},
	{"AE_REDY", "Aether", "Aura Burst will create a powerful wave of magic energy that will break anything that is breakable!", {waittime = -1}},		
})

AddMindscapeDialogSet(TUTORIAL_TEXT_17, {
	{"AE_REDY", "Aether", "Performing Aura Burst will also summon a magic barrier around my allies and myself!", {waittime = -1}},
	{"AE_REDY", "Aether", "This barrier can take up to two hits and anyone protected by this won't take hitstun when damaged!", {waittime = -1}},	
	{"IN_DETR", "Inazuma", "Looks like your Castle is just up ahead!", {waittime = -1}},	
	{"AE_GLEE", "Aether", "That was a lot of fun! Let's do this again sometime!", {waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_18, {
	{"AE_SHOK", "Aether", "Darn! Missed the platform...", {waittime = -1}},
	{"IN_DETR", "Inazuma", "To cross large gaps, I must use Bolt Stream and Electric Barrier to slow down my descent.",{waittime = -1}},
	{"IN_DETR", "Inazuma", "Timing is key.", {waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_19, {
	{"AE_SHOK", "Aether", "Whoops!...", {waittime = -1}},
	{"IN_DETR", "Inazuma", "I can cross bodies of water while in Neural Shock.",{waittime = -1}},
	{"IN_DETR", "Inazuma", "To enter Neural Shock, press and hold Jump after performing a Bolt Stream.", {waittime = -1}},
})

AddMindscapeDialogSet(TUTORIAL_TEXT_20, {
	{"AE_SHOK", "Aether", "Darn! Didn't make it!", {waittime = -1}},
	{"AE_REDY", "Aether", "I'll need to run around a bit in order to activate my Crystal Wings!",{waittime = -1}},
	{"AE_REDY", "Aether", "I can also use spells like Arcane Rush or Speedwave to boost myself up into the air!", {waittime = -1}},
})

// Used for Aether's Castle
// Make sure you use this with the stuff in AethersCastle.lua!

AddMindscapeDialogSet(3000, {
	{"none", "Toad", "Haha! I am Toad!", {waittime=-1}},
	{"skin", "skin", "Bruh", {waittime=-1}},
})
AddMindscapeDialogSet(3002, {
	{"none", "Toad", "I'm the best!", {waittime=-1}},
	{"skin", "skin", "That's nice", {waittime=-1}},
})
AddMindscapeDialogSet(3004, {
	{"none", "Toad", "AAAAAAAAAAAA", {waittime=-1}},
	{"skin", "skin", "Dude quiet down", {waittime=-1}},
})
AddMindscapeDialogSet(3006, {
	{"none", "Toad", "What's this icky paint-like goop?", {waittime=-1}},
	{"skin", "skin", "What's what", {waittime=-1}},
})
AddMindscapeDialogSet(3008, {
	{"none", "Toad", "Bingo! Hahaha", {waittime=-1}},
	{"skin", "skin", "No", {waittime=-1}},
})
AddMindscapeDialogSet(3010, {
	{"none", "Toad", "Time for adventure!", {waittime=-1}},
	{"skin", "skin", "Shut up", {waittime=-1}},
})
AddMindscapeDialogSet(3012, {
	{"none", "Toad", "OOOooooAAAAoOAOAOA", {waittime=-1}},
	//{"skin", "skin", "Die please", {waittime=-1}}, //Oops, one dialog too many
})

// Vio dialog with anyone
AddMindscapeDialog(3013, "VI_NORM", "Vio", "Aether won't let me sit on the chair anymore after what happened last time.", {waittime=-1})
AddMindscapeDialog(3014, "VI_NORM", "Vio", "Don't worry, that's just part of my body. Trust me.", {waittime=-1})
AddMindscapeDialog(3015, "VI_FLIP", "Vio", "Do you think Santa visits Elvoway?", {waittime=-1})
AddMindscapeDialog(3016, "VI_MINI", "Vio", "She can't get rid of all of us.", {waittime=-1})
AddMindscapeDialog(3017, "VI_BABY", "Vio", "She'll never see it coming.", {waittime=-1})
AddMindscapeDialog(3018, "VI_SPLT", "Vio", "Drive responsibly.", {waittime=-1})
AddMindscapeDialog(3019, "VI_NORM", "Vio", "I don't actually need to eat. It's just fun and I like tasting things.", {waittime=-1})
AddMindscapeDialogSet(3020, {
	{"VI_NORM", "Vio", "Oh, this? I'm not dancing.", {waittime=-1}},
	{"VI_NORM", "Vio", "I'm just trying to see how long it will take until the librarian asks me to leave.", {waittime=-1}},
})
AddMindscapeDialogSet(3022, {
	{"VI_NORM", "Vio", "Only Elvowain Magicians and those with special permission can set foot on castle grounds...", {waittime=-1}},
	{"VI_NORM", "Vio", "Don't tell Aether I'm here.", {waittime=-1}},
})

//Vio dialog with Aether
AddMindscapeDialogSet(nil, {
	{"AE_SHOK", "Aether", "Vio! What did we talk about!?", {waittime=-1}},
	{"VI_NORM", "Vio", "I'm not touching the chair! I'm just standing next to it!", {waittime=-1}},
}, "aether", 3013)
AddMindscapeDialogSet(nil, {
	{"AE_SHOK", "Aether", "What are you doing here!?", {waittime=-1}},
	{"AE_SHOK", "Aether", "You're getting your body slime all over the place!", {waittime=-1}},
	{"VI_NORM", "Vio", "Excuse you, my body isn't slime!", {waittime=-1}},
	{"AE_SHOK", "Aether", "Well, what are you made of then?", {waittime=-1}},
	{"VI_NORM", "Vio", "...", {waittime=-1}},
}, "aether", 3014)
AddMindscapeDialogSet(nil, {
	{"VI_FLIP", "Vio", "Hi mom.", {waittime=-1}},
	{"AE_GLEE", "Aether", "Hello Vio!", {waittime=-1}},
}, "aether", 3015)
AddMindscapeDialogSet(nil, {
	{"VI_MINI", "Vio", "We are legion.", {waittime=-1}},
}, "aether", 3016)
AddMindscapeDialogSet(nil, {
	{"VI_BABY", "Vio", "I am a plushie. Ignore me.", {waittime=-1}},
}, "aether", 3017)
AddMindscapeDialogSet(nil, {
	{"AE_SHOK", "Aether", "Are you okay!?", {waittime=-1}},
	{"VI_SPLT", "Vio", "Yeah.", {waittime=-1}},
}, "aether", 3018)
AddMindscapeDialogSet(nil, {
	{"VI_NORM", "Vio", "Remember that hot dog eating contest you had with Skyte?", {waittime=-1}},
	{"AE_SHOK", "Aether", "That was a magic duel. There weren't even hot dogs there!", {waittime=-1}},
}, "aether", 3019)
AddMindscapeDialogSet(nil, {
	{"AE_REDY", "Aether", "Doing some stretches?", {waittime=-1}},
	{"VI_NORM", "Vio", "No.", {waittime=-1}},
}, "aether", 3020)
AddMindscapeDialogSet(nil, {
	{"AE_SHOK", "Aether", "Vio? How did you get here?", {waittime=-1}},
	{"VI_NORM", "Vio", "I walked here.", {waittime=-1}},
	{"AE_SHOK", "Aether", "How???", {waittime=-1}},
	{"VI_NORM", "Vio", "...", {waittime=-1}},
}, "aether", 3022)

//Vio dialog with Inazuma
AddMindscapeDialogSet(nil, {
	{"VI_NORM", "Vio", "What if I put tacks on the chair?", {waittime=-1}},
	{"IN_DETR", "Inazuma", "You want Brightheart to kill you?", {waittime=-1}},
	{"VI_NORM", "Vio", "He can try.", {waittime=-1}},
}, "inazuma", 3013)
AddMindscapeDialogSet(nil, {
	{"VI_NORM", "Vio", "Hey Dad, are you here to give me a bath?", {waittime=-1}},
	{"IN_DAMN", "Inazuma", "No.", {waittime=-1}},
}, "inazuma", 3014)
AddMindscapeDialogSet(nil, {
	{"IN_DETR", "Inazuma", "Are you sure this is a good idea?", {waittime=-1}},
	{"VI_FLIP", "Vio", "I always have the best ideas!", {waittime=-1}},
	{"IN_DAMN", "Inazuma", "...", {waittime=-1}},
}, "inazuma", 3015)
AddMindscapeDialogSet(nil, {
	{"IN_DETR", "Inazuma", "What are you doing, Vio?", {waittime=-1}},
	{"VI_MINI", "Vio", "Asserting dominance.", {waittime=-1}},
	{"IN_DETR", "Inazuma", "No, you're not.", {waittime=-1}},
}, "inazuma", 3016)
AddMindscapeDialogSet(nil, {
	{"IN_ANGR", "Inazuma", "Get out of here, Vio!", {waittime=-1}},
	{"VI_BABY", "Vio", "Hey, how come I can't be in here but you can!?", {waittime=-1}},
	{"IN_DETR", "Inazuma", "...", {waittime=-1}},
	{"IN_ANGR", "Inazuma", "Just get out!", {waittime=-1}},
}, "inazuma", 3017)
AddMindscapeDialogSet(nil, {
	{"VI_SPLT", "Vio", "Hi Dad.", {waittime=-1}},
	{"IN_DAMN", "Inazuma", "Stop lying there, Vio. You'll cause an accident.", {waittime=-1}},
}, "inazuma", 3018)
AddMindscapeDialogSet(nil, {
	{"VI_NORM", "Vio", "Care to join me, Dad?", {waittime=-1}},
	{"IN_DETR", "Inazuma", "Not really.", {waittime=-1}},
	{"VI_NORM", "Vio", "What if I asked the Chef to bring in a wheel of cheese?", {waittime=-1}},
	{"IN_DETR", "Inazuma", "...", {waittime=-1}},
}, "inazuma", 3019)
AddMindscapeDialogSet(nil, {
	{"IN_DAMN", "Inazuma", "Vio, stop bothering the librarian.", {waittime=-1}},
	{"VI_NORM", "Vio", "Nuh uh.", {waittime=-1}},
}, "inazuma", 3020)
AddMindscapeDialogSet(nil, {
	{"VI_NORM", "Vio", "Hi Daddy.", {waittime=-1}},
	{"IN_DAMN", "Inazuma", "What do you want?", {waittime=-1}},
	{"VI_NORM", "Vio", "I wanted to see how you were doing!", {waittime=-1}},
	{"IN_DAMN", "Inazuma", "That's nice. Go away.", {waittime=-1}},
}, "inazuma", 3022)

// Maid / Knight Dialog
local castledialog = {
	//Normal Maid
	"Welcome to Castle Elvoway!", //3024
	"Feel free to roam about as you please!", //3025
	"Lovely weather we're having, wouldn't you agree?", //3026
	"Do not worry, if you accidentally fall, we will quickly escort you back to safety!", //3027
	{"Some of us are sisters. Some of us are using cloning magic.", //3028
		"Care to guess which one I am? Hehehe~"}, //3029
	"Oh goodness, there's so much work to be done!", //3030
	"Oh dear, I hope the King of Magic is doing well.", //3031
	
	//Portrait Room Maid
	"Welcome to the Portrait Room!", //3032
	{"These frames were crafted by our wonderful King of Magic!", //3033
		"They have the ability to show the viewer different futures!"}, //3034
	"Oh dear, this room gets dirty quite often...", //3035
	{"The foliage here is the result of powerful magic in this room.", //3036
		"It amplifies the plant's life force as well!"}, //3037
	{"Long ago, King Brightheart wanted to take down this room.", //3038
		"But our wondrous Princess Aether convinced him not to."}, //3039
	{"Princess Aether used to frequent this room.", //3040
		"She would soar close to the ceiling and play with us as we cleaned."}, //3041
	{"Unfortunately, these frames don't reveal the future to everyone.", //3042
		"They have a bit of sentience, you see.", //3043
		"They only reveal their pictures to those they deem worthy."}, //3044
		
	//
	"knightstart", //Knight dialog start indicator
	//

	//Normal Knight
	"...", //3045
	"...", //3046
	"...", //3047
	"...", //3048
	"...", //3049
	"...", //3050
	"Don't you have anything better to do?", //3051
	
	//Throne Knight
	{"King Brightheart is not here right now.", //3052
			"We will be protecting his throne as we await his return."}, //3053
	"Our all powerful and wise King of Magic is out on another expedition to claim more land from the miasma.", //3054
	{"King Brightheart is also known as the King of Magic.", //3055
			"By far, he is one of the strongest Battle Mages in Elvoway's recorded history.", //3056
			"He is second only to his daughter, Princess Aether, who was born as the Meridian of Magic."}, //3057
	"It is said that King Brightheart's power rivals several previous Meridians of Magic.", //3058
	{"The King of Magic is someone we all depend on.", //3059
			"His power protects us from the miasma and provides bountiful harvests every single year."}, //3060
	{"One day, Princess Aether will ascend to the throne and become our next ruler.", //3061
		"She will make a fine Queen one day...", //3062
		"Though I foresee her wanting a smaller chair when that time comes."}, //3063
	{"We see the shrine. We don't really care.", //3064
		"Each knight here on their own is fully capable of subduing someone using the Chaos Emeralds."}, //3065


	//Emerald Knight
	"This is the artifact room.", //3066
	{"Most of our artifacts are stored deeper within the room.", //3067
		"On display are artifacts that can easily be dealt with and contained if something were to happen."}, //3068
	{"That big gemstone over there is in fact an exact copy of the Master Emerald from the Chaos dimension.", //3069
		"It was forged by the Magic King himself in an experiment to understand Chaos Energy.", //3070
		"It now sits here as a trophy of his many accomplishments."}, //3071
	{"We've had thieves in the past that took note of how easily accessible this room is.", //3072
		"To date, none have succeeded in pilfering anything in here."}, //3073
	{"This room was built by the King of Magic.", //3074
		"Many artifacts and trophies of his exploits are housed deep within this sanctum."}, //3075
	"Many artifacts stored here are capable of mass destruction.", //3076
	"Access to further levels in the artifact room is denied. For now.", //3077
}

local castleskindialog = {
	//Aether's Alt Dialog
	{
		//Normal Maid
		"Welcome home, Princess Aether! We missed you very much!",
		"Feel free to roam about as you please, your majesty!!",
		"Lovely weather we're having!",
		{"Do not worry if you accidentally fall.",
			"We will quickly escort you back to safety just as we did when you were little!"},
		{"Some of us are sisters. Some of us are using cloning magic.",
			"I remember when you used cloning magic to play tricks on the Knights.",
			"You were filled with so much energy back then!"},
		0,
		{"Oh dear, your father has been gone for quite some time.",
			"I do hope he doesn't work too hard again."},

		//Portrait Room Maid
		"Welcome to the Portrait Room, Princess Aether!",
		{"As you already know, these frames were crafted by your Father, the wonderful King of Magic!",
			"They have the ability to show the viewer different futures!"},
		0,
		0,
		{"I remember when you confronted your father about taking down this room.",
			"You were such a gutsy little child, my princess."},
		{"Remember when we used to play together in this room, my princess?",
			"Oh, how I miss those days."},
		0,
		
		//
		"knightstart", //Knight dialog start indicator
		//

		//Normal Knight
		"Hail Princess Aether.",
		0,
		0,
		0,
		0,
		0,
		"I beg your pardon, Princess Aether, but I must remain vigilant in my task.",
		
		//Throne Knight
		{"Greetings, Princess Aether.",
			"Your father, King Brightheart, is not here right now."},
		0,
		{"You have grown up to become a powerful arch mage, Princess Aether.",
			"As to be expected from the daughter of the King of Magic."},
		{"It is said that King Brightheart's power rivals several previous Meridians of Magic.",
			"As the current Meridian of Magic and daughter of the King of Magic...",
			"You have the potential to surpass your father, Princess Aether."},
		{"Your father is someone we all depend on.",
			"His power protects us from the miasma and provides bountiful harvests every single year."},
		{"I look forward to the day you ascend to the throne, Princess Aether.",
			"We believe in your courage and wisdom to lead us as did your father."},
		"With you around, Princess Aether, only a fool would dare attempt to steal from the Artifact Room.",
		
		//Emerald Knight
		"Welcome to the artifact room, Princess Aether.",
		0,
		0,
		0,
		0,
		0,
		{"Your father told us that nobody is allowed into the deeper chambers of the artifact room.",
			"My apologies, Princess Aether, he said that this limitation includes even you."},
			
		//
		[0] = "aether", //Skin definition
		//
	},
	
	//Inazuma Alt Dialog !MAKE SURE HE'S LAST!
	{
		//Normal Maid
		"Welcome to Castle Elvoway.",
		"Feel free to roam about.",
		"Oh, this day was going fine earlier... ",
		{"You flew when you fought our king, didn't you?",
			"If you fall, just fly back up here yourself."},
		"Bless her soul, but I have no idea what our lovely princess sees in someone like you.",
		"Will you be leaving soon?", 
		{"Look what you've done! Now I have to clean up that spot you've made!",
			"Why must you be so filthy?"},
		
		//Portrait Room Maid
		"Welcome to the Portrait Room.",
		0,
		"Oh dear, this room gets dirty quite often...",
		0,
		0,
		0,
		0,
		
		//
		"knightstart", //Knight dialog start indicator
		//
		
		//Normal Knight
		0,
		0,
		0,
		0,
		0,
		0,
		"You bring dishonor to your family name.",
		
		//Throne Knight
		0,
		{"Our all powerful and wise King of Magic is out on another expedition to claim more land from the miasma.",
			"A noble act from a noble man. And you?"},
		{"King Brightheart is also known as the King of Magic.",
			"By far, he is one of the strongest Battle Mages in Elvoway's recorded history.",
			"He is second only to his daughter, Princess Aether, who was born as the Meridian of Magic.",
			"And you? A dishonorable cheat that would use any means necessary to win a fight."},
		0,
		0,
		{"One day, Princess Aether will ascend to the throne and become our next ruler.",
			"She will make a fine Queen one day...",
			"Though I will never accept you as our king if it comes down to that."},
		{"We do not fear you.",
			"Don't think fighting us will go the same way as it did last time."},
			
		//Emerald Knight
		//N/A
		
		//
		[0] = "inazuma", //Skin definition
		//
	},
}

//Add Maid / Knight Dialog
local castleid = 3024
local knight = false
for i, dialogs in ipairs(castledialog)
	if dialogs == "knightstart" then knight = true continue end
	
	//Get name and portrait
	local name = knight and "Knight" or "Maid"
	local portrait = knight and "EV_KNGH" or "EV_MAID"
	local set = {}
	
	//Add normal dialog
	if type(dialogs) != "table" then dialogs = {$} end
	for j, line in ipairs(dialogs)
		table.insert(set, {portrait, name, line, {waittime=-1}})
	end
	AddMindscapeDialogSet(castleid, set)
	
	//Add skin dialog
	for q, skindialog in ipairs(castleskindialog)
		local skin = skindialog[0]
		local skindialogs = skindialog[i]
		
		if not skindialogs then continue end
		if type(skindialogs) != "table" then skindialogs = {$} end
		if skin == "inazuma" and not knight then portrait = "EV_MADI" end //Maids hate inazuma
		
		set = {}
		for j, line in ipairs(skindialogs)
			table.insert(set, {portrait, name, line, {waittime=-1}})
		end
		AddMindscapeDialogSet(nil, set, skin, castleid)
	end
	
	castleid = $ + #dialogs
end

//Start the text
//Or stop the text if we give 0 or nil as a value
//Also handles skin alterate text
rawset(_G, "startmindtext", function(id, color)
	if not consoleplayer return end
	
	//Stop the text!
	if not id
		textid = 0
		textletter = 0
		textwait = 0
		textcolor = 0
		textdebug = false
		return
		
	//Debug text
	elseif id < 0
		id = -id
		textdebug = true
	end
	
	pressbuttons = {{[BT_JUMP] = -1, [BT_SPIN] = 0}, {[BT_JUMP] = -1, [BT_SPIN] = 0}}
	
	//Whoops, not valid!
	if type(mindtext[id]) != "table"
		print(id.." is not a valid text id!")
		textid = 0
		textletter = 0
		textwait = 0
		textdebug = false
		return
	end
	if textdebug then print(id) end //Print the ID in debug mode
	
	//Set up our variables
	textid = id
	textskin = nil
	textcolor = color or 0
	
	//Find skin alternate text
	for _, player in ipairs({consoleplayer, secondarydisplayplayer})
		if not (player and player.valid and textid == id) then continue end
		
		//Your skin
		local skin = skins[player.skin].name
		textskin = $ or skin
		
		//Change the text
		if not textdebug
			//Change our text to the skin's text if we can
			if mindtext[id][skin] != nil
				textid = mindtext[id][skin]
				
			//Extra Sonic support!
			elseif mindtext[id]["sonic"]
			and not (skins[skin].flags & SF_MACHINE)
			and (skin == "pointy" or skin == "werehog"
			or string.find(skin, "sonic"))
				textid = mindtext[id]["sonic"]
			end
			
			//Other players' skins
			for other in players.iterate
				if other == player continue end
				if other.mo and other.mo.valid
					if mindtext[id][skin.."&"..other.mo.skin] != nil
						textid = mindtext[id][skin.."&"..other.mo.skin]
						break
					end
				end
			end
			
			//We got changed to no text, let's leave!
			if not textid
				startmindtext(0)
				return
				
			//Update the skin if the text got changed
			elseif textid != id
				textskin = skin
			end
		end
	end
	
	textletter = 0
	textwait = 0
	texttime = leveltime
	
	//Get the text and make sure it doesn't go offscreen to the right
	text = mindtextvalue(textid, "text")
	length = string.len(text)
	
	//Get the other variables together
	flip = mindtextvalue(textid, "flip")
	numlines = mindtextvalue(textid, "numlines")
	
	x = flip and 93 or 61
	icnorigin_x = 6 -- szv: default was 36, some text found a way to clip off-screen
	snap = flip and V_SNAPTORIGHT or V_SNAPTOLEFT
end)

local textseen = {}

//Allow the message to appear midlevel via a Linedef Executor
//The line's front texture x offset determines the ID
//The text will also only show for each player once per level load, unless the
//no climb flag is ticked. In that case, it will show everytime the linedef is executed.
addHook("LinedefExecute", function(line, mobj)
	local id = line.frontside.textureoffset/FRACUNIT
	if mobj.player and (mobj.player == consoleplayer
	or mobj.player == secondarydisplayplayer)
	and (not textseen[id] or line.flags & ML_NOCLIMB)
		textseen[id] = true
		startmindtext(id)
	end
end, "MINDTEXT")

//Reset the text upon changing the level
addHook("MapChange", function(newmap)
	if not textdebug
		startmindtext()
	end
	textseen = {}
	
	if DeltaChars and mapheaderinfo[newmap].mindscape
		DeltaChars.DialogueBox.seenDialogue["emeralds"] = true
	end
end)

//The actual text box
hud.add(function(v, player, cam)
	//No text! Go away
	if not textid or menuactive or paused
	or (splitscreen and player != secondarydisplayplayer) return end
	
	//Make sure the text doesn't go offscreen
	local linesplit = mindtextsplit(text, "\n")
	local width = v.stringWidth(linesplit[#linesplit], V_ALLOWLOWERCASE)
	if not numlines
		local i = 0
		while i < #linesplit and i < 3
			i = $ + 1
			
			//If a line is too wide, move a word from that line to the next
			local splitline = linesplit[i]
			width = v.stringWidth(linesplit[i], V_ALLOWLOWERCASE)
			if width > (320 - x)
				local nextline = linesplit[i + 1] or ""
				local splittext1 = ""
				local splittext2 = nextline
				local spacesplit = mindtextsplit(splitline, " ")
				
				//Go over each word, moving anything offscreen to the next line
				for j, word in ipairs(spacesplit)
					local prevtext = splittext1
					if j > 1
						splittext1 = $ + " " + word
					else
						splittext1 = $ + word
					end
					
					width = v.stringWidth(splittext1, V_ALLOWLOWERCASE)
					if width > (320 - x - icnorigin_x)
						splittext1 = prevtext
						splittext2 = ""
						for n = j, #spacesplit
							if n > j
								splittext2 = $ + " " + spacesplit[n]
							else
								splittext2 = $ + spacesplit[n]
							end
						end
						
						splittext2 = $ + nextline
						break
					end
				end
				
				//Set the text
				linesplit[i] = splittext1
				linesplit[i + 1] = splittext2
			end
		end
		
		//Stitch the text back together
		text = ""
		for j, splittext in ipairs(linesplit)
			text = $ + splittext
			if j != #linesplit
				text = $ + "\n"
			end
		end
		numlines = #linesplit
		length = string.len(text)
		
		//Apply the new values to the mindtext so we don't need to do this again
		mindtext[textid][3] = text
		mindtext[textid][7] = numlines
	end
	
	//Draw the black box at the bottom
	v.drawScaled(0,150*FRACUNIT,FRACUNIT*3/2,
	v.cachePatch("SPECTILE"),V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_50TRANS)
	
	//What's your name, bud?
	local name = mindtextvalue(textid, "name")
	if name == "skin" and textskin != nil
		name = skins[textskin].realname
	end

	//Draw the name
	if flip
		v.drawString(229, 150, name, V_SNAPTOBOTTOM|V_SNAPTORIGHT|V_YELLOWMAP, "right")
	else
		v.drawString(x, 150, name, V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_YELLOWMAP)
	end
	
	//Get the image
	local image = mindtextvalue(textid, "image")
	
	//Support Inazuma's Silverhorns
	if name == "Inazuma" and string.sub(image, 1, 3) == "IN_"
	and Silverhorn.Inazuma.CheckSilverhorns(player)
		image = "IS_" + string.sub(image, 4)
	end
	
	//Support all the different Sonics!
	if name == "Sonic" and string.sub(image, 1, 3) == "SH_"
		if textskin == "xsonic"
			image = "XS_" + string.sub(image, 4)
		elseif (skins[textskin].height > skins["sonic"].height
		or textskin == "adventuresonic" or textskin == "modernsonic")
		and textskin != "werehog"
			image = "MS_" + string.sub(image, 4)
		end
	end
	
	//Draw the CSS over the background icon if we're told to!
	if image == "skin" and textskin != nil
		if flip
			//Background Icon
			v.drawScaled(284*FRACUNIT,150*FRACUNIT,FRACUNIT/5,
			v.cachePatch("TALKBOX"), V_SNAPTOBOTTOM|V_SNAPTORIGHT|V_FLIP)
			
			//CSS
			v.drawScaled(284*FRACUNIT,150*FRACUNIT,FRACUNIT/2,
			v.getSprite2Patch(textskin, SPR2_XTRA, false, 1),
			V_SNAPTOBOTTOM|V_SNAPTORIGHT|V_FLIP)
		else
			//Background Icon
			v.drawScaled(icnorigin_x*FRACUNIT,150*FRACUNIT,FRACUNIT/5,
			v.cachePatch("TALKBOX"), V_SNAPTOBOTTOM|V_SNAPTOLEFT)
			
			//CSS
			v.drawScaled(icnorigin_x*FRACUNIT,150*FRACUNIT,FRACUNIT/2,
			v.getSprite2Patch(textskin, SPR2_XTRA, false, 1),
			V_SNAPTOBOTTOM|V_SNAPTOLEFT)
		end
		
	//No image? 
	elseif image == "none"
		//Got it.
		
	//We got a normal icon instead? Let's draw that.
	elseif flip
		v.drawScaled(284*FRACUNIT,150*FRACUNIT,FRACUNIT/5,
		v.cachePatch(image), V_SNAPTOBOTTOM|V_SNAPTORIGHT|V_FLIP,
		v.getColormap(TC_DEFAULT, textcolor))
	else
		v.drawScaled(icnorigin_x*FRACUNIT,150*FRACUNIT,FRACUNIT/5,
		v.cachePatch(image), V_SNAPTOBOTTOM|V_SNAPTOLEFT,
		v.getColormap(TC_DEFAULT, textcolor))
	end
	
	//Draw the lines of text
	if textletter > 0
		v.drawString(x, 162,
		string.sub(text,1,min(textletter, length)),
		V_SNAPTOBOTTOM|snap|V_ALLOWLOWERCASE)
	end
	
	//Press a button to advance the text!
	//Also draw the indicator to press jump
	if textwait < 0
		v.drawString(x*FRACUNIT + width*FRACUNIT,
		150*FRACUNIT + numlines*12*FRACUNIT + (leveltime % 8)*INT16_MAX,
		" ", V_SNAPTOBOTTOM|snap|V_YELLOWMAP, "fixed")
		
		if pressbuttons[1][BT_JUMP] == 1 or pressbuttons[2][BT_JUMP] == 1
			textwait = 0
			texttime = leveltime + 1
		else
			return
		end
	end
	
	//Advance text multiple times per frame if we're behind
	//If we're ahead, don't bother advancing
	local advtimes = leveltime - texttime
	texttime = leveltime
	if advtimes <= 0 then return end
	
	//Advance the text!
	local dosound
	local i = 0
	while i < advtimes
		//Hold it! No text!
		if not textid then break end
		
		//Increase pressbuttons in case of lag
		if i
			if pressbuttons[1][BT_JUMP] then pressbuttons[1][BT_JUMP] = $ + 1 end
			if pressbuttons[2][BT_JUMP] then pressbuttons[2][BT_JUMP] = $ + 1 end
			if pressbuttons[1][BT_SPIN] then pressbuttons[1][BT_SPIN] = $ + 1 end
			if pressbuttons[2][BT_SPIN] then pressbuttons[2][BT_SPIN] = $ + 1 end
		end
		i = $ + 1 //Advance frame
		
		//Wait before advancing text
		if textwait and not pressbuttons[1][BT_SPIN]
		and not pressbuttons[2][BT_SPIN]
			textwait = $ - 1
			continue
		end
			
		//Advance it!
		if (pressbuttons[1][BT_SPIN] or pressbuttons[2][BT_SPIN])
		and textletter < length
			textletter = length
			dosound = sfx_pmdtx
		else
			textletter = $ + 1
		end
		local curletter = string.sub(text, textletter, textletter)
		
		//Advance to the next text box once we're done
		if textletter > length
			local nexttext = mindtextvalue(textid, "nexttext")
			if textdebug and mindtext[textid + 1]
				nexttext = textid + 1
			end
			
			startmindtext(nexttext, textcolor)
			
		//This character doesn't have a length, seems we should skip it
		elseif not v.stringWidth(curletter, V_SNAPTOBOTTOM|snap|V_ALLOWLOWERCASE)
			advtimes = $ + 1
			textwait = 0
			
		//Find out how long we wait before the next letter
		else
			if curletter == "." or curletter == "!" or curletter == "?"
				textwait = 15
			elseif curletter == ","
				textwait = 10
			else
				textwait = 0
			end
			dosound = sfx_pmdtx //Printing sound
			
			//The final letter takes longer than the others
			local waittime = mindtextvalue(textid, "waittime")
			if textletter == length and waittime != nil
				if waittime >= 0 and not textdebug
					textwait = $ + mindtextvalue(textid, "waittime")
				else
					textwait = -1
				end
			end
		end
	end
	
	//Play a sound
	if dosound
		S_StartSound(nil, dosound)
	end
end)

//Prevent movement during manually advance-able text, and detect jump inputs
addHook("PlayerCmd", function(player, cmd)
	if not (player == consoleplayer or player == secondarydisplayplayer) then return end
	local i = (player == secondarydisplayplayer and 2 or 1)
	
	if textid and (mindtextvalue(textid, "waittime") < 0 or textdebug)
		if pressbuttons[i][BT_JUMP] == -1
			if not (cmd.buttons & (BT_JUMP | BT_SPIN))
				pressbuttons[i][BT_JUMP] = 0
			end
		elseif cmd.buttons & (BT_JUMP | BT_SPIN)
			pressbuttons[i][BT_JUMP] = $ + 1
		else
			pressbuttons[i][BT_JUMP] = 0
		end
		
		if cmd.buttons & BT_SPIN
			pressbuttons[i][BT_SPIN] = $ + 1
		else
			pressbuttons[i][BT_SPIN] = 0
		end
		
		cmd.buttons = 0
		cmd.forwardmove = 0
		cmd.sidemove = 0
	elseif pressbuttons[i][BT_JUMP] > 0
		if not (cmd.buttons & (BT_JUMP | BT_SPIN))
			pressbuttons[i][BT_JUMP] = 0
		else
			cmd.buttons = $ &~ (BT_JUMP | BT_SPIN)
		end
	end
end)