-- ============================================================
-- Mindscape
-- Startup
-- by Altalune
-- ============================================================

-- Maps used for warping / special in the map set used
Silverhorn.SpMaps = {
	["Credits"] = "RZ"
}

-- Disables the stage title when the act is locked. Meant to be run once
function Silverhorn.FailedEntry_DisableMindscapeTitleOnMapLoad()
	if mapheaderinfo[gamemap].mindscape and (MindscapeSessionFlags & MS_FAILEDENTRY) then
		hud.disable("stagetitle")
	end
end

-- Re-enables the stage title when the map is forcibly left (single-use)
function Silverhorn.FailedEntry_EnableMindscapeTitleOnMapChange()
	if mapheaderinfo[gamemap].mindscape and (MindscapeSessionFlags & MS_INMINDSCAPE) then
		hud.enable("stagetitle")
	end
end

-- Force enables analog mode for a player and logs their old control mode
function Silverhorn.ForceEnableAnalogMode(player)
	if player
	and not (player.pflags & PF_ANALOGMODE)
	and not (player.pflags & PF_DIRECTIONCHAR)
		player.silverhornoldmode = player.pflags & (PF_ANALOGMODE|PF_DIRECTIONCHAR)
		player.pflags = $ | PF_ANALOGMODE | PF_DIRECTIONCHAR
	end
end

-- Restores a player's old control mode set above
function Silverhorn.UndoForceEnableAnalogMode(player)
	if player and player.silverhornoldmode != nil
		player.pflags = ($ &~ (PF_ANALOGMODE|PF_DIRECTIONCHAR)) | player.silverhornoldmode
		player.silverhornoldmode = nil
	end
end

-- Spawns players from the portals
function Silverhorn.SpawnMindscapePlayer(player)
	if not player.mo then return end

	local portal = P_SpawnMobjFromMobj(player.mo,0,0,0,MT_MINDSCAPE_PORTAL)
	-- portal.isintroportal = true -- TODO: if needed, let the portal object handle this later if it doesn't work out
	portal.scale = player.mo.scale
	portal.scalespeed = $/8
	portal.reactiontime = player.mo.scale
	portal.fuse = TICRATE*4
	portal.target = player.mo
	S_StartSound(portal, sfx_s3kbds, player)
	
	-- Colorize portal when we've beaten Mindscape already
	if not mapheaderinfo[gamemap].mindscape
	and (MindscapeSessionFlags & MS_BEATMINDSCAPE) then
		portal.color = SKINCOLOR_SUPERGOLD5
		portal.colorized = true
	end
	
	P_SetObjectMomZ(player.mo,8*FRACUNIT,false)
	P_InstaThrust(player.mo,player.mo.angle,8*player.mo.scale)
	player.mo.state = S_PLAY_SPRING
	player.mo.flags2 = $|MF2_DONTDRAW
end

-- ============================================
-- sub-sequence
-- ============================================

EventSH.new("intro_sq_inazuma_fx", {
function(c,e)
	if not c.mobj.valid then return end

	if leveltime % c.rate == 0 then
		if c.fxtype == "spark" then
			Silverhorn.Inazuma.SpawnTrail(c.mobj)
		elseif c.fxtype == "skid" then
			local fx = P_SpawnMobj(c.mobj.x, c.mobj.y, c.mobj.z, MT_SPINDUST)
			fx.momz = 1*FU
			P_InstaThrust(fx, c.mobj.angle+ANGLE_180, 2*FU)
		end
	end
	EventSH.waitUntil(e, not e)
end})

local function MSIntro_InazumaFX(mo, spawnrate, fxtype, usefx)
	if usefx then
		EventSH.start("intro_sq_inazuma_fx", {mobj=mo, fxtype=fxtype, rate=spawnrate})
	else
		EventSH.destroy("intro_sq_inazuma_fx")
	end
end

EventSH.new("intro_sq_faccshootlaser", {
function(c,e)
	c.time = 0
end,
function(c,e)
	if not c.mobj.valid then return end

	if (c.time % c.spawnrate == 0) then
		EvActionSH.Sq_CreateMobjActor(c, "facclaser", MT_THOK, {x=c.mobj.x,y=c.mobj.y,z=FU+2*FU}, function(mo)
			mo.intro = 1
			mo.state = S_FACCIOLO_LASER
			mo.tics = 2*TICRATE
			mo.flags = $1|MF_NOCLIP|MF_NOCLIPHEIGHT
			EvActionSH.setmobjscale(mo, FU)
		end)
		c.m_facclaser.spriteyscale = abs(c.mobj.floorz - c.mobj.ceilingz)/64
		-- Sound
		S_StartSoundAtVolume(nil, sfx_faclfr, 50)
		P_StartQuake(5*FU, 5)
	end

	c.time = $1+1
	EventSH.waitUntil(e, not e)
end})

local function MSIntro_FaccFireLasers(mo, amount, ended)
	if not ended then
		EventSH.start("intro_sq_faccshootlaser", {mobj=mo, spawnrate=amount})
	else
		EventSH.destroy("intro_sq_faccshootlaser")
	end
end

EventSH.new("intro_sq_faccshadow", {
function(c,e)
	if not c.mobj.valid then return end

	EvActionSH.Sq_CreateMobjActor(c, "faccshadow", MT_THOK, {x=c.mobj.x,y=c.mobj.y,z=c.mobj.z+50*FU}, function(mo)
		mo.intro = 1
		mo.color = SKINCOLOR_BLACK
		mo.tics = FU
		mo.spriteyoffset = -32*FU
		mo.frame = 0
		mo.blendmode = AST_SUBTRACT
		EvActionSH.setmobjscale(mo, 0)
	end)
	-- EvActionSH.setmobjscale(c.m_faccshadow, FU*3, FU/2)
	EvActionSH.movemobjto(c.m_faccshadow, c.mobj)
end,
function(c,e)
	-- pointless validity check
	if c.mobj.valid then
		-- alternate: Follow target
		c.m_faccshadow.scale = 4*FU + 2*cos(FixedAngle(16*FU)*leveltime)
		EvActionSH.movemobjto(c.m_faccshadow, {x=c.mobj.x,y=c.mobj.y,z=c.mobj.z+50*FU})
	end
	EventSH.waitUntil(e, not (c.mobj and c.mobj.valid))
end,
function(c,e)
	P_RemoveMobj(c.m_faccshadow)
end})

local function MSIntro_FaccShadow(mo, ended)
	if not ended then
		EventSH.start("intro_sq_faccshadow", {mobj=mo})
	else
		EventSH.destroy("intro_sq_faccshadow")
	end
end

EventSH.new("intro_sq_faccmovement", {
function(c,e)
	c.time = 0
end,
function(c,e)
	if not c.mobj.valid then return end

	EvActionSH.movemobjto(c.mobj, {
		x=c.pos.x+2*cos(FixedAngle(2*FU)*leveltime),
		y=c.pos.y+2*sin(FixedAngle(4*FU)*leveltime),
		z=c.pos.z+6*cos(FixedAngle(4*FU)*leveltime)})

	if (c.time % 6 == 0 and not (c.mobj.flags2 & MF2_DONTDRAW)) then
		local g = P_SpawnGhostMobj(c.mobj)
		g.momz = -FU/2
	end
	c.time = $1+1

	EventSH.waitUntil(e, not e)
end})

local function MSIntro_FaccMovement(mo, position, ended)
	if not ended then
		EventSH.start("intro_sq_faccmovement", {mobj=mo, pos=position})
	else
		EventSH.destroy("intro_sq_faccmovement")
	end
end

local function MSIntro_Cleanup()
	EventSH.read("sh_msintro", function(re)
		-- Cleanup objects if they exist
		for k, v in pairs(re) do
			if type(v) == "userdata" and v.valid then
				-- print(k) -- debug info
				v.fuse = 2
				-- P_RemoveMobj(v)
			end
		end
	end)
end


-- ============================================
-- sequence
-- ============================================
local Fan_Sector_Position = {x=-11296*FU, y=-10256*FU, z=3553*FU}

EventSH.new("sh_msintro_skip", {
function(c,e)
	-- Set timer max
	c.activetime = 0
	c.skipprogress = 0
	c.maxskipprogress = 2*TICRATE-15

	-- Display settings
	c.skipprompttext = "[SPIN]: Skip"
	c.skipprogresswidth = 85 -- progress bar width
	c.disp_y = 4 -- display start y
	c.maxdisp_y = 4 -- max display shown
	c.mindisp_y = -16 -- min display hidden

	-- Hidden settings
	c.skipshowntime = 6*TICRATE

	-- Enable graphic render
	EvActionSH.togglerender(c)
	EvActionSH.drawcineborders(c, 1, 16) -- temporary
	EvActionSH.rendertext(c, 4, "SkipPrompt", 300, c.disp_y, "", V_30TRANS|V_SNAPTOTOP|V_SNAPTORIGHT, "right")
	EvActionSH.renderfill(c, 3, "ProgressBar", 300, c.disp_y+12-4, 32, 2, 122|V_SNAPTOTOP|V_SNAPTORIGHT)
end,
function(c,e)
	c.activetime = $1 + 1

	EvActionSH.doforall(function(play)
		-- (re-thinking this- only the host should skip)
		if not (play == server or IsPlayerAdmin(play)) then return end

		-- Hold spin to skip
		if play.cmd.buttons & BT_SPIN then
			if c.activetime % 1 == 0 then
				c.skipprogress = $1+1
			end
			-- Shows the skip prompt
			if not (c.skipshowntime > TICRATE) then c.disp_y = c.maxdisp_y end
			c.skipshowntime = min($1+15, 3*TICRATE)
		else
			c.skipprogress = max(0, $1-1)
		end
	end)

	-- Skip hidden upwards display
	c.disp_y = (c.skipshowntime > TICRATE) and $1 or max(c.mindisp_y, $1-1)
	c.skipshowntime = max(0, $1-1)
	EvActionSH.editrender(c, "SkipPrompt").y = c.disp_y
	EvActionSH.editrender(c, "ProgressBar").y = c.disp_y+12-4
	EvActionSH.editrender(c, "SkipPrompt").text = c.skipprompttext

	-- Main progress display
	EvActionSH.editrender(c, "ProgressBar").x = 300-(c.skipprogress*c.skipprogresswidth/c.maxskipprogress)
	EvActionSH.editrender(c, "ProgressBar").width = 0+(c.skipprogress*c.skipprogresswidth/c.maxskipprogress)

	EventSH.waitUntil(e, c.skipprogress >= c.maxskipprogress)
end,
function(c,e)
	-- Timer reached, jump ahead in the intro (and should remove everything)
	MSIntro_Cleanup()

	EventSH.destroy("sh_msintro")
	EventSH.start("sh_msintro_wrapup", {fromcutscene=true})
end})



EventSH.new("sh_msintro", {
function(c,e)
	-- If you watched the intro already this session, don't play it again unless you get booted from the map
	if (MindscapeSessionFlags & MS_SEENINTRO) then
		EventSH.destroyself(e)
		EventSH.start("sh_msintro_wrapup")
	end

	-- Allow skipping
	if not (MindscapeSessionFlags & MS_FAILEDENTRY) then
		EventSH.start("sh_msintro_skip")
	end

	EvActionSH.SetInternalHudStatus({"time", "score", "rings", "lives"}, false)
end,
function(c,e)
	-- Failure or no?
	if (MindscapeSessionFlags & MS_FAILEDENTRY) then
		EvActionSH.startsound(nil, sfx_lose)
		EvActionSH.fadescreen("full", 31, 4)
		EventSH.wait(e, FRACUNIT, true)
	end
end,
-- Setup
function(c,e)
	EvActionSH.fadescreen("in", 31, 4)
	-- Make camera
	EvActionSH.Sq_MakeCamera(c, -20600*FU, -18600*FU, -504*FU)

	-- Map Anchor
	EvActionSH.coordinateanchor(c, "mindscape1", -27008*FU, -8768*FU, 2456*FU)
	c.entryspot = EvActionSH.coordinategetoffset(c, "mindscape1", -700*FU, 0, 256*FU)

	EvActionSH.coordinateanchor(c, "mindscape2", -23488*FU, -12800*FU, 1944*FU)
	EvActionSH.coordinateanchor(c, "mindscape3", -22016*FU, -18432*FU, -512*FU)

	EvActionSH.Sq_CreateActor(c, "inazuma", "inazuma", SPR_PLAY, 4*FU, 6*FU, SKINCOLOR_CERULEAN, function(mo)	mo.intro = 1 end)
	EvActionSH.Sq_CreateActor(c, "facciolo", nil, SPR_FACC, 4*FU, 6*FU, SKINCOLOR_SLATE, function(mo) mo.intro = 1 end)

	-- Set original spawn location
	EvActionSH.doforall(function(player)
		if (player and player.mo.valid) then
			player.mo.ms_originalspawn = {x = player.mo.x, y = player.mo.y,
			z = player.mo.z, angle = player.mo.angle}
		end
	end)
	-- Re-locate everyone to a fan sector
	-- EvActionSH.relocateplayers(Fan_Sector_Position)
	EvActionSH.relocateplayers_start(FRACUNIT, Fan_Sector_Position)

	if not (MindscapeSessionFlags & MS_FAILEDENTRY) then
		S_ChangeMusic("voidl", true, nil, 0, 0, 0, 4*MUSICRATE)
	end
end,
-- P1. PANEL 1
function(c,e)

	-- Starting Camera Position
	EvActionSH.setactivecamera(true, c.camera)
	c.camerastart = EvActionSH.coordinategetoffset(c, "mindscape1", (128)*FU, (128+64)*FU, -25*FU)
	-- c.camerastart_old = EvActionSH.coordinategetoffset(c, "mindscape1", 640*FU, 240*FU, -26*FU)

	EvActionSH.setcamera(true, c.camera, c.camerastart, 225*FU, 18*FU)
	-- EvActionSH.setcamera(true, c.camera, {x=-26369*FU,y=-8531*FU,z=2430*FU}, 225*FU, 18*FU)

	EvActionSH.Sq_CreateMobjActor(c, "portal", MT_FACCIOLO2_SPACEPORTAL, c.entryspot, function(mo)
		mo.intro = 1
		-- mo.renderflags = $1|RF_PAPERSPRITE
		mo.facc = 1
		mo.spriteyoffset = -32*FU
	end)
	EvActionSH.setmobjscale(c.m_portal, 0)
	EvActionSH.setmobjscale(c.m_portal, 2*FU, FU/32)
	EvActionSH.coordinateanchor(c, "entryportal", c.m_portal.x, c.m_portal.y, c.m_portal.z)
	EvActionSH.startsound(nil, sfx_s3kbds)
	-- P_MoveOrigin(c.ac_portal, 39167*FU, 57005*FU, 2430*FU)

	P_InstaThrust(c.camera, c.camera.angle, -FU/8)
end,
function(c,e)
	EventSH.wait(e, 2*TICRATE)
end,
function(c,e)
	local behind = EvActionSH.coordinategetoffset(c, "entryportal", -96*FU, 0, 0)
	P_SetOrigin(c.ac_inazuma, behind.x, behind.y, behind.z)
	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_FALL, 0, nil, {loopalways=true})

	c["landingspot"] = EvActionSH.coordinategetoffset(c, "mindscape1", -256*FU, 0, 0*FU)
	MobjMoverSH.moveto("inazuma_landing", c.ac_inazuma, c["landingspot"], {speed=16*FU, varc=96*FU})
	EvActionSH.startsound(nil, sfx_s1c3)
end,
function(c,e)
	MobjMoverSH.waitmovedone(e, "inazuma_landing")
end,
function(c,e)
	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_STND, nil, "E", {delay=4, loopalways=true})
	EvActionSH.setmobjscale(c.m_portal, 0, FU/50)
	EvActionSH.startsound(nil, sfx_portcl, 60)
end,
function(c,e)
	EventSH.wait(e, 15)
end,
function(c,e)
	EvActionSH.setmobjlookat(c.ac_inazuma, c.m_portal)
end,
function(c,e)
	EventSH.wait(e, 90)
end,
-- P1. PANEL 2
function(c,e)
	-- c.camerapanel2 = EvActionSH.coordinategetoffset(c, "mindscape1", -210*FU, -8*FU, 30*FU)
	-- EvActionSH.setcamera(true, c.camera, c.camerapanel2, 180*FU, 0)
	c.camerapanel2 = EvActionSH.coordinategetoffset(c, "mindscape1", -300*FU, 8*FU, 30*FU)
	EvActionSH.setcamera(true, c.camera, c.camerapanel2, 0, 0)
	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_FLY_, 0, "A", {delay=44})
	P_InstaThrust(c.camera, c.camera.angle, 0)
	startmindtext(1510)
end,
function(c,e)
	EventSH.wait(e, 6*TICRATE)
end,
-- P1. PANEL 3
function(c,e)
	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_FLY_, 0, nil, {delay=3})
	EventSH.wait(e, 1*TICRATE, true)
end,
function(c,e)
	c.camerapanel3 = EvActionSH.coordinategetoffset(c, "mindscape1", 1024*FU, 1024*FU, 64*FU)
	EvActionSH.setcamera(true, c.camera, c.camerapanel3, 90*FU, 15*FU)
	P_InstaThrust(c.camera, c.camera.angle-ANGLE_90, FU)
end,
function(c,e)
	EventSH.wait(e, 25)
end,
function(c,e)
	startmindtext(1511)
	EventSH.wait(e, 3*TICRATE, true)
end,
-- P1. PANEL 4
function(c,e)
	c.camerapanel4 = EvActionSH.coordinategetoffset(c, "mindscape1", -100*FU, -50*FU, 32*FU)
	EvActionSH.setcamera(true, c.camera, c.camerapanel4, 180*FU, 0)
	P_InstaThrust(c.camera, c.camera.angle, 0)
	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_MLEE, 0, "A")
	EventSH.wait(e, 25, true)
end,
function(c,e)
	startmindtext(1512)
	EventSH.wait(e, 2*TICRATE, true)
end,
function(c,e)
	-- TODO: emoticon
	EvActionSH.Sq_CreateMobjActor(c, "shockemote", MT_PUSH, {x=c.ac_inazuma.x, y=c.ac_inazuma.y-64*FU, z=c.ac_inazuma.z+64*FU}, function(mo)
		mo.intro = 1
		mo.fuse = 10
		mo.sprite = SPR_WHAT
	end)
	FrameAnimSH.add("ina", c.m_shockemote, SPR_WHAT, "A", "D", {delay=2})
	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_MLEE, 0, nil, {delay=4})
	EvActionSH.startsound(nil, sfx_alart)
	EventSH.wait(e, 50, true)
end,
-- P1. PANEL 5
function(c,e)
	-- S_ChangeMusic("wfly", true)
	EvActionSH.Sq_CreateValue(c, "faccappear", EvActionSH.coordinategetoffset(c, "mindscape1", 256*FU, 0*FU, 256*FU))
	-- EvActionSH.Sq_CreateActor(c, "facciolo", nil, SPR_FACC, 4*FU, 6*FU, SKINCOLOR_SLATE)
	EvActionSH.movemobjto(c.ac_facciolo, c.faccappear)
	EvActionSH.setmobjlookat(c.ac_facciolo, c.ac_inazuma)
	MSIntro_FaccMovement(c.ac_facciolo, c.faccappear)

	-- c.faccappear = EvActionSH.coordinategetoffset(c, "mindscape1", 256*FU, 0*FU, 256*FU)
	c.camerapanel5 = EvActionSH.coordinategetoffset(c, "mindscape1", 200*FU, 0*FU, 315*FU)
	c.camerapanel5_b = EvActionSH.coordinategetoffset(c, "mindscape1", -512*FU, 0*FU, 15*FU)
	EvActionSH.setcamera(true, c.camera, c.camerapanel5, 0, 0)
	P_InstaThrust(c.camera, c.camera.angle-ANGLE_180, FU/32)

	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_STND, nil, "E", {delay=5, loopalways=true})
	EvActionSH.setmobjlookat(c.ac_inazuma, c.ac_facciolo)
	startmindtext(1513)
	EventSH.wait(e, 35, true)
end,
function(c,e)
	MobjMoverSH.moveto("camera", c.camera, c.camerapanel5_b, {speed=30*FU, noangle=true, ease="inoutcubic"})
	-- MobjMoverSH.moveto("camera", c.camera, c.camerapanel5_b, {speed=25*FU, noangle=true, ease="inoutcubic"})
	EvActionSH.startsound(nil, sfx_facslm)
end,
function(c,e)
	MobjMoverSH.waitmovedone(e, "camera")
end,
function(c,e)
	P_InstaThrust(c.camera, c.camera.angle-ANGLE_180, FU/6)
	EventSH.wait(e, 2*TICRATE, true)
end,
-- P1. PANEL 6
function(c,e)
	EvActionSH.Sq_CreateValue(c, "camerapanel6", EvActionSH.coordinategetoffset(c, "mindscape1", 360*FU, 64*FU, 320*FU))
	EvActionSH.setcamera(true, c.camera, c.camerapanel6, 0, -10*FU)
	EvActionSH.setmobjlookat(c.camera, c.ac_inazuma)
	EvActionSH.setmobjeangle(c.camera, FixedAngle(8*FU), true)
	P_InstaThrust(c.camera, c.camera.angle, 0)
	-- EvActionSH.cameratrack(true, c.camera, c.ac_inazuma)
	startmindtext(1514)
	EventSH.wait(e, 3*TICRATE, true)
end,
function(c,e)
	startmindtext(1515)
	EventSH.wait(e, 3*TICRATE, true)
end,
function(c,e)
	startmindtext(1516)
	EventSH.wait(e, 4*TICRATE, true)
end,
-- P2. PANEL 1
function(c,e)
	EvActionSH.Sq_CreateValue(c, "camera2panel1", EvActionSH.coordinategetoffset(c, "mindscape1", 222*FU, -32*FU, 300*FU))
	EvActionSH.setcamera(true, c.camera, c.camera2panel1, 0, 0)
	EvActionSH.setmobjeangle(c.camera, FixedAngle(16*FU), true)
	EvActionSH.setmobjlookat(c.ac_facciolo, c.camera)
	EvActionSH.setmobjframe(c.ac_facciolo, R_Char2Frame("M"))
	startmindtext(1520)
	EventSH.wait(e, 5*TICRATE, true)
end,
-- P2. PANEL 2
function(c,e)
	EvActionSH.Sq_CreateValue(c, "camera2panel2", EvActionSH.coordinategetoffset(c, "mindscape1", -120*FU, 16*FU, 70*FU))
	EvActionSH.setcamera(true, c.camera, c.camera2panel2, 200*FU, -24*FU)
	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_TIRE, nil, "B", {delay=18, loopalways=true})
	-- EvActionSH.setmobjlookat(c.camera, c.ac_inazuma)
	P_InstaThrust(c.camera, c.camera.angle, FU/32)
	startmindtext(1521)
	S_FadeOutStopMusic(3*MUSICRATE)
	EventSH.wait(e, 4*TICRATE, true)
end,
-- Cut Point
-- P2. PANEL 3
function(c,e)
	S_ChangeMusic("wfly", true)

	EvActionSH.setmobjframe(c.ac_facciolo, R_Char2Frame("C"))
	EvActionSH.setmobjlookat(c.ac_facciolo, c.ac_inazuma)

	EvActionSH.Sq_CreateValue(c, "camera2panel3", EvActionSH.coordinategetoffset(c, "mindscape1", 64*FU, 0*FU, 280*FU))
	EvActionSH.Sq_CreateValue(c, "camera2panel4", EvActionSH.coordinategetoffset(c, "mindscape1", 300*FU, 64*FU, 350*FU))
	EvActionSH.setcamera(true, c.camera, c.camera2panel3, 0, 10*FU)
	P_InstaThrust(c.camera, c.camera.angle, 0)
	P_InstaThrust(c.camera, c.camera.angle-ANGLE_180, FU/32)
	startmindtext(1522)

	EventSH.wait(e, 1*TICRATE, true)
end,
function(c,e)
	EvActionSH.Sq_CreateMobjActor(c, "energyball", MT_FACCIOLO_DEATHBALL, {x=c.ac_facciolo.x,y=c.ac_facciolo.y,z=c.ac_facciolo.z+170*FU}, function(mo)
		mo.intro = 1
		mo.blendmode = AST_ADD
		mo.frame = $1|TR_TRANS20
		mo.colorized = true
		-- mo.color = SKINCOLOR_MAGENTA
		EvActionSH.setmobjscale(mo, 0)
		EvActionSH.setmobjscale(mo, 1*FU, FU/32)
	end)
	EvActionSH.startsound(c.ac_facciolo, sfx_dbchrg, 90)

	EvActionSH.Sq_CreateMobjActor(c, "tracker", MT_PUSH, {x=c.ac_facciolo.x,y=c.ac_facciolo.y,z=c.camera.z}, function(mo)
		mo.intro = 1
	end)
	EventSH.wait(e, 2*TICRATE, true)
end,
function(c,e)
	P_InstaThrust(c.camera, c.camera.angle, 0)

	MobjMoverSH.moveto("camera", c.camera, c.camera2panel4, {speed=8*FU, noangle=true, harc=-32*FU, varc=-32*FU, ease="inoutsine"})
	MobjMoverSH.moveto("tracker", c.m_tracker, {x=c.ac_inazuma.x+256*FU, y=c.ac_inazuma.y-96*FU, z=c.ac_inazuma.z+66*FU}, {speed=10*FU, noangle=true, ease="inoutsine"})
	-- EvActionSH.setcamera(true, c.camera, c.camera2panel4, -90*FU, 10*FU)
	-- EvActionSH.movemobjto(c.m_tracker, {x=c.ac_inazuma.x+256*FU, y=c.ac_inazuma.y-96*FU, z=c.ac_inazuma.z+66*FU})
end,
function(c,e)
	if EvActionSH.getmobjdistance(c.camera, c.camera2panel4) > 18*FU then
		EventSH.pause(e)
	else
		EventSH.resume(e)
	end
	EvActionSH.cameratrack(true, c.camera, c.m_tracker)
	-- EventSH.wait(e,3*TICRATE)
end,
-- P2. PANEL 4
function(c,e)
	startmindtext(1523)
	EventSH.wait(e, 2*TICRATE, true)
end,
function(c,e)
	S_StopSoundByID(c.ac_facciolo, sfx_dbchrg)
	S_StopSound(c.ac_facciolo)
	EvActionSH.startsound(c.ac_facciolo, sfx_dbfire)
	FrameAnimSH.add("fac", c.ac_facciolo, SPR_FACC, "C", "E", {delay=4})
	MobjMoverSH.moveto("eball", c.m_energyball, c.ac_inazuma, {speed=45*FU, varc=8*FU, ease="insine"})
	c.balldistance_ref = EvActionSH.getmobjdistance(c.m_energyball, c.ac_inazuma)
end,
function(c,e)
	EventSH.waitUntil(e, EvActionSH.getmobjdistance(c.m_energyball, c.ac_inazuma) <= c.balldistance_ref/2)
end,
-- P2. PANEL 5
function(c,e)
	P_RemoveMobj(c.m_energyball)
	EvActionSH.Sq_CreateValue(c, "camera2panel5", EvActionSH.coordinategetoffset(c, "mindscape1", 0*FU, 210*FU, 128*FU))
	EvActionSH.setcamera(true, c.camera, c.camera2panel5, 210*FU, -10*FU)

	EvActionSH.Sq_CreateMobjActor(c, "energyexplode", MT_PUSH, {x=c.ac_inazuma.x,y=c.ac_inazuma.y,z=c.ac_inazuma.z}, function(mo)
		mo.intro = 1
		mo.sprite = SPR_FFBL
		mo.spriteyoffset = -32*FU
		EvActionSH.setmobjscale(mo, 5*FU)
		FrameAnimSH.add("boom", mo, SPR_FFBL, "F", "K", {delay=18})
	end)
	S_StopSound(c.ac_facciolo)
	EvActionSH.startsound(c.ac_inazuma, sfx_dbboms)
	EvActionSH.startsound(c.ac_inazuma, sfx_zcrash)

	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_FALL, "A", "B", {delay=15})
	EvActionSH.setmobjeangle(c.ac_inazuma, FixedAngle(-25*FU))
	EvActionSH.setspriteroll(c.ac_inazuma, FixedAngle(-25*FU))
	EvActionSH.movemobjrelative(c.ac_inazuma, 0, 128*FU, 96*FU)
	c.ac_inazuma.momy = FU/12
	c.ac_inazuma.momz = FU/12
end,
function(c,e)
	EventSH.wait(e, 2*TICRATE)
end,
-- P2. PANEL 5 B
function(c,e)
	-- S_StopSound(c.ac_inazuma)
	MSIntro_FaccMovement(nil, nil, true)
	P_RemoveMobj(c.m_energyexplode)
	EvActionSH.Sq_CreateValue(c, "camera2panel5b", EvActionSH.coordinategetoffset(c, "mindscape2", 300*FU, 700*FU, -128*FU))
	EvActionSH.Sq_CreateValue(c, "camera2panel5c", EvActionSH.coordinategetoffset(c, "mindscape2", -128*FU, 720*FU, -64*FU))
	EvActionSH.setcamera(true, c.camera, c.camera2panel5b, 0, 30*FU)

	EvActionSH.Sq_CreateValue(c, "ina_fallback", EvActionSH.coordinategetoffset(c, "mindscape2", 1024*FU, 1024*FU, 720*FU))
	EvActionSH.Sq_CreateValue(c, "ina_fallback_b", EvActionSH.coordinategetoffset(c, "mindscape2", 1280*FU, 100*FU, 0))
	EvActionSH.Sq_CreateValue(c, "ina_fallback_c", EvActionSH.coordinategetoffset(c, "mindscape2", 100*FU, 840*FU, 0*FU))
	EvActionSH.Sq_CreateValue(c, "ina_fallback_d", EvActionSH.coordinategetoffset(c, "mindscape2", 0*FU, -256*FU, 0))
	EvActionSH.setspriteroll(c.ac_inazuma, 0)
	EvActionSH.stopmobjmomentum(c.ac_inazuma)
	EvActionSH.movemobjto(c.ac_inazuma, c.ina_fallback)
	MobjMoverSH.moveto("inazuma_fallback", c.ac_inazuma, c.ina_fallback_b, {speed=20*FU, varc=96*FU, ease="insine"})
	EvActionSH.startsound(nil, sfx_ebar, 90)
	P_InstaThrust(c.camera, c.camera.angle-ANGLE_180, FU/4)
	MSIntro_FaccFireLasers(c.ac_inazuma, 16)
end,
function(c,e)
	MobjMoverSH.waitmovedone(e, "inazuma_fallback")
end,
-- PANEL 5 C
function(c,e)
	MSIntro_FaccFireLasers(nil, nil, true)
	MobjMoverSH.moveto("inazuma_fallback", c.ac_inazuma, c.ina_fallback_c, {speed=40*FU, varc=50*FU})
	MobjMoverSH.moveto("camera", c.camera, c.camera2panel5c, {speed=25*FU, noangle=true, ease="insine"})
	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_TWIN, "A", "D", {delay=3, loop=5*TICRATE})
	MSIntro_InazumaFX(c.ac_inazuma, 1, "spark", true)
	EvActionSH.startsound(nil, sfx_boltst)
	MSIntro_FaccFireLasers(c.ac_inazuma, 5)
end,
function(c,e)
	MobjMoverSH.waitmovedone(e, "inazuma_fallback")
end,
function(c,e)
	MSIntro_InazumaFX(nil, nil, nil, false)
	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_GRND, "A", "B", {delay=3})
	EvActionSH.setmobjeangle(c.ac_inazuma, FixedAngle(-90*FU))
	c.ac_inazuma.momy = 2*FU
	EvActionSH.startsound(nil, sfx_ljmp)
	EventSH.wait(e, 10, true)
end,
function(c,e)
	c.ac_inazuma.momy = -50*FU
	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_TWIN, "A", "D", {delay=3, loop=5*TICRATE})
	MSIntro_InazumaFX(c.ac_inazuma, 1, "spark", true)
	EvActionSH.startsound(nil, sfx_boltst)
end,
function(c,e)
	MSIntro_FaccFireLasers(nil, nil, true)
	EventSH.wait(e, 1*TICRATE, true)
end,
-- P2. PANEL 5 D
function(c,e)
	P_InstaThrust(c.camera, c.camera.angle, 0)
	EvActionSH.Sq_CreateValue(c, "camera2panel5d", EvActionSH.coordinategetoffset(c, "mindscape3", 512*FU, -32*FU, 96*FU))
	EvActionSH.setcamera(true, c.camera, c.camera2panel5d, 0, 0)
	P_InstaThrust(c.camera, c.camera.angle-ANGLE_180, FU/4)

	EvActionSH.Sq_CreateValue(c, "startingarea_track", EvActionSH.coordinategetoffset(c, "mindscape3", 1408*FU, 0*FU, 0*FU))
	EvActionSH.cameratrack(true, c.camera, c.startingarea_track)

	EvActionSH.stopmobjmomentum(c.ac_inazuma)
	EvActionSH.movemobjto(c.ac_inazuma, EvActionSH.coordinategetoffset(c, "mindscape3", 0*FU, 0*FU, 0*FU))
	MobjMoverSH.moveto("inazuma_tostart", c.ac_inazuma, c.startingarea_track, {speed=30*FU, varc=50*FU})

	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_TWIN, "A", "D", {delay=3, loop=5*TICRATE})
	EvActionSH.startsound(nil, sfx_boltbl)
end,
function(c,e)
	EvActionSH.cameratrack(true, c.camera, c.startingarea_track)
	MobjMoverSH.waitmovedone(e, "inazuma_tostart")
end,
function(c,e)
	EvActionSH.Sq_CreateValue(c, "camera2panel5d", EvActionSH.coordinategetoffset(c, "mindscape3", 1450*FU, 256*FU, 96*FU))
	EvActionSH.setcamera(true, c.camera, c.camera2panel5d, -100*FU, -16*FU)

	MSIntro_InazumaFX(nil, nil, nil, false)
	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_SKID, "A", "A")
	c.ac_inazuma.momx = 8*FU
	MSIntro_InazumaFX(c.ac_inazuma, 3, "skid", true)
	EvActionSH.startsound(nil, sfx_skid)
	EventSH.wait(e, 1*TICRATE, true)
end,
function(c,e)
	EvActionSH.setmobjeangle(c.ac_inazuma, FixedAngle(180*FU))
	MSIntro_InazumaFX(nil, nil, nil, false)
	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_STND, nil, "E", {delay=4, loopalways=true})
	EventSH.wait(e, 1*TICRATE, true)
end,
-- P2. PANEL 6
function(c,e)
	EvActionSH.Sq_CreateValue(c, "faccstill", EvActionSH.coordinategetoffset(c, "mindscape3", 1300*FU, 0*FU, 256*FU))
	EvActionSH.movemobjto(c.ac_facciolo, c.faccstill)
	EvActionSH.setmobjlookat(c.ac_facciolo, c.ac_inazuma)
	EvActionSH.setmobjframe(c.ac_facciolo, R_Char2Frame("A"))
	MSIntro_FaccMovement(c.ac_facciolo, c.faccstill)

	EvActionSH.Sq_CreateValue(c, "camera2panel6a", EvActionSH.coordinategetoffset(c, "mindscape3", 1700*FU, 16*FU, 12*FU))
	EvActionSH.setcamera(true, c.camera, c.camera2panel6a, 180*FU, 30*FU)
	startmindtext(1524)
end,
function(c,e)
	EventSH.wait(e, 1*TICRATE)
end,
function(c,e)
	EvActionSH.stopmobjmomentum(c.camera)
	EvActionSH.coordinateanchor(c, "facclocation", c.ac_facciolo.x, c.ac_facciolo.y, c.ac_facciolo.z)
	EvActionSH.Sq_CreateValue(c, "camera2panel6b", EvActionSH.coordinategetoffset(c, "facclocation", 64*FU, 12*FU, 16*FU))
	EvActionSH.setcamera(true, c.camera, c.camera2panel6b, 200*FU, 25*FU)
	EventSH.wait(e, 4*TICRATE, true)
end,
-- P3. PANEL 1
function(c,e)
	EvActionSH.coordinateanchor(c, "facclocation", c.ac_facciolo.x, c.ac_facciolo.y, c.ac_facciolo.z)
	EvActionSH.Sq_CreateValue(c, "camera2panel6b", EvActionSH.coordinategetoffset(c, "facclocation", 64*FU, 12*FU, 16*FU))
	EvActionSH.setcamera(true, c.camera, c.camera2panel6b, 200*FU, 25*FU)

	EvActionSH.setmobjframe(c.ac_facciolo, R_Char2Frame("N"))
	MSIntro_FaccMovement(c.ac_facciolo, c.faccstill)
	startmindtext(1525)
	EventSH.wait(e, 4*TICRATE, true)
end,
function(c,e)
	startmindtext(1526)
	EventSH.wait(e, 4*TICRATE, true)
end,
-- P3. PANEL 2
function(c,e)
	EvActionSH.coordinateanchor(c, "inalocation", c.ac_inazuma.x, c.ac_inazuma.y, c.ac_inazuma.z)
	EvActionSH.Sq_CreateValue(c, "camera3panel2", EvActionSH.coordinategetoffset(c, "inalocation", -32*FU, -96*FU, 50*FU))
	EvActionSH.setcamera(true, c.camera, c.camera3panel2, 90*FU, -10*FU)

	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_TAL7, "A", "A")
	P_InstaThrust(c.camera, c.camera.angle-ANGLE_90, FU/32)
	startmindtext(1527)
	EventSH.wait(e, 3*TICRATE, true)
end,
function(c,e)
	startmindtext(1528)
	EventSH.wait(e, 3*TICRATE+10, true)
end,
-- P3. PANEL 3
function(c,e)
	EvActionSH.stopmobjmomentum(c.camera)
	EvActionSH.Sq_CreateValue(c, "camera3panel3", EvActionSH.coordinategetoffset(c, "inalocation", -80*FU, -60*FU, 25*FU))
	EvActionSH.setcamera(true, c.camera, c.camera3panel3, 45*FU, 0)

	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_TAL8, "A", "A")
	MSIntro_FaccMovement(nil, nil, true)
	EvActionSH.stopmobjmomentum(c.ac_facciolo)
	startmindtext(1529)
	EventSH.wait(e, 3*TICRATE, true)
end,
-- P3. PANEL 4
function(c,e)
	EvActionSH.Sq_CreateValue(c, "camera3panel4", EvActionSH.coordinategetoffset(c, "inalocation", 110*FU, 0, 12*FU))
	EvActionSH.setcamera(true, c.camera, c.camera3panel4, 180*FU, 32*FU)
	P_InstaThrust(c.camera, c.camera.angle, FU/64)

	FrameAnimSH.addspr2("ina", c.ac_inazuma, SPR2_STND, nil, "E", {delay=4, loopalways=true})
	EvActionSH.Sq_CreateValue(c, "facc_exit", EvActionSH.coordinategetoffset(c, "mindscape3", -3840*FU, 3840*FU, 5555*FU))
	EvActionSH.stopmobjmomentum(c.ac_facciolo)
	EvActionSH.setmobjframe(c.ac_facciolo, R_Char2Frame("A"))
end,
function(c,e)
	-- TODO: facciolo doesn't start moving for no reason sometimes, and I don't know why.
	MobjMoverSH.moveto("facc_tocore", c.ac_facciolo, c.facc_exit, {speed=20*FU, varc=32*FU, harc=-450*FU, noangle=true, ease="inquart"})
end,
function(c,e)
	startmindtext(1530)
	EventSH.wait(e, 2*TICRATE, true)
end,
function(c,e)
	EvActionSH.fadescreen("full", 0, 0)
end,
function(c,e)
	MSIntro_FaccShadow(c.ac_facciolo)
	EvActionSH.setmobjvisible(c.ac_facciolo, false)
	EvActionSH.startsound(nil, sfx_facwrp, 100)
	EventSH.wait(e, 5, true)
end,
function(c,e)
	EvActionSH.fadescreen("clear")
	EventSH.wait(e, 2*TICRATE, true)
end,
function(c,e)
	startmindtext(1531)
	EvActionSH.startsound(nil, sfx_facflt, 80)
end,
function(c,e)
	EventSH.wait(e, 4*TICRATE)
end,
function(c,e)
	EvActionSH.fadescreen("out", 31, 8)
	S_FadeOutStopMusic(4*MUSICRATE)
	EventSH.destroy("sh_msintro_skip") -- Skipping cutoff
end,
function(c,e)
	EventSH.wait(e, 6*TICRATE)
end,
function(c,e)
	EvActionSH.relocateplayers_end() -- End constant warp
	EventSH.wait(e, 1*TICRATE, true)
end,
function(c,e)
	-- MSIntro_FaccShadow(nil, true)

	Silverhorn.SaveValue("game", "ms_accessed", true)

	EventSH.start("sh_msintro_wrapup", {fromcutscene=true})
end})





EventSH.new("sh_msintro_wrapup", {
function(c,e)
end,
function(c,e)
	-- Cleanup objects if they exist
	MSIntro_Cleanup()
	startmindtext(nil)
	EvActionSH.relocateplayers_end() -- End constant warp
	
	EventSH.destroy("sh_msintro_skip") -- Skipping cutoff
end,
function(c,e)

	-- Set map anchor
	EvActionSH.coordinateanchor(c, "mindscape3", -22016*FU, -18432*FU, -512*FU)

	-- First entry from cutscene, or you skipped it (do this once)
	if (c.fromcutscene) and not (MindscapeSessionFlags & MS_SEENINTRO) then
		MindscapeSessionFlags = $ | MS_SEENINTRO --Intro's over, toggle that we've seen it!
		EvActionSH.fadescreen("in", 31, 1)
		EvActionSH.doforall(function(player)
			local mo = player.mo

			--Return to spawn point
			if (mo.ms_originalspawn) then
				P_SetOrigin(mo, mo.ms_originalspawn.x,
				mo.ms_originalspawn.y, mo.ms_originalspawn.z)
				mo.angle = mo.ms_originalspawn.angle
			end
			-- EvActionSH.relocateplayers(EvActionSH.coordinategetoffset(c, "mindscape3", 1500*FU, 0, 0*FU), FixedAngle(180*FU))

			-- Hop out of portals if you aren't inazuma
			if not (mo.skin == "inazuma") then
				Silverhorn.SpawnMindscapePlayer(player)
			else
				mo.z = mo.floorz
				mo.state = S_PLAY_STND
			end
		end)
		-- Post-intro text
		startmindtext(7)
	end

	EvActionSH.SetInternalHudStatus({"time", "score", "rings", "lives"}, true)
	EvActionSH.stopcameraview(true)

	if (c.fromcutscene or multiplayer) then
		S_ChangeMusic(Silverhorn.UltimateMode() and "mndult" or "mindsc", true)
	end
end})









-- ============================================
-- Boss EV
-- 'Sequence' is root. Every thing else after
-- references this area sometimes
-- ============================================

Silverhorn.MindscapeSectorCollection = {}
Silverhorn.MindscapeSpecialParamNum = 11854
Silverhorn.MindscapeSpecialCollection = {}

-- Saves the brightness levels of the stage on map load
function Silverhorn.SaveMindscapeBrightnessLevels(mapnum)
	-- Save every map sector when mindscape is loaded
	if (mapheaderinfo[mapnum].mindscape) then
		-- Reset list
		Silverhorn.MindscapeSectorCollection = {}
		-- Fill a custom list with only the sector number, and lightlevel
		for sector in sectors.iterate do
			-- Skip invalid
			if not (sector and sector.valid) then continue end
			-- Insert sector into collection
			table.insert(Silverhorn.MindscapeSectorCollection, {sector=sector, sectornum = #sector, sectortag = sector.tag, lightlevel = sector.lightlevel})
		end
	else
		Silverhorn.MindscapeSectorCollection = {}
	end
end

-- Reloads previous light levels
function Silverhorn.ReloadMindscapeBrightnessLevels(mapnum)
	-- Ony run when mindscape is loaded
	if (mapheaderinfo[mapnum].mindscape) then
		for i=1, #Silverhorn.MindscapeSectorCollection do
			local saved_sector = Silverhorn.MindscapeSectorCollection[i]
			if (saved_sector and saved_sector.valid) then
				saved_sector.sector.lightlevel = saved_sector.lightlevel
			end
		end
	end
end

-- Collects Special map object types for Collapse
function Silverhorn.CollectMindscapeSpecialObjects(mapnum)
	-- Gather axis points in mindscape
	if (mapheaderinfo[mapnum].mindscape) then
		-- Reset list
		Silverhorn.MindscapeSpecialCollection = {}

		for mthing in mapthings.iterate do

			if not (mthing and mthing.valid) then continue end
			if not mthing.mobj then continue end

			-- Gets altviewman objects to use as references, etc
			if (mthing.mobj.type == MT_ALTVIEWMAN or mthing.mobj.type == MT_TELEPORTMAN or mthing.mobj.type == MT_AXIS)
			and (mthing.angle >= 800 and mthing.angle <= 999) then

				local mocfg = {}

				-- Grabs an (existing) linedef special to act as extra object parameters
				local lineparams = lines[P_FindSpecialLineFromTag(Silverhorn.MindscapeSpecialParamNum, mthing.angle)]

				mocfg["id"] = mthing.angle -- identifier tag (udmf pls)
				mocfg["mobj"] = mthing.mobj -- reference the mobj
				mocfg["x"] = mthing.mobj.x
				mocfg["y"] = mthing.mobj.y
				mocfg["z"] = mthing.mobj.z
				mocfg["valid"] = mthing.mobj.valid
				-- mocfg["valid"] = true

				if lineparams then
					-- Line angle
					local v1 = lineparams.v1
					local v2 = lineparams.v2
					mocfg["angle"] = R_PointToAngle2(v1.x, v1.y, v2.x, v2.y)
					-- Linedef execution tag
					mocfg["exectag"] = lineparams.frontside.textureoffset/FU
				end
				-- Update table
				table.insert(Silverhorn.MindscapeSpecialCollection, mocfg)
			end
		end
	else
		-- Clear instead
		Silverhorn.MindscapeSpecialCollection = {}
	end
end

-- Returns the map thing mobj from the collection that was made on load
function Silverhorn.MS_GetSpecialMobj(id)
	for i = 1, #Silverhorn.MindscapeSpecialCollection do
		if (Silverhorn.MindscapeSpecialCollection[i].id == id) then
			return (Silverhorn.MindscapeSpecialCollection[i])
		end
	end
end

-- Setup for Phase 2 ending
function Silverhorn.Phase2Destruction(mo)

	EventSH.start("ph2_final_d", {boss=mo})
end