local configPlayers = {}
local execCallbacks = {}
local connectCallbacks = {}

local function valid(mo)
	return mo and mo.valid
end

// hooks a function to be called on the client's end whenever the client joins a new game
function Silverhorn.AddExecCallback(func)
	table.insert(execCallbacks, func)
end

// hooks a function to be called when a client connects to the game
// acts as a substutite for PlayerJoin hooks since they aren't called when players rejoin games within the rejointimeout buffer
// which is useful, but I'd really like another hook for when someone joins a netgame at all...
function Silverhorn.AddPlayerConnectCallback(func)
	table.insert(connectCallbacks, func)
end

local function ExecTicker()
	if isserver ~= (consoleplayer == server) // the client is player 1 while the gamestate is being received, avoid sending commands during this time
		return
	end
	
	for i, player in ipairs({consoleplayer, secondarydisplayplayer})
		if not valid(configPlayers[i])
		and valid(player)
			configPlayers[i] = player
			for _, callback in ipairs(execCallbacks)
				callback(player, i)
			end
		end
	end
end

local prevQuitTimes = {}
local function PlayerConnectTicker()
	for i = 0, #players - 1
		local player = players[i]
		if not player
			prevQuitTimes[i] = INT32_MAX
			continue
		end
		if prevQuitTimes[i] and player.quittime == 0
			for _, callback in ipairs(connectCallbacks)
				callback(player)
			end
		end
		prevQuitTimes[i] = player.quittime
	end
end

addHook("ThinkFrame", do
	PlayerConnectTicker()
	ExecTicker()
end)

addHook("NetVars", function(net)
	prevQuitTimes = net(prevQuitTimes)
end)
