/*
Changeable button
	by Lach
	
v1
*/

local MAXSPLITSCREENPLAYERS = 2
local DEFAULTVALUE = BT_CUSTOM2
local COMMANDNAME = "silverhorn_dancebutton"
local COMMANDDESCRIPTION = "Sets which button activates Inazuma's and Aether's dances."
local VARNAME = "silverhorn_dancebutton"
local SAVENAME = "dancebutton"

local NOBUTTON = 1 << 31
local BUTTONNAMES = {
	// not including the weapon buttons here since those affect Aether's spells
	[BT_ATTACK] = "Ring Toss",
	[BT_TOSSFLAG] = "Toss Flag",
	[BT_FIRENORMAL] = "Ring Toss Normal",
	[BT_CUSTOM1] = "Custom 1",
	[BT_CUSTOM2] = "Custom 2",
	[BT_CUSTOM3] = "Custom 3",
	[NOBUTTON] = "none"
}

local function SimplifyString(str)
	return str:lower():gsub(" ", "")
end

// return a button constant based on the given string argument
local function GetButtonValue(arg)
	local num = tonumber(arg)
	
	// is this a button constant?
	if num
		if BUTTONNAMES[num]
			return num
		end
		return
	end
	
	// is this the name of a button?
	arg = SimplifyString($)
	for i, name in pairs(BUTTONNAMES)
		if SimplifyString(name) == arg
			return i
		end
	end
	
	// does this match a BT_ constant name?
	arg = $:upper()
	if arg:sub(1, 3) ~= "BT_"
		arg = "BT_"..arg
	end
	
	num = _G[arg]
	
	if num and BUTTONNAMES[num]
		return num
	end
	
	return
end

// return a button constant based on the value read from the file
local function ReadButton(playernum)
	local val = Silverhorn.LocalData["setting"][SAVENAME + playernum]
	if val != nil
		val = GetButtonValue(val)
	end
	return val or DEFAULTVALUE
end

// save the button values to the file
local function WriteButton(val, playernum)
	local button = val or DEFAULTVALUE
	Silverhorn.SaveValue("setting", SAVENAME + playernum, button)
end

// set this player's button to the given button constant, and save
local function SetButton(player, n, save)
	player[VARNAME] = n
	if save
	and (not netgame or player == consoleplayer)
		WriteButton(player[VARNAME], player == secondarydisplayplayer and 2 or 1)
	end
end

local function ChangeButton(player, ...)
	local silent = false
	local save = true
	local args = {...}
	local arg = ""
	
	for i, word in ipairs(args)
		word = $:upper()
		if word == "-SILENT"
			silent = true
		elseif word == "-NOSAVE"
			save = false
		else
			arg = $ .. word .. " "
		end
	end
	
	arg = $:sub(1, $:len() - 1)
	
	if not arg:len()
		if not silent
			local val = player[VARNAME]
			local buttonname
			if val
				buttonname = BUTTONNAMES[val] or "none"
			else
				buttonname = BUTTONNAMES[DEFAULTVALUE]
			end
			CONS_Printf(splitscreen and consoleplayer or player,
				COMMANDDESCRIPTION..
				"\nCurrently set to "..buttonname..".")
		end
		return
	end
	
	if arg == "DEFAULT"
	or tonumber(arg) == 0
		SetButton(player, 0, save)
		return
	end
	
	if arg == "NONE"
		SetButton(player, NOBUTTON, save)
		return
	end
	
	local val = GetButtonValue(arg)
	
	if val
		SetButton(player, val, save)
		return
	end
	
	if not silent
		CONS_Printf(splitscreen and consoleplayer or player,
			"Not a valid button!")
	end
end

COM_AddCommand(COMMANDNAME..2, ChangeButton, COM_SPLITSCREEN)
COM_AddCommand(COMMANDNAME, ChangeButton)

Silverhorn.AddExecCallback(function(player, playerNum)
	if player[VARNAME] ~= nil
		return
	end
	
	local text = COMMANDNAME
	if playerNum == 2
		text = $..2
	end
	text = $ .. " -silent -nosave " .. ReadButton(playerNum)
	CommandQueue_Add(text)
end)

Silverhorn.AddPlayerConnectCallback(function(player)
	player[VARNAME] = nil
end)
