/*
SilverhornChars functions

beta
*/

local Silverhorn = Silverhorn // performance hax

local SUPER_MUSIC = "_super" // generic super music we replace with the Ultimate music
local ULTIMATE_MUSIC = "ultinz" // music lump to play while Ultimate

local function valid(mo)
	return mo and mo.valid
end

local cv_renderer = CV_FindVar("renderer")

Silverhorn.RENDERER_NONE = 0
Silverhorn.RENDERER_SOFTWARE = 1
Silverhorn.RENDERER_OPENGL = 2
function Silverhorn.GetRenderer()
	if cv_renderer == nil
		return Silverhorn.RENDERER_NONE
	end
	return cv_renderer.value
end

// are we in ultimate mode?
function Silverhorn.UltimateMode()
	return ultimatemode or gametype == GT_UCOOP or gametype == GT_TUCOOP
end

// handles busting @fof
Silverhorn.MobjBustFOF = function(mo, fof)
	local line = fof.master
	
	// bust the FOF
	EV_CrumbleChain(fof)
	
	// run a linedef executor?
	if line.flags & ML_EFFECT5
		P_LinedefExecute(P_AproxDistance(line.dx, line.dy) >> FRACBITS, mo, fof.target)
	end
end

// returns whether @player is in their Sol form
function Silverhorn.IsSol(player, solonly)
	return (player.solchar and player.solchar.istransformed)
	or (not solonly and player.powers[pw_super]
	and ((player.hyper and player.hyper.transformed)
	or (MRCE_isHyper and MRCE_isHyper(player))))
end

// transforms @player into their Sol form
function Silverhorn.SolTransformation(player, animation)
	local sol = player.solchar
	local mo = player.mo
	
	if sol and mo and not sol.istransformed
		sol.istransformed = 1
		player.powers[pw_extralife] = 0
		player.powers[pw_sneakers] = 0
		player.powers[pw_invulnerability] = 0
		if animation == false then return end
		
		if P_IsObjectOnGround(mo)
			mo.z = $ + P_MobjFlip(mo)*mo.scale
		end
		A_ForceStop(mo)
		mo.state = S_PLAY_SUPER_TRANS1
		sol.spindown = true
		sol.emeraldcircle = 1
		S_StartSound(mo, sfx_secir)
	end
end

// moved Ultimate Inazuma's pulse effect into a separate function so it can be used with Sol forms
// must supply a player, the object to manipulated for the pulse effect, and a color
function Silverhorn.PulseAura(player, horn, color)
	local mo = player.mo
	horn.flags2 = $ & ~MF2_LINKDRAW
	horn.sprite = SPR_PLAY
	horn.frame = mo.frame|((leveltime%18/2)<<FF_TRANSSHIFT)
	horn.sprite2 = mo.sprite2
	horn.scale = FixedMul(mo.scale,((leveltime%18/2)+32)*FRACUNIT/32)
	horn.color = color
	horn.colorized = true
	horn.blendmode = AST_ADD
	A_CapeChase(horn,1-((leveltime%18/2)<<FRACBITS),0)
	if horn and horn.valid
		if displayplayer and displayplayer.valid
			P_MoveOrigin(horn,mo.x+P_ReturnThrustX(nil,displayplayer.cmd.angleturn<<16,FRACUNIT),
			mo.y+P_ReturnThrustY(nil,displayplayer.cmd.angleturn<<16,FRACUNIT),horn.z)
		else
			P_MoveOrigin(horn,mo.x+P_ReturnThrustX(nil,player.cmd.angleturn<<16,FRACUNIT),
			mo.y+P_ReturnThrustY(nil,player.cmd.angleturn<<16,FRACUNIT),horn.z)
		end
	end
end

function Silverhorn.CheckIncest(player)
    local name = player.name:lower():gsub("%A", "")
	local skinname = skins[player.skin].name
    if (skinname == "aether" and name:find("inazuma") and name:find("sister"))
    or (skinname == "inazuma" and name:find("aether") and name:find("brother"))
        return true
    end
    return false
end

// Sol form aura pulse
addHook("PostThinkFrame", function()
	for player in players.iterate
		local aura = player.silverhornaura
		
		if not player continue end
		local mo = player and player.mo
		
		if valid(mo) and mo.health and Silverhorn.CheckIncest(player)
			P_DamageMobj(mo, nil, nil, 1, DMG_INSTAKILL)
		end
		
		// Guard frames (Based on code by Fav)
		if valid(mo) and (player.inazuma or player.aether)
		and (player.guard or (mo.silverhornguard
		and (player.panim >= PA_IDLE and player.panim <= PA_WALK)))
			mo.state = S_TAILSOVERLAY_RUN
			player.panim = PA_IDLE
			mo.tics = -1
			if not (player.inazuma and Silverhorn.Inazuma.CheckSilverhorns(player))
				mo.sprite2 = SPR2_GARD
			end
			
			mo.frame = $ &~ FF_FRAMEMASK
			if player.guard == 2
				mo.silverhornguard = TICRATE*2/3
				mo.frame = $ | B
			elseif mo.silverhornguard == 1
				mo.silverhornguard = 0
				mo.state = S_PLAY_STND
			elseif mo.silverhornguard
				mo.silverhornguard = $ - 1
				mo.frame = $ | B
			end
		elseif valid(mo) and mo.silverhornguard
			mo.silverhornguard = $ - 1
		end
		
		if not Silverhorn.IsSol(player)
		or not (player.inazuma or player.aether)
		or (player.solchar and player.solchar.istransformed == 1)
		or not valid(mo) or not mo.health
			if valid(aura)
				P_RemoveMobj(aura)
				player.silverhornaura = nil
			end
			continue
		end
		
		if not valid(aura)
			aura = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_OVERLAY)
			aura.target = mo
			aura.tracer = mo
			aura.state = S_INVISIBLE
			player.silverhornaura = aura
		end
		
		aura.skin = mo.skin
		Silverhorn.PulseAura(player, aura, solchars[mo.skin][1])
	end
end)

--Push code
local PUSH_SPRITE2 = SPR2_FIRE
addHook("MobjMoveBlocked", function(mo, mobj, line)
	local player = mo.player
	if (mo.skin != "aether" and mo.skin != "inazuma")
	or not player
	or not P_IsObjectOnGround(mo)
	or not player.panim or player.panim > PA_RUN
	or not (player.cmd.forwardmove or player.cmd.sidemove)
	then return end
	
	local tics = 7
	
	--Face object we're pushing
	if mobj
		local inputangle = player.cmd.angleturn<<16 + R_PointToAngle2(0, 0, player.cmd.forwardmove*FRACUNIT, -player.cmd.sidemove*FRACUNIT)
		local mox = mo.x - mo.momx - P_ReturnThrustX(mo, inputangle, mo.radius)
		local moy = mo.y - mo.momy - P_ReturnThrustY(mo, inputangle, mo.radius)
		player.drawangle = R_PointToAngle2(mox, moy, mobj.x, mobj.y)
		player.powers[pw_justsprung] = tics --Prevent changing angle
		
	--Wall logic
	elseif line
		if player.speed > mo.scale*2 --Going too fast! Abort!
			return
		end
		tics = 2
	end
	
	--Push animation
	--This is a bit of a hack to save having to do any state code, but it works
	local spr2super = mo.sprite2 & FF_SPR2SUPER
	mo.state = S_PLAY_GASP --Dummy state that transitions to walk
	player.panim = player.speed > FixedMul(player.runspeed, mo.scale) and PA_RUN or PA_WALK
	mo.tics = tics
	
	--Push frame
	local frames = skins[player.skin].sprites[PUSH_SPRITE2].numframes
	mo.sprite2 = PUSH_SPRITE2 | spr2super
	mo.frame = ($ &~ FF_FRAMEMASK) | ((leveltime/8 + #player) % frames)
end)

--Don't load in the MusicChange hook in srb2p
--We'd only be increasing the chances of crashing
--Especially bad when you can't save in srb2p
if srb2p then return end

--Check for if we should play the Ultimate music
local doultimatemusic = false
local function CheckUltimateMusic(player)
	if not player and player.valid
	and player.mo and player.mo.valid
	and player.inazuma then return end
	
	if Silverhorn.Inazuma.IsUltimate(player)
		doultimatemusic = true
	end
end

addHook("MusicChange", function(oldmusic, newmusic, flags, looping)
	--When turned into Ultimate, replace the super music with Ultimate's theme
	if newmusic == SUPER_MUSIC then
		doultimatemusic = false
		if not multiplayer
			for player in players.iterate
				CheckUltimateMusic(player)
			end
		else
			for i, player in ipairs({consoleplayer, secondarydisplayplayer})
				CheckUltimateMusic(player)
			end
		end
		
		if doultimatemusic then return ULTIMATE_MUSIC end
		
	--Turns off music name when the intro is loaded (prevents immediate interruption of level music)
	elseif newmusic == "MINDSC" and mapheaderinfo[gamemap].mindscape then
		if not (MindscapeSessionFlags & MS_SEENINTRO) then
			return true
		elseif consoleplayer
		and consoleplayer.starpostnum >= Silverhorn.FINALSTARPOSTID
			return true
		elseif Silverhorn.UltimateMode() then
			return "MNDULT"
		end
	end
end)