//Sky Kill by SMS Alfredo
//Makes any sector with a sky act as a death pit
//Add it to your map and put this in the SOC header: Lua.skykill = true
//To add exceptions for certain tag numbers, follow this format:
/*
Lua.skykillex = 3, 6, 2, 9
*/

if SkyKill then return end
rawset(_G, "SkyKill", true)

local exceptionlist = nil
local lastmap = nil
local MAXEXCEPTIONS = 99

local function SkyKill_GetExceptionTag(tag, taglist)
	local found_tag = false

	for _, extag in ipairs(exceptionlist) do
		--print("Exception: " .. extag .. "/ Sector: " .. sector.tag)
		for i = 1, #taglist, 1 do
			if (extag == taglist[i]) then
				found_tag = true
				break
			end
		end
		if found_tag then break end
	end
	return found_tag
end

addHook("PostThinkFrame", do
	if not mapheaderinfo[gamemap].skykill then return end

	if not exceptionlist or lastmap != gamemap
		lastmap = gamemap
		exceptionlist = {}

		local exc = 1
		-- Split the tags by delimiter (you can temporarily exclude by just putting anything that isn't a number)
		local exclusions = mapheaderinfo[gamemap]["skykillex"]
		exclusions = $ and string.gmatch($, "[^,]+")
		if exclusions
			for nx in exclusions do
				if exc > (MAXEXCEPTIONS) then
					print("Sky Kill does not support more than " + MAXEXCEPTIONS + " exceptions!")
					break
				end

				local exception = tonumber(nx)

				if exception then
					if exception < 0 then exception = $ + UINT16_MAX + 1 end
					-- exceptionlist[exception] = true
					table.insert(exceptionlist, exception)
				else
					continue
				end
				exc = $1 + 1
			end
		end
	end

	for player in players.iterate
		if player.pflags & PF_GODMODE or player.playerstate then continue end
		local mo = player.mo
		if not (mo and mo.health and P_IsObjectOnGround(mo)
		and mo.eflags & MFE_ONGROUND) then continue end
		local sector = mo.subsector and mo.subsector.sector

		-- (Account for sector taglists)
		if not sector or SkyKill_GetExceptionTag(sector.tag, sector.taglist) then continue end
		-- if not sector or exceptionlist[sector.tag] then continue end

		if P_MobjFlip(mo) > 0
			if string.upper(sector.floorpic) == "F_SKY1"
			and mo.z <= sector.floorheight
				P_DamageMobj(mo, nil, nil, 1, DMG_DEATHPIT)
			end
		elseif string.upper(sector.ceilingpic) == "F_SKY1"
		and mo.z + mo.height >= sector.ceilingheight
			P_DamageMobj(mo, nil, nil, 1, DMG_DEATHPIT)
		end
	end
end)