//Sky Save by SMS Alfredo
//Makes any sector with a sky act as a teleporter
//Add it to your map and put this in the SOC header: Lua.skysave = true
//To add exceptions for certain tag numbers, follow this format:
/*
Lua.skysaveex1 = 3
Lua.skysaveex2 = 6
Lua.skysaveex3 = 2
Lua.skysaveex4 = 9
*/

//If you want to use a whilelist rather than a blacklist, use the above to
//determine the section tags that Sky Save will EXCLUSIVELY work on.
//To do so, set Lua.skysave to whitelist instead of true.

//To make the player teleport to a specific location every time in a map,
//specify the X, Y, and Z pos in order like so: Lua.skysavepos = 960, 720, 1024

if SkySave then return end
rawset(_G, "SkySave", true)

local exceptionlist = nil
local lastmap = nil
local MAXEXCEPTIONS = 99

local function SkySave_GetExceptionTag(tag, taglist)
	local found_tag = false

	for _, extag in ipairs(exceptionlist) do
		--print("Exception: " .. extag .. "/ Sector: " .. sector.tag)
		for i = 1, #taglist, 1 do
			if (extag == taglist[i]) then
				found_tag = true
				break
			end
		end
		if found_tag then break end
	end
	
	if mapheaderinfo[gamemap].skysave == "whitelist"
		return not found_tag
	else
		return found_tag
	end
end

addHook("PostThinkFrame", do
	if not mapheaderinfo[gamemap].skysave then return end

	if not exceptionlist or lastmap != gamemap
		lastmap = gamemap
		exceptionlist = {}

		local exc = 1
		-- Split the tags by delimiter (you can temporarily exclude by just putting anything that isn't a number)
		local exclusions = mapheaderinfo[gamemap]["skysaveex"]
		exclusions = $ and string.gmatch($, "[^,]+")
		if exclusions
			for nx in exclusions do
				if exc > (MAXEXCEPTIONS) then
					print("Sky Save does not support more than " + MAXEXCEPTIONS + " exceptions!")
					break
				end

				local exception = tonumber(nx)

				if exception then
					if exception < 0 then exception = $ + UINT16_MAX + 1 end
					table.insert(exceptionlist, exception)
				else
					continue
				end
				exc = $1 + 1
			end
		end
	end

	for player in players.iterate
		local mo = player.mo
		if player.playerstate or not (mo and mo.valid) then continue end
		
		-- During respawn
		if mo.pitrespawn
			mo.pitrespawn = $ - 1
			if mo.pitrespawn == 8 and player == displayplayer
				P_FlashPal(player, PAL_WHITE, 16)
			end
			
			P_ResetPlayer(player)
			if mo.state != S_PLAY_STUN
				mo.state = S_PLAY_STUN
			end
			
			if not mo.pitrespawn
				mo.flags = mo.pitrespawnflags
				A_ForceStop(mo)
				
				local positions = mapheaderinfo[gamemap]["skysavepos"]
				positions = $ and string.gmatch($, "[^,]+")
				if positions
					local i = 1
					for pos in positions do
						if i == 1
							mo.pitrespawnx = pos*FRACUNIT
						elseif i == 2
							mo.pitrespawny = pos*FRACUNIT
						else
							mo.pitrespawnz = pos*FRACUNIT
						end
						i = $ + 1
					end
				end
				
				P_SetOrigin(mo, mo.pitrespawnx,
				mo.pitrespawny, mo.pitrespawnz + mo.height*2*P_MobjFlip(mo))
				S_StartSound(mo, sfx_prloop)
				
				local ring = P_SpawnMobjFromMobj(mo, 0,0,0, MT_THROWNEXPLOSION)
				ring.target = mo
				ring.fuse = 1
				P_KillMobj(ring)
				
				if not player.awayviewtics
					player.awayviewtics = 1
					player.awayviewtics = 0
				end
			end
			continue
		end
		
		-- Must be grounded for the rest of the code!
		if not (P_IsObjectOnGround(mo) and mo.eflags & MFE_ONGROUND)
		or (mo.floorrover and P_CanRunOnWater(player, mo.floorrover))
		or (mo.ceilingrover and P_CanRunOnWater(player, mo.ceilingrover))
		or (Silverhorn and mapheaderinfo[gamemap].mindscape
		and Silverhorn.Inazuma.IsUltimate(player))
		then continue end
		
		local sector = mo.subsector and mo.subsector.sector
		
		-- Start teleport if we fall on sky
		if sector and mo.pitrespawnx != nil
		and not (mo.flags & (MF_NOCLIP | MF_NOCLIPHEIGHT | MF_NOCLIPTHING))
		and not SkySave_GetExceptionTag(sector.tag, sector.taglist)
		and (P_MobjFlip(mo) > 0
		and string.upper(sector.floorpic) == "F_SKY1"
		and mo.z <= sector.floorheight)
		or (P_MobjFlip(mo) < 0
		and string.upper(sector.ceilingpic) == "F_SKY1"
		and mo.z + mo.height >= sector.ceilingheight)
			mo.pitrespawn = TICRATE
			player.powers[pw_nocontrol] = mo.pitrespawn
			S_StartSound(mo, sfx_s3k51)
			
			P_ResetPlayer(player)
			mo.state = S_PLAY_STUN
			
			mo.pitrespawnflags = mo.flags
			mo.flags = $ | MF_NOCLIP | MF_NOCLIPHEIGHT | MF_NOCLIPTHING
			
			A_ForceStop(mo)
			mo.momz = abs(mo.height)*-P_MobjFlip(mo)
			
		-- Get last standing position for teleport
		elseif not (mo.standingslope and mo.standingslope.valid
		and abs(mo.standingslope.zangle) < ANGLE_45)
			mo.pitrespawnx = mo.x
			mo.pitrespawny = mo.y
			mo.pitrespawnz = mo.z
		end
	end
end)