/*
Inazuma & Aether
Tag Team Swap Mechanic
	by SMS Alfredo
	& Lach
*/

local Silverhorn = Silverhorn // performance hax
local Inazuma = Silverhorn.Inazuma
local Aether = Silverhorn.Aether

// if a player is in any of these states, they will be set to S_PLAY_STND instead
local STATE_BLACKLIST = {
	S_CHRISPY_DANCE,
	S_INAZUMA_DASH,
	S_PLAY_GASP
}

// swap these pflags, leave the rest alone
local PFLAG_WHITELIST = PF_JUMPED|PF_NOJUMPDAMAGE|PF_THOKKED|PF_SPINNING|PF_STARTDASH|PF_SHIELDABILITY|PF_GLIDING|PF_BOUNCING|PF_CANCARRY

// @s1 and @s2 can be players, mobjs, tables, etc.
// @field is a string, the name of the field to be swapped
local function SwapStructField(s1, s2, field)
	s1[field], s2[field] = $2, $1
end

// blocks a character switch if any player in @playas is in the spell menu
// otherwise, resets spells and flight
local function ProcessAetherStuff(...)
	local playas = {...}
	for i, player in ipairs(playas)
		local aether = player.aether
		if not aether then continue end
		if aether.halted
			return false
		end
	end
	
	for i, player in ipairs(playas)
		local aether = player.aether
		if not aether then continue end
		if aether.flight
			aether.flight = 0
			aether.wings = nil
		end
		for j = 1, SPL_MAX
			DeactivateSpell(player, j, ABORT_TAGTEAM)
		end
	end
	return true
end

// ensures all Inazuma/Aether overlays are retargeted to their new players
local function ShiftTargets(...)
	for i, player in ipairs({...})
		local objs = {}
		local mo = player.mo
		local inazuma = player.inazuma
		local aether = player.aether
		
		if inazuma
			table.insert(objs, inazuma.aura) // immuinity aura
			table.insert(objs, inazuma.boltstreamoverlay) // bolt stream
		end
		
		if aether
			if aether.wings
				for _, wing in ipairs(aether.wings)
					table.insert(objs, wing)
				end
			end
		end
		
		table.insert(objs, mo.aetherbarrier) // aura barrier
		table.insert(objs, mo.aura) // speedwave
		
		for i, obj in ipairs(objs)
			if obj.valid
				obj.target = player.mo
			end
		end
	end
end

addHook("MapLoad", do
	local player1 = players[0]
	local mo1 = player1 and player1.mo
	local player2 = players[1]
	local mo2 = player2 and player2.mo
	
	if not netgame and not multiplayer
	and mo1 and (mo1.skin == "inazuma" or mo1.skin == "aether")
	and mo2 and (mo2.skin == "inazuma" or mo2.skin == "aether")
	and mo1.skin == mo2.skin
		if mo1.skin == "aether"
			R_SetPlayerSkin(player2, "inazuma")
		else
			R_SetPlayerSkin(player2, "aether")
		end
	end
end)

addHook("PreThinkFrame", do
	if netgame or multiplayer
		return
	end
	
	local player1 = players[0]
	local mo1 = player1 and player1.mo
	local player2 = players[1]
	local mo2 = player2 and player2.mo
	
	if mo1 and mo1.health
	and (mo1.skin == "inazuma" or mo1.skin == "aether")
		if not (player1.cmd.buttons & BT_CUSTOM1)
		or player1.pflags & PF_STASIS or player1.exiting
			player1.silverhornswapready = true
			player1.silverhornswapping = false
		elseif player1.silverhornswapready then
			player1.silverhornswapping = true
			player1.silverhornswapready = false
		else
			player1.silverhornswapping = false
		end
		
		if not (mo2 and mo2.health
		and (mo2.skin == "inazuma" or mo2.skin == "aether"))
		then return end
		
		if P_PlayerInPain(player1) or mo1.state == S_INAZUMA_SMACK
		or (mo1.state >= S_PLAY_SUPER_TRANS1 and mo1.state <= S_PLAY_SUPER_TRANS6)
		or (player1.solchar and player1.solchar.istransformed == 1)
		or player1.powers[pw_carry]
		or P_PlayerInPain(player2) or mo2.state == S_INAZUMA_SMACK
		or (mo2.state >= S_PLAY_SUPER_TRANS1 and mo2.state <= S_PLAY_SUPER_TRANS6)
		or (player2.solchar and player2.solchar.istransformed == 1)
		or player2.powers[pw_carry]
			if player1.silverhornswapping
			and not mapheaderinfo[gamemap].silverhorn_endingmap
				S_StartSound(mo1, sfx_pfail, player1)
			end
			return
		end
		
		// check ultimate transformation for later
		local transforming = player1.inazuma and (player1.inazuma.transtime
		or player1.inazuma.flags & (Inazuma.FLAGS.TRANSFORMING|Inazuma.FLAGS.ULTIMATESLUMBER))
		
		// set lightning color
		if player2.inazuma_lightningcolor != player1.inazuma_lightningcolor
			player2.inazuma_lightningcolor = player1.inazuma_lightningcolor
		end
		
		// set super
		if player1.powers[pw_super] and not (player1.inazuma
		and player1.inazuma.flags & Inazuma.FLAGS.ULTIMATESLUMBER)
			if mo1.state == S_PLAY_SUPER_TRANS1
			and mo2.state ~= S_PLAY_SUPER_TRANS1
				mo2.state = S_PLAY_SUPER_TRANS1
			end
			player2.charflags = $ | SF_SUPER
			player2.powers[pw_super] = player1.powers[pw_super]
			player2.rings = player1.rings
		end
		
		// set sol
		if player1.solchar and player1.solchar.istransformed and player2.solchar
			player2.solemeralds = player1.solemeralds
			player2.rings = player1.rings
			if not player2.solchar.istransformed
				Silverhorn.SolTransformation(player2, player1.solchar.istransformed == 1)
			end
		end
		
		// bot command stuff
		if player2.bot == 1
			// mimic player 1 during sol transformation
			if player2.solchar and player2.solchar.istransformed == 1
				player2.cmd.buttons = player1.cmd.buttons
			end
			
			// set slow fall
			if not P_IsObjectOnGround(mo1) and not P_IsObjectOnGround(mo2)
			and ((mo1.skin == "inazuma" and player1.inazuma
			and player1.inazuma.flags & Inazuma.FLAGS.BARRIER)
			or (mo1.skin == "aether" and player1.aether and player1.aether.slowfall))
				if mo2.skin == "inazuma"
					player2.cmd.buttons = $ | BT_SPIN
				elseif not (player2.aether and player2.aether.flight)
					if player2.pflags & PF_STARTJUMP
						player2.cmd.buttons = $ &~ BT_JUMP
					else
						player2.cmd.buttons = $ | BT_JUMP
					end
				end
			elseif mo2.skin == "aether" and (player2.aether and player2.aether.slowfall)
			and player1.powers[pw_carry] == CR_PLAYER
				player2.cmd.buttons = $ | BT_JUMP
			end
			
			// no jumping allowed during ultimate transformation or while being carried!
			if (player1.inazuma and transforming)
			or player2.powers[pw_carry] == CR_PLAYER
				player2.cmd.buttons = $ &~ BT_JUMP
			end
		end
		
		// start a swap
		if player1.silverhornswapping and not transforming
		and ProcessAetherStuff(player1, player2)
			S_StartSound(mo1, sfx_pswap)
			
			// swap colors
			SwapStructField(player1, player2, "skincolor")
			SwapStructField(mo1, mo2, "color")
			
			// swap skins
			local xmomSpeed
			if player1.isxmomentum
				xmomSpeed = player1.normalspeed
			end
			
			SwapStructField(mo1, mo2, "skin")
			R_SetPlayerSkin(player1, mo1.skin)
			R_SetPlayerSkin(player2, mo2.skin)
			
			if xmomSpeed ~= nil
				player1.normalspeed = xmomSpeed
				player2.normalspeed = xmomSpeed
			end
			
			// swap HAX
			SwapStructField(player1, player2, "hasnomomentum")
			SwapStructField(player1, player2, "blazespindash")
			
			// swap inazuma table
			SwapStructField(player1, player2, "inazuma")
			mo2.inazuma = player2.inazuma
			if player2.inazuma != nil
				player2.inazuma.player = player2
				player2.inazuma.mo = mo2
				
				// I'm not sure what the original lines here were supposed to do, but R_SetPlayerSkin woulda set the normalspeed back to the default for the skin, so I'm just gonna simplify it
				Inazuma.SetNormalspeed(player2, Inazuma.GetModifiedSpeed(player2, skins[player2.skin].normalspeed))
				
				if player2.inazuma.flags&Inazuma.FLAGS.ULTIMATE
					player2.charflags = $ & ~SF_NOSUPERSPRITES
				end
			end
			mo1.inazuma = player1.inazuma
			if player1.inazuma != nil
				player1.inazuma.player = player1
				player1.inazuma.mo = mo1
				if player1.aether != nil
				and player1.aether.dashmode >= Aether.DASHMODE_TIME
					Inazuma.SetNormalspeed(player1, Inazuma.GetModifiedSpeed(player1, skins[player1.skin].normalspeed) + Inazuma.NEURALSHOCK_ADDSPEED)
					player1.inazuma.neuralshock = Inazuma.NEURALSHOCK_TIME
				else
					Inazuma.SetNormalspeed(player1, Inazuma.GetModifiedSpeed(player1, skins[player1.skin].normalspeed))
				end
				if player1.inazuma.flags&Inazuma.FLAGS.ULTIMATE
					player1.charflags = $ & ~SF_NOSUPERSPRITES
				end
			end
			
			//cancel BattleMod abilities
			player1.actionstate = 0
			player1.actiontime = 0
			player2.actionstate = 0
			player2.actiontime = 0
			
			// swap aether table
			SwapStructField(player1, player2, "aether")
			if player1.aether != nil and player2.inazuma != nil
			and player2.inazuma.neuralshock
				player1.aether.dashmode = Aether.DASHMODE_TIME
			end
			
			// swap state
			SwapStructField(mo1, mo2, "state")
			
			// swap pflags
			local pflags = player1.pflags & PFLAG_WHITELIST
			player1.pflags = $ & ~PFLAG_WHITELIST | (player2.pflags & PFLAG_WHITELIST)
			player2.pflags = $ & ~PFLAG_WHITELIST | pflags
			
			// don't retain PF_JUMPED on the ground!
			if player1.pflags & PF_JUMPED and P_IsObjectOnGround(mo1)
				player1.pflags = $ & ~PF_JUMPED
			end
			if player2.pflags & PF_JUMPED and P_IsObjectOnGround(mo2)
				player2.pflags = $ & ~PF_JUMPED
			end
			
			// swap character-specific fields
			SwapStructField(player1, player2, "inazumasilverhorn")
			SwapStructField(player1, player2, "inazuma_barrier")
			SwapStructField(player1, player2, "isinaura") // speedwave
			SwapStructField(player1, player2, "wasinaura")
			SwapStructField(mo1, mo2, "aura")
			SwapStructField(mo1, mo2, "aetherbarrier") // aura barrier
			
			// reset object targets
			ShiftTargets(player1, player2)

			P_MoveOrigin(mo2, mo1.x, mo1.y, mo1.z)
			mo2.momx = mo1.momx
			mo2.momy = mo1.momy
			P_Thrust(mo2, player1.drawangle,
			min(-8*mo1.scale,FixedMul(player1.speed/-2,mo1.scale)))
			mo2.momz = mo1.momz
			mo2.angle = mo1.angle
			player2.drawangle = player1.drawangle
			
			// prevent raising errors when switching into certain states
			for i = 0, 1
				local mo = players[i].mo
				for _, state in ipairs(STATE_BLACKLIST)
					if mo.state == state
						if P_IsObjectOnGround(mo)
							mo.state = S_PLAY_STND
						else
							mo.state = S_PLAY_FALL
						end
					end
				end
			end
		end
	end
end)