-- make textures float
local function getTrueTag(tag)
	if (tag >= 0) then return tag end	-- it's normal
	return FRACUNIT + tag
end
local FLOATINGTEXTURE_TAG = 60002	-- what tag will be detected to give the texture floating functionality

addHook("MapLoad", function(gamemap)
	if server and server.valid then server.floatingtextures = {} end	-- flush table
	if not (mapheaderinfo[gamemap].floatingtextures) then return end
	
	for i = 0, #lines - 1
		local line = lines[i]
		if not (line and line.valid) then continue end
		
		//print("line tag: "..line.tag)
		//print("line (true) tag: "..getTrueTag(line.tag))
		if (getTrueTag(line.tag) != FLOATINGTEXTURE_TAG) then continue end	-- only pickup lines with this tag
		
		if not (line.backside) then return end	-- it NEEDS the backside
		local floatheight = line.backside.textureoffset
		local totaltics = line.backside.rowoffset / FRACUNIT
		if not (floatheight and totaltics) then continue end	-- we need both of these
		
		local floatinfo = {
			starty = line.frontside.rowoffset,	-- starting position
			floatheight = floatheight,	-- how high it'll float up
			totaltics = totaltics,	-- how long it takes to float (in tics)
		}
		//print("FLOATINFO["..i.."]:")
		//print("starty: "..floatinfo["starty"])
		//print("floatheight: "..floatinfo["floatheight"])
		//print("totaltics: "..floatinfo["totaltics"])
		server.floatingtextures[i] = floatinfo
	end
end)

addHook("ThinkFrame", do
	if not (mapheaderinfo[gamemap].floatingtextures) then return end
	
	if not (server and server.valid and server.floatingtextures) then return end	-- table doesn't exist yet
	Silverhorn.perfStats:start("Floating textures")
	for i, floatinfo in pairs(server.floatingtextures)
		local easingtics = floatinfo["totaltics"]
		local maxeasing = FRACUNIT*2
		local easing = ease.inoutsine(((leveltime % easingtics) * maxeasing) / easingtics, FRACUNIT)
		
		-- set rowoffset
		//print("easing: "..easing)
		local line = lines[i]
		if not (line and line.valid) then continue end
		line.frontside.rowoffset = floatinfo["starty"] + FixedMul(easing, floatinfo["floatheight"])
	end
	Silverhorn.perfStats:stop("Floating textures")
end)