// Varren's SpriteGroup3D, with a few tweaks
-- Builds up a mobj sprite group from a list of mobjs
rawset(_G, "P_BuildSpriteGroup", function(source, grouplist)

	-- source mobj is not valid
	if not source.valid then return end

	source.sprgroup = grouplist
end)

-- Updates the position and callback functions of the sprite group
rawset(_G, "P_UpdateSpriteGroup", function(source, func)

	-- source mobj is not valid
	if not source.valid then return end

	-- Run through the entire source spritegroup
	for i=1,#source.sprgroup do

		local groupmobj = source.sprgroup[i].mobj
		if not (groupmobj and groupmobj.valid) then continue end
		
		local state = source.sprgroup[i].state or 0
		local offset = source.sprgroup[i].offset or {x = 0, y = 0, z = 0} -- TODO: be able to exclude each axis; default to 0
		local direction = source.sprgroup[i].angle or 0
		local rotation = source.sprgroup[i].rotation or 0
		
		-- Go to the state assigned
		if state
			groupmobj.state = state
		end

		-- Follow the source angle + independent angle
		groupmobj.angle = source.angle+FixedAngle(direction*FRACUNIT)
		
		-- Set scale to source scale
		groupmobj.scale = source.scale

		-- Run a callback function to edit one or all items
		if (func) then
			do
				func(source.sprgroup[i])
			end
		end
		
		if not (groupmobj and groupmobj.valid) then continue end
		
		local height = P_MobjFlip(source) < 0 and groupmobj.height or 0
		
		-- Update the position of all group items to be relative to the mobj angle + offsets (and scaling!)
		P_MoveOrigin(
			groupmobj,
			source.x+FixedMul(offset.x*cos(source.angle+FixedAngle(rotation*FRACUNIT)), groupmobj.scale),
			source.y+FixedMul(offset.y*sin(source.angle+FixedAngle(rotation*FRACUNIT)), groupmobj.scale),
			source.z+offset.z*FRACUNIT-height)
	end
end)

rawset(_G, "CreatePaperspriteCircle", function(mo, num, radius, _type, _state, zoffs, _id)
	local t = {}
	
	for i = 0, 360 - 360/num, 360/num
		local flip = (i >= 360/2) and -1 or 1
		
		local b = {
			mobj = P_SpawnMobjFromMobj(mo, 0, 0, 0, _type),
			state = _state,
			offset = {x = flip*radius, y = flip*radius, z = zoffs},
			angle = i,
			rotation = i + 90*flip,
			id = _id
		}
		
		table.insert(t, b)
	end
	
	return t
end)

rawset(_G, "MergeTables", function(t1, t2)
	local n = #t1
	
	for k, v in ipairs(t2)
		t1[n + k] = v
	end
	
	return t1
end)

local function EntityHeights(e)
	local _type = userdataType(e)
	
	if _type == "mobj_t"
		return e.z, e.z + e.height
	elseif _type == "sector_t"
		return e.floorheight, e.ceilingheight
	elseif _type == "ffloor_t"
		return e.bottomheight, e.topheight
	end
end

rawset(_G, "HeightCheck", function(a, b)
	local abottom, atop = EntityHeights(a)
	local bbottom, btop = EntityHeights(b)
	
	if abottom > btop return false end
	if atop < bbottom return false end
	if bbottom > atop return false end
	if btop < abottom return false end
	
	return true
end)

// GetRelativeControlDirection
rawset(_G, "GetRelativeControlDirection", function(p)
	if not p.aether return end
	if not p.aether.cmd return end
	
	local cmd = p.aether.cmd
	if not (cmd.forwardmove or cmd.sidemove) then return end
	
	return R_PointToAngle2(0,0, cmd.forwardmove*FRACUNIT, cmd.sidemove*FRACUNIT)
end)